
/*

  KLayout Layout Viewer
  Copyright (C) 2006-2007 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/



#include "dbEdgeProcessor.h"
#include "dbLayout.h"
#include "tlTimer.h"
#include "gsi.h"

#include <vector>
#include <deque>

namespace db
{

// -------------------------------------------------------------------------------
//  Some utilities ..

static inline db::Coord edge_ymin (const db::Edge &e) 
{
  return std::min (e.p1 ().y (), e.p2 ().y ());
}

static inline db::Coord edge_ymax (const db::Edge &e) 
{
  return std::max (e.p1 ().y (), e.p2 ().y ());
}

static inline db::Coord edge_xmin (const db::Edge &e) 
{
  return std::min (e.p1 ().x (), e.p2 ().x ());
}

static inline db::Coord edge_xmax (const db::Edge &e) 
{
  return std::max (e.p1 ().x (), e.p2 ().x ());
}

// HINT: it is important that this method delivers exactly (!) the same x for the same edge (after normalization to dy()>0) and same y!
static inline double edge_xaty (db::Edge e, db::Coord y)
{
  if (e.p1 ().y () > e.p2 ().y ()) {
    e.swap_points ();
  }
  if (y <= e.p1 ().y ()) {
    return e.p1 ().x ();
  } else if (y >= e.p2 ().y ()) {
    return e.p2 ().x ();
  } else {
    return double (e.p1 ().x ()) + double (e.dx ()) * double (y - e.p1 ().y ()) / double (e.dy ());
  }
}

static inline db::Coord edge_xmin_at_yinterval (const db::Edge &e, int y1, int y2) 
{
  if (e.dx () == 0) {
    return e.p1 ().x ();
  } else if (e.dy () == 0) {
    return std::min (e.p1 ().x (), e.p2 ().x ());
  } else {
    return db::Coord (floor (edge_xaty (e, ((e.dy () < 0) ^ (e.dx () < 0)) == 0 ? y1 : y2)));
  }
}

static inline db::Coord edge_xmax_at_yinterval (const db::Edge &e, int y1, int y2) 
{
  if (e.dx () == 0) {
    return e.p1 ().x ();
  } else if (e.dy () == 0) {
    return std::max (e.p1 ().x (), e.p2 ().x ());
  } else {
    return db::Coord (ceil (edge_xaty (e, ((e.dy () < 0) ^ (e.dx () < 0)) != 0 ? y1 : y2)));
  }
}

struct ProjectionCompare
{
  ProjectionCompare (const db::Edge &e)
    : m_e (e) { }

  bool operator () (const db::Point &a, const db::Point &b) const
  {
    return db::sprod (m_e, db::Edge (m_e.p1 (), a)) < db::sprod (m_e, db::Edge (m_e.p1 (), b));
  }

private:
  db::Edge m_e;
};

struct PolyMapCompare
{
  PolyMapCompare (db::Coord y)
    : m_y (y) { }

  bool operator() (const std::pair<db::Edge, size_t> &a, const std::pair<db::Edge, size_t> &b) const
  {
    //  simple cases ..
    if (a.first.dx () == 0 && b.first.dx () == 0) {
      return a.first.p1 ().x () < b.first.p1 ().x ();
    } else if (edge_xmax (a.first) < edge_xmin (b.first)) {
      return true;
    } else if (edge_xmin (a.first) > edge_xmax (b.first)) {
      return false;
    } else {

      //  complex case:
      double xa = edge_xaty (a.first, m_y);
      double xb = edge_xaty (b.first, m_y);

      if (xa != xb) {
        return xa < xb;
      } else {

        //  compare by angle of normalized edges

        db::Edge ea (a.first);
        db::Edge eb (b.first);

        if (ea.dy () < 0) {
          ea.swap_points ();
        }
        if (eb.dy () < 0) {
          eb.swap_points ();
        }

        return db::vprod_sign (ea, eb) < 0;

      }
    }
  }

private:
  db::Coord m_y;
};

// -------------------------------------------------------------------------------
//  PolygonGenerator implementation

struct PGPoint
{
  PGPoint () : point (), contour (0), first (false) { }
  PGPoint (const db::Point &p, size_t i, bool f) : point (p), contour (i), first (f) { }

  db::Point point;
  size_t contour;
  bool first;
};

class PGPolyContour
{
public:
  typedef std::deque <db::Point> contour_type;
  typedef contour_type::const_iterator const_iterator;
  typedef contour_type::iterator iterator;

  PGPolyContour ()
    : m_is_hole (false), m_next (-1), m_last (-1)
  {
    // ...
  }

  PGPolyContour (const PGPolyContour &d)
    : m_contour (d.m_contour), m_is_hole (d.m_is_hole), m_next (d.m_next), m_last (d.m_last)
  {
    // ...
  }

  PGPolyContour &operator= (const PGPolyContour &d)
  {
    if (this != &d) {
      m_contour = d.m_contour;
      m_is_hole = d.m_is_hole;
      m_next = d.m_next;
      m_last = d.m_last;
    }
    return *this;
  }

  const_iterator begin () const { return m_contour.begin (); }
  const_iterator end () const { return m_contour.end (); }
  iterator begin () { return m_contour.begin (); }
  iterator end () { return m_contour.end (); }
  const db::Point &front () const { return m_contour.front (); }
  db::Point &front () { return m_contour.front (); }
  const db::Point &back () const { return m_contour.back (); }
  db::Point &back () { return m_contour.back (); }
  void push_back (const db::Point &p) { m_contour.push_back (p); }
  void push_front (const db::Point &p) { m_contour.push_front (p); }
  bool empty () const { return m_contour.empty (); }
  size_t size () const { return m_contour.size (); }

  void last (long n) 
  {
    m_last = n;
  }

  long last () const
  {
    return m_last;
  }

  void next (long n) 
  {
    m_next = n;
  }

  long next () const
  {
    return m_next;
  }

  bool is_hole () const
  {
    return m_is_hole;
  }

  void is_hole (bool hole)
  {
    m_is_hole = hole;
  }

  void clear () 
  { 
    m_next = -1;
    m_last = -1;
    m_contour.clear (); 
  }

  void erase (iterator from, iterator to) 
  {
    m_contour.erase (from, to);
  }

  template <class I>
  void insert (iterator at, I from, I to) 
  {
    m_contour.insert (at, from, to);
  }

private:
  contour_type m_contour;
  bool m_is_hole;
  long m_next;
  long m_last;
};


class PGContourList 
{
public:
  PGContourList () 
    : m_free_contours (-1)
  { }

  PGPolyContour &operator[] (size_t n) 
  {
    return m_contours [n];
  }
  
  const PGPolyContour &operator[] (size_t n) const
  {
    return m_contours [n];
  }
  
  size_t size () const
  {
    return m_contours.size ();
  }

  size_t allocate ()
  {
    size_t index;

    if (m_free_contours >= 0) {
      index = m_free_contours;
      m_free_contours = m_contours [m_free_contours].next ();
      m_contours [index].next (-1);
    } else {
      index = m_contours.size ();
      m_contours.push_back (PGPolyContour ());
    }

    return index;
  }

  void join (size_t n1, size_t n2)
  {
    PGPolyContour &c1 = m_contours [n1];
    PGPolyContour &c2 = m_contours [n2];

    if (c1.next () < 0) {
      c1.next (c2.next ());
      c1.last (c2.last ());
    } else if (c2.next () >= 0) {
      m_contours [c1.last ()].next (c2.next ());
      c1.last (c2.last ());
    }

    c2.clear ();
    c2.next (m_free_contours);
    m_free_contours = n2;
  }

  void free (size_t n)
  {
    PGPolyContour &c = m_contours [n];
    c.clear ();
    c.next (m_free_contours);
    m_free_contours = n;
  }

  void free_all (size_t n)
  {
    long i = long (n);
    do {
      long ii = m_contours [i].next ();
      free (i);
      i = ii;
    } while (i >= 0);
  }

  void clear ()
  {
    m_free_contours = -1;
    m_contours.clear ();
  }

  void append (size_t what, size_t to)
  {
    m_contours [m_contours [to].next () < 0 ? to : m_contours [to].last ()].next (what);

    long l = m_contours [what].last ();
    m_contours [to].last (l < 0 ? what : l);
  }

private:
  long m_free_contours;
  std::vector <PGPolyContour> m_contours;
};


//#define DEBUG_POLYGON_GENERATOR

PolygonGenerator::PolygonGenerator (PolygonSink &psink, bool resolve_holes, bool min_coherence) 
  : EdgeSink (), 
    mp_contours (new PGContourList ()),
    m_y (std::numeric_limits<db::Coord>::min ()),
    m_open_pos (m_open.end ()), 
    mp_psink (&psink),
    mp_spsink (0),
    m_resolve_holes (resolve_holes),
    m_min_coherence (min_coherence),
    m_compress (true)
{ 
  //  .. nothing yet ..
}

PolygonGenerator::PolygonGenerator (SimplePolygonSink &spsink, bool min_coherence) 
  : EdgeSink (), 
    mp_contours (new PGContourList ()),
    m_y (std::numeric_limits<db::Coord>::min ()),
    m_open_pos (m_open.end ()), 
    mp_psink (0),
    mp_spsink (&spsink),
    m_resolve_holes (true),
    m_min_coherence (min_coherence),
    m_compress (true)
{ 
  //  .. nothing yet ..
}

PolygonGenerator::~PolygonGenerator ()
{
  delete mp_contours;
  mp_contours = 0;
}

void
PolygonGenerator::start ()
{
  if (mp_psink) {
    mp_psink->start ();
  }
  if (mp_spsink) {
    mp_spsink->start ();
  }
}

void 
PolygonGenerator::flush ()
{
#ifdef DEBUG_POLYGON_GENERATOR
  for (open_map_iterator_type i = m_open.begin (); i != m_open.end (); ++i) { 
    printf ("%d:%s%c ", i->contour, i->point.to_string().c_str(), i == m_open_pos ? '*' : ' '); 
  } 
  printf ("\n");
#endif
  tl_assert (m_open.empty ());

  mp_contours->clear ();
  m_open.clear ();

  if (mp_psink) {
    mp_psink->flush ();
  }
  if (mp_spsink) {
    mp_spsink->flush ();
  }
}

void 
PolygonGenerator::begin_scanline (db::Coord y)
{
#ifdef DEBUG_POLYGON_GENERATOR
  printf ("begin y=%d\n", y);
#endif
  m_open_pos = m_open.begin ();
  m_y = y;

#ifdef DEBUG_POLYGON_GENERATOR
  printf ("m_open=");
  for (open_map_type::const_iterator o = m_open.begin (); o != m_open.end (); ++o) {
    printf ("%d:%s ", o->contour, o->point.to_string().c_str()); 
  } 
  printf ("\n");
  printf ("contours:\n"); 
  for (size_t j = 0; j < mp_contours->size (); ++j) { 
    printf ("c%d: ", j); 
    for (size_t i = 0; i < (*mp_contours)[j].size (); ++i) { 
      printf ("%s ", ((*mp_contours)[j].begin () + i)->to_string().c_str ()); 
    } 
    printf ("\n"); 
  }
#endif
}

void 
PolygonGenerator::end_scanline (db::Coord y)
{
  join_contours (std::numeric_limits<db::Coord>::max ());
}

void 
PolygonGenerator::crossing_edge (const db::Edge &e)
{
#ifdef DEBUG_POLYGON_GENERATOR
  printf ("xing(%s)\n", e.to_string().c_str());
#endif
  join_contours (std::numeric_limits<db::Coord>::max ());
  ++m_open_pos;
}

void 
PolygonGenerator::skip_n (size_t n)
{
  join_contours (std::numeric_limits<db::Coord>::max ());
#ifdef DEBUG_POLYGON_GENERATOR
  printf ("skip(%d)\n", n);
#endif
  while (n-- > 0) {
    ++m_open_pos;
  }
}

void 
PolygonGenerator::put (const db::Edge &e)
{
#ifdef DEBUG_POLYGON_GENERATOR
  printf ("put(%s) y=%d m_open(%d)=", e.to_string().c_str(),m_y,std::distance (m_open.begin (), m_open_pos));
  for (open_map_iterator_type i = m_open.begin (); i != m_open.end (); ++i) { 
    printf ("%d:%s%c ", i->contour, i->point.to_string().c_str(), i == m_open_pos ? '*' : ' '); 
  } 
  printf ("\n");
#endif

  if (m_open_pos != m_open.end ()) {
    db::Coord x;
    if (e.p1 ().y () == m_y && e.p2 ().y () == m_y) {
      x = std::min (e.p1 ().x (), e.p2 ().x ());
    } else if (e.p1 ().y () == m_y) {
      x = e.p1 ().x ();
    } else {
      x = e.p2 ().x ();
    }
    join_contours (x);
  }

  if (m_open_pos != m_open.end () && e.p1 ().y () == m_y && m_open_pos->point == e.p1 () && (! m_min_coherence || e.dy () == 0)) {

    PGPolyContour &c = (*mp_contours) [m_open_pos->contour];
    tl_assert (c.back () == e.p1 ());
    c.push_back (e.p2 ());
    m_open_pos->point = e.p2 ();

    if (e.p2 ().y () > m_y) {
      ++m_open_pos;
    }

  } else if (m_open_pos != m_open.end () && e.p2 ().y () == m_y && m_open_pos->point == e.p2 () && (m_min_coherence || e.dy () == 0)) {

    PGPolyContour &c = (*mp_contours) [m_open_pos->contour];
    tl_assert (c.front () == e.p2 ());
    c.push_front (e.p1 ());
    m_open_pos->point = e.p1 ();

    if (e.p1 ().y () > m_y) {
      ++m_open_pos;
    }

  } else {

    bool hole = (e.dy () < 0);

    size_t poly_index = mp_contours->allocate ();

    (*mp_contours) [poly_index].is_hole (hole);
    (*mp_contours) [poly_index].push_back (e.p1 ());
    (*mp_contours) [poly_index].push_back (e.p2 ());

#ifdef DEBUG_POLYGON_GENERATOR
    printf ("create %s %d\n", hole ? "hole" : "hull", poly_index);
#endif
    m_open.insert (m_open_pos, PGPoint (hole ? e.p1 () : e.p2 (), poly_index, true));
    m_open.insert (m_open_pos, PGPoint (hole ? e.p2 () : e.p1 (), poly_index, false));

    --m_open_pos;

  }
#ifdef DEBUG_POLYGON_GENERATOR
  for (open_map_iterator_type i = m_open.begin (); i != m_open.end (); ++i) { 
    printf ("%d:%s%c ", i->contour, i->point.to_string().c_str(), i == m_open_pos ? '*' : ' '); 
  } 
  printf ("\n");
#endif

}

void 
PolygonGenerator::join_contours (db::Coord x) 
{
  while (m_open_pos != m_open.end ()) {

    open_map_iterator_type n = m_open_pos;
    ++n;
    if (n == m_open.end () || m_open_pos->point.y () != m_y || m_open_pos->point != n->point || m_open_pos->point.x () > x) {
      return;
    }
    
    open_map_iterator_type nn = n;
    open_map_iterator_type next = ++nn;

    //  don't join, except -/+ pair (for max. coherence) or +/- pair (for min coherence)
    bool minus0 = (m_open_pos->first == (*mp_contours) [m_open_pos->contour].is_hole ());
    bool minus1 = (n->first == (*mp_contours) [n->contour].is_hole ());
    if (! (minus0 == !m_min_coherence && minus1 == m_min_coherence)) {

      //  join a southward pair consisting of the next and second-next edge. 
      if (nn != m_open.end () && m_open_pos->point == nn->point) {
        next = m_open_pos;
        m_open_pos = n;
        n = nn;
        ++nn;
      } else if (m_open_pos->point.x () == x) {
        return;
      }

    }

    size_t i1 = m_open_pos->contour;
    size_t i2 = n->contour;
#ifdef DEBUG_POLYGON_GENERATOR
    printf ("join %d and %d\n", i1, i2);
    for (open_map_iterator_type i = m_open.begin (); i != m_open.end (); ++i) { 
      printf ("%d:%s%c%c%c ", i->contour, i->point.to_string().c_str(), i == m_open_pos ? '*' : ' ', i == n ? '+' : ' ', i == nn ? '#' : ' '); 
    } 
    printf ("\n");
    printf ("--> input contours:\n"); 
    for (size_t j = 0; j < mp_contours->size (); ++j) { 
      printf ("--> c%d: ", j); 
      for (size_t i = 0; i < (*mp_contours)[j].size (); ++i) { 
        printf ("%s ", ((*mp_contours)[j].begin () + i)->to_string().c_str ()); 
      } 
      printf ("\n"); 
    }
#endif

    tl_assert (i1 < mp_contours->size ());
    tl_assert (i2 < mp_contours->size ());

    PGPolyContour &c1 = (*mp_contours) [i1];
    PGPolyContour &c2 = (*mp_contours) [i2];

    if (i1 != i2) {

      tl_assert (! c2.empty ());
      tl_assert (! c1.empty ());

      if (! m_open_pos->first && ! n->first) {

        //  remove c1 from list of contours, join with c2
        if (c2.is_hole ()) {
          c2.insert (c2.end (), c1.begin () + 1, c1.end ());
        } else {
          c2.insert (c2.begin (), c1.begin (), c1.end () - 1);
        }

        mp_contours->join (i2, i1);

        open_map_iterator_type o = m_open_pos;
        do {
          --o;
        } while (o != m_open.begin () && o->contour != i1);
        tl_assert (o != m_open.begin ());
        o->contour = i2;
        o->first = false;

      } else {

        //  remove c1 from list of contours, join with c2
        if (c2.is_hole ()) {  // yes! c2 is correct!
          c1.insert (c1.end (), c2.begin () + 1, c2.end ());
        } else {
          c1.insert (c1.begin (), c2.begin (), c2.end () - 1);
        }

        mp_contours->join (i1, i2);

        open_map_iterator_type o = n;
        do {
          ++o;
        } while (o != m_open.end () && o->contour != i2);
        tl_assert (o != m_open.end ());
        o->contour = i1;

        if (m_open_pos->first && n->first) {
          o->first = true;
        }

      }

    } else {

      if (! c1.is_hole ()) {
#ifdef DEBUG_POLYGON_GENERATOR
        printf ("finish %d (hull)\n", i1);
#endif

        produce_poly (c1);

        //  remove this contour plus all associated holes from the contour list
        mp_contours->free_all (i1);

      } else if (m_resolve_holes) {

        //  join with next contour by creating a stitch line
        tl_assert (nn != m_open.end ());

        size_t inext = nn->contour;
        PGPolyContour &cnext = (*mp_contours) [inext];

        tl_assert (cnext.size () >= 2);

        //  compute intersection point with next edge
        db::Edge enext (cnext.front (), cnext.begin () [1]);
        db::Coord xnext = db::coord_traits<db::Coord>::rounded (edge_xaty (enext, m_y));
        db::Point pnext (xnext, m_y);

        tl_assert (c1.size () >= 2);
        cnext.front () = pnext;
        if ((c1.begin () + 1)->y () == m_y) {
          cnext.insert (cnext.begin (), c1.begin () + 1, c1.end ());
        } else {
          cnext.insert (cnext.begin (), c1.begin (), c1.end ());
        }
        cnext.push_front (pnext);
        if (enext.p1 () != pnext) {
          cnext.push_front (enext.p1 ());
        }

        mp_contours->free (i1);

      } else {

        //  join with next contour by inserting as a hole
        tl_assert (nn != m_open.end ());

        mp_contours->append (i1, nn->contour);

      }

    }

    m_open.erase (m_open_pos);
    m_open.erase (n);

    m_open_pos = next;
#ifdef DEBUG_POLYGON_GENERATOR
    printf ("--> output contours:\n"); 
    for (size_t j = 0; j < mp_contours->size (); ++j) { 
      printf ("--> c%d: ", j); 
      for (size_t i = 0; i < (*mp_contours)[j].size (); ++i) { 
        printf ("%s ", ((*mp_contours)[j].begin () + i)->to_string().c_str ()); 
      } 
      printf ("\n"); 
    }
#endif

  }
}

bool
PolygonGenerator::ms_compress = true;

void 
PolygonGenerator::produce_poly (const PGPolyContour &c)
{
  size_t n = 0;
  for (long inext = c.next (); inext >= 0; inext = (*mp_contours) [inext].next ()) {
    ++n;
  }

  if (mp_psink) {

    PGPolyContour::const_iterator p0 = c.begin ();
    PGPolyContour::const_iterator p1 = c.end ();
    // remove duplicate point at end
    tl_assert (p0 != p1);
    --p1;
    tl_assert (*p1 == *p0);

    if (n == 0 && m_poly.holes () == 0) {
      //  fast mode ..
      m_poly.assign_hull (p0, p1, m_compress && ms_compress /*compress*/);
    } else {

      m_poly.clear (n);
      m_poly.assign_hull (p0, p1, m_compress && ms_compress /*compress*/);

      for (long inext = c.next (); inext >= 0; inext = (*mp_contours) [inext].next ()) {

        tl_assert ((*mp_contours) [inext].is_hole ());

        PGPolyContour::const_iterator p0 = (*mp_contours) [inext].begin ();
        PGPolyContour::const_iterator p1 = (*mp_contours) [inext].end ();
        // remove duplicate point at end
        tl_assert (p0 != p1);
        --p1;
        tl_assert (*p1 == *p0);

        m_poly.insert_hole (p0, p1, m_compress && ms_compress /*compress*/);

      }

    }

    mp_psink->put (m_poly);

  } 

  if (mp_spsink) {

    tl_assert (n == 0);  //  there should not be holes since we forced resolve_holes to true ...
    m_spoly.assign_hull (c.begin (), c.end (), m_compress && ms_compress /*compress*/);

    mp_spsink->put (m_spoly);

  } 

}


struct WorkEdge
{
  WorkEdge (const db::Edge &e, EdgeProcessor::property_type p = 0) 
    : edge (e), data (0), prop (p)
  { }

  WorkEdge &operator= (const db::Edge &e) 
  {
    edge = e;
    return *this;
  }

  void add_cutpoint (const db::Point &p)
  {
    std::vector <db::Point> *cut_points = (std::vector <db::Point> *) data;
    if (! cut_points) {
      cut_points = new std::vector <db::Point> ();
      data = cut_points;
    }
    cut_points->push_back (p);
  }

  std::vector <db::Point> *detach_cutpoints ()
  {
    std::vector <db::Point> *cut_points = (std::vector <db::Point> *) data;
    data = 0;
    return cut_points;
  }

  db::Edge edge;
  void *data;
  db::EdgeProcessor::property_type prop;
};

struct EdgeYMinCompare 
{
  bool operator() (const WorkEdge &a, const WorkEdge &b) const
  {
    db::Coord ya = edge_ymin (a.edge);
    db::Coord yb = edge_ymin (b.edge);
    if (ya != yb) {
      return ya < yb;
    } else {
      return a.edge < b.edge;
    }
  }
};

struct EdgeXMinCompare 
{
  bool operator() (const WorkEdge &a, const WorkEdge &b) const
  {
    db::Coord xa = edge_xmin (a.edge);
    db::Coord xb = edge_xmin (b.edge);
    if (xa != xb) {
      return xa < xb;
    } else {
      return a.edge < b.edge;
    }
  }
};

struct EdgeXMinAtYIntervalCompare
{
  EdgeXMinAtYIntervalCompare (db::Coord y1, db::Coord y2)
    : m_y1 (y1), m_y2 (y2)
  {
    // .. nothing yet ..
  }

  bool operator() (const WorkEdge &a, const WorkEdge &b) const
  {
    if (edge_xmax (a.edge) < edge_xmin (b.edge)) {
      return true;
    } else if (edge_xmin (a.edge) >= edge_xmax (b.edge)) {
      return false;
    } else {
      db::Coord xa = edge_xmin_at_yinterval (a.edge, m_y1, m_y2);
      db::Coord xb = edge_xmin_at_yinterval (b.edge, m_y1, m_y2);
      if (xa != xb) {
        return xa < xb;
      } else {
        return a.edge < b.edge;
      }
    }
  }

public:
  db::Coord m_y1, m_y2;
};

struct EdgeXAtYCompare
{
  EdgeXAtYCompare (db::Coord y)
    : m_y (y) { }

  bool operator() (const WorkEdge &a, const WorkEdge &b) const
  {
    //  simple cases ..
    if (a.edge.dx () == 0 && b.edge.dx () == 0) {
      return a.edge.p1 ().x () < b.edge.p1 ().x ();
    } else if (edge_xmax (a.edge) < edge_xmin (b.edge)) {
      return true;
    } else if (edge_xmin (a.edge) > edge_xmax (b.edge)) {
      return false;
    } else {

      //  complex case:
      //  HINT: "volatile" forces xa and xb into memory and disables FPU register optimisation.
      //  That way, we can exactly compare doubles afterwards.
      volatile double xa = edge_xaty (a.edge, m_y);
      volatile double xb = edge_xaty (b.edge, m_y);

      if (xa != xb) {
        return xa < xb;
      } else {

        //  compare by angle of normalized edges

        db::Edge ea (a.edge);
        db::Edge eb (b.edge);

        if (ea.dy () < 0) {
          ea.swap_points ();
        }
        if (eb.dy () < 0) {
          eb.swap_points ();
        }

        return db::vprod_sign (ea, eb) < 0;

      }
    }
  }

  bool equal (const WorkEdge &a, const WorkEdge &b) const
  {
    //  simple cases ..
    if (a.edge.dx () == 0 && b.edge.dx () == 0) {
      return a.edge.p1 ().x () == b.edge.p1 ().x ();
    } else if (edge_xmax (a.edge) < edge_xmin (b.edge)) {
      return false;
    } else if (edge_xmin (a.edge) > edge_xmax (b.edge)) {
      return false;
    } else {

      //  complex case:
      //  HINT: "volatile" forces xa and xb into memory and disables FPU register optimisation.
      //  That way, we can exactly compare doubles afterwards.
      volatile double xa = edge_xaty (a.edge, m_y);
      volatile double xb = edge_xaty (b.edge, m_y);

      if (xa != xb) {
        return false;
      } else {

        //  compare by angle of normalized edges

        db::Edge ea (a.edge);
        db::Edge eb (b.edge);

        if (ea.dy () < 0) {
          ea.swap_points ();
        }
        if (eb.dy () < 0) {
          eb.swap_points ();
        }

        return db::vprod_sign (ea, eb) == 0;

      }
    }
  }

private:
  db::Coord m_y;
};

// -------------------------------------------------------------------------------
//  SimpleMerge implementation

SimpleMerge::SimpleMerge (int mode)
  : m_wc_n (0), m_wc_s (0), m_mode (mode)
{
  // .. nothing yet ..
}

void  
SimpleMerge::reset ()
{
  m_wc_n = m_wc_s = 0;
}

void 
SimpleMerge::reserve (size_t)
{
  // .. nothing yet ..
}

inline bool
SimpleMerge::is_true (int wc) const
{
  if (m_mode > 0) {
    return wc >= m_mode;
  } else if (m_mode < 0) {
    return wc <= m_mode || -wc <= m_mode;
  } else {
    return false;
  }
}

int 
SimpleMerge::edge (bool north, bool enter, property_type)
{
  int *wc = north ? &m_wc_n : &m_wc_s;
  bool t0 = is_true (*wc);
  if (enter) {
    ++*wc;
  } else {
    --*wc;
  }
  bool t1 = is_true (*wc);
  if (t1 && ! t0) {
    return 1;
  } else if (! t1 && t0) {
    return -1;
  } else {
    return 0;
  }
}

int 
SimpleMerge::compare_ns () const
{
  if (is_true (m_wc_s) && ! is_true (m_wc_n)) {
    return -1;
  } else if (! is_true (m_wc_s) && is_true (m_wc_n)) {
    return 1;
  } else {
    return 0;
  }
}

bool 
SimpleMerge::is_reset () const
{
  return (m_wc_n == 0 && m_wc_s == 0);
}

// -------------------------------------------------------------------------------
//  MergeOp implementation

//#define DEBUG_MERGEOP

MergeOp::MergeOp (unsigned int min_wc)
  : m_wc_n (0), m_wc_s (0), m_min_wc (min_wc), m_zeroes (0)
{
  //  .. nothing yet ..
}

void  
MergeOp::reset ()
{
  m_wcv_n.clear ();
  m_wcv_s.clear ();
  m_wc_n = 0;
  m_wc_n = 0;
  m_zeroes = 0;
}

void 
MergeOp::reserve (size_t n)
{
  m_wcv_n.clear ();
  m_wcv_s.clear ();
  m_wcv_n.resize (n, 0);
  m_wcv_s.resize (n, 0);
  m_zeroes = 2 * n;
}

static inline 
bool result_by_mode (int wc, unsigned int min_wc)
{
  return wc > int (min_wc);
}

int 
MergeOp::edge (bool north, bool enter, property_type p)
{
  tl_assert (p < m_wcv_n.size () && p < m_wcv_s.size ());

  int *wcv = north ? &m_wcv_n [p] : &m_wcv_s [p];
  int *wc = north ? &m_wc_n : &m_wc_s;

  bool inside_before = (*wcv != 0);
  *wcv += (enter ? 1 : -1);
  bool inside_after = (*wcv != 0);
  m_zeroes += (!inside_after) - (!inside_before);
#ifdef DEBUG_MERGEOP
  printf ("north=%d, enter=%d, prop=%d -> %d\n", north, enter, p, m_zeroes);
#endif
  tl_assert (long (m_zeroes) >= 0);

  bool res_before = result_by_mode (*wc, m_min_wc);
  if (inside_before != inside_after) {
    *wc += (inside_after - inside_before);
  }
  bool res_after = result_by_mode (*wc, m_min_wc);

  return res_after - res_before;
}

int 
MergeOp::compare_ns () const
{
  return result_by_mode (m_wc_n, m_min_wc) - result_by_mode (m_wc_s, m_min_wc);
}

// -------------------------------------------------------------------------------
//  BooleanOp implementation

//#define DEBUG_BOOLEAN

BooleanOp::BooleanOp (BoolOp mode)
  : m_wc_na (0), m_wc_nb (0), m_wc_sa (0), m_wc_sb (0), m_mode (mode), m_zeroes (0)
{
  //  .. nothing yet ..
}

void  
BooleanOp::reset ()
{
  m_wcv_n.clear ();
  m_wcv_s.clear ();
  m_wc_na = m_wc_sa = 0;
  m_wc_nb = m_wc_sb = 0;
  m_zeroes = 0;
}

void 
BooleanOp::reserve (size_t n)
{
  m_wcv_n.clear ();
  m_wcv_s.clear ();
  m_wcv_n.resize (n, 0);
  m_wcv_s.resize (n, 0);
  m_zeroes = 2 * n;
}

static inline 
bool result_by_mode (int wca, int wcb, db::BooleanOp::BoolOp mode)
{
  switch (mode) {
  case BooleanOp::And:
    return (wca != 0) && (wcb != 0);
  case BooleanOp::ANotB:
    return (wca != 0) && ! (wcb != 0);
  case BooleanOp::BNotA:
    return ! (wca != 0) && (wcb != 0);
  case BooleanOp::Xor:
    return ((wca != 0) && ! (wcb != 0)) || (! (wca != 0) && (wcb != 0));
  case BooleanOp::Or:
    return (wca != 0) || (wcb != 0);
  default:
    return false;
  }
}

int 
BooleanOp::edge (bool north, bool enter, property_type p)
{
  tl_assert (p < m_wcv_n.size () && p < m_wcv_s.size ());

  int *wcv = north ? &m_wcv_n [p] : &m_wcv_s [p];
  int *wca = north ? &m_wc_na : &m_wc_sa;
  int *wcb = north ? &m_wc_nb : &m_wc_sb;

  bool inside_before = (*wcv != 0);
  *wcv += (enter ? 1 : -1);
  bool inside_after = (*wcv != 0);
  m_zeroes += (!inside_after) - (!inside_before);
#ifdef DEBUG_BOOLEAN
  printf ("north=%d, enter=%d, prop=%d -> %d\n", north, enter, p, m_zeroes);
#endif
  tl_assert (long (m_zeroes) >= 0);

  bool res_before = result_by_mode (*wca, *wcb, m_mode);
  if (inside_before != inside_after) {
    if ((p % 2) == 0) {
      *wca += (inside_after - inside_before);
    } else {
      *wcb += (inside_after - inside_before);
    }
  }
  bool res_after = result_by_mode (*wca, *wcb, m_mode);

  return res_after - res_before;
}

int 
BooleanOp::compare_ns () const
{
  return result_by_mode (m_wc_na, m_wc_nb, m_mode) - result_by_mode (m_wc_sa, m_wc_sb, m_mode);
}

// -------------------------------------------------------------------------------
//  EdgeProcessor implementation

//#define DEBUG_EDGE_PROCESSOR

EdgeProcessor::EdgeProcessor ()
{
  mp_work_edges = new std::vector <WorkEdge> ();
}

EdgeProcessor::~EdgeProcessor ()
{
  if (mp_work_edges) {
    delete mp_work_edges;
    mp_work_edges = 0;
  }
}

void 
EdgeProcessor::reserve (size_t n)
{
  mp_work_edges->reserve (n);
}

void 
EdgeProcessor::insert (const db::Edge &e, EdgeProcessor::property_type p)
{
  if (e.p1 () != e.p2 ()) {
    mp_work_edges->push_back (WorkEdge (e, p));
  }
}

void 
EdgeProcessor::insert (const db::Polygon &q, EdgeProcessor::property_type p)
{
  for (db::Polygon::polygon_edge_iterator e = q.begin_edge (); ! e.at_end (); ++e) {
    insert (*e, p);
  }
}

void 
EdgeProcessor::clear ()
{
  mp_work_edges->clear ();
}

void 
EdgeProcessor::process (db::EdgeSink &es, EdgeEvaluatorBase &op)
{
  tl::SelfTimer timer (tl::verbosity () >= 21, "EdgeProcessor: process");

  //  step 1: preparation

  if (mp_work_edges->empty ()) {
    es.start ();
    es.flush ();
    return;
  }

  property_type n_props = 0;
  for (std::vector <WorkEdge>::iterator e = mp_work_edges->begin (); e != mp_work_edges->end (); ++e) {
    if (e->prop > n_props) {
      n_props = e->prop;
    }
  }
  ++n_props;

  bool snapped = true;
  while (snapped) {

    tl::SelfTimer timer1 (tl::verbosity () >= 41, "EdgeProcessor: iteration");

    //  step 2: find intersections
    std::sort (mp_work_edges->begin (), mp_work_edges->end (), EdgeYMinCompare ());

    db::Coord y = edge_ymin ((*mp_work_edges) [0].edge);
    std::vector <WorkEdge>::iterator future = mp_work_edges->begin ();
    for (std::vector <WorkEdge>::iterator current = mp_work_edges->begin (); current != mp_work_edges->end (); ) {

      size_t n = std::distance (current, future);
      db::Coord yy = y;

      //  use as many scanlines as to fetch approx. 50% new edges into the scanline (this
      //  is an empirically determined factor)
      do {

        while (future != mp_work_edges->end () && edge_ymin (future->edge) <= yy) {
          ++future;
        }

        if (future != mp_work_edges->end ()) {
          yy = edge_ymin (future->edge);
        } else {
          yy = std::numeric_limits <db::Coord>::max ();
        }

      } while (future != mp_work_edges->end () && std::distance (current, future) < long (n + n / 2));

      if (current != future) {

        std::sort (current, future, EdgeXMinAtYIntervalCompare (y, yy));

#ifdef DEBUG_EDGE_PROCESSOR
        printf ("y=%d..%d\n", y, yy);
        printf ("edges:"); 
        for (std::vector <std::pair <db::Edge, void *> >::iterator c1 = current; c1 != future; ++c1) { 
          printf (" %s", c1->edge.to_string().c_str ()); 
        } 
        printf ("\n");
#endif
        db::Coord x = edge_xmin_at_yinterval (current->edge, y, yy);

        std::vector <WorkEdge>::iterator f = current;
        for (std::vector <WorkEdge>::iterator c = current; c != future; ) {

          while (f != future && edge_xmin_at_yinterval (f->edge, y, yy) <= x) {
            ++f;
          }

          db::Coord xx = std::numeric_limits <db::Coord>::max ();
          if (f != future) {
            xx = edge_xmin_at_yinterval (f->edge, y, yy);
          }

#ifdef DEBUG_EDGE_PROCESSOR
          printf ("edges %d..%d:", x, xx); 
          for (std::vector <std::pair <db::Edge, void *> >::iterator c1 = c; c1 != f; ++c1) { o
            printf (" %s", c1->edge.to_string().c_str ()); o
          } 
          printf ("\n");
#endif
          if (std::distance (c, f) > 1) {

            db::Box cell (x, y, xx, yy);

            for (std::vector <WorkEdge>::iterator c1 = c; c1 != f - 1; ++c1) {
              for (std::vector <WorkEdge>::iterator c2 = c1 + 1; c2 != f; ++c2) {

                if (c1->edge.parallel (c2->edge) && c1->edge.side_of (c2->edge.p1 ()) == 0) {

                  //  both edges are coincident - produce the ends of the edges involved as cut points
                  if (cell.contains (c2->edge.p1 ()) && c1->edge.contains (c2->edge.p1 ()) && c1->edge.p1 () != c2->edge.p1 () && c1->edge.p2 () != c2->edge.p1 ()) {
                    c1->add_cutpoint (c2->edge.p1 ());
                  }
                  if (cell.contains (c2->edge.p2 ()) && c1->edge.contains (c2->edge.p2 ()) && c1->edge.p1 () != c2->edge.p2 () && c1->edge.p2 () != c2->edge.p2 ()) {
                    c1->add_cutpoint (c2->edge.p2 ());
                  }
                  if (cell.contains (c1->edge.p1 ()) && c2->edge.contains (c1->edge.p1 ()) && c2->edge.p1 () != c1->edge.p1 () && c2->edge.p2 () != c1->edge.p1 ()) {
                    c2->add_cutpoint (c1->edge.p1 ());
                  }
                  if (cell.contains (c1->edge.p2 ()) && c2->edge.contains (c1->edge.p2 ()) && c2->edge.p1 () != c1->edge.p2 () && c2->edge.p2 () != c1->edge.p2 ()) {
                    c2->add_cutpoint (c1->edge.p2 ());
                  }

                } else if (c1->edge.p1 () != c2->edge.p1 () && c1->edge.p2 () != c2->edge.p1 () &&
                           c1->edge.p1 () != c2->edge.p2 () && c1->edge.p2 () != c2->edge.p2 ()) {

                  std::pair <bool, db::Point> cp = c1->edge.intersect_point (c2->edge);
                  if (cp.first && cell.contains (cp.second)) {

                    //  add a cut point to c1 and c2
                    c1->add_cutpoint (cp.second);
                    c2->add_cutpoint (cp.second);

                  }

                }

              }
            }

          }

          x = xx;
          for (std::vector <WorkEdge>::iterator cc = c; cc != f; ++cc) {
            if (edge_xmax (cc->edge) < x || edge_xmax_at_yinterval (cc->edge, y, yy) < x) {
              if (c != cc) {
                std::swap (*cc, *c);
              }
              ++c;
            }
          }

        }

      }

      y = yy;
      for (std::vector <WorkEdge>::iterator c = current; c != future; ++c) {
        if (edge_ymax (c->edge) <= y) {
          if (current != c) {
            std::swap (*current, *c);
          }
          ++current;
        }
      }
      
    }

    //  step 3: create new edges from the ones with cutpoints
    
    snapped = false;
    
    size_t n_work = mp_work_edges->size ();
    size_t nw = 0;
    for (size_t n = 0; n < n_work; ++n) {

      WorkEdge &ew = (*mp_work_edges) [n];

      std::vector <db::Point> *cut_points = ew.detach_cutpoints ();
      if (cut_points) {

        db::Edge e = ew.edge;
        property_type p = ew.prop;
        std::sort (cut_points->begin (), cut_points->end (), ProjectionCompare (e));

        //  check, if any rounding occured. In this case, we need another iteration
        for (std::vector <db::Point>::iterator cp = cut_points->begin (); cp != cut_points->end () && ! snapped; ++cp) {
          if (db::vprod_sign (e, db::Edge (e.p1 (), *cp)) != 0) {
            snapped = true;
          }
        }

        if (e.p1 () != cut_points->front ()) {
          if (nw <= n) {
            (*mp_work_edges) [nw++] = WorkEdge (db::Edge (e.p1 (), cut_points->front ()), p);
          } else {
            mp_work_edges->push_back (WorkEdge (db::Edge (e.p1 (), cut_points->front ()), p));
          }
        }
        for (std::vector <db::Point>::iterator cp = cut_points->begin () + 1; cp != cut_points->end (); ++cp) {
          if (cp [-1] != cp [0]) {
            if (nw <= n) {
              (*mp_work_edges) [nw++] = WorkEdge (db::Edge (cp [-1], cp [0]), p);
            } else {
              mp_work_edges->push_back (WorkEdge (db::Edge (cp [-1], cp [0]), p));
            }
          }
        }
        if (cut_points->back () != e.p2 ()) {
          if (nw <= n) {
            (*mp_work_edges) [nw++] = WorkEdge (db::Edge (cut_points->back (), e.p2 ()), p);
          } else {
            mp_work_edges->push_back (WorkEdge (db::Edge (cut_points->back (), e.p2 ()), p));
          }
        }

        delete cut_points;

      } else {
        if (nw < n) {
          (*mp_work_edges) [nw] = (*mp_work_edges) [n];
        }
        ++nw;
      }

    }

    if (nw != n_work) {
      mp_work_edges->erase (mp_work_edges->begin () + nw, mp_work_edges->begin () + n_work);
    }

    // HINT: to test single iteration mode: 
    //   snapped = false;
  }

  tl::SelfTimer timer2 (tl::verbosity () >= 41, "EdgeProcessor: production");

  //  step 4: remove all horizontal edges

  std::vector <WorkEdge>::iterator cw = mp_work_edges->begin (); 
  for (std::vector <WorkEdge>::iterator current = mp_work_edges->begin (); current != mp_work_edges->end (); ++current) {
    if (current->edge.dy () != 0) {
      if (current != cw) {
        *cw = *current;
      }
      ++cw;
    }
  }
  
  mp_work_edges->erase (cw, mp_work_edges->end ());

  //  step 5: compute the result edges 
  
  es.start (); // call this as late as possible. This way, input containers can be identical with output containers ("clear" is done after the input is read)

  op.reset ();
  op.reserve (n_props);

  std::sort (mp_work_edges->begin (), mp_work_edges->end (), EdgeYMinCompare ());

  db::Coord y = edge_ymin ((*mp_work_edges) [0].edge);
  size_t skip_unit = 1;

  std::vector <WorkEdge>::iterator future = mp_work_edges->begin ();
  for (std::vector <WorkEdge>::iterator current = mp_work_edges->begin (); current != mp_work_edges->end (); ) {

    std::vector <WorkEdge>::iterator f0 = future;
    while (future != mp_work_edges->end () && edge_ymin (future->edge) <= y) {
      tl_assert (future->data == 0); // HINT: for development
      ++future;
    }
    std::sort (f0, future, EdgeXAtYCompare (y));

    db::Coord yy = std::numeric_limits <db::Coord>::max ();
    if (future != mp_work_edges->end ()) {
      yy = edge_ymin (future->edge);
    }
    for (std::vector <WorkEdge>::const_iterator c = current; c != future; ++c) {
      if (edge_ymax (c->edge) > y) {
        yy = std::min (yy, edge_ymax (c->edge));
      }
    }

    db::Coord ysl = y;
    es.begin_scanline (y);

    tl_assert (op.is_reset ()); // HINT: for development

    if (current != future) {

      std::inplace_merge (current, f0, future, EdgeXAtYCompare (y));
#ifdef DEBUG_EDGE_PROCESSOR
      printf ("y=%d ", y);
      for (std::vector <WorkEdge>::iterator c = current; c != future; ++c) { 
        printf ("%d-", long (c->data)); 
      } 
      printf ("\n");
#endif

      db::Coord hx = 0;
      int ho = 0;

      size_t new_skip_unit = std::distance (current, future);

      for (std::vector <WorkEdge>::iterator c = current; c != future; ) {

        size_t skip = size_t (c->data) % skip_unit;
        size_t skip_res = size_t (c->data) / skip_unit;
#ifdef DEBUG_EDGE_PROCESSOR
        printf ("X %d->%d,%d\n", int(c->data), skip, skip_res);
#endif

        if (skip != 0 && (c + skip >= future || (c + skip)->data != 0)) {

          tl_assert (c + skip <= future);

          es.skip_n (skip_res);

          c->data = (void *) (skip + new_skip_unit * skip_res);

          //  skip this interval - has not changed
          c += skip;

        } else {

          std::vector <WorkEdge>::iterator c0 = c;
          size_t n_res = 0;

          do {

            c->data = 0;
            std::vector <WorkEdge>::iterator f = c + 1;

            //  HINT: "volatile" forces x and xx into memory and disables FPU register optimisation.
            //  That way, we can exactly compare doubles afterwards.
            volatile double x = edge_xaty (c->edge, y);

            while (f != future) {
              volatile double xx = edge_xaty (f->edge, y);
              if (xx != x) {
                break;
              }
              f->data = 0;
              ++f;
            }

            //  compute edges that occure at this vertex
            
            bool vertex = false;
            
            for (std::vector <WorkEdge>::iterator cc = c; cc != f; ) {

              std::vector <WorkEdge>::iterator e = mp_work_edges->end ();

              int pn = 0, ps = 0;

              do {

                if (e == mp_work_edges->end () && edge_ymax (cc->edge) > y) {
                  e = cc;
                }

                if (edge_ymax (cc->edge) > y) {
                  pn += op.edge (true, cc->edge.dy () > 0, cc->prop);
                }
                if (edge_ymin (cc->edge) < y) {
                  ps += op.edge (false, cc->edge.dy () > 0, cc->prop);
                }

                ++cc;

              } while (cc != f && EdgeXAtYCompare (y).equal (cc [-1], *cc));

              if (! vertex && (ps != 0 || pn != 0)) {

                if (ho != 0) {
                  db::Edge he (db::Point (hx, y), db::Point (db::coord_traits<db::Coord>::rounded (x), y));
                  if (ho > 0) {
                    he.swap_points ();
                  }
                  es.put (he);
#ifdef DEBUG_EDGE_PROCESSOR
                  printf ("put(%s)\n", he.to_string().c_str());
#endif
                } 

                vertex = true;

              }

              if (e != mp_work_edges->end ()) {

                db::Edge edge (e->edge);

                if ((pn > 0 && edge.dy () < 0) || (pn < 0 && edge.dy () > 0)) {
                  edge.swap_points ();
                }

                if (pn != 0) {
                  ++n_res;
                  if (edge_ymin (edge) == y) {
                    es.put (edge);
#ifdef DEBUG_EDGE_PROCESSOR
                    printf ("put(%s)\n", edge.to_string().c_str());
#endif
                  } else {
                    es.crossing_edge (edge);
#ifdef DEBUG_EDGE_PROCESSOR
                    printf ("xing(%s)\n", edge.to_string().c_str());
#endif
                  }
                }

              }

            }

            if (vertex) {
              hx = db::coord_traits<db::Coord>::rounded (x);
              ho = op.compare_ns ();
            }

            c = f;

          } while (c != future && ! op.is_reset ());

          //  TODO: assert that there is no overflow here:
          c0->data = (void *) (std::distance (c0, c) + new_skip_unit * n_res);

        }

      }

      skip_unit = new_skip_unit;

      y = yy;

#ifdef DEBUG_EDGE_PROCESSOR
      for (std::vector <WorkEdge>::iterator c = current; c != future; ++c) { 
        printf ("%d-", long (c->data)); 
      } 
      printf ("\n");
#endif
      std::vector <WorkEdge>::iterator c0 = current;
      std::vector <WorkEdge>::iterator last_interval = future;
      current = future;

      bool valid = true;

      for (std::vector <WorkEdge>::iterator c = future; c != c0; ) {

        --c;

        bool start_interval = (c->data != 0);
        void *data = c->data;
        c->data = 0;

        db::Coord ymax = edge_ymax (c->edge);
        if (ymax >= y) {
          --current;
          if (current != c) {
            *current = *c;
          }
        }
        if (ymax <= y) {
          //  an edge ends now. The interval is not valid, i.e. cannot be skipped easily.
          valid = false;
        }

        if (start_interval && current != future) {
          current->data = valid ? data : 0;
          last_interval = current;
          valid = true;
        }

      }
#ifdef DEBUG_EDGE_PROCESSOR
      for (std::vector <WorkEdge>::iterator c = current; c != future; ++c) { 
        printf ("%d-", long (c->data)); 
      } 
      printf ("\n");
#endif
    
    }

    tl_assert (op.is_reset ()); // HINT: for development (second)

    es.end_scanline (ysl);

  }

  es.flush ();

}

void
EdgeProcessor::simple_merge (const std::vector<db::Edge> &in, std::vector <db::Edge> &edges)
{
  clear ();
  reserve (in.size ());
  insert_sequence (in.begin (), in.end ());

  db::SimpleMerge op;
  db::EdgeContainer out (edges);
  process (out, op);
}

void
EdgeProcessor::simple_merge (const std::vector<db::Edge> &in, std::vector <db::Polygon> &polygons, bool resolve_holes, bool min_coherence)
{
  clear ();
  reserve (in.size ());
  insert_sequence (in.begin (), in.end ());

  db::SimpleMerge op;
  db::PolygonContainer pc (polygons);
  db::PolygonGenerator out (pc, resolve_holes, min_coherence);
  process (out, op);
}

void
EdgeProcessor::simple_merge (const std::vector<db::Polygon> &in, std::vector <db::Edge> &edges)
{
  clear ();
  reserve (count_edges (in));
  for (std::vector<db::Polygon>::const_iterator q = in.begin (); q != in.end (); ++q) {
    insert (*q);
  }

  db::SimpleMerge op;
  db::EdgeContainer out (edges);
  process (out, op);
}

void
EdgeProcessor::simple_merge (const std::vector<db::Polygon> &in, std::vector <db::Polygon> &polygons, bool resolve_holes, bool min_coherence)
{
  clear ();
  reserve (count_edges (in));
  for (std::vector<db::Polygon>::const_iterator q = in.begin (); q != in.end (); ++q) {
    insert (*q);
  }

  db::SimpleMerge op;
  db::PolygonContainer pc (polygons);
  db::PolygonGenerator out (pc, resolve_holes, min_coherence);
  process (out, op);
}

void
EdgeProcessor::merge (const std::vector<db::Polygon> &in, std::vector <db::Edge> &edges, unsigned int min_wc)
{
  clear ();
  reserve (count_edges (in));

  size_t n = 0;
  for (std::vector<db::Polygon>::const_iterator q = in.begin (); q != in.end (); ++q, ++n) {
    insert (*q, n);
  }

  db::MergeOp op (min_wc);
  db::EdgeContainer out (edges);
  process (out, op);
}

void
EdgeProcessor::merge (const std::vector<db::Polygon> &in, std::vector <db::Polygon> &polygons, unsigned int min_wc, bool resolve_holes, bool min_coherence)
{
  clear ();
  reserve (count_edges (in));

  size_t n = 0;
  for (std::vector<db::Polygon>::const_iterator q = in.begin (); q != in.end (); ++q, ++n) {
    insert (*q, n);
  }

  db::MergeOp op (min_wc);
  db::PolygonContainer pc (polygons);
  db::PolygonGenerator out (pc, resolve_holes, min_coherence);
  process (out, op);
}

void
EdgeProcessor::size (const std::vector<db::Polygon> &in, db::Coord dx, db::Coord dy, std::vector <db::Edge> &out, unsigned int mode)
{
  clear ();
  reserve (count_edges (in));

  size_t n = 0;
  for (std::vector<db::Polygon>::const_iterator q = in.begin (); q != in.end (); ++q, n += 2) {
    insert (*q, n);
  }

  std::vector <db::Polygon> tmp; // TODO: should be local to processor to save reallocation time?

  //  produce polygons - this way the contours are formed and sizing can be applied
  db::BooleanOp op (db::BooleanOp::Or);
  db::PolygonContainer pc (tmp);
  db::PolygonGenerator pg (pc, false /*don't resolve holes*/, false /*min. coherence*/);
  process (pg, op);

  //  2nd step: size and merge everything into result vector
  for (std::vector <db::Polygon>::iterator p = tmp.begin (); p != tmp.end (); ++p) {

    clear ();

    p->size (dx, dy, mode);
    insert (*p, 0);
    p->clear ();

    //  produce polygons
    db::SimpleMerge op (1 /*wc>0*/);
    db::EdgeContainer pg (out);
    process (pg, op);

  }
}

void
EdgeProcessor::size (const std::vector<db::Polygon> &in, db::Coord dx, db::Coord dy, std::vector <db::Polygon> &out, unsigned int mode, bool resolve_holes, bool min_coherence)
{
  clear ();
  reserve (count_edges (in));

  size_t n = 0;
  for (std::vector<db::Polygon>::const_iterator q = in.begin (); q != in.end (); ++q, n += 2) {
    insert (*q, n);
  }

  std::vector <db::Polygon> tmp; // TODO: should be local to processor to save reallocation time?

  //  produce polygons - this way the contours are formed and sizing can be applied
  db::BooleanOp op (db::BooleanOp::Or);
  db::PolygonContainer pc (tmp);
  db::PolygonGenerator pg (pc, false /*don't resolve holes*/, min_coherence);
  process (pg, op);

  //  2nd step: size and merge everything into result vector
  for (std::vector <db::Polygon>::iterator p = tmp.begin (); p != tmp.end (); ++p) {

    clear ();

    p->size (dx, dy, mode);
    insert (*p, 0);
    p->clear ();

    //  produce polygons
    db::SimpleMerge op (1 /*wc>0*/);
    db::PolygonContainer pc (out);
    db::PolygonGenerator pg (pc, resolve_holes, min_coherence);
    process (pg, op);

  }
}

void 
EdgeProcessor::boolean (const std::vector<db::Polygon> &a, const std::vector<db::Polygon> &b, std::vector <db::Edge> &out, int mode)
{
  clear ();
  reserve (count_edges (a) + count_edges (b));

  size_t n;
  
  n = 0;
  for (std::vector<db::Polygon>::const_iterator q = a.begin (); q != a.end (); ++q, n += 2) {
    insert (*q, n);
  }

  n = 1;
  for (std::vector<db::Polygon>::const_iterator q = b.begin (); q != b.end (); ++q, n += 2) {
    insert (*q, n);
  }

  db::BooleanOp op ((db::BooleanOp::BoolOp) mode);
  db::EdgeContainer ec (out);
  process (ec, op);
}

void 
EdgeProcessor::boolean (const std::vector<db::Polygon> &a, const std::vector<db::Polygon> &b, std::vector <db::Polygon> &out, int mode, bool resolve_holes, bool min_coherence)
{
  clear ();
  reserve (count_edges (a) + count_edges (b));

  size_t n;
  
  n = 0;
  for (std::vector<db::Polygon>::const_iterator q = a.begin (); q != a.end (); ++q, n += 2) {
    insert (*q, n);
  }

  n = 1;
  for (std::vector<db::Polygon>::const_iterator q = b.begin (); q != b.end (); ++q, n += 2) {
    insert (*q, n);
  }

  db::BooleanOp op ((db::BooleanOp::BoolOp) mode);
  db::PolygonContainer pc (out);
  db::PolygonGenerator pg (pc, resolve_holes, min_coherence);
  process (pg, op);
}

void 
EdgeProcessor::boolean (const std::vector<db::Edge> &a, const std::vector<db::Edge> &b, std::vector <db::Edge> &out, int mode)
{
  clear ();
  reserve (a.size () + b.size ());

  insert_sequence (a.begin (), a.end (), 0);
  insert_sequence (b.begin (), b.end (), 1);

  db::BooleanOp op ((db::BooleanOp::BoolOp) mode);
  db::EdgeContainer ec (out);
  process (ec, op);
}

void 
EdgeProcessor::boolean (const std::vector<db::Edge> &a, const std::vector<db::Edge> &b, std::vector <db::Polygon> &out, int mode, bool resolve_holes, bool min_coherence)
{
  clear ();
  reserve (a.size () + b.size ());

  insert_sequence (a.begin (), a.end (), 0);
  insert_sequence (b.begin (), b.end (), 1);

  db::BooleanOp op ((db::BooleanOp::BoolOp) mode);
  db::PolygonContainer pc (out);
  db::PolygonGenerator pg (pc, resolve_holes, min_coherence);
  process (pg, op);
}

// -------------------------------------------------------------------------------
//  ShapeProcessor implementation

ShapeProcessor::ShapeProcessor ()
{
  // .. nothing yet ..
}

void 
ShapeProcessor::clear ()
{
  m_processor.clear ();
}

void 
ShapeProcessor::reserve (size_t n)
{
  m_processor.reserve (n);
}

void 
ShapeProcessor::process (db::EdgeSink &es, EdgeEvaluatorBase &op)
{
  m_processor.process (es, op);
}

size_t
ShapeProcessor::count_edges (const db::Shape &shape)
{
  size_t n = 0;

  if (shape.is_polygon ()) {

    //  KLUDGE: Shape should have a point_count property (or similar)
    for (db::Shape::polygon_edge_iterator e = shape.begin_edge (); ! e.at_end (); ++e) {
      ++n;
    }

  } else if (shape.is_path ()) {

    //  KLUDGE: This is rather inefficient ..
    db::Polygon poly;
    shape.polygon (poly);
    for (db::Polygon::polygon_edge_iterator e = poly.begin_edge (); ! e.at_end (); ++e) {
      ++n;
    }

  } else if (shape.is_box ()) {
    n += 4;
  }

  return n;
}

void
ShapeProcessor::merge (const std::vector<db::Shape> &in, const std::vector<db::CplxTrans> &trans, std::vector <db::Edge> &out_edges, unsigned int min_wc)
{
  clear ();

  size_t e = 0;
  for (std::vector<db::Shape>::const_iterator i = in.begin (); i != in.end (); ++i) {
    e += count_edges (*i);
  }
  reserve (e + e / 4); // heuristic reserve for crossing points

  size_t n = 0;
  for (std::vector<db::Shape>::const_iterator i = in.begin (); i != in.end (); ++i, ++n) {
    if (n < trans.size ()) {
      insert (*i, db::ICplxTrans (trans [n]), n);
    } else {
      insert (*i, n);
    }
  }

  db::MergeOp op (min_wc);
  db::EdgeContainer out (out_edges);
  process (out, op);
}

void
ShapeProcessor::merge (const std::vector<db::Shape> &in, const std::vector<db::CplxTrans> &trans, std::vector <db::Polygon> &out_polygons, unsigned int min_wc, bool resolve_holes, bool min_coherence)
{
  clear ();

  size_t e = 0;
  for (std::vector<db::Shape>::const_iterator i = in.begin (); i != in.end (); ++i) {
    e += count_edges (*i);
  }
  reserve (e + e / 4); // heuristic reserve for crossing points

  size_t n = 0;
  for (std::vector<db::Shape>::const_iterator i = in.begin (); i != in.end (); ++i, ++n) {
    if (n < trans.size ()) {
      insert (*i, db::ICplxTrans (trans [n]), n);
    } else {
      insert (*i, n);
    }
  }

  db::MergeOp op (min_wc);
  db::PolygonContainer pc (out_polygons);
  db::PolygonGenerator out (pc, resolve_holes, min_coherence);
  process (out, op);
}

void
ShapeProcessor::boolean (const std::vector<db::Shape> &in_a, const std::vector<db::CplxTrans> &trans_a,
                         const std::vector<db::Shape> &in_b, const std::vector<db::CplxTrans> &trans_b,
                         int mode, std::vector <db::Edge> &out_edges)
{
  clear ();

  size_t e = 0;
  for (std::vector<db::Shape>::const_iterator i = in_a.begin (); i != in_a.end (); ++i) {
    e += count_edges (*i);
  }
  for (std::vector<db::Shape>::const_iterator i = in_b.begin (); i != in_b.end (); ++i) {
    e += count_edges (*i);
  }
  reserve (e + e / 4); // heuristic reserve for crossing points

  size_t n;
  n = 0;
  for (std::vector<db::Shape>::const_iterator i = in_a.begin (); i != in_a.end (); ++i, ++n) {
    if (n < trans_a.size ()) {
      insert (*i, db::ICplxTrans (trans_a [n]), n * 2);
    } else {
      insert (*i, n * 2);
    }
  }
  n = 0;
  for (std::vector<db::Shape>::const_iterator i = in_b.begin (); i != in_b.end (); ++i, ++n) {
    if (n < trans_b.size ()) {
      insert (*i, db::ICplxTrans (trans_b [n]), n * 2 + 1);
    } else {
      insert (*i, n * 2 + 1);
    }
  }

  db::BooleanOp op ((db::BooleanOp::BoolOp) mode);
  db::EdgeContainer out (out_edges);
  process (out, op);
}

void
ShapeProcessor::boolean (const std::vector<db::Shape> &in_a, const std::vector<db::CplxTrans> &trans_a,
                         const std::vector<db::Shape> &in_b, const std::vector<db::CplxTrans> &trans_b,
                         int mode, std::vector <db::Polygon> &out_polygons, bool resolve_holes, bool min_coherence)
{
  clear ();

  size_t e = 0;
  for (std::vector<db::Shape>::const_iterator i = in_a.begin (); i != in_a.end (); ++i) {
    e += count_edges (*i);
  }
  for (std::vector<db::Shape>::const_iterator i = in_b.begin (); i != in_b.end (); ++i) {
    e += count_edges (*i);
  }
  reserve (e + e / 4); // heuristic reserve for crossing points

  size_t n;
  n = 0;
  for (std::vector<db::Shape>::const_iterator i = in_a.begin (); i != in_a.end (); ++i, ++n) {
    if (n < trans_a.size ()) {
      insert (*i, db::ICplxTrans (trans_a [n]), n * 2);
    } else {
      insert (*i, n * 2);
    }
  }
  n = 0;
  for (std::vector<db::Shape>::const_iterator i = in_b.begin (); i != in_b.end (); ++i, ++n) {
    if (n < trans_b.size ()) {
      insert (*i, db::ICplxTrans (trans_b [n]), n * 2 + 1);
    } else {
      insert (*i, n * 2 + 1);
    }
  }

  db::BooleanOp op ((db::BooleanOp::BoolOp) mode);
  db::PolygonContainer pc (out_polygons);
  db::PolygonGenerator out (pc, resolve_holes, min_coherence);
  process (out, op);
}

void 
ShapeProcessor::size (const std::vector<db::Shape> &in, const std::vector<db::CplxTrans> &trans,
                      db::Coord dx, db::Coord dy, std::vector <db::Polygon> &out, unsigned int mode, bool resolve_holes, bool min_coherence)
{
  //  1st step: merge input
  clear ();

  size_t e = 0;
  for (std::vector<db::Shape>::const_iterator i = in.begin (); i != in.end (); ++i) {
    e += count_edges (*i);
  }
  reserve (e + e / 4); // heuristic reserve for crossing points

  size_t n;
  n = 0;
  for (std::vector<db::Shape>::const_iterator i = in.begin (); i != in.end (); ++i, ++n) {
    if (n < trans.size ()) {
      insert (*i, db::ICplxTrans (trans [n]), n * 2);
    } else {
      insert (*i, n * 2);
    }
  }

  std::vector <db::Polygon> tmp; // TODO: should be local to processor to save reallocation time?

  //  produce polygons - this way the contours are formed and sizing can be applied
  db::BooleanOp op (db::BooleanOp::Or);
  db::PolygonContainer pc (tmp);
  db::PolygonGenerator pg (pc, false /*don't resolve holes*/, false /*max. coherence to make as many holes as possible*/);
  process (pg, op);

  //  2nd step: size and merge everything into result vector
  for (std::vector <db::Polygon>::iterator p = tmp.begin (); p != tmp.end (); ++p) {

    clear ();
  
    p->size (dx, dy, mode);
    m_processor.insert (*p, 0);
    p->clear ();

    //  produce polygons
    db::SimpleMerge op (1 /*wc>0*/);
    db::PolygonContainer pc (out);
    db::PolygonGenerator pg (pc, resolve_holes, min_coherence);
    process (pg, op);

  }
}

void 
ShapeProcessor::size (const std::vector<db::Shape> &in, const std::vector<db::CplxTrans> &trans,
                      db::Coord dx, db::Coord dy, std::vector <db::Edge> &out, unsigned int mode)
{
  //  1st step: merge input
  clear ();

  size_t e = 0;
  for (std::vector<db::Shape>::const_iterator i = in.begin (); i != in.end (); ++i) {
    e += count_edges (*i);
  }
  reserve (e + e / 4); // heuristic reserve for crossing points

  size_t n;
  n = 0;
  for (std::vector<db::Shape>::const_iterator i = in.begin (); i != in.end (); ++i, ++n) {
    if (n < trans.size ()) {
      insert (*i, db::ICplxTrans (trans [n]), n * 2);
    } else {
      insert (*i, n * 2);
    }
  }

  std::vector <db::Polygon> tmp; // TODO: should be local to processor to save reallocation time?

  //  produce polygons - this way the contours are formed and sizing can be applied
  db::BooleanOp op (db::BooleanOp::Or);
  db::PolygonContainer pc (tmp);
  db::PolygonGenerator pg (pc, false /*don't resolve holes*/, false /*max. coherence to make as many holes as possible*/);
  process (pg, op);

  //  2nd step: size and merge everything into result vector
  for (std::vector <db::Polygon>::iterator p = tmp.begin (); p != tmp.end (); ++p) {

    clear ();
  
    p->size (dx, dy, mode);
    m_processor.insert (*p, 0);
    p->clear ();

    //  produce polygons
    db::SimpleMerge op (1 /*wc>0*/);
    db::EdgeContainer ec (out);
    process (ec, op);

  }
}

//  TODO: these method should be part of db::Cell
void 
ShapeProcessor::collect_shapes_hier (const db::CplxTrans &tr, const db::Layout &layout, const db::Cell &cell, unsigned int layer, std::vector<db::Shape> &shapes, std::vector<db::CplxTrans> &trans) const
{
  for (db::Shapes::shape_iterator s = cell.shapes (layer).begin (db::ShapeIterator::All); ! s.at_end (); ++s) {
    shapes.push_back (*s);
    trans.push_back (tr);
  }

  for (db::Cell::const_iterator i = cell.begin (); ! i.at_end (); ++i) {
    for (db::CellInstArray::iterator a = i->begin (); ! a.at_end (); ++a) {
      collect_shapes_hier (tr * i->complex_trans (*a), layout, layout.cell (i->cell_index ()), layer, shapes, trans);
    }
  }
}

//  TODO: these method should be part of db::Cell
size_t 
ShapeProcessor::count_shapes_hier (const db::Layout &layout, const db::Cell &cell, unsigned int layer, std::map<db::cell_index_type, size_t> &cache) const
{
  std::map<db::cell_index_type, size_t>::iterator c = cache.find (cell.cell_index ());

  if (c != cache.end ()) {
    return c->second;
  } else {

    size_t n = cell.shapes (layer).size ();
    for (db::Cell::const_iterator i = cell.begin (); ! i.at_end (); ++i) {
      n += count_shapes_hier (layout, layout.cell (i->cell_index ()), layer, cache) * i->size ();
    }

    cache.insert (std::make_pair (cell.cell_index (), n));
    return n;

  }
}

void 
ShapeProcessor::boolean (const db::Layout &layout_in_a, const db::Cell &cell_in_a, unsigned int layer_in_a, 
                         const db::Layout &layout_in_b, const db::Cell &cell_in_b, unsigned int layer_in_b, 
                         db::Shapes &shapes, int mode, bool with_sub_hierarchy, bool resolve_holes, bool min_coherence)
{
  std::vector<db::Shape> in_a, in_b;
  std::vector<db::CplxTrans> trans_a, trans_b;

  if (with_sub_hierarchy) {
 
    std::map<db::cell_index_type, size_t> ca, cb;
    size_t na = count_shapes_hier (layout_in_a, cell_in_a, layer_in_a, ca);
    size_t nb = count_shapes_hier (layout_in_b, cell_in_b, layer_in_b, cb);

    in_a.reserve (na); 
    in_b.reserve (nb);

    trans_a.reserve (na); 
    trans_b.reserve (nb);

    collect_shapes_hier (db::CplxTrans (), layout_in_a, cell_in_a, layer_in_a, in_a, trans_a);
    collect_shapes_hier (db::CplxTrans (), layout_in_b, cell_in_b, layer_in_b, in_b, trans_b);

  } else {

    size_t na = cell_in_a.shapes (layer_in_a).size ();
    size_t nb = cell_in_b.shapes (layer_in_b).size ();

    in_a.reserve (na); 
    in_b.reserve (nb);

    for (db::Shapes::shape_iterator s = cell_in_a.shapes (layer_in_a).begin (db::ShapeIterator::All); ! s.at_end (); ++s) {
      in_a.push_back (*s);
    }

    for (db::Shapes::shape_iterator s = cell_in_b.shapes (layer_in_b).begin (db::ShapeIterator::All); ! s.at_end (); ++s) {
      in_b.push_back (*s);
    }

  }

  clear ();

  size_t e = 0;
  for (std::vector<db::Shape>::const_iterator i = in_a.begin (); i != in_a.end (); ++i) {
    e += count_edges (*i);
  }
  for (std::vector<db::Shape>::const_iterator i = in_b.begin (); i != in_b.end (); ++i) {
    e += count_edges (*i);
  }
  reserve (e + e / 4); // heuristic reserve for crossing points

  size_t n;
  n = 0;
  for (std::vector<db::Shape>::const_iterator i = in_a.begin (); i != in_a.end (); ++i, ++n) {
    if (n < trans_a.size ()) {
      insert (*i, db::ICplxTrans (trans_a [n]), n * 2);
    } else {
      insert (*i, n * 2);
    }
  }
  n = 0;
  for (std::vector<db::Shape>::const_iterator i = in_b.begin (); i != in_b.end (); ++i, ++n) {
    if (n < trans_b.size ()) {
      insert (*i, db::ICplxTrans (trans_b [n]), n * 2 + 1);
    } else {
      insert (*i, n * 2 + 1);
    }
  }

  db::BooleanOp op ((db::BooleanOp::BoolOp) mode);
  db::ShapeGenerator sg (shapes, true /*clear shapes*/);
  db::PolygonGenerator out (sg, resolve_holes, min_coherence);
  process (out, op);
}

void 
ShapeProcessor::size (const db::Layout &layout_in, const db::Cell &cell_in, unsigned int layer_in, 
                      db::Shapes &out, db::Coord dx, db::Coord dy, unsigned int mode, bool with_sub_hierarchy, bool resolve_holes, bool min_coherence)
{
  std::vector<db::Shape> in;
  std::vector<db::CplxTrans> trans;

  if (with_sub_hierarchy) {

    std::map<db::cell_index_type, size_t> c;
    size_t n = count_shapes_hier (layout_in, cell_in, layer_in, c);

    in.reserve (n); 
    trans.reserve (n); 

    collect_shapes_hier (db::CplxTrans (), layout_in, cell_in, layer_in, in, trans);

  } else {

    size_t n = cell_in.shapes (layer_in).size ();

    in.reserve (n); 

    for (db::Shapes::shape_iterator s = cell_in.shapes (layer_in).begin (db::ShapeIterator::All); ! s.at_end (); ++s) {
      in.push_back (*s);
    }

  }

  //  1st step: merge input
  clear ();

  size_t e = 0;
  for (std::vector<db::Shape>::const_iterator i = in.begin (); i != in.end (); ++i) {
    e += count_edges (*i);
  }
  reserve (e + e / 4); // heuristic reserve for crossing points

  size_t n;
  n = 0;
  for (std::vector<db::Shape>::const_iterator i = in.begin (); i != in.end (); ++i, ++n) {
    if (n < trans.size ()) {
      insert (*i, db::ICplxTrans (trans [n]), n * 2);
    } else {
      insert (*i, n * 2);
    }
  }

  std::vector <db::Polygon> tmp; // TODO: should be local to processor to save reallocation time?

  //  produce polygons - this way the contours are formed and sizing can be applied
  db::BooleanOp op (db::BooleanOp::Or);
  db::PolygonContainer pc (tmp);
  db::PolygonGenerator pg (pc, false /*don't resolve holes*/, false /*max. coherence to make as many holes as possible*/);
  process (pg, op);

  //  2nd step: size and merge everything into result vector
  out.clear ();

  for (std::vector <db::Polygon>::iterator p = tmp.begin (); p != tmp.end (); ++p) {

    clear ();
  
    p->size (dx, dy, mode);
    m_processor.insert (*p);
    p->clear ();

    //  produce polygons
    db::SimpleMerge op (1 /*wc>0*/);
    db::ShapeGenerator sg (out, false /*don't clear shapes (done before)*/);
    db::PolygonGenerator pg (sg, resolve_holes, min_coherence);
    process (pg, op);

  }

}

void 
ShapeProcessor::merge (const db::Layout &layout_in, const db::Cell &cell_in, unsigned int layer_in, 
                       db::Shapes &shapes, bool with_sub_hierarchy, unsigned int min_wc, bool resolve_holes, bool min_coherence)
{
  std::vector<db::Shape> in;
  std::vector<db::CplxTrans> trans;

  if (with_sub_hierarchy) {

    std::map<db::cell_index_type, size_t> c;
    size_t n = count_shapes_hier (layout_in, cell_in, layer_in, c);

    in.reserve (n); 
    trans.reserve (n); 

    collect_shapes_hier (db::CplxTrans (), layout_in, cell_in, layer_in, in, trans);

  } else {

    size_t n = cell_in.shapes (layer_in).size ();

    in.reserve (n); 

    for (db::Shapes::shape_iterator s = cell_in.shapes (layer_in).begin (db::ShapeIterator::All); ! s.at_end (); ++s) {
      in.push_back (*s);
    }

  }

  clear ();

  size_t e = 0;
  for (std::vector<db::Shape>::const_iterator i = in.begin (); i != in.end (); ++i) {
    e += count_edges (*i);
  }
  reserve (e + e / 4); // heuristic reserve for crossing points

  size_t n;
  n = 0;
  for (std::vector<db::Shape>::const_iterator i = in.begin (); i != in.end (); ++i, ++n) {
    if (n < trans.size ()) {
      insert (*i, db::ICplxTrans (trans [n]), n);
    } else {
      insert (*i, n);
    }
  }

  db::MergeOp op (min_wc);
  db::ShapeGenerator sg (shapes, true /*clear shapes*/);
  db::PolygonGenerator out (sg, resolve_holes, min_coherence);
  process (out, op);
}

} // namespace db

