
/*

  KLayout Layout Viewer
  Copyright (C) 2006-2007 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#include "dbInstances.h"
#include "dbLayout.h"

namespace db
{

Instances::cell_inst_wp_tree_type        Instances::ms_empty_wp_tree        = Instances::cell_inst_wp_tree_type ();
Instances::cell_inst_tree_type           Instances::ms_empty_tree           = Instances::cell_inst_tree_type ();
Instances::stable_cell_inst_wp_tree_type Instances::ms_empty_stable_wp_tree = Instances::stable_cell_inst_wp_tree_type ();
Instances::stable_cell_inst_tree_type    Instances::ms_empty_stable_tree    = Instances::stable_cell_inst_tree_type ();

// -------------------------------------------------------------------------------------
//  local classes

/**
 *  @brief A base class for instance operations 
 *
 *  This class is used for the Op classes for the undo/redo queuing mechanism.
 */
class InstOpBase
  : public db::Op
{
public:
  InstOpBase () : db::Op () { }

  virtual void undo (Instances *instances) = 0;
  virtual void redo (Instances *instances) = 0;
};

/**
 *  @brief A undo/redo queue object for the instances
 *
 *  This class is used internally to queue an insert or erase operation
 *  into the db::Object manager's undo/redo queue.
 */
template <class Inst>
class InstOp
  : public InstOpBase
{
public:
  typedef db::Box box_type;
  typedef db::box_convert<Inst> box_convert_type;
  typedef db::box_tree<box_type, Inst, box_convert_type> tree_type;

  InstOp (bool insert, const Inst &sh)
    : m_insert (insert)
  {
    m_insts.push_back (sh);
  }
  
  template <class Iter>
  InstOp (bool insert, Iter from, Iter to)
    : m_insert (insert)
  {
    size_t n = 0;
    for (Iter i = from; i != to; ++i) {
      ++n;
    }
    m_insts.reserve (n);
    for (Iter i = from; i != to; ++i) {
      m_insts.push_back (*i);
    }
  }

  template <class Iter>
  InstOp (bool insert, Iter from, Iter to, bool dummy)
    : m_insert (insert)
  {
    m_insts.reserve (std::distance (from, to));
    for (Iter i = from; i != to; ++i) {
      m_insts.push_back (**i);
    }
  }

  virtual void undo (Instances *insts)
  {
    if (m_insert) {
      erase (insts);
    } else {
      insert (insts);
    }
  }

  virtual void redo (Instances *insts)
  {
    if (m_insert) {
      insert (insts);
    } else {
      erase (insts);
    }
  }

private:
  bool m_insert;
  std::vector<Inst> m_insts;

  void insert (Instances *insts);
  void erase (Instances *insts);
};

template <class Inst> 
void 
InstOp<Inst>::insert (Instances *insts)
{
  insts->insert (m_insts.begin (), m_insts.end ());
}

template <class Inst> 
void 
InstOp<Inst>::erase (Instances *insts)
{
  //  because we use the stable box_tree, which is only used in editable mode, we must make sure
  //  we have editable mode here.
  db::check_editable_mode ();

  if (((const Instances *) insts)->stable_inst_tree (typename Inst::tag ()).size () <= m_insts.size ()) {

    //  If all shapes are to be removed, just clear the instances
    insts->clear (typename Inst::tag ());

  } else {

    std::sort (m_insts.begin (), m_insts.end ());

    //  This is not quite effective but seems to be the simpliest way
    //  of implementing this: search for each element and erase these.
    //  The alternative would be to store the iterator along with the object.
    std::vector<typename tree_type::const_iterator> to_erase;
    to_erase.reserve (m_insts.size ());
    for (typename tree_type::const_iterator linst = ((const Instances *) insts)->stable_inst_tree (typename Inst::tag ()).begin (); linst != ((const Instances *) insts)->stable_inst_tree (typename Inst::tag ()).end (); ++linst) {
      typename std::vector<Inst>::iterator i = std::lower_bound (m_insts.begin (), m_insts.end (), *linst);
      if (i != m_insts.end () && *i == *linst) {
        to_erase.push_back (linst);
      }
    }

    insts->erase_positions (typename Inst::tag (), to_erase.begin (), to_erase.end ());

  }
}

// -------------------------------------------------------------------------------------
//  ParentInstRep implementation

Instance
ParentInstRep::child_inst () const
{
  return mp_layout->cell (this->m_parent_cell_index).sorted_inst_ptr (this->m_index);
}

const ParentInstRep::basic_inst_type *
ParentInstRep::basic_child_inst () const
{
  return mp_layout->cell (this->m_parent_cell_index).basic_sorted_inst_ptr (this->m_index);
}

ParentInstRep::cell_inst_array_type  
ParentInstRep::inst () const
{
  //  create a new parent instance by cloning and inverting the array
  basic_inst_type ci (*basic_child_inst ());
  ci.object () = cell_inst_type (this->m_parent_cell_index);
  ci.invert ();
  return ci;
}

// -------------------------------------------------------------------------------------
//  ParentInstIterator implementation

ParentInstIterator &
ParentInstIterator::operator++() 
{
  cell_index_type ci = m_rep.basic_child_inst ()->object ().cell_index ();
  m_rep.inc ();

  if (m_rep.index () == mp_layout->cell (m_rep.parent_cell_index ()).cell_instances () ||
      m_rep.basic_child_inst ()->object ().cell_index () != ci) {
    ++m_iter;
    if (m_iter != m_end) {
      m_rep = parent_inst_type (m_iter->parent_cell_index (), m_iter->index ());
    } else {
      m_rep = parent_inst_type ();
    }
  }

  return *this;
}

// -------------------------------------------------------------------------------------
//  Instance implementation

Instance::cell_inst_array_type::iterator 
Instance::begin_touching (const cell_inst_array_type::box_type &b, const layout_type *g) const 
{
  db::box_convert<cell_inst_type> bc (*g);
  return cell_inst ().begin_touching (b, bc);
}

std::string
Instance::to_string () const
{
  const cell_inst_array_type &ci = cell_inst ();

  std::string r = "cell_index=" + tl::to_string (ci.object ().cell_index ());

  if (ci.is_complex ()) {
    r += " " + ci.complex_trans ().to_string ();
  } else {
    r += " " + (*ci.begin ()).to_string ();
  }

  db::point<coord_type> a, b;
  unsigned long amax, bmax;
  if (ci.is_regular_array (a, b, amax, bmax)) {
    r += " array=(" + a.to_string () + "," + b.to_string () + " " + tl::to_string (amax) + "x" + tl::to_string (bmax) + ")";
  }

  if (has_prop_id ()) {
    r += " prop_id=" + tl::to_string (prop_id ());
  }

  return r;
}

// -------------------------------------------------------------------------------------
//  instance_iterator implementation

template <class Traits>
void
instance_iterator<Traits>::release_iter ()
{ 
  if (m_type == TInstance) {
    if (m_stable) {
      if (m_with_props) {
        ((stable_iter_wp_type *) m_generic.pstable_iter)->~stable_iter_wp_type ();
      } else {
        ((stable_iter_type *) m_generic.stable_iter)->~stable_iter_type ();
      }
    } else {
      if (m_with_props) {
        ((iter_wp_type *) m_generic.piter)->~iter_wp_type ();
      } else {
        ((iter_type *) m_generic.iter)->~iter_type ();
      }
    }
  }
}

template <class Traits>
void
instance_iterator<Traits>::make_iter ()
{ 
  if (m_type == TInstance) {
    if (m_stable) {
      if (m_with_props) {
        new ((stable_iter_wp_type *) m_generic.pstable_iter) stable_iter_wp_type ();
      } else {
        new ((stable_iter_type *) m_generic.stable_iter) stable_iter_type ();
      }
    } else {
      if (m_with_props) {
        new ((iter_wp_type *) m_generic.piter) iter_wp_type ();
      } else {
        new ((iter_type *) m_generic.iter) iter_type ();
      }
    }
    m_traits.init (this); 
  }
}

template <class Traits>
bool
instance_iterator<Traits>::operator== (const instance_iterator<Traits> &d) const
{
  if (! (m_type == d.m_type && m_stable == d.m_stable && m_with_props == d.m_with_props)) {
    return false;
  }
  if (m_type == TNull) {
    return true;
  } else {
    if (m_stable) {
      if (m_with_props) {
        return (get_stable_wp_iter () == d.get_stable_wp_iter ());
      } else {
        return (get_stable_iter () == d.get_stable_iter ());
      }
    } else {
      if (m_with_props) {
        return (get_wp_iter () == d.get_wp_iter ());
      } else {
        return (get_iter () == d.get_iter ());
      }
    }
  }
}

template <class Traits>
instance_iterator<Traits> &
instance_iterator<Traits>::operator= (const instance_iterator<Traits> &iter)
{
  if (this != &iter) {

    release_iter ();

    m_type = iter.m_type;
    m_stable = iter.m_stable;
    m_with_props = iter.m_with_props;
    m_traits = iter.m_traits;

    if (m_type == TInstance) {

      if (m_stable) {
        if (m_with_props) {
          new ((stable_iter_wp_type *) m_generic.pstable_iter) stable_iter_wp_type (*(stable_iter_wp_type *) iter.m_generic.pstable_iter);
        } else {
          new ((stable_iter_type *) m_generic.stable_iter) stable_iter_type (*(stable_iter_type *) iter.m_generic.stable_iter);
        }
      } else {
        if (m_with_props) {
          new ((iter_wp_type *) m_generic.piter) iter_wp_type (*(iter_wp_type *) iter.m_generic.piter);
        } else {
          new ((iter_type *) m_generic.iter) iter_type (*(iter_type *) iter.m_generic.iter);
        }
      }

      update_ref ();

    }

  }

  return *this;
}

template <class Traits>
instance_iterator<Traits> &
instance_iterator<Traits>::operator++() 
{
  if (m_type == TInstance) {
    if (m_stable) {
      if (m_with_props) {
        ++*((stable_iter_wp_type *) m_generic.pstable_iter);
      } else {
        ++*((stable_iter_type *) m_generic.stable_iter);
      }
    } else {
      if (m_with_props) {
        ++*((iter_wp_type *) m_generic.piter);
      } else {
        ++*((iter_type *) m_generic.iter);
      }
    }
    make_next ();
    update_ref ();
  }
  return *this;
}

template <class Traits>
void
instance_iterator<Traits>::make_next () 
{
  while (true) {
    if (m_stable) {
      if (m_with_props) {
        if (! get_stable_wp_iter ().at_end ()) {
          return;
        }
      } else {
        if (! get_stable_iter ().at_end ()) {
          return;
        }
      }
    } else {
      if (m_with_props) {
        if (! get_wp_iter ().at_end ()) {
          return;
        }
      } else {
        if (! get_iter ().at_end ()) {
          return;
        }
      }
    }
    release_iter ();
    m_with_props = ! m_with_props;
    if (! m_with_props) {
      m_type = TNull;
      return;
    }
    make_iter ();
  } 
}

template <class Traits>
void
instance_iterator<Traits>::update_ref ()
{
  if (m_type == TInstance) {
    if (m_stable) {
      if (m_with_props) {
        m_ref = m_traits.instance_from_stable_iter (get_stable_wp_iter ());
      } else {
        m_ref = m_traits.instance_from_stable_iter (get_stable_iter ());
      }
    } else {
      if (m_with_props) {
        m_ref = *get_wp_iter ();
      } else {
        m_ref = *get_iter ();
      }
    }
  } else {
    m_ref = value_type ();
  }
}

template class instance_iterator<NormalInstanceIteratorTraits>;
template class instance_iterator<TouchingInstanceIteratorTraits>;
template class instance_iterator<OverlappingInstanceIteratorTraits>;

// -------------------------------------------------------------------------------------
//  NormalInstanceIteratorTraits implementation

NormalInstanceIteratorTraits::NormalInstanceIteratorTraits ()
  : mp_insts (0)
{ }

NormalInstanceIteratorTraits::NormalInstanceIteratorTraits (const instances_type *insts)
  : mp_insts (insts)
{ }

void 
NormalInstanceIteratorTraits::init (instance_iterator<NormalInstanceIteratorTraits> *iter) const
{
  tl_assert (mp_insts != 0);
  if (iter->m_stable) {
    if (iter->m_with_props) {
      cell_inst_wp_array_type::tag tag = cell_inst_wp_array_type::tag ();
      *((stable_iter_wp_type *) iter->m_generic.pstable_iter) = mp_insts->stable_inst_tree (tag).begin_flat ();
    } else {
      cell_inst_array_type::tag tag = cell_inst_array_type::tag ();
      *((stable_iter_type *) iter->m_generic.stable_iter) = mp_insts->stable_inst_tree (tag).begin_flat ();
    }
  } else {
    if (iter->m_with_props) {
      cell_inst_wp_array_type::tag tag = cell_inst_wp_array_type::tag ();
      *((iter_wp_type *) iter->m_generic.piter) = iter_wp_type (mp_insts->inst_tree (tag).begin (), mp_insts->inst_tree (tag).end ());
    } else {
      cell_inst_array_type::tag tag = cell_inst_array_type::tag ();
      *((iter_type *) iter->m_generic.iter) = iter_type (mp_insts->inst_tree (tag).begin (), mp_insts->inst_tree (tag).end ());
    }
  }
}

// -------------------------------------------------------------------------------------
//  TouchingInstanceIteratorTraits implementation

TouchingInstanceIteratorTraits::TouchingInstanceIteratorTraits ()
  : mp_insts (0), mp_layout (0)
{ }

TouchingInstanceIteratorTraits::TouchingInstanceIteratorTraits (const instances_type *insts, const box_type &box, const layout_type *layout)
  : mp_insts (insts), m_box (box), mp_layout (layout)
{ }

void 
TouchingInstanceIteratorTraits::init (instance_iterator<TouchingInstanceIteratorTraits> *iter) const
{
  tl_assert (mp_insts != 0);
  if (iter->m_stable) {
    if (iter->m_with_props) {
      cell_inst_wp_array_type::tag tag = cell_inst_wp_array_type::tag ();
      *((stable_iter_wp_type *) iter->m_generic.pstable_iter) = mp_insts->stable_inst_tree (tag).begin_touching (m_box, cell_inst_wp_array_box_converter (*mp_layout));
    } else {
      cell_inst_array_type::tag tag = cell_inst_array_type::tag ();
      *((stable_iter_type *) iter->m_generic.stable_iter) = mp_insts->stable_inst_tree (tag).begin_touching (m_box, cell_inst_array_box_converter (*mp_layout));
    }
  } else {
    if (iter->m_with_props) {
      cell_inst_wp_array_type::tag tag = cell_inst_wp_array_type::tag ();
      *((iter_wp_type *) iter->m_generic.piter) = mp_insts->inst_tree (tag).begin_touching (m_box, cell_inst_wp_array_box_converter (*mp_layout));
    } else {
      cell_inst_array_type::tag tag = cell_inst_array_type::tag ();
      *((iter_type *) iter->m_generic.iter) = mp_insts->inst_tree (tag).begin_touching (m_box, cell_inst_array_box_converter (*mp_layout));
    }
  }
}


// -------------------------------------------------------------------------------------
//  OverlappingInstanceIteratorTraits implementation

OverlappingInstanceIteratorTraits::OverlappingInstanceIteratorTraits ()
  : mp_insts (0), mp_layout (0)
{ }

OverlappingInstanceIteratorTraits::OverlappingInstanceIteratorTraits (const instances_type *insts, const box_type &box, const layout_type *layout)
  : mp_insts (insts), m_box (box), mp_layout (layout)
{ }

void 
OverlappingInstanceIteratorTraits::init (instance_iterator<OverlappingInstanceIteratorTraits> *iter) const
{
  tl_assert (mp_insts != 0);
  if (iter->m_stable) {
    if (iter->m_with_props) {
      cell_inst_wp_array_type::tag tag = cell_inst_wp_array_type::tag ();
      *((stable_iter_wp_type *) iter->m_generic.pstable_iter) = mp_insts->stable_inst_tree (tag).begin_overlapping (m_box, cell_inst_wp_array_box_converter (*mp_layout));
    } else {
      cell_inst_array_type::tag tag = cell_inst_array_type::tag ();
      *((stable_iter_type *) iter->m_generic.stable_iter) = mp_insts->stable_inst_tree (tag).begin_overlapping (m_box, cell_inst_array_box_converter (*mp_layout));
    }
  } else {
    if (iter->m_with_props) {
      cell_inst_wp_array_type::tag tag = cell_inst_wp_array_type::tag ();
      *((iter_wp_type *) iter->m_generic.piter) = mp_insts->inst_tree (tag).begin_overlapping (m_box, cell_inst_wp_array_box_converter (*mp_layout));
    } else {
      cell_inst_array_type::tag tag = cell_inst_array_type::tag ();
      *((iter_type *) iter->m_generic.iter) = mp_insts->inst_tree (tag).begin_overlapping (m_box, cell_inst_array_box_converter (*mp_layout));
    }
  }
}

// -------------------------------------------------------------------------------------
//  ChildCellIterator implementation

ChildCellIterator::ChildCellIterator ()
  : m_iter (), m_end (), mp_insts (0)
{ }

ChildCellIterator::ChildCellIterator (const instances_type *insts)
  : m_iter (insts->begin_sorted_insts ()),
    m_end (insts->end_sorted_insts ()),
    mp_insts (insts)
{ }

cell_index_type 
ChildCellIterator::operator* () const
{
  return (*m_iter)->object ().cell_index ();
}

ChildCellIterator &
ChildCellIterator::operator++() 
{
  cell_index_type ci = operator* ();
  do {
    ++m_iter;
  } while (m_iter != m_end && operator* () == ci);
  return *this;
}

// -------------------------------------------------------------------------------------
//  Instances implementation

Instances::Instances (cell_type *cell)
  : mp_cell (cell)
{
  m_generic.any = 0;
  m_generic_wp.any = 0;
}

Instances::~Instances ()
{
  do_clear_insts ();
}

Instances &
Instances::operator= (const Instances &d)
{
  if (this != &d) {
    clear_insts ();
    if (db::editable_mode ()) {
      if (! d.stable_inst_tree (cell_inst_array_type::tag ()).empty ()) {
        stable_inst_tree (cell_inst_array_type::tag ()) = d.stable_inst_tree (cell_inst_array_type::tag ());
      }
      if (! d.stable_inst_tree (cell_inst_wp_array_type::tag ()).empty ()) {
        stable_inst_tree (cell_inst_wp_array_type::tag ()) = d.stable_inst_tree (cell_inst_wp_array_type::tag ());
      }
    } else {
      if (! d.inst_tree (cell_inst_array_type::tag ()).empty ()) {
        inst_tree (cell_inst_array_type::tag ()) = d.inst_tree (cell_inst_array_type::tag ());
      }
      if (! d.inst_tree (cell_inst_wp_array_type::tag ()).empty ()) {
        inst_tree (cell_inst_wp_array_type::tag ()) = d.inst_tree (cell_inst_wp_array_type::tag ());
      }
    }
    m_parent_insts = d.m_parent_insts;
  }
  return *this;
}

template <class Tag, class I>
void 
Instances::erase_positions (Tag tag, I first, I last)
{
  db::check_editable_mode ("Instances::erase");

  mp_cell->invalidate_insts ();  //  HINT: must come before the change is done!
  if (mp_cell->manager () && mp_cell->manager ()->transacting ()) {
    mp_cell->manager ()->queue (mp_cell, new db::InstOp<typename Tag::object_type> (false /*not insert*/, first, last, true /*dummy*/));
  }

  typedef db::box_tree<box_type, typename Tag::object_type, db::box_convert<typename Tag::object_type> > tree_type;
  tree_type &t = stable_inst_tree (tag);
  t.erase_positions (first, last);
}

template <class InstArray>
Instances::instance_type
Instances::insert (const InstArray &inst)
{
  if (mp_cell->manager () && mp_cell->manager ()->transacting ()) {
    mp_cell->manager ()->queue (mp_cell, new db::InstOp<InstArray> (true /*insert*/, inst));
  }
  mp_cell->invalidate_insts ();
  if (db::editable_mode ()) {
    return instance_type (stable_inst_tree (typename InstArray::tag ()).insert (inst));
  } else {
    return instance_type (*(inst_tree (typename InstArray::tag ()).insert (inst)));
  }
}

template <class I>
void 
Instances::insert (I from, I to)
{
  typedef std::iterator_traits<I> it_traits;
  typedef typename it_traits::value_type value_type;

  if (mp_cell->manager () && mp_cell->manager ()->transacting ()) {
    db::check_editable_mode ("Instances::insert");
    mp_cell->manager ()->queue (mp_cell, new db::InstOp<typename it_traits::value_type> (true /*insert*/, from, to));
  }
  mp_cell->invalidate_insts ();
  if (db::editable_mode ()) {
    stable_inst_tree (typename value_type::tag ()).insert (from, to);
  } else {
    inst_tree (typename value_type::tag ()).insert (from, to);
  }
}

template <class Tag>
bool 
Instances::is_valid_by_tag (Tag tag, const instance_type &ref) const
{
  db::check_editable_mode ("Instances::is_valid");
  return ref.basic_iter (tag)->is_valid ();
}

bool
Instances::is_valid (const instance_type &ref) const
{
  if (ref.has_prop_id ()) {
    return is_valid_by_tag (cell_inst_wp_array_type::tag (), ref);
  } else {
    return is_valid_by_tag (cell_inst_array_type::tag (), ref);
  }
}

template <class InstArray>
void 
Instances::replace (const InstArray *replace, const InstArray &with)
{
  db::check_editable_mode ("Instances::replace");

  if (mp_cell->manager () && mp_cell->manager ()->transacting ()) {
    mp_cell->manager ()->queue (mp_cell, new db::InstOp<InstArray> (false /*not insert*/, *replace));
    mp_cell->manager ()->queue (mp_cell, new db::InstOp<InstArray> (true /*insert*/, with));
  }
  mp_cell->invalidate_insts ();
  //  HINT: this only works because we know our box trees well:
  *((InstArray *)replace) = with;
}

template <class Tag, class Tree>
void 
Instances::erase_inst_by_iter (Tag tag, Tree &t, typename Tree::const_iterator iter)
{
  if (iter.vector () != &t.objects ()) {
    throw tl::Exception (tl::translate ("Trying to erase an object from a list that it does not belong to"));
  }

  mp_cell->invalidate_insts ();

  if (mp_cell->manager () && mp_cell->manager ()->transacting ()) {
    db::check_editable_mode ("Instances::erase");
    mp_cell->manager ()->queue (mp_cell, new db::InstOp<typename Tag::object_type> (false /*not insert*/, *iter));
  }

  t.erase (iter.to_non_const ());
}

template <class InstArray>
Instances::instance_type 
Instances::replace (const instance_type &ref, const InstArray &inst)
{
  db::check_editable_mode ("Instances::replace");  //  do this here for correct function name, since it may be mapped to erase and insert ..
  const InstArray *cp = ref.basic_ptr (typename InstArray::tag ());
  if (cp) {
    if (ref.basic_iter (typename InstArray::tag ())->vector () != &(((const Instances *) this)->stable_inst_tree (typename InstArray::tag ()).objects ())) {
      throw tl::Exception (tl::translate ("Trying to replace an object in a list that it does not belong to"));
    }
    //  in-place replacement 
    replace (cp, inst);
    return ref;
  } else if (ref.has_prop_id ()) {
    //  not an in-place replacement - erase and insert
    //  HINT: in this case the current reference has a property while the given "inst" must be 
    //  one without properties - keep the properties as they are.
    db::PropertiesRepository::properties_id_type pid = ref.prop_id ();
    erase (ref);
    return insert (cell_inst_wp_array_type (inst, pid));
  } else {
    //  not an in-place replacement - erase and insert
    erase (ref);
    return insert (inst);
  }
}

template <class Tag>
void 
Instances::erase_inst_by_tag (Tag tag, const typename Tag::object_type &obj)
{
  db::check_editable_mode ("Instances::erase");

  mp_cell->invalidate_insts ();

  if (mp_cell->manager () && mp_cell->manager ()->transacting ()) {
    mp_cell->manager ()->queue (mp_cell, new db::InstOp<typename Tag::object_type> (false /*not insert*/, obj));
  }

  typedef db::box_tree<box_type, typename Tag::object_type, db::box_convert<typename Tag::object_type> > tree_type;
  tree_type &t = stable_inst_tree (tag);
  t.erase (t.iterator_from_pointer (const_cast <typename Tag::object_type *> (&obj)));
}

template <class Tag>
void 
Instances::erase_insts_by_tag (Tag tag, std::vector<instance_type>::const_iterator s1, std::vector<instance_type>::const_iterator s2)
{
  typedef db::box_tree<box_type, typename Tag::object_type, db::box_convert<typename Tag::object_type> > tree_type;
  tree_type &t = stable_inst_tree (tag);

  std::vector<typename tree_type::iterator> iters;
  iters.reserve (std::distance (s1, s2));

  for (std::vector<instance_type>::const_iterator s = s1; s != s2; ++s) {
    iters.push_back (t.iterator_from_pointer (const_cast <typename Tag::object_type *> (s->basic_ptr (tag))));
  }

  erase_positions (Tag (), iters.begin (), iters.end ());
}

void 
Instances::clear_insts ()
{
  mp_cell->invalidate_insts ();
  if (mp_cell->manager () && mp_cell->manager ()->transacting ()) {
    db::check_editable_mode ("Instances::clear_insts");
    const Instances *const_this = this;
    if (! const_this->stable_inst_tree (cell_inst_array_type::tag ()).empty ()) {
      mp_cell->manager ()->queue (mp_cell, new db::InstOp<cell_inst_array_type> (false /*not insert*/, const_this->stable_inst_tree (cell_inst_array_type::tag ()).begin (), const_this->stable_inst_tree (cell_inst_array_type::tag ()).end ()));
    }
    if (! const_this->stable_inst_tree (cell_inst_wp_array_type::tag ()).empty ()) {
      mp_cell->manager ()->queue (mp_cell, new db::InstOp<cell_inst_wp_array_type> (false /*not insert*/, const_this->stable_inst_tree (cell_inst_wp_array_type::tag ()).begin (), const_this->stable_inst_tree (cell_inst_wp_array_type::tag ()).end ()));
    }
  }
  do_clear_insts ();
}

void 
Instances::clear (Instances::cell_inst_array_type::tag) 
{
  mp_cell->invalidate_insts ();
  if (m_generic.any) {
    if (db::editable_mode ()) {
      delete m_generic.stable_tree;
    } else {
      delete m_generic.unstable_tree;
    }
    m_generic.any = 0;
  }
}

void 
Instances::clear (Instances::cell_inst_wp_array_type::tag) 
{
  mp_cell->invalidate_insts ();
  if (m_generic_wp.any) {
    if (db::editable_mode ()) {
      delete m_generic_wp.stable_tree;
    } else {
      delete m_generic_wp.unstable_tree;
    }
    m_generic_wp.any = 0;
  }
}

Instances::instance_type
Instances::instance_from_pointer (const basic_inst_type *p) const
{
  if (db::editable_mode ()) {
    if (stable_inst_tree (cell_inst_array_type::tag ()).is_member_of (p)) {
      return instance_type (stable_inst_tree (cell_inst_array_type::tag ()).iterator_from_pointer (static_cast <const cell_inst_array_type *> (p)));
    }
    if (stable_inst_tree (cell_inst_wp_array_type::tag ()).is_member_of (p)) {
      return instance_type (stable_inst_tree (cell_inst_wp_array_type::tag ()).iterator_from_pointer (static_cast <const cell_inst_wp_array_type *> (p)));
    }
  } else {
    if (p < &*(inst_tree (cell_inst_array_type::tag ()).end ()) && p >= &*(inst_tree (cell_inst_array_type::tag ()).begin ())) {
      return instance_type (*static_cast <const cell_inst_array_type *> (p));
    }
    if (p < &*(inst_tree (cell_inst_wp_array_type::tag ()).end ()) && p >= &*(inst_tree (cell_inst_wp_array_type::tag ()).begin ())) {
      return instance_type (*static_cast <const cell_inst_wp_array_type *> (p));
    }
  }
  return Instance ();
}

bool 
Instances::empty () const
{
  if (db::editable_mode ()) {
    return ((! m_generic.stable_tree || m_generic.stable_tree->empty ()) &&
            (! m_generic_wp.stable_tree || m_generic_wp.stable_tree->empty ()));
  } else {
    return ((! m_generic.unstable_tree || m_generic.unstable_tree->empty ()) &&
            (! m_generic_wp.unstable_tree || m_generic_wp.unstable_tree->empty ()));
  }
}

void 
Instances::erase (const Instance &ref)
{
  if (ref.is_null ()) {
    //  .. nothing ..
  } else if (ref.has_prop_id ()) {
    erase_inst_by_tag (cell_inst_wp_array_type::tag (), ref);
  } else {
    erase_inst_by_tag (cell_inst_array_type::tag (), ref);
  } 
}

void 
Instances::erase (const Instances::const_iterator &e)
{
  if (e.at_end ()) {
    //  .. nothing ..
  } else if (e->has_prop_id ()) {
    erase_inst_by_tag (cell_inst_wp_array_type::tag (), e);
  } else {
    erase_inst_by_tag (cell_inst_array_type::tag (), e);
  } 
}

void 
Instances::erase_insts (const std::vector<Instance> &instances)
{
  for (std::vector<instance_type>::const_iterator i = instances.begin (); i != instances.end (); ) {

    std::vector<instance_type>::const_iterator inext = i;
    while (inext != instances.end () && inext->has_prop_id () == i->has_prop_id ()) {
      ++inext;
    }

    if (i->has_prop_id ()) {
      erase_insts_by_tag (cell_inst_wp_array_type::tag (), i, inext);
    } else {
      erase_insts_by_tag (cell_inst_array_type::tag (), i, inext);
    }

    i = inext;

  }
}

void 
Instances::count_parent_insts (std::vector <size_t> &count) const
{
  cell_index_type last_ci = (cell_index_type) -1l;
  
  for (sorted_inst_iterator c = begin_sorted_insts (); c != end_sorted_insts (); ++c) {
    cell_index_type ci = (*c)->object ().cell_index ();
    if (ci != last_ci) {
      last_ci = ci;
      ++count [ci];
    }
  }
}

/**
 *  @brief A helper function that compares child instances
 */
template <class I>
struct cell_inst_compare_f
{
  bool operator() (const I *a, const I *b) const
  {
    return a->raw_less (*b);
  }
};

void 
Instances::sort_child_insts () 
{
  m_insts_by_cell_index = sorted_inst_vector ();
  m_insts_by_cell_index.reserve (cell_instances ());

  //  HINT: we do not use the flat iterator since this would required a "made" index
  //  which is not available in some cases.
  if (db::editable_mode ()) {
    if (m_generic.any) {
      for (stable_cell_inst_tree_type::const_iterator i = m_generic.stable_tree->begin (); i != m_generic.stable_tree->end (); ++i) {
        m_insts_by_cell_index.push_back (&*i);
      }
    }
    if (m_generic_wp.any) {
      for (stable_cell_inst_wp_tree_type::const_iterator i = m_generic_wp.stable_tree->begin (); i != m_generic_wp.stable_tree->end (); ++i) {
        m_insts_by_cell_index.push_back (&*i);
      }
    }
  } else {
    if (m_generic.any) {
      for (cell_inst_tree_type::const_iterator i = m_generic.unstable_tree->begin (); i != m_generic.unstable_tree->end (); ++i) {
        m_insts_by_cell_index.push_back (&*i);
      }
    }
    if (m_generic_wp.any) {
      for (cell_inst_wp_tree_type::const_iterator i = m_generic_wp.unstable_tree->begin (); i != m_generic_wp.unstable_tree->end (); ++i) {
        m_insts_by_cell_index.push_back (&*i);
      }
    }
  }

  std::sort (m_insts_by_cell_index.begin (), m_insts_by_cell_index.end (), cell_inst_compare_f<basic_inst_type> ());
}

void 
Instances::make_index ()
{
  if (db::editable_mode ()) {
    if (m_generic.any) {
      m_generic.stable_tree->make_index ();
    }
    if (m_generic_wp.any) {
      m_generic_wp.stable_tree->make_index ();
    }
  }
}

void 
Instances::sort_inst_tree (const Layout *g)
{
  if (m_generic.any) {
    if (db::editable_mode ()) {
      m_generic.stable_tree->sort (db::box_convert<cell_inst_array_type> (*g));
    } else {
      m_generic.unstable_tree->sort (db::box_convert<cell_inst_array_type> (*g));
      //  since we use unstable instance trees in non-editable mode, we need to resort the child instances in this case
      sort_child_insts ();
    }
  }
  if (m_generic_wp.any) {
    if (db::editable_mode ()) {
      m_generic_wp.stable_tree->sort (db::box_convert<cell_inst_wp_array_type> (*g));
    } else {
      m_generic_wp.unstable_tree->sort (db::box_convert<cell_inst_wp_array_type> (*g));
      //  since we use unstable instance trees in non-editable mode, we need to resort the child instances in this case
      sort_child_insts ();
    }
  }

}

void 
Instances::update_relations (Layout *g, cell_index_type cell_index)
{
  cell_index_type last_ci = (cell_index_type) -1l;

  size_t idx = 0;
  for (sorted_inst_iterator c = begin_sorted_insts (); c != end_sorted_insts (); ++c, ++idx) {
    cell_index_type ci = (*c)->object ().cell_index ();
    if (ci != last_ci) {
      last_ci = ci;
      g->cell (ci).instances ().m_parent_insts.push_back (parent_inst_type (cell_index, idx));
    }
  }
}

size_t 
Instances::child_cells () const
{
  size_t n = 0;
  for (ChildCellIterator i = begin_child_cells (); ! i.at_end (); ++i) {
    ++n;
  }
  return n;
}

size_t 
Instances::cell_instances () const
{
  if (db::editable_mode ()) {
    return (m_generic.stable_tree ? m_generic.stable_tree->size () : 0) +
           (m_generic_wp.stable_tree ? m_generic_wp.stable_tree->size () : 0);
  } else {
    return (m_generic.unstable_tree ? m_generic.unstable_tree->size () : 0) +
           (m_generic_wp.unstable_tree ? m_generic_wp.unstable_tree->size () : 0);
  }
}

void
Instances::collect_mem_stat (db::MemStatistics &m) const
{
  m.cell_info (sizeof (mp_cell), sizeof (mp_cell));
  m.cell_info (sizeof (m_parent_insts), sizeof (m_parent_insts));        //  need more: see below
  m.cell_info (sizeof (m_generic), sizeof (m_generic));  //  need more: see below
  m.cell_info (sizeof (m_generic_wp), sizeof (m_generic_wp));        //  need more: see below
  m.cell_info (sizeof (m_insts_by_cell_index), sizeof (m_insts_by_cell_index));        //  need more: see below

  if (db::editable_mode ()) {
    if (m_generic.stable_tree) {
      m.inst_trees (*m_generic.stable_tree);
    }
    if (m_generic_wp.stable_tree) {
      m.inst_trees (*m_generic_wp.stable_tree);
    }
  } else {
    if (m_generic.unstable_tree) {
      m.inst_trees (*m_generic.unstable_tree);
    }
    if (m_generic_wp.unstable_tree) {
      m.inst_trees (*m_generic_wp.unstable_tree);
    }
  }

  m.instances (m_parent_insts);
  m.instances (m_insts_by_cell_index);
  m.instances (-sizeof (m_parent_insts), -sizeof (m_parent_insts));        //  the actual object is counted for cell, not for instances (see above)
}

Instances::instance_type 
Instances::replace_prop_id (const instance_type &ref, db::PropertiesRepository::properties_id_type prop_id)
{
  db::check_editable_mode ("Instances::replace_prop_id");
  if (! ref.is_null ()) {
    cell_inst_wp_array_type new_inst (ref.cell_inst (), prop_id);
    return replace (ref, new_inst);
  } else {
    return ref;
  }
}

void 
Instances::do_clear_insts ()
{
  if (m_generic.any) {
    if (db::editable_mode ()) {
      delete m_generic.stable_tree;
    } else {
      delete m_generic.unstable_tree;
    }
    m_generic.any = 0;
  }
  if (m_generic_wp.any) {
    if (db::editable_mode ()) {
      delete m_generic_wp.stable_tree;
    } else {
      delete m_generic_wp.unstable_tree;
    }
    m_generic_wp.any = 0;
  }
}

void
Instances::undo (db::Op *op)
{
  db::InstOpBase *instop = dynamic_cast<InstOpBase *> (op);
  if (instop) {
    instop->undo (this);
  } 
}

void
Instances::redo (db::Op *op)
{
  //  actions are only queued by the instance list - this is should be 
  //  responsible for the handling of the latter.
  //  HACK: this is not really a nice concept, but it saves us a pointer to the manager.
  db::InstOpBase *instop = dynamic_cast<InstOpBase *> (op);
  if (instop) {
    instop->redo (this);
  } 
}

Instances::instance_type 
Instances::do_insert (const Instances::instance_type &ref, 
                      tl::func_delegate_base <db::cell_index_type> &im,
                      tl::func_delegate_base <db::PropertiesRepository::properties_id_type> &pm)
{
  if (! ref.has_prop_id ()) {
    cell_inst_array_type inst = *ref.basic_ptr (cell_inst_array_type::tag ());
    inst.object () = cell_inst_type (im (ref.cell_index ()));
    return insert (inst);
  } else {
    cell_inst_wp_array_type inst_wp = *ref.basic_ptr (cell_inst_wp_array_type::tag ());
    inst_wp.object () = cell_inst_type (im (ref.cell_index ()));
    inst_wp.properties_id (pm (ref.prop_id ()));
    return insert (inst_wp);
  }
}

//  explicit instantiations
template void Instances::replace<> (const Instances::cell_inst_array_type *replace, const Instances::cell_inst_array_type &with);
template void Instances::replace<> (const Instances::cell_inst_wp_array_type *replace, const Instances::cell_inst_wp_array_type &with);
template Instance Instances::replace<> (const Instance &replace, const Instances::cell_inst_wp_array_type &with);
template Instance Instances::replace<> (const Instance &replace, const Instances::cell_inst_array_type &with);

}

