
/*

  KLayout Layout Viewer
  Copyright (C) 2006-2007 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#include "dbLayoutUtils.h"

namespace db
{

// ------------------------------------------------------------------------------------
//  Recursive shape iterator implementation

RecursiveShapeIterator::RecursiveShapeIterator ()
{
  //  anything. Not necessary reasonable.
  m_layer = 0;
  mp_layout = 0;
  mp_cell = 0;
  m_overlapping = false;
  m_depth = std::numeric_limits<unsigned int>::max (); // all
  m_shape_flags = shape_iterator::All;
  mp_shape_prop_sel = 0;
  m_shape_inv_prop_sel = false;
}

RecursiveShapeIterator::RecursiveShapeIterator (const layout_type &layout, const cell_type &cell, unsigned int layer, const box_type &region, bool overlapping)
  : m_box_convert (layout, layer)
{
  m_layer = layer;
  m_region = region;
  mp_layout = &layout;
  mp_cell = &cell;
  m_overlapping = overlapping;
  init ();
}

RecursiveShapeIterator::RecursiveShapeIterator (const layout_type &layout, const cell_type &cell, unsigned int layer)
  : m_box_convert (layout, layer)
{
  m_layer = layer;
  m_region = box_type::world ();
  mp_layout = &layout;
  mp_cell = &cell;
  m_overlapping = false;
  init ();
}

void 
RecursiveShapeIterator::init ()
{
  m_depth = std::numeric_limits<unsigned int>::max (); // all
  m_shape_flags = shape_iterator::All;
  mp_shape_prop_sel = 0;
  m_shape_inv_prop_sel = false;

  new_cell ();
  next_shape ();
}

void 
RecursiveShapeIterator::next_shape ()
{
  while (m_shape.at_end ()) {

    if (! m_inst.at_end () && m_inst_iterators.size () < m_depth) {

      if (mp_layout->cell (m_inst->cell_index ()).bbox (m_layer).empty ()) {

        ++m_inst;
        if (! m_inst.at_end ()) {
          if (m_local_region != box_type::world ()) {
            m_inst_array = m_inst->cell_inst ().begin_touching (m_local_region, m_box_convert);
          } else {
            m_inst_array = m_inst->cell_inst ().begin ();
          }
        }

      } else {
        down ();
      }

    } else {

      //  no more instances: up and next instance
      if (m_inst_iterators.empty ()) {
        //  nothing left:
        return;
      }

      up ();

      ++m_inst_array;
      if (m_inst_array.at_end ()) {
        ++m_inst;
        if (! m_inst.at_end ()) {
          if (m_local_region != box_type::world ()) {
            m_inst_array = m_inst->cell_inst ().begin_touching (m_local_region, m_box_convert);
          } else {
            m_inst_array = m_inst->cell_inst ().begin ();
          }
        }
      }

    }

  }
}

void
RecursiveShapeIterator::down ()
{
  m_trans_stack.push_back (m_trans);
  m_cells.push_back (mp_cell);

  m_inst_iterators.push_back (m_inst);
  m_inst_array_iterators.push_back (m_inst_array);

  mp_cell = &mp_layout->cell (m_inst->cell_index ());
  m_trans = m_trans * m_inst->complex_trans (*m_inst_array);

  new_cell ();
}

void
RecursiveShapeIterator::up ()
{
  m_inst = m_inst_iterators.back ();
  m_inst_array = m_inst_array_iterators.back ();
  m_inst_iterators.pop_back ();
  m_inst_array_iterators.pop_back ();

  m_trans = m_trans_stack.back ();
  m_trans_stack.pop_back ();
  mp_cell = m_cells.back ();
  m_cells.pop_back ();

  m_local_region = box_type::world ();
  if (m_region != m_local_region) {
    m_local_region = itrans ().inverted () * m_region;
  }
}

void 
RecursiveShapeIterator::new_cell ()
{
  //  don't transform the world region, since transformation of that region might not work properly
  m_local_region = box_type::world ();
  if (m_region != m_local_region) {
    m_local_region = itrans ().inverted () * m_region;
  }

  if (! m_overlapping) {
    m_shape = mp_cell->shapes (m_layer).begin_touching (m_local_region, m_shape_flags, mp_shape_prop_sel, m_shape_inv_prop_sel); 
  } else {
    m_shape = mp_cell->shapes (m_layer).begin_overlapping (m_local_region, m_shape_flags, mp_shape_prop_sel, m_shape_inv_prop_sel); 
  }

  m_inst = mp_cell->begin_touching (m_local_region);
  if (! m_inst.at_end ()) {
    if (m_local_region != box_type::world ()) {
      m_inst_array = m_inst->cell_inst ().begin_touching (m_local_region, m_box_convert); 
    } else {
      m_inst_array = m_inst->cell_inst ().begin (); 
    }
  }
}

// ------------------------------------------------------------------------------------
//  PropertyMapper implementation

PropertyMapper::PropertyMapper (db::Layout &target, const db::Layout &source)
  : mp_target (&target), mp_source (&source)
{
  //  .. nothing yet ..
}

/**
 *  @brief Instantiate a property mapper for mapping of property ids from the source to the target layout
 *
 *  This version does not specify a certain source or target layout. These must be set with the
 *  set_source or set_target methods.
 */
PropertyMapper::PropertyMapper ()
  : mp_target (0), mp_source (0)
{
  //  .. nothing yet ..
}

/**
 *  @brief Specify the source layout
 */
void 
PropertyMapper::set_source (const db::Layout &source)
{
  if (&source != mp_source) {
    m_prop_id_map.clear ();
    mp_source = &source;
  }
}

/**
 *  @brief Specify the target layout
 */
void 
PropertyMapper::set_target (db::Layout &target)
{
  if (&target != mp_target) {
    m_prop_id_map.clear ();
    mp_target = &target;
  }
}

/**
 *  @brief The actual mapping function
 */
db::Layout::properties_id_type 
PropertyMapper::operator() (db::Layout::properties_id_type source_id)
{
  tl_assert (mp_source != 0);
  tl_assert (mp_target != 0);

  std::map <db::Layout::properties_id_type, db::Layout::properties_id_type>::const_iterator p = m_prop_id_map.find (source_id);

  if (p == m_prop_id_map.end ()) {
    db::Layout::properties_id_type new_id = mp_target->properties_repository ().translate (mp_source->properties_repository (), source_id);
    m_prop_id_map.insert (std::make_pair (source_id, new_id));
    return new_id;
  } else {
    return p->second;
  }
}

}

