
/*

  KLayout Layout Viewer
  Copyright (C) 2006-2007 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#include "dbOASISWriter.h"
#include "dbEdgeProcessor.h"
#include "tlStream.h"
#include "tlUtils.h"

namespace db
{

// ---------------------------------------------------------------------------------
//  Generic delivery of shapes (specialized to compressing / non-compressing variants)

namespace {

template <class Sh>
class ShapeDelivery
{
public:
  ShapeDelivery ()
  {
    //  .. nothing yet ..
  }

  virtual ~ShapeDelivery ()
  {
    //  .. nothing yet ..
  }

  virtual bool at_end () const = 0;

  virtual void operator () (Sh &shape, db::Repetition &rep, db::PropertiesRepository::properties_id_type &prop_id) = 0;

  virtual ShapeDelivery &operator++ () 
  {
    inc ();
    return *this;
  }

protected:
  virtual void inc () = 0;
};

/**
 *  @brief Non-compressing standard shape delivery
 */
template <class Sh>
class StandardShapeDelivery
  : public ShapeDelivery<Sh>
{
public:
  StandardShapeDelivery (const db::Shapes &shapes, unsigned int type)
    : ShapeDelivery<Sh> (), 
      m_shape_iterator (shapes.begin (type))
  {
    //  .. nothing yet ..
  }

  virtual bool at_end () const 
  {
    return m_shape_iterator.at_end ();
  }

  virtual void operator () (Sh &shape, db::Repetition &rep, db::PropertiesRepository::properties_id_type &prop_id)
  {
    m_shape_iterator->instantiate (shape);
    prop_id = m_shape_iterator->prop_id ();
    rep = db::Repetition ();
  }

  virtual void inc ()
  {
    ++m_shape_iterator;
  }

private:
  db::ShapeIterator m_shape_iterator;
};

/**
 *  @brief Compare operator for points, distinct x clustered (with same y)
 */
struct point_cmp_x
{
  bool operator() (const db::Point &a, const db::Point &b) 
  {
    if (a.y () != b.y ()) {
      return a.y () < b.y ();
    } else {
      return a.x () < b.x ();
    }
  }
};

/**
 *  @brief Compare operator for points, distinct y clustered (with same x)
 */
struct point_cmp_y
{
  bool operator() (const db::Point &a, const db::Point &b) 
  {
    if (a.x () != b.x ()) {
      return a.x () < b.x ();
    } else {
      return a.y () < b.y ();
    }
  }
};

/**
 *  @brief Compare operator for points/abstract repetition pair with configurable point compare operator
 */
template <class PC>
struct rep_point_cmp
{
  bool operator () (const std::pair <db::Point, std::pair <db::Coord, int> > &a, const std::pair <db::Point, std::pair <db::Coord, int> > &b)
  {
    if (a.second != b.second) {
      return a.second < b.second;
    }
    PC pc;
    return pc (a.first, b.first);
  }
};

const unsigned int max_level = 10;

/**
 *  @brief Compressing shape delivery
 */
template <class Sh>
class CompressingShapeDelivery
  : public ShapeDelivery<Sh>
{
public:
  /** 
   *  @brief Constructor
   *
   *  @param shapes The shapes container from which to take the shapes
   *  @param type The shape type (ShapeIterator types)
   *  @param level The compression level 
   *
   *  Allowed levels are:
   *    0   - simple
   *    1   - 2d arrays
   *    2++ - search for 2nd, 3rd ... order neighbors
   */
  CompressingShapeDelivery (const db::Shapes &shapes, unsigned int type, unsigned int level)
    : ShapeDelivery<Sh> (),
      m_ntot (0), m_nrep (0), m_nnonrep (0), m_level (level < max_level ? level : max_level)
  {
    //  count the instances per compression slot
    std::map <key_type, size_t> counts;
    for (db::ShapeIterator shape_iterator (shapes, type); ! shape_iterator.at_end (); ++shape_iterator) {

      std::pair <key_type, size_t> key;
      shape_iterator->instantiate (key.first.first);
      key.first.second = shape_iterator->prop_id ();
      key.second = 0;

      db::Disp tr;
      key.first.first.reduce (tr);

      counts.insert (key).first->second += 1;

      ++m_ntot;

    }

    //  prepare vectors for the displacements
    for (typename std::map <key_type, size_t>::const_iterator c = counts.begin (); c != counts.end (); ++c) {
      std::pair <key_type, value_type> key;
      key.first = c->first;
      m_normalized.insert (key).first->second.first.reserve (c->second);
    }

    //  collect the normalized shapes (with properties)
    for (db::ShapeIterator shape_iterator (shapes, type); ! shape_iterator.at_end (); ++shape_iterator) {

      key_type key;
      shape_iterator->instantiate (key.first);
      key.second = shape_iterator->prop_id ();

      db::Disp tr;
      key.first.reduce (tr);

      m_normalized.find (key)->second.first.push_back (tr.disp ());

    }

    //  produce the repetitions
    
    for (typename std::map <key_type, value_type>::iterator n = m_normalized.begin (); n != m_normalized.end (); ++n) {

      disp_vector displacements;
      disp_vector::iterator d;

      typedef std::vector <std::pair <db::Point, std::pair <db::Coord, int> > > tmp_rep_vector;
      tmp_rep_vector repetitions;
      tmp_rep_vector::iterator rw;

      //  First pass: try compression in x direction, second pass: try compression in y direction
      for (int xpass = 1; xpass >= 0; --xpass) {
      
        displacements.clear ();
        repetitions.clear ();

        displacements.swap (n->second.first);
        if (xpass) {
          std::sort (displacements.begin (), displacements.end (), point_cmp_x ());
        } else {
          std::sort (displacements.begin (), displacements.end (), point_cmp_y ());
        }

        disp_vector::iterator dwindow = displacements.begin ();
        for (d = displacements.begin (); d != displacements.end (); ) {

          if (level <= 1) {

            disp_vector::iterator dd = d;
            ++dd;

            db::Point dxy;
            int nxy = 1;

            if (dd != displacements.end ()) {

              dxy = xpass ? db::Point (dd->x () - d->x (), 0) : db::Point (0, dd->y () - d->y ());
              while (dd != displacements.end () && *dd == dd[-1] + dxy) {
                ++dd;
                ++nxy;
              } 

            }

            if (nxy < 2) {

              n->second.first.push_back (*d++);
              if (! xpass) {
                ++m_nnonrep;
              }

            } else {

              if (level > 0) {
                repetitions.push_back (std::make_pair (*d, std::make_pair (xpass ? dxy.x () : dxy.y (), nxy)));
              } else {
                n->second.second.push_back (std::make_pair (*d, db::Repetition (new RegularRepetition (dxy, db::Point (), nxy, 1))));
              }

              d = dd;

            }

          } else {

            //  collect the nearest neighbor distances and counts for 2..level order neighbors
            int nxy_max = 1;
            unsigned int nn_max = 0;

            //  move the window of identical x/y coordinates if necessary
            if (d == dwindow) {
              for (dwindow = d + 1; dwindow != displacements.end () && (xpass ? (dwindow->y () == d->y ()) : (dwindow->x () == d->x ())); ++dwindow) 
                ;
            }

            for (unsigned int nn = 0; nn < level; ++nn) {

              disp_vector::iterator dd = d + (nn + 1);
              if (dd >= dwindow) {
                break;
              }

              db::Point dxy = xpass ? db::Point (dd->x () - d->x (), 0) : db::Point (0, dd->y () - d->y ());

              int nxy = 2;
              while (dd != dwindow) {
                disp_vector::iterator df = std::lower_bound (dd + 1, dwindow, *dd + dxy);
                if (df == dwindow || *df != *dd + dxy) {
                  break;
                }
                ++nxy;
                dd = df;
              }

              if (nxy > nxy_max) {
                nxy_max = nxy;
                nn_max = nn;
              }

            }

            if (nxy_max < 2) {

              //  no candidate found - just keep that one
              n->second.first.push_back (*d++);
              if (! xpass) {
                ++m_nnonrep;
              }

            } else {

              //  take out the ones of this sequence from the list
              db::Point dxy_max = xpass ? db::Point ((d + nn_max + 1)->x () - d->x (), 0) : db::Point (0, (d + nn_max + 1)->y () - d->y ());

              disp_vector::iterator dd = d;
              for (int nxy = 1; nxy < nxy_max; ++nxy) {
                disp_vector::iterator df = std::lower_bound (dd + 1, dwindow, *dd + dxy_max);
                tl_assert (df != dwindow);
                tl_assert (*df == *dd + dxy_max);
                std::rotate (dd + 1, df, df + 1);
                ++dd;
              }

              repetitions.push_back (std::make_pair (*d, std::make_pair (xpass ? dxy_max.x () : dxy_max.y (), nxy_max)));

              d += nxy_max;

            }

          }

        }

        //  Try to compact these repetitions further, y direction first, then x direction
        for (int ypass2 = 1; ypass2 >= 0; --ypass2) {
        
          if (ypass2) {
            std::sort (repetitions.begin (), repetitions.end (), rep_point_cmp<point_cmp_y> ());
          } else {
            std::sort (repetitions.begin (), repetitions.end (), rep_point_cmp<point_cmp_x> ());
          }

          rw = repetitions.begin ();
          for (tmp_rep_vector::const_iterator r = repetitions.begin (); r != repetitions.end (); ) {

            tmp_rep_vector::const_iterator rr = r;
            ++rr;

            db::Point dxy2;
            if (rr != repetitions.end ()) {
              dxy2 = ypass2 ? db::Point (0, rr->first.y () - r->first.y ()) : db::Point (rr->first.x () - r->first.x (), 0);
            }
            int nxy2 = 1;

            db::Point dxy2n (dxy2);
            while (rr != repetitions.end () && rr->second == r->second && rr->first == r->first + dxy2n) {
              ++nxy2;
              ++rr;
              dxy2n += dxy2;
            }

            if (nxy2 < 2 && ypass2) {
              *rw++ = *r;
            } else {
              db::Point a (xpass ? r->second.first : 0, xpass ? 0 : r->second.first);
              n->second.second.push_back (std::make_pair (r->first, db::Repetition (new RegularRepetition (a, dxy2, r->second.second, nxy2))));
              ++m_nrep;
            }

            r = rr;

          }

          repetitions.erase (rw, repetitions.end ());

        }

      }

    }

    //  take all irregular displacements with size > 1 and make one irregular repetitions from them.
    for (typename std::map <key_type, value_type>::iterator n = m_normalized.begin (); n != m_normalized.end (); ++n) {

      if (n->second.first.size () > 1) {

        //  need to normalize?
        db::Point p0 = n->second.first [0];
        if (p0 != db::Point ()) {
          for (std::vector<db::Point>::iterator p = n->second.first.begin (); p != n->second.first.end (); ++p) {
            *p -= p0;
          }
        }
          
        n->second.second.push_back (std::make_pair (p0, Repetition ()));

        IrregularRepetition *iterated_rep = new IrregularRepetition ();
        iterated_rep->points ().swap (n->second.first);

        n->second.second.back ().second.set_base (iterated_rep);

      }

    }

    m_norm_iter = m_normalized.begin ();
    if (! at_end ()) {
      m_disp_iter = m_norm_iter->second.first.begin ();
      m_rep_iter = m_norm_iter->second.second.begin ();
      tl_assert (m_disp_iter != m_norm_iter->second.first.end () || m_rep_iter != m_norm_iter->second.second.end ());
    }
  }

  virtual bool at_end () const 
  {
    return m_norm_iter == m_normalized.end ();
  }

  virtual void operator () (Sh &shape, db::Repetition &rep, db::PropertiesRepository::properties_id_type &prop_id)
  {
    shape = m_norm_iter->first.first;
    prop_id = m_norm_iter->first.second;

    if (m_disp_iter != m_norm_iter->second.first.end ()) {
      rep = db::Repetition ();
      shape.transform (db::Trans (*m_disp_iter));
    } else {
      tl_assert (m_rep_iter != m_norm_iter->second.second.end ());
      shape.transform (db::Trans (m_rep_iter->first));
      rep = m_rep_iter->second;
    }
  }

  virtual void inc ()
  {
    if (m_disp_iter != m_norm_iter->second.first.end ()) {
      ++m_disp_iter;
    } else {
      ++m_rep_iter;
    }

    if (m_disp_iter == m_norm_iter->second.first.end () && m_rep_iter == m_norm_iter->second.second.end ()) {

      ++m_norm_iter;

      if (! at_end ()) {
        m_disp_iter = m_norm_iter->second.first.begin ();
        m_rep_iter = m_norm_iter->second.second.begin ();
        tl_assert (m_disp_iter != m_norm_iter->second.first.end () || m_rep_iter != m_norm_iter->second.second.end ());
      }

    }
  }

private:
  typedef std::vector<db::Point> disp_vector;
  typedef std::vector<std::pair<db::Point, db::Repetition> > rep_vector;
  typedef std::pair<disp_vector, rep_vector> value_type;
  typedef std::pair<Sh, db::PropertiesRepository::properties_id_type> key_type;

  std::map <key_type, value_type> m_normalized;
  typename std::map <key_type, value_type>::const_iterator m_norm_iter;
  disp_vector::const_iterator m_disp_iter;
  rep_vector::const_iterator m_rep_iter;

  size_t m_ntot, m_nrep, m_nnonrep;
  unsigned int m_level;
};

/**
 *  @brief Delegate for all kind of deliveries
 */
template <class Sh>
class GenericShapeDelivery
  : public ShapeDelivery<Sh>
{
public:
  GenericShapeDelivery (const db::Shapes &shapes, unsigned int type, int compression_level)
    : ShapeDelivery<Sh> ()
  {
    if (compression_level <= 0) {
      mp_base = new StandardShapeDelivery<Sh> (shapes, type);
    } else {
      mp_base = new CompressingShapeDelivery<Sh> (shapes, type, compression_level - 1); 
    }
  }

  ~GenericShapeDelivery ()
  {
    delete mp_base;
    mp_base = 0;
  }

  virtual bool at_end () const 
  {
    return mp_base->at_end ();
  }

  virtual void operator () (Sh &shape, db::Repetition &rep, db::PropertiesRepository::properties_id_type &prop_id) 
  {
    (*mp_base) (shape, rep, prop_id);
  }

  virtual void inc ()
  {
    ++(*mp_base);
  }

private:
  ShapeDelivery<Sh> *mp_base;
};

}

// ---------------------------------------------------------------------------------
//  Generic delivery of instances (specialized to compressing / non-compressing variants)

namespace {

class InstDelivery
{
public:
  InstDelivery ()
  {
    //  .. nothing yet ..
  }

  virtual ~InstDelivery ()
  {
    //  .. nothing yet ..
  }

  virtual bool at_end () const = 0;

  virtual void operator () (db::CellInstArray &inst, db::Repetition &rep, db::PropertiesRepository::properties_id_type &prop_id) = 0;

  virtual InstDelivery &operator++ () 
  {
    inc ();
    return *this;
  }

protected:
  virtual void inc () = 0;
};

/**
 *  @brief Non-compressing standard instance delivery
 */
class StandardInstDelivery
  : public InstDelivery
{
public:
  StandardInstDelivery (const db::Cell &cell)
    : InstDelivery (), 
      m_inst_iterator (cell.begin ())
  {
    //  .. nothing yet ..
  }

  virtual bool at_end () const 
  {
    return m_inst_iterator.at_end ();
  }

  virtual void operator () (db::CellInstArray &inst, db::Repetition &rep, db::PropertiesRepository::properties_id_type &prop_id)
  {
    if (m_inst_iterator->is_complex ()) {
      inst = db::CellInstArray (db::CellInst (m_inst_iterator->cell_index ()), m_inst_iterator->complex_trans ());
    } else {
      inst = db::CellInstArray (db::CellInst (m_inst_iterator->cell_index ()), m_inst_iterator->front ());
    }

    db::Point a, b;
    unsigned long amax, bmax;

    bool is_reg = m_inst_iterator->is_regular_array (a, b, amax, bmax);
    if (is_reg && (amax > 1 || bmax > 1)) {

      rep = Repetition (new RegularRepetition (a, b, amax, bmax));

    } else {

      std::vector<db::Point> points;
      bool is_iterated = m_inst_iterator->is_iterated_array (& points); // @@@ 
      if (is_iterated && points.size () > 1) {

        //  need to normalize?
        db::Point p0 = points [0];
        if (p0 != db::Point ()) {
          inst.transform (db::Trans (p0));
          for (std::vector<db::Point>::iterator p = points.begin (); p != points.end (); ++p) {
            *p -= p0;
          }
        }
        
        IrregularRepetition *iterated_rep = new IrregularRepetition ();
        rep = Repetition (iterated_rep);
        iterated_rep->points ().swap (points);

      }

    }

    prop_id = m_inst_iterator->prop_id ();
  }

  virtual void inc ()
  {
    ++m_inst_iterator;
  }

private:
  db::Cell::const_iterator m_inst_iterator;
};

/**
 *  @brief Compressing instance delivery
 */
class CompressingInstDelivery
  : public InstDelivery
{
public:
  CompressingInstDelivery (const db::Cell &cell)
    : InstDelivery ()
  {
    //  count the instances per compression slot
    std::map <key_type, std::pair<size_t, size_t> > counts;
    for (db::Cell::const_iterator inst_iterator = cell.begin (); ! inst_iterator.at_end (); ++inst_iterator) {

      std::pair <key_type, std::pair<size_t, size_t> > key;
      key.first.second = inst_iterator->prop_id ();
      key.second = std::make_pair (size_t (0), size_t (0));

      if (inst_iterator->is_complex ()) {
        key.first.first = db::CellInstArray (db::CellInst (inst_iterator->cell_index ()), inst_iterator->complex_trans ());
      } else {
        key.first.first = db::CellInstArray (db::CellInst (inst_iterator->cell_index ()), inst_iterator->front ());
      }

      db::Point a, b;
      unsigned long amax, bmax;
      bool is_reg = inst_iterator->is_regular_array (a, b, amax, bmax);

      //  reduce by displacement and insert into key map
      key.first.first.transform (db::Trans (-inst_iterator->front ().disp ()));
      std::map <key_type, std::pair<size_t, size_t> >::iterator entry = counts.insert (key).first;

      //  only consider non-regular arrays for compression (count number of slots)
      if (! is_reg || (amax <= 1 && bmax <= 1)) {
        entry->second.first += inst_iterator->size ();
      } else {
        entry->second.second += 1;
      }

    }

    //  prepare vectors for the displacements
    for (std::map <key_type, std::pair<size_t, size_t> >::const_iterator c = counts.begin (); c != counts.end (); ++c) {
      std::pair <key_type, value_type> key;
      key.first = c->first;
      m_normalized.insert (key).first->second.first.reserve (c->second.first);
      m_normalized.insert (key).first->second.second.reserve (c->second.second + 1 /*potentially for the irregular rep.*/); 
    }

    //  collect the normalized shapes (with properties)
    for (db::Cell::const_iterator inst_iterator = cell.begin (); ! inst_iterator.at_end (); ++inst_iterator) {

      key_type key;
      key.second = inst_iterator->prop_id ();

      if (inst_iterator->is_complex ()) {
        key.first = db::CellInstArray (db::CellInst (inst_iterator->cell_index ()), inst_iterator->complex_trans ());
      } else {
        key.first = db::CellInstArray (db::CellInst (inst_iterator->cell_index ()), inst_iterator->front ());
      }

      db::Point a, b;
      unsigned long amax, bmax;
      bool is_reg = inst_iterator->is_regular_array (a, b, amax, bmax);

      //  reduce by displacement
      key.first.transform (db::Trans (-inst_iterator->front ().disp ()));

      if (! is_reg || (amax <= 1 && bmax <= 1)) {
        //  consider all instances of non-regular arrays for compression
        for (db::CellInstArray::iterator a = inst_iterator->begin (); ! a.at_end (); ++a) {
          m_normalized.find (key)->second.first.push_back ((*a).disp ());
        }
      } else {
        //  keep regular repetitions as such ..
        m_normalized.find (key)->second.second.push_back (std::make_pair (inst_iterator->front ().disp (), Repetition (new RegularRepetition (a, b, amax, bmax))));
      }

    }

    //  sort all irregular displacements with size > 1 and make one irregular repetitions from it.
    for (std::map <key_type, value_type>::iterator n = m_normalized.begin (); n != m_normalized.end (); ++n) {

      if (n->second.first.size () > 1) {

        std::sort (n->second.first.begin (), n->second.first.end ());

        //  need to normalize?
        db::Point p0 = n->second.first [0];
        if (p0 != db::Point ()) {
          for (std::vector<db::Point>::iterator p = n->second.first.begin (); p != n->second.first.end (); ++p) {
            *p -= p0;
          }
        }
          
        n->second.second.push_back (std::make_pair (p0, Repetition ()));

        IrregularRepetition *iterated_rep = new IrregularRepetition ();
        iterated_rep->points ().swap (n->second.first);

        n->second.second.back ().second.set_base (iterated_rep);

      }

    }

    m_norm_iter = m_normalized.begin ();
    if (! at_end ()) {
      m_disp_iter = m_norm_iter->second.first.begin ();
      m_rep_iter = m_norm_iter->second.second.begin ();
      tl_assert (m_disp_iter != m_norm_iter->second.first.end () || m_rep_iter != m_norm_iter->second.second.end ());
    }
  }

  virtual bool at_end () const 
  {
    return m_norm_iter == m_normalized.end ();
  }

  virtual void operator () (db::CellInstArray &inst, db::Repetition &rep, db::PropertiesRepository::properties_id_type &prop_id)
  {
    inst = m_norm_iter->first.first;
    prop_id = m_norm_iter->first.second;

    if (m_disp_iter != m_norm_iter->second.first.end ()) {

      rep = db::Repetition ();
      inst.transform (db::Trans (*m_disp_iter));

    } else {

      tl_assert (m_rep_iter != m_norm_iter->second.second.end ());
      inst.transform (db::Trans (m_rep_iter->first));
      rep = m_rep_iter->second;

    }
  }

  virtual void inc ()
  {
    if (m_disp_iter != m_norm_iter->second.first.end ()) {
      ++m_disp_iter;
    } else {
      ++m_rep_iter;
    }

    if (m_disp_iter == m_norm_iter->second.first.end () && m_rep_iter == m_norm_iter->second.second.end ()) {

      ++m_norm_iter;

      if (! at_end ()) {
        m_disp_iter = m_norm_iter->second.first.begin ();
        m_rep_iter = m_norm_iter->second.second.begin ();
        tl_assert (m_disp_iter != m_norm_iter->second.first.end () || m_rep_iter != m_norm_iter->second.second.end ());
      }

    }
  }

private:
  typedef std::vector<db::Point> disp_vector;
  typedef std::vector<std::pair<db::Point, db::Repetition> > rep_vector;
  typedef std::pair<disp_vector, rep_vector> value_type;
  typedef std::pair<db::CellInstArray, db::PropertiesRepository::properties_id_type> key_type;

  std::map <key_type, value_type> m_normalized;
  std::map <key_type, value_type>::const_iterator m_norm_iter;
  disp_vector::const_iterator m_disp_iter;
  rep_vector::const_iterator m_rep_iter;
};

/**
 *  @brief Delegate for all kind of deliveries
 */
class GenericInstDelivery
  : public InstDelivery
{
public:
  GenericInstDelivery (const db::Cell &cell, int compression_level)
    : InstDelivery ()
  {
    if (compression_level <= 0) {
      mp_base = new StandardInstDelivery (cell);
    } else {
      mp_base = new CompressingInstDelivery (cell); 
    }
  }

  ~GenericInstDelivery ()
  {
    delete mp_base;
    mp_base = 0;
  }

  virtual bool at_end () const 
  {
    return mp_base->at_end ();
  }

  virtual void operator () (db::CellInstArray &inst, db::Repetition &rep, db::PropertiesRepository::properties_id_type &prop_id) 
  {
    (*mp_base) (inst, rep, prop_id);
  }

  virtual void inc ()
  {
    ++(*mp_base);
  }

private:
  InstDelivery *mp_base;
};

}

// ---------------------------------------------------------------------------------
//  OASISWriter implementation

OASISWriter::OASISWriter ()
  : mp_stream (0),
    m_progress (tl::translate ("Writing OASIS file"), 0, 10000)
{
  m_progress.set_format (tl::translate ("%.0f MB"));
  m_progress.set_res (1.0);
  m_progress.set_unit (1024 * 1024);
}

void 
OASISWriter::write_byte (unsigned char b)
{
  mp_stream->put ((const char *) &b, 1);
}

void 
OASISWriter::write (long n)
{
  if (n < 0) {
    write (((unsigned long) (-n) << 1) | 1);
  } else {
    write ((unsigned long) n << 1);
  }
}

void 
OASISWriter::write (unsigned long n)
{
  do {
    unsigned char b = n & 0x7f;
    n >>= 7;
    if (n > 0) {
      b |= 0x80;
    }
    mp_stream->put ((const char *) &b, 1);
  } while (n > 0);
}

void
OASISWriter::write (float d) 
{
  if (fabs (d) >= 0.5 && fabs (floor (d + 0.5) - d) < 1e-6 && fabs (d) < double (std::numeric_limits<long>::max ())) {

    //  whole number (negative or positive)
    if (d < 0.0) {
      write_byte (1);
      write ((unsigned long) floor (-d + 0.5));
    } else { 
      write_byte (0);
      write ((unsigned long) floor (d + 0.5));
    }

  } else {

    write_byte (6);

    //  4-Byte IEEE real
    union {
      float d;
      uint32_t i;
    } f2i;

    f2i.d = d;
    uint32_t i = f2i.i;
    char b[sizeof (f2i.i)];
    for (unsigned int n = sizeof (f2i.i); n > 0; --n) {
      b[n - 1] = i & 0xff; 
      i >>= 8;
    }
    mp_stream->put (b, sizeof (f2i.i));

  }
}

void
OASISWriter::write (double d) 
{
  if (fabs (d) >= 0.5 && fabs (floor (d + 0.5) - d) < 1e-10 && fabs (d) < double (std::numeric_limits<long>::max ())) {

    //  whole number (negative or positive)
    if (d < 0.0) {
      write_byte (1);
      write ((unsigned long) floor (-d + 0.5));
    } else { 
      write_byte (0);
      write ((unsigned long) floor (d + 0.5));
    }

  } else {

    write_byte (7);

    //  8-Byte IEEE real
    union {
      double d;
      uint64_t i;
    } f2i;

    f2i.d = d;
    uint64_t i = f2i.i;
    char b[sizeof (f2i.i)];
    for (unsigned int n = 0; n < sizeof (f2i.i); n++) {
      b[n] = i & 0xff; 
      i >>= 8;
    }
    mp_stream->put (b, sizeof (f2i.i));

  }
}

void 
OASISWriter::write_bstring (const char *s)
{
  unsigned long l = strlen (s);
  write (l);
  mp_stream->put (s, l);
}

void  
OASISWriter::write_astring (const char *s)
{
  unsigned long l = strlen (s);

  write (l);

  bool valid = true;
  for (const char *c = s; *c && valid; ++c) {
    if (*c < 0x20 || *c > 0x7e) {
      valid = false;
    }
  }

  if (valid) {
    mp_stream->put (s, l);
  } else {
    for (const char *c = s; *c; ++c) {
      if (*c < 0x20 || *c > 0x7e) {
        write_byte ('*');
      } else {
        mp_stream->put (c, 1);
      }
    }
  }
}

void  
OASISWriter::write_nstring (const char *s)
{
  unsigned long l = strlen (s);

  if (!l) {
    write_astring ("*");
  } else {

    write (l);

    bool valid = true;
    for (const char *c = s; *c && valid; ++c) {
      if (*c < 0x21 || *c > 0x7e) {
        valid = false;
      }
    }

    if (valid) {
      mp_stream->put (s, l);
    } else {
      for (const char *c = s; *c; ++c) {
        if (*c < 0x21 || *c > 0x7e) {
          write_byte ('*');
        } else {
          mp_stream->put (c, 1);
        }
      }
    }

  }
}

template <class R>
inline R safe_scale (double sf, R value)
{
  double i = floor (sf * value + 0.5);
  if (i < double (std::numeric_limits<R>::min ())) {
    throw tl::Exception ("Scaling failed: coordinate underflow");
  }
  if (i > double (std::numeric_limits<R>::max ())) {
    throw tl::Exception ("Scaling failed: coordinate overflow");
  }
  return R (i);
}

void
OASISWriter::write_gdelta (const db::Point &p, double sf)
{
  db::Coord x = p.x ();
  db::Coord y = p.y ();

  if (sf != 1.0) {
    x = safe_scale (sf, x);
    y = safe_scale (sf, y);
  }

  if (x == -y || x == y || x == 0 || y == 0) {

    unsigned long dir = 0;
    unsigned long l = 0;

    if (x > 0) {
      l = x;
      if (y == 0) {
        dir = 0;
      } else if (y < 0) {
        dir = 7;
      } else {
        dir = 4;
      }
    } else if (x == 0) {
      if (y < 0) {
        l = -y;
        dir = 3;
      } else {
        l = y;
        dir = 1;
      }
    } else if (x < 0) {
      l = -x;
      if (y == 0) {
        dir = 2;
      } else if (y < 0) {
        dir = 6;
      } else {
        dir = 5;
      }
    }

    write ((l << 4) | (dir << 1));

  } else {

    unsigned long d;
    if (x < 0) {
      d = ((unsigned long) -x << 2) | 3;
    } else {
      d = ((unsigned long) x << 2) | 1;
    }
    write (d);
    write (y);

  }
}

void
OASISWriter::write_coord (db::Coord c, double sf)
{
  if (sf == 1.0) {
    return write (c);
  } else {
    return write (safe_scale (sf, c));
  }
}

void
OASISWriter::write_ucoord (db::Coord c, double sf)
{
  if (sf == 1.0) {
    return write ((unsigned long) c);
  } else {
    return write (safe_scale (sf, (unsigned long) c));
  }
}

void
OASISWriter::write_coord (db::Coord c)
{
  if (m_sf == 1.0) {
    return write (c);
  } else {
    return write (safe_scale (m_sf, c));
  }
}

void
OASISWriter::write_ucoord (db::Coord c)
{
  if (m_sf == 1.0) {
    return write ((unsigned long) c);
  } else {
    return write (safe_scale (m_sf, (unsigned long) c));
  }
}

static bool 
is_valid_prop (const db::LayerProperties &prop, void *)
{
  return prop.layer >= 0 && prop.datatype >= 0;
}

void
OASISWriter::emit_property_def (const db::Layout &layout, db::PropertiesRepository::properties_id_type prop_id, unsigned int &propname_id, unsigned int &propstring_id)
{
  const db::PropertiesRepository::properties_set &props = layout.properties_repository ().properties (prop_id);
  for (db::PropertiesRepository::properties_set::const_iterator p = props.begin (); p != props.end (); ++p) {

    const tl::Variant &name = layout.properties_repository ().prop_name (p->first);
    const char *name_str = "S_GDS_PROPERTY";
    if (! name.is_long () || (name.to_long () >= 0x8000 && name.to_long () < 0)) {
      name_str = name.to_string ();
    }
    if (m_propnames.insert (std::make_pair (name_str, propname_id)).second) {
      write_byte (7);
      write_nstring (name_str);
      ++propname_id;
    }

    if (p->second.is_string ()) {
      if (m_propstrings.insert (std::make_pair (p->second.to_string (), propstring_id)).second) {
        write_byte (9);
        write_bstring (p->second.to_string ());
        ++propstring_id;
      }
    }

  }
}

void 
OASISWriter::write (const db::Layout &layout, tl::OutputStream &stream, const db::SaveLayoutOptions &options) throw (tl::Exception)
{
  const OASISWriterOptions *oasis_options = dynamic_cast <const OASISWriterOptions *> (options.get_specific_options ());
  if (oasis_options) {
    m_options = *oasis_options;
  } else {
    m_options = OASISWriterOptions ();
  }

  mp_stream = &stream;

  double dbu = (options.dbu () == 0.0) ? layout.dbu () : options.dbu ();
  m_sf = options.scale_factor () * (layout.dbu () / dbu);
  if (fabs (m_sf - 1.0) < 1e-9) {
    //  to avoid rounding problems, set to 1.0 exactly if possible.
    m_sf = 1.0;
  }

  std::vector <std::pair <unsigned int, db::LayerProperties> > layers;
  options.get_valid_layers (layout, layers, &is_valid_prop);

  std::vector <db::cell_index_type> cells;
  options.get_cells (layout, cells);

  //  write header

  char magic[] = "%SEMI-OASIS\015\012";
  mp_stream->put (magic, sizeof (magic) - 1);

  //  START record
  write_byte (1); 
  write_bstring ("1.0");
  write (1.0 / dbu);
  write_byte (0);  //  offset-flag
  //  offset table:
  for (unsigned int i = 0; i < 12; ++i) {
    write_byte (0);
  }

  //  Prepare name tables
  unsigned int id;
 
  m_textstrings.clear ();
  m_propnames.clear ();
  m_propstrings.clear ();

  for (std::vector<db::cell_index_type>::const_iterator cell = cells.begin (); cell != cells.end (); ++cell) {
    //  CELLNAME (explicit)
    write_byte (4);
    write_nstring (layout.cell_name (*cell));
    write ((unsigned long) *cell);
  }

  //  build text string table

  id = 0;
  for (std::vector<db::cell_index_type>::const_iterator cell = cells.begin (); cell != cells.end (); ++cell) {

    const db::Cell &cref (layout.cell (*cell));
    for (std::vector <std::pair <unsigned int, db::LayerProperties> >::const_iterator l = layers.begin (); l != layers.end (); ++l) {
      db::ShapeIterator shape (cref.shapes (l->first).begin (db::ShapeIterator::Texts));
      while (! shape.at_end ()) {
        if (m_textstrings.insert (std::make_pair (shape->text_string (), id)).second) {
          write_byte (5);
          write_astring (shape->text_string ());
          ++id;
          m_progress.set (mp_stream->pos ());
        }
        ++shape;
      }
    }

  }

  //  build property name and value string tables

  unsigned int propname_id = 0;
  unsigned int propstring_id = 0;
  std::set <unsigned long> prop_ids_done;

  for (std::vector<db::cell_index_type>::const_iterator cell = cells.begin (); cell != cells.end (); ++cell) {

    const db::Cell &cref (layout.cell (*cell));

    for (db::Cell::const_iterator inst = cref.begin (); ! inst.at_end (); ++inst) {
      if (inst->has_prop_id () && inst->prop_id () != 0 && prop_ids_done.find (inst->prop_id ()) == prop_ids_done.end ()) {
        prop_ids_done.insert (inst->prop_id ());
        emit_property_def (layout, inst->prop_id (), propname_id, propstring_id);
        m_progress.set (mp_stream->pos ());
      }
    }

    for (std::vector <std::pair <unsigned int, db::LayerProperties> >::const_iterator l = layers.begin (); l != layers.end (); ++l) {
      db::ShapeIterator shape (cref.shapes (l->first).begin (db::ShapeIterator::Boxes | db::ShapeIterator::Polygons | db::ShapeIterator::Edges | db::ShapeIterator::Paths | db::ShapeIterator::Texts));
      while (! shape.at_end ()) {
        if (shape->has_prop_id () && shape->prop_id () != 0 && prop_ids_done.find (shape->prop_id ()) == prop_ids_done.end ()) {
          prop_ids_done.insert (shape->prop_id ());
          emit_property_def (layout, shape->prop_id (), propname_id, propstring_id);
          m_progress.set (mp_stream->pos ());
        }
        ++shape;
      }
    }

  }

  //  write layernames table
  
  for (std::vector <std::pair <unsigned int, db::LayerProperties> >::const_iterator l = layers.begin (); l != layers.end (); ++l) {

    if (! l->second.name.empty ()) {

      //  write mappings to text layer and shape layers
      write_byte (11);
      write_nstring (l->second.name.c_str ());
      write_byte (3);
      write ((unsigned long) l->second.layer);
      write_byte (3);
      write ((unsigned long) l->second.datatype);

      write_byte (12);
      write_nstring (l->second.name.c_str ());
      write_byte (3);
      write ((unsigned long) l->second.layer);
      write_byte (3);
      write ((unsigned long) l->second.datatype);

      m_progress.set (mp_stream->pos ());

    }

  }

  //  body

  for (std::vector<db::cell_index_type>::const_iterator cell = cells.begin (); cell != cells.end (); ++cell) {

    //  reset modal variables
    mm_repetition.reset ();
    mm_placement_cell.reset ();
    mm_placement_x = 0;
    mm_placement_y = 0;
    mm_layer.reset ();
    mm_datatype.reset ();
    mm_textlayer.reset ();
    mm_texttype.reset ();
    mm_text_x = 0;
    mm_text_y = 0;
    mm_text_string.reset ();
    mm_geometry_x = 0;
    mm_geometry_y = 0;
    mm_geometry_w.reset ();
    mm_geometry_h.reset ();
    mm_polygon_point_list.reset ();
    mm_path_halfwidth.reset ();
    mm_path_start_extension.reset ();
    mm_path_end_extension.reset ();
    mm_path_point_list.reset ();
    mm_ctrapezoid_type.reset ();
    mm_circle_radius.reset ();
    mm_last_property_name.reset ();
    mm_last_property_is_sprop.reset ();
    mm_last_value_list.reset ();

    m_progress.set (mp_stream->pos ());

    //  cell header 

    write_byte (13);  // CELL
    write ((unsigned long) *cell);

    //  cell body 
    const db::Cell &cref (layout.cell (*cell));

    //  instances
    write_insts (layout, cref);

    //  shapes
    for (std::vector <std::pair <unsigned int, db::LayerProperties> >::const_iterator l = layers.begin (); l != layers.end (); ++l) {
      write_texts (layout, cref, l->second, l->first);
      write_polygons (layout, cref, l->second, l->first);
      write_edges (layout, cref, l->second, l->first);
      write_paths (layout, cref, l->second, l->first);
      write_boxes (layout, cref, l->second, l->first);
      m_progress.set (mp_stream->pos ());
    }

    //  end of cell

  }

  //  END record

  write_byte (2);
  unsigned int pad = 256 - 4;
  write (pad);
  for (unsigned int i = 0; i < pad; ++i) {
    write_byte (0);
  }
  write_byte (0);

  m_progress.set (mp_stream->pos ());

}

void 
OASISWriter::write (const Repetition &rep)
{
  if (mm_repetition == rep) {
    write_byte (0);  // reuse
  } else {

    mm_repetition = rep;

    db::Point a, b;
    size_t amax, bmax;

    bool is_reg = rep.is_regular (a, b, amax, bmax);
    const std::vector<db::Point> *iterated = rep.is_iterated ();

    if (iterated) {

      tl_assert (iterated->size () >= 2);
      tl_assert ((*iterated)[0] == db::Point ());

      //  extract common grid
      db::Coord g = 0;
      for (std::vector<db::Point>::const_iterator p = iterated->begin () + 1; p != iterated->end (); ++p) {

        db::Coord x = safe_scale (m_sf, p->x ());
        if (x < 0) {
          x = -x;
        }
        if (x != 0) {
          g = (g == 0) ? x : tl::lcd (g, x);
        }

        db::Coord y = safe_scale (m_sf, p->y ());
        if (y < 0) {
          y = -y;
        }
        if (y != 0) {
          g = (g == 0) ? y : tl::lcd (g, y);
        }

      }

      if (g == 0) {
        write_byte (10);
        write (iterated->size () - 2);
        g = 1;
      } else {
        write_byte (11);
        write (iterated->size () - 2);
        write_ucoord (g, 1.0);
      }

      db::Point last_point;
      for (std::vector<db::Point>::const_iterator p = iterated->begin () + 1; p != iterated->end (); ++p) {
        db::Point s (safe_scale (m_sf, p->x ()), safe_scale (m_sf, p->y ()));
        db::Point delta = s - last_point;
        last_point = s;
        write_gdelta (db::Point (delta.x () / g, delta.y () / g), 1.0);
      }

    } else {

      tl_assert (is_reg);

      //  currently there are only regular repetitions
      //  TODO: optimize for orthogonal cases 
      tl_assert (is_reg);
      tl_assert (amax >= 2 || bmax >= 2);

      if (amax == 1 || bmax == 1) {

        if (bmax == 1) {
          b = a;
          bmax = amax;
        }

        if (b.x () == 0 && b.y () >= 0) {
          write_byte (3);
          write (bmax - 2);
          write_ucoord (b.y ());
        } else if (b.y () == 0 && b.x () >= 0) {
          write_byte (2);
          write (bmax - 2);
          write_ucoord (b.x ());
        } else {
          write_byte (9);
          write (bmax - 2);
          write_gdelta (b);
        }

      } else if (b.x () == 0 && b.y () >= 0 && a.y () == 0 && a.x () >= 0) {

        write_byte (1);
        write (amax - 2);
        write (bmax - 2);
        write_ucoord (a.x ());
        write_ucoord (b.y ());

      } else if (b.y () == 0 && b.x () >= 0 && a.x () == 0 && a.y () >= 0) {

        write_byte (1);
        write (bmax - 2);
        write (amax - 2);
        write_ucoord (b.x ());
        write_ucoord (a.y ());

      } else {

        write_byte (8);
        write (amax - 2);
        write (bmax - 2);
        write_gdelta (a);
        write_gdelta (b);

      }

    }

  }
}

void 
OASISWriter::write_insts (const db::Layout &layout, const db::Cell &cell)
{
  for (GenericInstDelivery inst_delivery (cell, m_options.compression_level); ! inst_delivery.at_end (); ++inst_delivery) {

    Repetition rep;
    db::PropertiesRepository::properties_id_type prop_id;
    db::CellInstArray inst;

    inst_delivery (inst, rep, prop_id);

    m_progress.set (mp_stream->pos ());

    db::Trans t = inst.front ();

    unsigned char info = 0x40;  // by reference number
    if (mm_placement_cell != inst.object ().cell_index ()) {
      info |= 0x80;
    }
    if (mm_placement_x != t.disp ().x ()) {
      info |= 0x20;
    }
    if (mm_placement_y != t.disp ().y ()) {
      info |= 0x10;
    }
    if (rep != Repetition ()) {
      info |= 0x08;
    }

    if (t.is_mirror ()) {
      info |= 0x01;
    }

    if (inst.is_complex ()) {
      write_byte (18);
      write_byte (info | 0x06);
    } else {
      write_byte (17);
      write_byte (info | ((t.rot () & 0x03) << 1));
    }

    if (info & 0x80) {
      mm_placement_cell = inst.object ().cell_index ();
      write ((unsigned long) mm_placement_cell.get ());
    }

    if (inst.is_complex ()) {
      write (inst.complex_trans ().mag ());
      write (inst.complex_trans ().angle ());
    }
    
    if (info & 0x20) {
      mm_placement_x = t.disp ().x ();
      write_coord (mm_placement_x.get ());
    }
    if (info & 0x10) {
      mm_placement_y = t.disp ().y ();
      write_coord (mm_placement_y.get ());
    }

    if (info & 0x08) {
      write (rep);
    }

    if (prop_id != 0) {
      write_props (layout, layout.properties_repository ().properties (prop_id));
    }

  }
}

void
OASISWriter::write_props (const db::Layout &layout, const db::PropertiesRepository::properties_set &props)
{
  std::vector<tl::Variant> pv_list; 

  for (db::PropertiesRepository::properties_set::const_iterator p = props.begin (); p != props.end (); ++p) {

    m_progress.set (mp_stream->pos ());

    const tl::Variant &name = layout.properties_repository ().prop_name (p->first);

    const char *name_str = "S_GDS_PROPERTY";
    bool sflag = true;

    pv_list.clear ();
    const std::vector<tl::Variant> *pvl = &pv_list;

    if (! name.is_long () || (name.to_long () >= 0x8000 && name.to_long () < 0)) {

      name_str = name.to_string ();
      sflag = false;

      if (p->second.is_list ()) {
        pvl = &p->second.get_list ();
      } else if (!p->second.is_nil ()) {
        pv_list.reserve (1);
        pv_list.push_back (p->second);
      }

    } else {

      pv_list.reserve (2);
      pv_list.push_back (name.to_long ());
      pv_list.push_back (p->second.to_string ());

    }

    bool same_name = (mm_last_property_name == name_str);
    bool same_value = (mm_last_value_list == *pvl);
    bool same_sflag = (mm_last_property_is_sprop == sflag);

    if (same_name && same_value && same_sflag) {
      write_byte (29); // repeat property 
    } else {

      write_byte (28);

      unsigned char info = sflag ? 1 : 0;
      if (same_value) {
        info |= 0x08;
      } else {
        if (pvl->size () >= 15) {
          info |= 0xf0;
        } else {
          info |= ((unsigned char)pvl->size ()) << 4;
        }
      }

      if (! same_name) {
        info |= 0x06;
      }

      write_byte (info);

      if (! same_name) {
        std::map <std::string, unsigned long>::const_iterator pni = m_propnames.find (name_str);
        tl_assert (pni != m_propnames.end ());
        write (pni->second);
        mm_last_property_name = name_str;
      }

      if (! same_value) {

        if (pvl->size () >= 15) {
          write ((unsigned long) pvl->size ());
        }

        //  write property values
        for (unsigned long i = 0; i < pvl->size (); ++i) {
          const tl::Variant &v = (*pvl)[i];
          if (v.is_double ()) {
            write (v.to_double ());
          } else if (v.is_long ()) {
            write_byte (9);
            write (v.to_long ());
          } else {
            const char *pvs = v.to_string ();
            std::map <std::string, unsigned long>::const_iterator pvi = m_propstrings.find (pvs);
            if (pvi != m_propstrings.end ()) {
              write_byte (14);
              write (pvi->second);
            } else {
              write_byte (11);
              write_bstring (pvs);
            }
          }
        }

        mm_last_value_list = *pvl;

      }

      mm_last_property_is_sprop = sflag;

    }

  }
}

void
OASISWriter::write_pointlist (const std::vector<db::Point> &pointlist, bool for_polygons)
{
  tl_assert ((for_polygons && pointlist.size () > 1) || (! for_polygons && pointlist.size () > 0));

  //  determine type: 0 (manhattan, horizontal first), 1 (manhattan, vert. first), -1 other
  db::Point plast (0, 0);
  int type = -1;
  int hvlast = -1;
  for (std::vector<db::Point>::const_iterator p = pointlist.begin (); p != pointlist.end (); ++p) {
    int hv = -1;
    if (p->x () == plast.x ()) {
      hv = 1;
    } else if (p->y () == plast.y ()) {
      hv = 0;
    } else {
      type = -1;
      break;
    }
    if (type == -1) {
      type = hv;
    } else if (hv != !hvlast) {
      type = -1;
      break;
    }
    hvlast = hv;
    plast = *p;
  }

  //  test last displacement for polygons
  if (for_polygons && type >= 0) {
    if (plast.x () == 0) {
      if (hvlast != 0) {
        type = -1;
      }
    } else if (plast.y () == 0) {
      if (hvlast != 1) {
        type = -1;
      }
    } else {
      type = -1;
    }
  }

  if (type == 0 || type == 1) {

    //  manhattan pointlist
    write_byte (type);
    size_t implicit = for_polygons ? 1 : 0;
    write ((unsigned long) (pointlist.size () - implicit));

    db::Point plast (0, 0);
    for (std::vector<db::Point>::const_iterator p = pointlist.begin (); p != pointlist.end () - implicit; ++p) {
      db::Coord d = p->x () - plast.x ();
      if (d == 0) {
        d = p->y () - plast.y ();
      }
      write_coord (d);
      plast = *p;
    }

  } else {

    //  generic pointlist
    write_byte (4);
    write ((unsigned long) pointlist.size ());
    db::Point plast (0, 0);
    for (std::vector<db::Point>::const_iterator p = pointlist.begin (); p != pointlist.end (); ++p) {
      write_gdelta (*p - plast);
      plast = *p;
    }

  }
}

void 
OASISWriter::write_texts (const db::Layout &layout, const db::Cell &cell, const db::LayerProperties &lprops, unsigned int lindex)
{
  int layer = lprops.layer;
  int datatype = lprops.datatype;

  for (GenericShapeDelivery<db::Text> shape_delivery (cell.shapes (lindex), db::ShapeIterator::Texts, m_options.compression_level); ! shape_delivery.at_end (); ++shape_delivery) {

    m_progress.set (mp_stream->pos ());

    db::Text text;
    db::PropertiesRepository::properties_id_type prop_id;
    db::Repetition rep;
    shape_delivery (text, rep, prop_id);

    db::Trans trans = text.trans ();
    std::map <std::string, unsigned long>::const_iterator ts = m_textstrings.find (text.string ());
    tl_assert (ts != m_textstrings.end ());
    unsigned long text_id = ts->second;

    unsigned char info = 0x20;

    if (mm_text_string != text.string ()) {
      info |= 0x40;
    }
    if (mm_textlayer != layer) {
      info |= 0x01;
    }
    if (mm_texttype != datatype) {
      info |= 0x02;
    }
    if (mm_text_x != trans.disp ().x ()) {
      info |= 0x10;
    }
    if (mm_text_y != trans.disp ().y ()) {
      info |= 0x08;
    }
    if (! rep.is_singular ()) {
      info |= 0x04;
    }

    write_byte (19);
    write_byte (info);
    if (info & 0x40) {
      mm_text_string = text.string ();
      write ((unsigned long) text_id);
    }
    if (info & 0x01) {
      mm_textlayer = layer;
      write ((unsigned long) layer);
    }
    if (info & 0x02) {
      mm_texttype = datatype;
      write ((unsigned long) datatype);
    }
    if (info & 0x10) {
      mm_text_x = trans.disp ().x ();
      write_coord (mm_text_x.get ());
    }
    if (info & 0x08) {
      mm_text_y = trans.disp ().y ();
      write_coord (mm_text_y.get ());
    }

    if (info & 0x04) {
      write (rep);
    }

    if (prop_id != 0) {
      write_props (layout, layout.properties_repository ().properties (prop_id));
    }

  }
}

void 
OASISWriter::write_polygon (int layer, int datatype, const db::Layout &layout, const db::Polygon &polygon, db::PropertiesRepository::properties_id_type prop_id, const db::Repetition &rep, std::vector<db::Point> &pointlist)
{
  //  @@@ how to deal with max vertex count?
 
  db::Polygon::polygon_contour_iterator e = polygon.begin_hull ();
  db::Point start = *e;
  pointlist.assign (++e, polygon.end_hull ());

  for (std::vector<db::Point>::iterator p = pointlist.begin (); p != pointlist.end (); ++p) {
    *p -= start;
  }

  if (pointlist.size () < 2) {
    throw tl::Exception (tl::translate ("Polygons with less than three points cannot be written to OASIS files"));
  }

  unsigned char info = 0x00;

  if (mm_layer != layer) {
    info |= 0x01;
  }
  if (mm_datatype != datatype) {
    info |= 0x02;
  }
  if (mm_geometry_x != start.x ()) {
    info |= 0x10;
  }
  if (mm_geometry_y != start.y ()) {
    info |= 0x08;
  }
  if (mm_polygon_point_list != pointlist) {
    info |= 0x20;
  }
  if (! rep.is_singular ()) {
    info |= 0x04;
  }

  write_byte (21);
  write_byte (info);

  if (info & 0x01) {
    mm_layer = layer;
    write ((unsigned long) layer);
  }
  if (info & 0x02) {
    mm_datatype = datatype;
    write ((unsigned long) datatype);
  }
  if (info & 0x20) {
    mm_polygon_point_list.swap (pointlist);
    write_pointlist (mm_polygon_point_list.get (), true /*for polygons*/);
  }
  if (info & 0x10) {
    mm_geometry_x = start.x ();
    write_coord (start.x ());
  }
  if (info & 0x08) {
    mm_geometry_y = start.y ();
    write_coord (start.y ());
  }
  if (info & 0x04) {
    write (rep);
  }

  if (prop_id != 0) {
    write_props (layout, layout.properties_repository ().properties (prop_id));
  }
}

void 
OASISWriter::write_polygons (const db::Layout &layout, const db::Cell &cell, const db::LayerProperties &lprops, unsigned int lindex)
{
  int layer = lprops.layer;
  int datatype = lprops.datatype;

  std::vector<db::Point> pointlist;

  for (GenericShapeDelivery<db::Polygon> shape_delivery (cell.shapes (lindex), db::ShapeIterator::Polygons, m_options.compression_level); ! shape_delivery.at_end (); ++shape_delivery) {

    m_progress.set (mp_stream->pos ());

    db::Polygon polygon;
    db::PropertiesRepository::properties_id_type prop_id;
    db::Repetition rep;
    shape_delivery (polygon, rep, prop_id);

    if (polygon.holes () > 0) {

      //  resolve holes ...
      std::vector<db::Polygon> polygons;

      db::EdgeProcessor ep;
      ep.insert_sequence (polygon.begin_edge ());
      db::PolygonContainer pc (polygons);
      db::PolygonGenerator out (pc, true /*resolve holes*/, false /*max coherence*/);
      db::SimpleMerge op;
      ep.process (out, op);

      for (std::vector<db::Polygon>::const_iterator p = polygons.begin (); p != polygons.end (); ++p) {
        write_polygon (layer, datatype, layout, *p, prop_id, rep, pointlist);
      }

    } else {
      write_polygon (layer, datatype, layout, polygon, prop_id, rep, pointlist);
    }
   
  }
}

void 
OASISWriter::write_boxes (const db::Layout &layout, const db::Cell &cell, const db::LayerProperties &lprops, unsigned int lindex)
{
  int layer = lprops.layer;
  int datatype = lprops.datatype;

  for (GenericShapeDelivery<db::Box> shape_delivery (cell.shapes (lindex), db::ShapeIterator::Boxes, m_options.compression_level); ! shape_delivery.at_end (); ++shape_delivery) {

    m_progress.set (mp_stream->pos ());

    db::Box box;
    db::PropertiesRepository::properties_id_type prop_id;
    db::Repetition rep;
    shape_delivery (box, rep, prop_id);

    unsigned char info = 0x00;
    
    if (mm_layer != layer) {
      info |= 0x01;
    }
    if (mm_datatype != datatype) {
      info |= 0x02;
    }

    if (box.width () == box.height ()) {
      info |= 0x80;  // square
    } else {
      if (mm_geometry_h != box.height ()) {
        info |= 0x20;
      }
    }
    if (mm_geometry_w != box.width ()) {
      info |= 0x40;
    }

    if (mm_geometry_x != box.left ()) {
      info |= 0x10;
    }
    if (mm_geometry_y != box.bottom ()) {
      info |= 0x08;
    }
    
    if (! rep.is_singular ()) {
      info |= 0x04;
    }

    write_byte (20);
    write_byte (info);

    if (info & 0x01) {
      mm_layer = layer;
      write ((unsigned long) layer);
    }
    if (info & 0x02) {
      mm_datatype = datatype;
      write ((unsigned long) datatype);
    }

    mm_geometry_w = box.width ();
    mm_geometry_h = box.height ();

    if (info & 0x40) {
      write_ucoord (mm_geometry_w.get ());
    }
    if (info & 0x20) {
      write_ucoord (mm_geometry_h.get ());
    }

    if (info & 0x10) {
      mm_geometry_x = box.left ();
      write_coord (mm_geometry_x.get ());
    }
    if (info & 0x08) {
      mm_geometry_y = box.bottom ();
      write_coord (mm_geometry_y.get ());
    }

    if (info & 0x04) {
      write (rep);
    }

    if (prop_id != 0) {
      write_props (layout, layout.properties_repository ().properties (prop_id));
    }

  }
}

void 
OASISWriter::write_paths (const db::Layout &layout, const db::Cell &cell, const db::LayerProperties &lprops, unsigned int lindex)
{
  int layer = lprops.layer;
  int datatype = lprops.datatype;

  std::vector<db::Point> pointlist;

  for (GenericShapeDelivery<db::Path> shape_delivery (cell.shapes (lindex), db::ShapeIterator::Paths, m_options.compression_level); ! shape_delivery.at_end (); ++shape_delivery) {

    m_progress.set (mp_stream->pos ());

    db::Path path;
    db::PropertiesRepository::properties_id_type prop_id;
    db::Repetition rep;
    shape_delivery (path, rep, prop_id);

    std::pair<db::Path::coord_type, db::Path::coord_type> ext (0, 0);
    //  for round paths, circles are places to mimic the extensions
    if (! path.round ()) {
      ext = path.extensions ();
    }

    db::Path::iterator e = path.begin ();
    db::Point start = *e;
    pointlist.assign (++e, path.end ());
    db::Point end = pointlist.back ();

    for (std::vector<db::Point>::iterator p = pointlist.begin (); p != pointlist.end (); ++p) {
      *p -= start;
    }

    if (pointlist.size () < 1) {
      throw tl::Exception (tl::translate ("Paths with less than two points cannot be written to OASIS files"));
    }

    db::Coord hw = path.width () / 2;
    if (hw * 2 != path.width ()) {
      throw tl::Exception (tl::translate ("Paths with odd width cannot be written to OASIS files"));
    }

    unsigned char info = 0x00;

    if (mm_layer != layer) {
      info |= 0x01;
    }
    if (mm_datatype != datatype) {
      info |= 0x02;
    }
    if (mm_geometry_x != start.x ()) {
      info |= 0x10;
    }
    if (mm_geometry_y != start.y ()) {
      info |= 0x08;
    }
    if (mm_path_point_list != pointlist) {
      info |= 0x20;
    }
    if (mm_path_start_extension != ext.first || mm_path_end_extension != ext.second) {
      info |= 0x80;
    }
    if (mm_path_halfwidth != hw) {
      info |= 0x40;
    }

    if (! rep.is_singular ()) {
      info |= 0x04;
    }

    write_byte (22);
    write_byte (info);

    if (info & 0x01) {
      mm_layer = layer;
      write ((unsigned long) layer);
    }
    if (info & 0x02) {
      mm_datatype = datatype;
      write ((unsigned long) datatype);
    }
    if (info & 0x40) {
      mm_path_halfwidth = hw;
      write_ucoord (hw);
    }

    if (info & 0x80) {

      unsigned char ext_scheme = 0;
      if (mm_path_start_extension == ext.first) {
        //  00
      } else if (ext.first == 0) {
        ext_scheme |= 0x04;
      } else if (ext.first == hw) {
        ext_scheme |= 0x08;
      } else {
        ext_scheme |= 0x0c;
      }
      if (mm_path_end_extension == ext.second) {
        //  00
      } else if (ext.second == 0) {
        ext_scheme |= 0x01;
      } else if (ext.second == hw) {
        ext_scheme |= 0x02;
      } else {
        ext_scheme |= 0x03;
      }

      write_byte (ext_scheme);

      if ((ext_scheme & 0x0c) == 0x0c) {
        write_coord (ext.first);
      }
      if ((ext_scheme & 0x03) == 0x03) {
        write_coord (ext.second);
      }

      mm_path_start_extension = ext.first;
      mm_path_end_extension = ext.second;

    }

    if (info & 0x20) {
      mm_path_point_list.swap (pointlist);
      write_pointlist (mm_path_point_list.get (), false /*=for paths*/);
    }
    if (info & 0x10) {
      mm_geometry_x = start.x ();
      write_coord (start.x ());
    }
    if (info & 0x08) {
      mm_geometry_y = start.y ();
      write_coord (start.y ());
    }

    if (info & 0x04) {
      write (rep);
    }

    if (prop_id != 0) {
      write_props (layout, layout.properties_repository ().properties (prop_id));
    }

    if (path.round ()) {

      //  write two circles at the path ends to mimic the round path ends.

      unsigned char info = 0;
      if (mm_circle_radius != hw) {
        info |= 0x20;
      }
      if (mm_geometry_x != start.x ()) {
        info |= 0x10;
      }
      if (mm_geometry_y != start.y ()) {
        info |= 0x08;
      }

      if (! rep.is_singular ()) {
        info |= 0x04;
      }

      write_byte (27);
      write_byte (info);

      if (info & 0x20) {
        mm_circle_radius = hw;
        write_ucoord (hw);
      }
      if (info & 0x10) {
        mm_geometry_x = start.x ();
        write_coord (start.x ());
      }
      if (info & 0x08) {
        mm_geometry_y = start.y ();
        write_coord (start.y ());
      }

      if (info & 0x04) {
        write (rep);
      }

      if (prop_id != 0) {
        write_props (layout, layout.properties_repository ().properties (prop_id));
      }

      info = 0;
      if (mm_geometry_x != end.x ()) {
        info |= 0x10;
      }
      if (mm_geometry_y != end.y ()) {
        info |= 0x08;
      }

      if (! rep.is_singular ()) {
        info |= 0x04;
      }

      write_byte (27);
      write_byte (info);

      if (info & 0x10) {
        mm_geometry_x = end.x ();
        write_coord (end.x ());
      }
      if (info & 0x08) {
        mm_geometry_y = end.y ();
        write_coord (end.y ());
      }

      if (info & 0x04) {
        write (rep);
      }

      if (prop_id != 0) {
        write_props (layout, layout.properties_repository ().properties (prop_id));
      }

    }

  }
}

void 
OASISWriter::write_edges (const db::Layout &layout, const db::Cell &cell, const db::LayerProperties &lprops, unsigned int lindex)
{
  int layer = lprops.layer;
  int datatype = lprops.datatype;

  db::ShapeIterator shape;
  shape = cell.shapes (lindex).begin (db::ShapeIterator::Edge);

  std::vector<db::Point> pointlist;

  while (! shape.at_end ()) {

    m_progress.set (mp_stream->pos ());

    db::Edge e (shape->edge ());
    pointlist.reserve (1);
    pointlist.erase (pointlist.begin (), pointlist.end ());
    pointlist.push_back (e.p2 () - e.p1 ());

    unsigned char info = 0x00;

    if (mm_layer != layer) {
      info |= 0x01;
    }
    if (mm_datatype != datatype) {
      info |= 0x02;
    }
    if (mm_geometry_x != e.p1 ().x ()) {
      info |= 0x10;
    }
    if (mm_geometry_y != e.p1 ().y ()) {
      info |= 0x08;
    }
    if (mm_path_point_list != pointlist) {
      info |= 0x20;
    }
    if (mm_path_start_extension != 0 || mm_path_end_extension != 0) {
      info |= 0x80;
    }
    if (mm_path_halfwidth != 0) {
      info |= 0x40;
    }

    write_byte (21);
    write_byte (info);

    if (info & 0x01) {
      mm_layer = layer;
      write ((unsigned long) layer);
    }
    if (info & 0x02) {
      mm_datatype = datatype;
      write ((unsigned long) datatype);
    }
    if (info & 0x40) {
      mm_path_halfwidth = 0;
      write (0);
    }

    if (info & 0x80) {
      write_byte (0x05);  // flush
      mm_path_start_extension = 0;
      mm_path_end_extension = 0;
    }

    if (info & 0x20) {
      mm_path_point_list = pointlist;
      write_pointlist (pointlist, false /*=for paths*/);
    }
    if (info & 0x10) {
      mm_geometry_x = e.p1 ().x ();
      write_coord (e.p1 ().x ());
    }
    if (info & 0x08) {
      mm_geometry_y = e.p1 ().y ();
      write_coord (e.p1 ().y ());
    }

    if (shape->has_prop_id () && shape->prop_id () != 0) {
      write_props (layout, layout.properties_repository ().properties (shape->prop_id ()));
    }

    ++shape;

  }
}

}

