
/*

  KLayout Layout Viewer
  Copyright (C) 2006-2007 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#include "dbPCellHeader.h"
#include "dbPCellVariant.h"

namespace db
{

// ----------------------------------------------------------------------------------------
//  PCellParametersCompareFunc implementation

bool 
PCellParametersCompareFunc::operator() (const pcell_parameters_type *a, const pcell_parameters_type *b)
{
  if (a->size () != b->size ()) {
    return a->size () < b->size ();
  }

  for (size_t i = 0; i < a->size (); ++i) {
    // @@@ KLUDGE: string comparison is slow ...
    int s = strcmp ((*a)[i].to_string (), (*b)[i].to_string ());
    if (s != 0) {
      return s < 0;
    }
  }

  return 0;
}
  
// ----------------------------------------------------------------------------------------
//  PCellHeader implementation

PCellHeader::PCellHeader (size_t pcell_id, const std::string &name, PCellDeclaration *declaration)
  : mp_declaration (declaration), m_pcell_id (pcell_id), m_name (name)
{
  if (mp_declaration) {
    mp_declaration->add_ref ();
    m_layer_declarations = mp_declaration->get_layer_declarations ();
  }
}

PCellHeader::PCellHeader (const PCellHeader &d)
  : mp_declaration (d.mp_declaration), m_pcell_id (d.m_pcell_id), m_name (d.m_name)
{
  if (mp_declaration) {
    mp_declaration->add_ref ();
    m_layer_declarations = mp_declaration->get_layer_declarations ();
  }
}

PCellHeader::~PCellHeader ()
{
  if (mp_declaration) {
    mp_declaration->release_ref ();
  }
  mp_declaration = 0;
}

void 
PCellHeader::declaration (PCellDeclaration *declaration)
{
  if (mp_declaration) {
    mp_declaration->release_ref ();
  }
  mp_declaration = declaration;
  if (declaration) {
    mp_declaration->add_ref ();
    m_layer_declarations = declaration->get_layer_declarations ();
  } else {
    m_layer_declarations.clear ();
  }
}

std::vector<unsigned int> 
PCellHeader::get_layer_indices (db::Layout &layout)
{
  if (! declaration ()) {
    return std::vector<unsigned int> ();
  }

  bool reuse_layer_list = (m_layer_indices.size () == m_layer_declarations.size ());
  for (size_t i = 0; i < m_layer_indices.size () && reuse_layer_list; ++i) {
    reuse_layer_list = layout.is_valid_layer (m_layer_indices[i]) 
                         && layout.get_properties (m_layer_indices[i]).log_equal (m_layer_declarations[i]);
  }

  if (! reuse_layer_list) {

    m_layer_indices.clear ();

    for (size_t i = 0; i < m_layer_declarations.size (); ++i) {

      bool found = false;
      for (size_t l = 0; l < layout.layers () && !found; ++l) {
        if (layout.is_valid_layer (l) && layout.get_properties (l).log_equal (m_layer_declarations[i])) {
          m_layer_indices.push_back (l);
          found = true;
        }
      }

      if (! found) {
        m_layer_indices.push_back (layout.insert_layer (m_layer_declarations[i]));
      }

    }

  }

  return m_layer_indices;
}

PCellVariant *
PCellHeader::get_variant (db::Layout &layout, const pcell_parameters_type &parameters)
{
  variant_map_t::iterator v = m_variant_map.find (&parameters);
  if (v != m_variant_map.end ()) {
    return v->second;
  } else {
    return 0;
  }
}

void 
PCellHeader::unregister_variant (PCellVariant *variant)
{
  variant_map_t::iterator v = m_variant_map.find (&variant->parameters ());
  tl_assert (v != m_variant_map.end ());
  m_variant_map.erase (v);
}

void 
PCellHeader::register_variant (PCellVariant *variant)
{
  variant_map_t::iterator v = m_variant_map.find (&variant->parameters ());
  tl_assert (v == m_variant_map.end ());
  m_variant_map.insert (std::make_pair (&variant->parameters (), variant));
}
  
}

