
/*

  KLayout Layout Viewer
  Copyright (C) 2006-2007 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#ifndef HDR_dbPolygonTools
#define HDR_dbPolygonTools

#include "dbPolygon.h"

#include <vector>
#include <limits>
#include <algorithm>

namespace db {

//  internal: compare edges by higher y coordinate
template <class E>
struct inside_poly_test_edge_max_compare_f
{
  bool operator() (const E &e1, const E &e2)
  {
    return std::max (e1.p1 ().y (), e1.p2 ().y ()) < std::max (e2.p1 ().y (), e2.p2 ().y ());
  }
};

/**
 *  @brief An inside test operator
 *
 *  This class allows to efficiently test whether multiple points are inside a given polygon.
 *  Since the test is efficiently implemented when the polygon edges are sorted, the sorting 
 *  and memory allocation step is performed once in the test operator's constructor while 
 *  each individual test is performed efficiently.
 */

template <class P>
class inside_poly_test 
{
public:
  typedef typename P::point_type point_type;
  typedef typename P::coord_type coord_type;
  typedef typename db::edge<coord_type> edge_type;

  /**
   *  @brief Constructor
   */
  inside_poly_test (const P &polygon)
  {
    m_edges.reserve (polygon.vertices ());
    for (typename P::polygon_edge_iterator e = polygon.begin_edge (); ! e.at_end (); ++e) {
      m_edges.push_back (*e);
    }
    std::sort (m_edges.begin (), m_edges.end (), inside_poly_test_edge_max_compare_f<edge_type> ());
  }

  /**
   *  @brief Actual test
   *
   *  This function returns 1, if the point is inside (not on)
   *  the polygon. It returns 0, if the point is on the polygon and -1
   *  if outside. 
   */
  int operator() (const point_type &pt) const
  {
    int wrapcount_left = 0;

    typename std::vector<edge_type>::const_iterator e = std::lower_bound (m_edges.begin (), m_edges.end (), edge_type (pt, pt), inside_poly_test_edge_max_compare_f<edge_type> ());
    while (e != m_edges.end () && pt.y () <= std::max (e->p1 ().y (), e->p2 ().y ())) {

      if ((*e).p1 ().y () <= pt.y () && (*e).p2 ().y () > pt.y ()) {
        int side = (*e).side_of (pt);
        if (side < 0) { 
          ++wrapcount_left;
        } else if (side == 0) {
          //  "on" the line is excluded in the predicate
          return 0;
        }
      } else if ((*e).p2 ().y () <= pt.y () && (*e).p1 ().y () > pt.y ()) {
        int side = (*e).side_of (pt);
        if (side > 0) { 
          --wrapcount_left;
        } else if (side == 0) {
          //  "on" the line is excluded in the predicate
          return 0;
        }
      } else if ((*e).p1 ().y () == pt.y () && (*e).p2 ().y () == pt.y () &&
                 (((*e).p1 ().x () <= pt.x () && (*e).p2 ().x () >= pt.x ()) ||
                  ((*e).p2 ().x () <= pt.x () && (*e).p1 ().x () >= pt.x ()))) {
        //  "on" the horizontal line is excluded in the predicate
        return 0;
      }

      ++e;

    }

    return (wrapcount_left != 0) ? 1 : -1;
  }

private:
  std::vector<edge_type> m_edges;
};

/**
 *  @brief A helper structure describing an edge cutting the cut line.
 */
template <class PointType>
struct cut_polygon_cutting_edge
{
  typedef typename PointType::coord_type coord_type;
  typedef typename db::coord_traits<coord_type>::area_type projection_type;

  cut_polygon_cutting_edge (unsigned int c, unsigned int n, projection_type p, const PointType &pt, bool l) 
    : contour (int (c)), index (int (n)), projected (p), point (pt), leave (l)
  { }

  bool operator< (const cut_polygon_cutting_edge &o) const 
  { 
    return projected < o.projected; 
  }

  int contour, index;
  projection_type projected;
  PointType point;
  bool leave;
};

/**
 *  @brief Polygon cut function
 *
 *  This functions cuts a polygon at the given line (given by an edge)
 *  and produces all parts of the polygon that are "right" of the line given by "line".
 */
template <class PolygonType, class OutputIter, class EdgeType>
void cut_polygon (const PolygonType &input, const EdgeType &line, OutputIter right_of_line)
{
  typedef typename PolygonType::point_type point_type;
  typedef typename PolygonType::coord_type coord_type;
  typedef typename PolygonType::contour_type contour_type;
  typedef db::edge<coord_type> edge_type;
  typedef cut_polygon_cutting_edge<point_type> cutting_edge_type;
  typedef typename db::coord_traits<coord_type>::area_type projection_type;

  bool do_hole_assignment = (input.holes () > 0);
  std::vector <PolygonType> hull_polygons;
  std::vector <PolygonType> hole_polygons;

  std::vector<cutting_edge_type> cutting_edges;

  for (unsigned int nc = 0; nc < input.holes () + 1; ++nc) {

    const contour_type &contour = input.contour (nc);
    if (contour.size () == 0) {
      continue;
    }

    bool any = false;

    int nn = contour.size () - 1;
    point_type last_pt = contour[nn];
    for (unsigned int n = 0; n < contour.size (); ++n) {

      edge_type e (last_pt, contour[n]);
      last_pt = e.p2 ();

      std::pair <bool, point_type> ip = line.crossed_by_point (e);
      if (ip.first) {

        int s1 = line.side_of (e.p1 ());
        int s2 = line.side_of (e.p2 ());

        projection_type p = db::sprod (ip.second, line.p2 (), line.p1 ());

        if (s1 < 0 && s2 >= 0) {
          // right -> left or on edge
          cutting_edges.push_back (cutting_edge_type (nc, nn, p, ip.second, true /*leave*/));
          any = true;
        }

        if (s1 >= 0 && s2 < 0) {
          // left or on edge -> right 
          cutting_edges.push_back (cutting_edge_type (nc, nn, p, ip.second, false /*enter*/));
        }

      }

      nn = n;

    }

    if (! any && line.side_of (contour[0]) < 0) {

      if (do_hole_assignment) {

        if (nc == 0) {
          //  the hull is fully on the right side -> just output the input polygon and that's it.
          *right_of_line++ = input;
          return;
        } else {
          //  remember hole contours for later assignment
          hole_polygons.push_back (PolygonType ());
          hole_polygons.back ().assign_hull (contour.begin (), contour.end ());
        }

      } else {
        PolygonType poly;
        poly.assign_hull (contour.begin (), contour.end ());
        *right_of_line++ = poly;
      }
    }

  }

  //  we use a stable sort here to keep the leave/enter order as it was for identical projected points.
  //  This has an particular effect for hole connection edges.
  std::stable_sort (cutting_edges.begin (), cutting_edges.end ());

  int wc = 0;
  for (typename std::vector<cutting_edge_type>::iterator c = cutting_edges.begin (); c != cutting_edges.end (); ++c) {

    wc += c->leave ? -1 : 1;
    if (abs (wc) > 1) {

      cutting_edge_type c1 (*c);

      int wc_next = wc;

      typename std::vector<cutting_edge_type>::iterator cc = c;
      while (wc_next != 0) {

        ++cc;
        tl_assert (cc != cutting_edges.end ());

        wc_next += cc->leave ? -1 : 1;
        std::swap (c1, *cc);

      }

      *c = c1;
      wc = 0;

    }

  }

  tl_assert (cutting_edges.size () % 2 == 0);

  std::vector<point_type> contour_points;
  typedef std::map <std::pair<int, int>, std::pair<typename std::vector<cutting_edge_type>::iterator, typename std::vector<cutting_edge_type>::iterator> > cut_point_map_type;
  cut_point_map_type cut_points;

  for (typename std::vector<cutting_edge_type>::iterator c = cutting_edges.begin (); c != cutting_edges.end (); c += 2) {
    cut_points.insert (std::make_pair (std::make_pair (c[0].contour, c[0].index), std::make_pair (c, c + 1)));
    cut_points.insert (std::make_pair (std::make_pair (c[1].contour, c[1].index), std::make_pair (c + 1, c)));
  }

  for (typename std::vector<cutting_edge_type>::iterator c = cutting_edges.begin (); c != cutting_edges.end (); c += 2) {

    if (c->contour >= 0) {

      typename std::vector<cutting_edge_type>::iterator c1 = c;
      typename std::vector<cutting_edge_type>::iterator c2 = c + 1;
      tl_assert (c1->leave != c2->leave);

      if (! c1->leave) {
        std::swap (c1, c2);
      }

      contour_points.clear ();
      bool is_hull = false;

      do {

        tl_assert (c1->contour >= 0 && c2->contour >= 0);

        contour_points.push_back (c1->point);
        contour_points.push_back (c2->point);

        int n = c2->index;
        int n0 = n;
        int nc = c2->contour;
        const contour_type &contour = input.contour (nc);

        if (nc == 0) {
          is_hull = true;
        }

        c1->contour = -1;
        c2->contour = -1;

        ++n;
        if (n == int (contour.size ())) {
          n = 0;
        }

        while (n != n0) {

          contour_points.push_back (contour[n]);

          typename cut_point_map_type::iterator cp = cut_points.find (std::make_pair (nc, n));
          if (cp != cut_points.end ()) {

            c1 = cp->second.first;
            c2 = cp->second.second;

            tl_assert (c1->leave == true);

            break;
          }

          ++n;
          if (n == int (contour.size ())) {
            n = 0;
          }

        }

        tl_assert (n != n0);

      } while (c1 != c && c2 != c);

      //  Hint: the algorithm tends to create spikes for hole insertion edges.
      //  Therefore we used "remove reflected" on the assignment.
      if (do_hole_assignment) {

        if (is_hull) {
          hull_polygons.push_back (PolygonType ());
          hull_polygons.back ().assign_hull (contour_points.begin (), contour_points.end (), true, true /*remove reflected*/);
        } else {
          hole_polygons.push_back (PolygonType ());
          hole_polygons.back ().assign_hull (contour_points.begin (), contour_points.end (), true, true /*remove reflected*/);
        }

      } else {
        PolygonType poly;
        poly.assign_hull (contour_points.begin (), contour_points.end (), true, true /*remove reflected*/);
        //  it might happen in some cases, that cut pieces may vanish (i.e. all points on a line). Thus we check, if that
        //  is the case and do not produce a polygon then.
        if (poly.vertices () > 0) {
          *right_of_line++ = poly;
        }
      }

    }

  }

  //  do hole assignment
  for (typename std::vector<PolygonType>::iterator hull = hull_polygons.begin (); hull != hull_polygons.end (); ++hull) {

    //  it might happen in some cases, that cut pieces may vanish (i.e. all points on a line). Thus we check, if that
    //  is the case and do not produce a polygon then.
    if (hull->vertices () > 0) {

      db::inside_poly_test<PolygonType> inside_hull (*hull);
      for (typename std::vector<PolygonType>::iterator hole = hole_polygons.begin (); hole != hole_polygons.end (); ++hole) {
        if (hole->hull ().size () > 0 && inside_hull (hole->hull ()[0])) {
          hull->insert_hole (hole->hull ().begin (), hole->hull ().end ());
          *hole = PolygonType ();
        }
      }

      *right_of_line++ = *hull;

    }

  }

  // check for non-assigned holes
  for (typename std::vector<PolygonType>::iterator hole = hole_polygons.begin (); hole != hole_polygons.end (); ++hole) {
    tl_assert (hole->hull ().size () == 0);
  }
}

/**
 *  @brief Split a polygon into two or more parts 
 *
 *  This function splits a polygon into parts using some heuristics to determine a "suitable" cut line.
 *  The cut line is choosen through a vertex close to a center (either horizontal or vertical). The splitting
 *  is supposed to create smaller parts with less vertices or a better area ratio of polygon to bounding box area.
 *
 *  @param polygon The input polygon
 *  @param output The parts 
 */
template <class PolygonType>
void
split_polygon (const PolygonType &polygon, std::vector<PolygonType> &output)
{
  typedef typename PolygonType::coord_type coord_type;
  typedef typename PolygonType::point_type point_type;
  typedef typename PolygonType::box_type box_type;
  typedef db::edge<coord_type> edge_type;

  box_type bbox = polygon.box ();
  box_type b1, b2;

  coord_type x = bbox.center ().x ();
  coord_type xx = x;
  bool xx_set = false;

  coord_type y = bbox.center ().y ();
  coord_type yy = y;
  bool yy_set = false;

  for (typename PolygonType::polygon_contour_iterator e = polygon.begin_hull (); e != polygon.end_hull (); ++e) {
    if ((*e).x () != bbox.left () && (*e).x () != bbox.right () && (abs ((*e).x () - x) < abs (xx - x) || ! xx_set)) {
      xx = (*e).x ();
      xx_set = true;
    }
    if ((*e).y () != bbox.top () && (*e).y () != bbox.bottom () && (abs ((*e).y () - y) < abs (yy - y) || ! yy_set)) {
      yy = (*e).y ();
      yy_set = true;
    }
  }

  if (! xx_set && ! yy_set) {
    if (bbox.width () > bbox.height ()) { 
      xx_set = true;
    } else {
      yy_set = true;
    }
  } else if (xx_set && yy_set) {
    //  an empiric threshold for splitting polygons in one direction: don't split along the long
    //  axis for polygons with a aspect ratio (of the bounding box) of larger than 3
    if (bbox.width () > 3 * bbox.height ()) {
      yy_set = false;
    } else if (bbox.height () > 3 * bbox.width ()) {
      xx_set = false;
    }
  }

  std::vector <PolygonType> xx_polygons;
  size_t xx_n = std::numeric_limits<size_t>::max ();
  if (xx_set) {

    db::cut_polygon (polygon, edge_type (point_type (xx, 0), point_type (xx, 1)), std::back_inserter (xx_polygons));
    db::cut_polygon (polygon, edge_type (point_type (xx, 1), point_type (xx, 0)), std::back_inserter (xx_polygons));
    
    xx_n = 0;
    for (typename std::vector <PolygonType>::const_iterator p = xx_polygons.begin (); p != xx_polygons.end (); ++p) {
      xx_n += p->vertices ();
    }

  }

  std::vector <PolygonType> yy_polygons;
  size_t yy_n = std::numeric_limits<size_t>::max ();
  if (yy_set) {

    db::cut_polygon (polygon, edge_type (point_type (0, yy), point_type (1, yy)), std::back_inserter (yy_polygons));
    db::cut_polygon (polygon, edge_type (point_type (1, yy), point_type (0, yy)), std::back_inserter (yy_polygons));
    
    yy_n = 0;
    for (typename std::vector <PolygonType>::const_iterator p = yy_polygons.begin (); p != yy_polygons.end (); ++p) {
      yy_n += p->vertices ();
    }

  }

  if (xx_n < yy_n) {
    output.swap (xx_polygons);
  } else {
    output.swap (yy_polygons);
  }
}

}

#endif

