
/*

  KLayout Layout Viewer
  Copyright (C) 2006-2007 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#include "edtFinder.h"

namespace edt
{

// -------------------------------------------------------------

//  max. number of tries in single-click selection before giving up
static int point_sel_tests = 10000;

//  max. number of tries in single-click selection before giving up
static int inst_point_sel_tests = 10000;

// ----------------------------------------------------------------------
//  Finder implementation

Finder::Finder (bool point_mode, bool top_level_sel)
  : m_min_level (0), m_max_level (0),
    mp_layout (0), m_point_mode (point_mode), m_top_level_sel (top_level_sel)
{
  m_distance = std::numeric_limits<double>::max ();
}

Finder::~Finder ()
{
  //  .. nothing yet ..
}

bool 
Finder::closer (double d)
{
  if (d <= m_distance) {
    m_distance = d;
    return true;
  } else {
    return false;
  }
}

void 
Finder::start (lay::LayoutView *view, const lay::CellView &cv, unsigned int cv_index, const std::vector<db::CplxTrans> &trans, const db::Box &region, int min_level, int max_level, int layer)
{
  m_layer = (unsigned int) layer;
  m_region = region; 
  mp_layout = &cv->layout ();
  mp_view = view;
  m_cv_index = cv_index;
  m_min_level = std::max (0, min_level);
  m_max_level = std::max (m_min_level, std::min (max_level, m_top_level_sel ? ((int) cv.specific_path ().size () + 1) : max_level));

  if (layer >= 0) {
    m_box_convert = db::box_convert <db::CellInst> (*mp_layout, (unsigned int) layer);
    m_cell_box_convert = db::box_convert <db::Cell> ((unsigned int) layer);
  } else {
    m_box_convert = db::box_convert <db::CellInst> (*mp_layout);
    m_cell_box_convert = db::box_convert <db::Cell> ();
  }

  m_path.erase (m_path.begin (), m_path.end ());

  for (std::vector<db::CplxTrans>::const_iterator t = trans.begin (); t != trans.end (); ++t) {
    do_find (*cv.cell (), cv.specific_path ().size (), *t * cv.context_trans ());
  }
}

unsigned int
Finder::test_edge (const db::DEdge &edg, double &distance, bool &match)
{
  unsigned int ret = 0;

  //  we hit the region with the edge end points - take the closest vertex
  if (m_region.contains (edg.p1 ()) || m_region.contains (edg.p2 ())) {

    double d1 = edg.p1 ().distance (m_region.center ());
    double d2 = edg.p2 ().distance (m_region.center ());

    //  snap to the point - nothing can get closer
    distance = 0.0;
    if (d1 < d2) {
      ret = 1;
    } else {
      ret = 2;
    }

    match = true;

  }
  
  //  if the edge cuts through the active region: test the
  //  edge as a whole
  if (ret == 0 && edg.clipped (m_region).first) {
    double d = edg.distance_abs (m_region.center ());
    if (! match || d < distance) {
      distance = d;
      ret = 3;
    }
    match = true;
  }

  return ret;
}

void
Finder::do_find (const db::Cell &cell, int level, const db::CplxTrans &t)
{
  if (! (level >= m_max_level || ! (t * m_cell_box_convert (cell)).touches (m_region) || mp_view->is_cell_hidden (cell.cell_index (), m_cv_index))) {

    db::Box touch_box (db::Box::from_double (t.inverted () * m_region)); 

    if (level >= m_min_level) {
      visit_cell (cell, touch_box, t, level);
    }

    db::Cell::touching_iterator inst = cell.begin_touching (touch_box); 
    while (! inst.at_end ()) {

      const db::CellInstArray &cell_inst = inst->cell_inst ();
      for (db::CellInstArray::iterator p = cell_inst.begin_touching (touch_box, m_box_convert); ! p.at_end (); ++p) {

        m_path.push_back (db::InstElement (*inst, p));

        do_find (mp_layout->cell (cell_inst.object ().cell_index ()), 
                 level + 1,
                 t * cell_inst.complex_trans (*p));

        m_path.pop_back ();

      }

      ++inst;

    }

  }
}

// -------------------------------------------------------------
//  ShapeFinder implementation

ShapeFinder::ShapeFinder (bool point_mode, bool top_level_sel, db::ShapeIterator::flags_type flags)
  : Finder (point_mode, top_level_sel), m_flags (flags), mp_prop_sel (0), m_inv_prop_sel (false), m_progress (tl::translate ("Selecting ..."))
{
  m_progress.set_unit (1000);
  m_progress.set_format ("");
  m_tries = point_sel_tests;
}

bool
ShapeFinder::find (lay::LayoutView *view, const db::DBox &region_mu)
{
  for (lay::LayerPropertiesConstIterator lprop = view->begin_layers (); ! lprop.at_end (); ++lprop) {
    if (lprop->is_visual ()) {
      find (view, *lprop, region_mu);
    }
  }

  return ! m_founds.empty ();
}

bool 
ShapeFinder::find (lay::LayoutView *view, const lay::LayerProperties &lprops, const db::DBox &region_mu)
{
  m_cv_index = lprops.cellview_index ();

  const lay::CellView &cv = view->cellview (m_cv_index);
  if (! cv.is_valid ()) {
    return false;
  }

  m_topcell = cv.cell_index ();

  double dbu = cv->layout ().dbu ();
  db::Box region = db::Box::from_double (region_mu * (1.0 / dbu));

  mp_prop_sel = &lprops.prop_sel ();
  m_inv_prop_sel = lprops.inverse_prop_sel ();

  int ctx_path_length = int (cv.specific_path ().size ());

  int min_level = view->get_min_hier_levels ();
  int max_level = view->get_max_hier_levels ();
  if (lprops.hier_levels ().has_from_level ()) {
    min_level = lprops.hier_levels ().from_level (ctx_path_length);
  }
  if (lprops.hier_levels ().has_to_level ()) {
    max_level = lprops.hier_levels ().to_level (ctx_path_length);
  }

  //  actually find
  try {
    start (view, cv, m_cv_index, lprops.trans (), region, min_level, max_level, int (lprops.layer_index ()));
  } catch (StopException) {
    // ..
  }

  //  return true if anything was found
  return ! m_founds.empty ();
}

void 
ShapeFinder::visit_cell (const db::Cell &cell, const db::Box &search_box, const db::CplxTrans &t, int /*level*/)
{
  if (! point_mode ()) {

    ++m_progress;

    const db::Shapes &shapes = cell.shapes (layer ());

    db::ShapeIterator shape = shapes.begin_touching (search_box, m_flags, mp_prop_sel, m_inv_prop_sel);
    while (! shape.at_end ()) {

      ++m_progress;

      //  in box mode, just test the boxes
      if (shape->bbox ().inside (search_box)) {

        //  in point mode just store that found that has the least "distance"
        m_founds.push_back (edt::ObjectInstPath ());
        m_founds.back ().set_cv_index (m_cv_index);
        m_founds.back ().set_topcell (m_topcell);
        m_founds.back ().assign_path (path ().begin (), path ().end ());
        m_founds.back ().set_layer (layer ());
        m_founds.back ().set_shape (*shape);

      }

      ++shape;

    } 

  } else {

    if (--m_tries < 0) {
      throw StopException ();
    }

    const db::Shapes &shapes = cell.shapes (layer ());

    db::ShapeIterator shape = shapes.begin_touching (search_box, m_flags, mp_prop_sel, m_inv_prop_sel);
    while (! shape.at_end ()) {

      bool match = false;
      double d = std::numeric_limits<double>::max ();

      if (--m_tries < 0) {
        throw StopException ();
      }

      db::Point point (search_box.center ());

      //  in point mode, test the edges and use a "closest" criterion
      if (shape->is_polygon ()) {

        for (db::Shape::polygon_edge_iterator e = shape->begin_edge (); ! e.at_end (); ++e) {
          test_edge (t * *e, d, match);
        }

        //  test if inside the polygon
        if (! match && inside_poly (shape->begin_edge (), point) >= 0) { 
          d = t.ctrans (poly_dist (shape->begin_edge (), point)); 
          match = true;
        }

      } else if (shape->is_path ()) {

        //  test the "spine"
        db::Shape::point_iterator pt = shape->begin_point (); 
        if (pt != shape->end_point ()) {
          db::Point p (*pt);
          ++pt;
          for (; pt != shape->end_point (); ++pt) {
            test_edge (t * db::Edge (p, *pt), d, match);
            p = *pt;
          }
        }

        //  convert to polygon and test those edges
        db::Polygon poly;
        shape->polygon (poly);
        for (db::Polygon::polygon_edge_iterator e = poly.begin_edge (); ! e.at_end (); ++e) {
          test_edge (t * *e, d, match);
        }

        //  test if inside the polygon
        if (! match && inside_poly (poly.begin_edge (), point) >= 0) { 
          d = t.ctrans (poly_dist (poly.begin_edge (), point));
          match = true;
        }

      } else if (shape->is_box ()) {

        const db::Box &box = shape->box ();

        //  convert to polygon and test those edges
        db::Polygon poly (box);
        for (db::Polygon::polygon_edge_iterator e = poly.begin_edge (); ! e.at_end (); ++e) {
          test_edge (t * *e, d, match);
        }

        if (! match && box.contains (search_box.center ())) {
          d = t.ctrans (poly_dist (poly.begin_edge (), point)); 
          match = true;
        }

      } else if (shape->is_text ()) {

        db::Point tp (shape->text_trans () * db::Point ());
        if (search_box.contains (tp)) {
          d = tp.distance (search_box.center ());
          match = true;
        }

      }

      if (match && closer (d)) {

        //  in point mode just store that found that has the least "distance"
        if (m_founds.empty ()) {
          m_founds.push_back (edt::ObjectInstPath ());
        }

        m_founds.back ().set_cv_index (m_cv_index);
        m_founds.back ().set_topcell (m_topcell);
        m_founds.back ().assign_path (path ().begin (), path ().end ());
        m_founds.back ().set_layer (layer ());
        m_founds.back ().set_shape (*shape);

      }

      ++shape;

    }

  }

}

// -------------------------------------------------------------
//  InstFinder implementation

InstFinder::InstFinder (bool point_mode, bool top_level_sel, bool full_arrays, bool enclose_inst)
  : Finder (point_mode, top_level_sel), 
    m_full_arrays (full_arrays), 
    m_enclose_insts (enclose_inst), 
    m_progress (tl::translate ("Selecting ..."))
{
  m_progress.set_unit (1000);
  m_progress.set_format ("");
  m_tries = inst_point_sel_tests;
}

bool
InstFinder::find (lay::LayoutView *view, const db::DBox &region_mu)
{
  std::set< std::pair<db::CplxTrans, int> > variants = view->cv_transform_variants ();
  for (std::set< std::pair<db::CplxTrans, int> >::const_iterator v = variants.begin (); v != variants.end (); ++v) {
    find (view, v->second, v->first, region_mu);
  }

  return ! m_founds.empty ();
}

bool 
InstFinder::find (lay::LayoutView *view, unsigned int cv_index, const db::CplxTrans &trans, const db::DBox &region_mu)
{
  const lay::CellView &cv = view->cellview (cv_index);
  if (! cv.is_valid ()) {
    return false;
  }

  m_cv_index = cv_index;
  m_topcell = cv.cell ()->cell_index ();
  mp_view = view;

  double dbu = cv->layout ().dbu ();
  db::Box region = db::Box::from_double (region_mu * (1.0 / dbu));

  //  actually find
  try {
    std::vector<db::CplxTrans> tv;
    tv.push_back (trans);
    start (view, cv, cv_index, tv, region, view->get_min_hier_levels (), view->get_max_hier_levels ());
  } catch (StopException) {
    // ..
  }

  //  return true if anything was found
  return ! m_founds.empty ();
}

void 
InstFinder::visit_cell (const db::Cell &cell, const db::Box &search_box, const db::CplxTrans &t, int level)
{
  if (! point_mode ()) {

    ++m_progress;

    //  shortcut for cases where nothing can be found
    if (level != max_level () - 1 && (mp_view->hidden_cells ().size () <= m_cv_index || mp_view->hidden_cells () [m_cv_index].empty ())) {
      return;
    }

    //  look for instances to check here ..
    db::Cell::touching_iterator inst = cell.begin_touching (search_box); 
    while (! inst.at_end ()) {

      const db::CellInstArray &cell_inst = inst->cell_inst ();

      ++m_progress;

      //  just consider the instances exactly at the last level of 
      //  hierarchy (this is where the boxes are drawn) or if of cells that
      //  are hidden.
      if (level == max_level () - 1 || mp_view->is_cell_hidden (cell_inst.object ().cell_index (), m_cv_index)) {

        db::box_convert <db::CellInst> bc (layout ());
        for (db::CellInstArray::iterator p = cell_inst.begin_touching (search_box, bc); ! p.at_end (); ++p) {
        
          ++m_progress;

          const db::DBox box = cell_inst.complex_trans (*p) * layout ().cell (cell_inst.object ().cell_index ()).bbox ();

          //  in box mode, just test the boxes
          if (! m_enclose_insts || box.inside (db::DBox (search_box))) {

            //  in point mode just store that found that has the least "distance"
            m_founds.push_back (edt::ObjectInstPath ());
            m_founds.back ().set_cv_index (m_cv_index);
            m_founds.back ().set_topcell (m_topcell);
            m_founds.back ().assign_path (path ().begin (), path ().end ());

            //  add the selected instance as the last element of the path
            db::InstElement el;
            el.inst_ptr = *inst;
            if (! m_full_arrays) {
              el.array_inst = p;
            }
            m_founds.back ().add_path (el);

            //  in "full arrays" mode, a single reference to that array is sufficient
            if (m_full_arrays) {
              break;
            }

          }

        }

      }

      ++inst;

    }

  } else {

    if (--m_tries < 0) {
      throw StopException ();
    }

    //  look for instances to check here ..
    db::Cell::touching_iterator inst = cell.begin_touching (search_box); 
    while (! inst.at_end ()) {

      if (--m_tries < 0) {
        throw StopException ();
      }

      const db::CellInstArray &cell_inst = inst->cell_inst ();

      //  just consider the instances exactly at the last level of 
      //  hierarchy (this is where the boxes are drawn) or if of cells that
      //  are hidden.
      if (level == max_level () - 1 || mp_view->is_cell_hidden (cell_inst.object ().cell_index (), m_cv_index)) {

        db::box_convert <db::CellInst> bc (layout ());
        for (db::CellInstArray::iterator p = cell_inst.begin_touching (search_box, bc); ! p.at_end (); ++p) {
        
          if (--m_tries < 0) {
            throw StopException ();
          }

          bool match = false;
          double d = std::numeric_limits<double>::max ();

          const db::DBox box = cell_inst.complex_trans (*p) * layout ().cell (cell_inst.object ().cell_index ()).bbox ();

          //  convert to polygon and test those edges
          db::DPolygon poly (box);
          for (db::DPolygon::polygon_edge_iterator e = poly.begin_edge (); ! e.at_end (); ++e) {
            test_edge (t * *e, d, match);
          }

          if (! match && box.contains (search_box.center ())) {
            d = t.ctrans (poly_dist (poly.begin_edge (), db::DPoint (search_box.center ()))); 
            match = true;
          }

          if (match && closer (d)) {

            //  in point mode just store that found that has the least "distance"
            if (m_founds.empty ()) {
              m_founds.push_back (edt::ObjectInstPath ());
            }

            m_founds.back ().set_cv_index (m_cv_index);
            m_founds.back ().set_topcell (m_topcell);
            m_founds.back ().assign_path (path ().begin (), path ().end ());

            //  add the selected instance as the last element of the path
            db::InstElement el;
            el.inst_ptr = *inst;
            if (! m_full_arrays) {
              el.array_inst = p;
            }
            m_founds.back ().add_path (el);

          }

        }

      }

      ++inst;

    }

  }

}

} // namespace edt

