
/*

  KLayout Layout Viewer
  Copyright (C) 2006-2007 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#include <QtGui/QInputDialog>

#include "dbEdgeProcessor.h"
#include "layExceptions.h"
#include "layLayoutView.h"
#include "layDialogs.h"
#include "edtPlugin.h"
#include "edtMainService.h"
#include "edtService.h"
#include "edtServiceImpl.h"
#include "edtConfig.h"
#include "edtDialogs.h"
#include "edtEditorOptionsPages.h"

namespace edt
{

// -----------------------------------------------------------------------------
//  Main Service implementation

MainService::MainService (db::Manager *manager, lay::LayoutView *view, lay::MainWindow *mw)
  : lay::Plugin (view),
    lay::Editable (view),
    db::Object (manager),
    mp_view (view),
    mp_main_window (mw),
    m_flatten_insts_levels (std::numeric_limits<int>::max ()),
    m_flatten_prune (false),
    m_align_hmode (0), m_align_vmode (0), m_align_visible_layers (false)
{ 
  //  collect the options pages and build the options dialog
  std::vector<edt::Service *> edt_services = mp_view->get_plugins <edt::Service> ();
}

MainService::~MainService ()
{
  //  .. nothing yet ..
}

void 
MainService::menu_activated (const std::string &symbol)
{
  if (symbol == "edt::descend") {
    cm_descend ();
  } else if (symbol == "edt::ascend") {
    cm_ascend ();
  } else if (symbol == "edt::edit_options") {
    cm_edit_options ();
  } else if (symbol == "edt::sel_align") {
    cm_align ();
  } else if (symbol == "edt::sel_round_corners") {
    cm_round_corners ();
  } else if (symbol == "edt::sel_size") {
    cm_size ();
  } else if (symbol == "edt::sel_union") {
    cm_union ();
  } else if (symbol == "edt::sel_intersection") {
    cm_intersection ();
  } else if (symbol == "edt::sel_difference") {
    cm_difference ();
  } else if (symbol == "edt::sel_change_layer") {
    cm_change_layer ();
  } else if (symbol == "edt::sel_flatten_insts") {
    cm_flatten_insts ();
  } else if (symbol == "edt::sel_move_hier_up") {
    cm_move_hier_up ();
  } else if (symbol == "edt::sel_make_cell") {
    cm_make_cell ();
  } else if (symbol == "edt::sel_make_cell_variants") {
    cm_make_cell_variants ();
  }
}

/**
 *  @brief A helper class to determine the common part of a set of instance elements
 */
class CommonInsts
{
public:
  CommonInsts () 
    : m_valid (true), m_first (true), m_ambigous (false), m_cv_index (0) 
  { 
    //  .. nothing yet ..  
  }

  void add (const ObjectInstPath &path, unsigned int n)
  {
    if (! m_valid) {

      //  don't do anything

    } else if (m_first) {

      m_common_inst.reserve (n);
      unsigned int nn = 0;
      for (ObjectInstPath::iterator p = path.begin (); p != path.end () && nn < n; ++nn) {
        m_common_inst.push_back (*p);
      }
      m_cv_index = path.cv_index ();
      m_first = false;

    } else if (m_cv_index != path.cv_index ()) {

      m_valid = false;

    } else {

      std::vector<db::InstElement>::iterator i1 = m_common_inst.begin ();
      ObjectInstPath::iterator i2 = path.begin ();
      while (i1 != m_common_inst.end () && i2 != path.end () && *i1 == *i2) {
        ++i1; ++i2;
      }
      if (i1 != m_common_inst.end ()) {
        m_ambigous = true;
        m_common_inst.erase (i1, m_common_inst.end ());
      }

    }
  }

  const std::vector<db::InstElement> &common_path () const
  {
    return m_common_inst;
  }

  bool valid () const
  {
    return m_valid;
  }

  bool anything () const
  {
    return ! m_first;
  }

  bool ambigous () const
  {
    return m_ambigous && m_common_inst.empty ();
  }

  bool empty () const
  {
    return m_common_inst.empty ();
  }

  unsigned int size () const
  {
    return m_common_inst.size ();
  }

  unsigned int cv_index () const
  {
    return m_cv_index;
  }

private:
  std::vector<db::InstElement> m_common_inst;
  bool m_valid;
  bool m_first;
  bool m_ambigous;
  unsigned int m_cv_index;
};


void  
MainService::cm_descend ()
{
  CommonInsts common_inst;

  std::vector<edt::Service *> edt_services = view ()->get_plugins <edt::Service> ();
  for (std::vector<edt::Service *>::const_iterator es = edt_services.begin (); es != edt_services.end () && common_inst.valid (); ++es) {
    for (edt::Service::objects::const_iterator sel = (*es)->selection ().begin (); sel != (*es)->selection ().end () && common_inst.valid (); ++sel) {
      common_inst.add (*sel, 1);
    }
  }

  //  cannot descend - we are at the lowest level already
  if (common_inst.empty ()) {
    return;
  }

  if (! common_inst.anything ()) {
    throw tl::Exception (tl::translate ("Select an object to determine into which instance to descend").c_str ());
  }
  if (! common_inst.valid () || common_inst.ambigous ()) {
    throw tl::Exception (tl::translate ("Selection is ambigous - cannot determine into which instance to descend").c_str ());
  }

  //  remove the common path and create a new set of selections

  std::vector< std::vector<ObjectInstPath> > new_selections;
  new_selections.reserve (edt_services.size ());

  for (std::vector<edt::Service *>::const_iterator es = edt_services.begin (); es != edt_services.end (); ++es) {

    new_selections.push_back (std::vector<ObjectInstPath> ());
    new_selections.back ().reserve ((*es)->selection ().size ());

    for (edt::Service::objects::const_iterator sel = (*es)->selection ().begin (); sel != (*es)->selection ().end (); ++sel) {

      new_selections.back ().push_back (*sel);
      edt::ObjectInstPath &new_sel = new_selections.back ().back ();
      new_sel.remove_front (common_inst.size ());
      //  it may happen that this way we dive into the instance selected. The resulting selection
      //  is not pointing to any instance any more and must be discarded therefore:
      if (new_sel.is_cell_inst () && new_sel.begin () == new_sel.end ()) {
        new_selections.back ().pop_back ();
      }

    }

  }

  //  this will clear the selection:
  view ()->descend (common_inst.common_path (), common_inst.cv_index ());
  view ()->set_current_cell_path (common_inst.cv_index (), view ()->cellview (common_inst.cv_index ()).combined_unspecific_path ());

  //  set the new selections
  unsigned int index = 0;
  for (std::vector<edt::Service *>::const_iterator es = edt_services.begin (); es != edt_services.end (); ++es, ++index) {
    (*es)->set_selection (new_selections [index].begin (), new_selections [index].end ());
  }

}

void  
MainService::cm_ascend ()
{
  //  add one path component and create a new set of selections

  std::vector<edt::Service *> edt_services = view ()->get_plugins <edt::Service> ();

  std::vector< std::vector<ObjectInstPath> > new_selections;
  new_selections.reserve (edt_services.size ());
  for (std::vector<edt::Service *>::const_iterator es = edt_services.begin (); es != edt_services.end (); ++es) {
    new_selections.push_back (std::vector<ObjectInstPath> ());
    new_selections.back ().insert (new_selections.back ().end (), (*es)->selection ().begin (), (*es)->selection ().end ());
  }

  //  this will clear the selection:
  for (int cv_index = 0; cv_index < int (view ()->cellviews ()); ++cv_index) {

    db::InstElement removed = view ()->ascend (cv_index);
    if (removed != db::InstElement ()) {

      db::cell_index_type new_top = view ()->cellview (cv_index).cell_index ();
      view ()->set_current_cell_path (cv_index, view ()->cellview (cv_index).combined_unspecific_path ());

      //  create and the new selections
      unsigned int index = 0;
      for (std::vector<edt::Service *>::const_iterator es = edt_services.begin (); es != edt_services.end (); ++es, ++index) {
        for (std::vector<ObjectInstPath>::iterator sel = new_selections [index].begin (); sel != new_selections [index].end (); ++sel) {
          if (int (sel->cv_index ()) == cv_index) {
            sel->insert_front (new_top, removed);
          }
        }
      }

    }

  }

  unsigned int index = 0;
  for (std::vector<edt::Service *>::const_iterator es = edt_services.begin (); es != edt_services.end (); ++es, ++index) {
    (*es)->set_selection (new_selections [index].begin (), new_selections [index].end ());
  }

}

void  
MainService::cm_flatten_insts ()
{
  tl_assert (view ()->is_editable ());

  lay::FlattenInstOptionsDialog options_dialog (view (), false /*don't allow prunining*/);

  if (options_dialog.exec (m_flatten_insts_levels, m_flatten_prune) && m_flatten_insts_levels != 0) {

    view ()->cancel_edits ();

    manager ()->transaction (tl::translate ("Flatten instances"));

    std::vector<edt::Service *> edt_services = view ()->get_plugins <edt::Service> ();
    for (std::vector<edt::Service *>::const_iterator es = edt_services.begin (); es != edt_services.end (); ++es) {

      for (edt::Service::objects::const_iterator r = (*es)->selection ().begin (); r != (*es)->selection ().end (); ++r) {

        const lay::CellView &cv = view ()->cellview (r->cv_index ());
        if (cv.is_valid ()) {

          db::Cell &target_cell = cv->layout ().cell (r->cell_index ());
          if (r->is_cell_inst () && target_cell.is_valid (r->back ().inst_ptr)) {

            //  because we select whole arrays in editable mode, we can iterator over them
            db::CellInstArray cell_inst = r->back ().inst_ptr.cell_inst ();
            for (db::CellInstArray::iterator a = cell_inst.begin (); ! a.at_end (); ++a) {
              cv->layout ().flatten (cv->layout ().cell (r->cell_index_tot ()), target_cell, cell_inst.complex_trans (*a), m_flatten_insts_levels < 0 ? m_flatten_insts_levels : m_flatten_insts_levels - 1);
            }

            target_cell.erase (r->back ().inst_ptr);

          }

        }

      }

    }

    //  The selection is no longer valid
    view ()->clear_selection ();

    manager ()->commit ();

  }
}

void  
MainService::cm_move_hier_up ()
{
  view ()->cancel_edits ();

  manager ()->transaction (tl::translate ("Move up in hierarchy"));

  std::vector<edt::Service *> edt_services = view ()->get_plugins <edt::Service> ();
  for (std::vector<edt::Service *>::const_iterator es = edt_services.begin (); es != edt_services.end (); ++es) {

    std::vector<ObjectInstPath> new_selection;
    new_selection.reserve ((*es)->selection ().size ());

    for (edt::Service::objects::const_iterator r = (*es)->selection ().begin (); r != (*es)->selection ().end (); ++r) {

      const lay::CellView &cv = view ()->cellview (r->cv_index ());
      if (cv.is_valid ()) {

        db::Cell &target_cell = *cv.cell ();

        if (r->is_cell_inst ()) {

          db::Instance new_inst = target_cell.insert (r->back ().inst_ptr);
          new_inst = target_cell.transform (new_inst, db::ICplxTrans (r->trans ()));

          new_selection.push_back (ObjectInstPath ());
          new_selection.back ().set_topcell (r->topcell ());
          new_selection.back ().set_cv_index (r->cv_index ());
          new_selection.back ().add_path (db::InstElement (new_inst, db::CellInstArray::iterator ()));

        } else {

          db::Shapes &target_shapes = target_cell.shapes (r->layer ());
          db::Shape new_shape = target_shapes.insert (r->shape ());
          new_shape = target_shapes.transform (new_shape, db::ICplxTrans (r->trans ()));

          new_selection.push_back (ObjectInstPath ());
          new_selection.back ().set_topcell (r->topcell ());
          new_selection.back ().set_cv_index (r->cv_index ());
          new_selection.back ().set_layer (r->layer ());
          new_selection.back ().set_shape (new_shape);

        }

      }

    }
          
    //  delete all the objects currently selected and set the new selection
    (*es)->del_selected ();

    (*es)->set_selection (new_selection.begin (), new_selection.end ());

  }

  manager ()->commit ();
}

void  
MainService::cm_make_cell_variants ()
{
  //  .. not implemented yet ..
}

void  
MainService::cm_make_cell ()
{
  tl_assert (view ()->is_editable ());

  int cv_index = -1;
  std::vector<edt::Service *> edt_services = view ()->get_plugins <edt::Service> ();
  for (std::vector<edt::Service *>::const_iterator es = edt_services.begin (); es != edt_services.end (); ++es) {
    for (edt::Service::objects::const_iterator r = (*es)->selection ().begin (); r != (*es)->selection ().end (); ++r) {
      if (cv_index < 0) {
        cv_index = r->cv_index ();
      } else if (cv_index != int (r->cv_index ())) {
        throw tl::Exception (tl::translate ("The selection must not contain objects from different layouts for 'make cell'"));
      }
    }
  }

  if (cv_index >= 0) {

    MakeCellOptionsDialog dialog (view ());

    const lay::CellView &cv = view ()->cellview (cv_index);

    if (dialog.exec (cv->layout (), m_make_cell_name)) {

      //  Compute the selection's bbox to establish a good origin for the new cell
      db::Box selection_bbox; 
      db::box_convert<db::CellInst> bc (cv->layout ());
      for (std::vector<edt::Service *>::const_iterator es = edt_services.begin (); es != edt_services.end (); ++es) {
        for (edt::Service::objects::const_iterator r = (*es)->selection ().begin (); r != (*es)->selection ().end (); ++r) {
          if (r->is_cell_inst ()) {
            selection_bbox += db::ICplxTrans (r->trans ()) * r->back ().bbox (bc);
          } else {
            selection_bbox += db::ICplxTrans (r->trans ()) * r->shape ().bbox ();
          }
        }
      }

      if (selection_bbox.empty ()) {
        throw tl::Exception (tl::translate ("The selection is empty. Cannot create a cell from an empty selection."));
      }

      view ()->cancel_edits ();

      manager ()->transaction (tl::translate ("Make cell from selection"));

      db::cell_index_type target_ci = cv->layout ().add_cell (m_make_cell_name.c_str ());
      //  create target cell
      db::Cell &target_cell = cv->layout ().cell (target_ci);

      //  create target cell instance
      db::Instance target_cell_inst = cv.cell ()->insert (db::CellInstArray (db::CellInst (target_ci), db::Trans (selection_bbox.lower_left ())));
      db::ICplxTrans to = db::ICplxTrans (db::Trans (-selection_bbox.lower_left ()));

      for (std::vector<edt::Service *>::const_iterator es = edt_services.begin (); es != edt_services.end (); ++es) {

        for (edt::Service::objects::const_iterator r = (*es)->selection ().begin (); r != (*es)->selection ().end (); ++r) {

          if (r->is_cell_inst ()) {

            db::Instance new_inst = target_cell.insert (r->back ().inst_ptr);
            target_cell.transform (new_inst, to * db::ICplxTrans (r->trans ()));

          } else {

            db::Shapes &target_shapes = target_cell.shapes (r->layer ());
            db::Shape new_shape = target_shapes.insert (r->shape ());
            target_shapes.transform (new_shape, to * db::ICplxTrans (r->trans ()));

          }

        }

        //  delete all the objects currently selected and set the new selection
        (*es)->del_selected ();

        //  establish the new instance as selection for the instance service
        std::vector<ObjectInstPath> new_selection;
        if ((*es)->flags () == db::ShapeIterator::Nothing) {
          new_selection.push_back (ObjectInstPath ());
          new_selection.back ().set_topcell (cv.cell_index ());
          new_selection.back ().set_cv_index (cv_index);
          new_selection.back ().add_path (db::InstElement (target_cell_inst));
        }
        (*es)->set_selection (new_selection.begin (), new_selection.end ());

      }

      manager ()->commit ();

    }

  }

}

/**
 *  @brief Extract a corner radius from a contour
 *
 *  The corners must be formed by soft bending edges. 
 *
 *  @param from, to The iterators describing the contour
 *  @param r The radius (in dbu units) extracted (if return value is true)
 *  @param n The number of points extracted (if return value is true)
 *  @param new_pts The contour without the rounded corners (if return value is true)
 *  @return True, if the extraction was successful
 */
static bool extract_rad_from_contour (db::Polygon::polygon_contour_iterator from, db::Polygon::polygon_contour_iterator to, double &rinner, double &router, unsigned int &n, std::vector <db::Point> &new_pts, bool fallback = false)
{
  db::Polygon::polygon_contour_iterator p0 = from;

  db::Polygon::polygon_contour_iterator p1 = p0;
  ++p1;
  if (p1 == to) {
    p1 = from;
  }
  db::Polygon::polygon_contour_iterator p2 = p1;

  const double cos_thr = 0.8;
  const double circle_segment_thr = 2.5;

  //  search for the first circle segment (where cos(a) > cos_thr) 
  double ls_inner = 0.0, ls_outer = 0.0;
  unsigned long n_ls_inner = 0, n_ls_outer = 0;

  if (! fallback) {

    do {

      ++p2;
      if (p2 == to) {
        p2 = from;
      }

      db::Edge ep (*p0, *p1);
      db::Edge e (*p1, *p2);

      bool inner = (db::vprod_sign (ep, e) > 0);
      double &ls = inner ? ls_inner : ls_outer;
      unsigned long &n_ls = inner ? n_ls_inner : n_ls_outer;

      if (db::sprod (ep, e) > cos_thr * e.double_length () * ep.double_length ()) {
        ls += std::min (e.double_length (), ep.double_length ());
        ++n_ls;
      }

      p0 = p1;
      p1 = p2;

    } while (p0 != from);

    if (n_ls_inner > 0) {
      ls_inner /= n_ls_inner;
    }
    if (n_ls_outer > 0) {
      ls_outer /= n_ls_outer;
    }

  }

  bool found = false;

  //  search for the first circle segment (where cos(a) > cos_thr) 

  p0 = from;

  p1 = p0;
  ++p1;
  if (p1 == to) {
    p1 = from;
  }
  p2 = p1;

  do {

    ++p2;
    if (p2 == to) {
      p2 = from;
    }

    db::Edge ep (*p0, *p1);
    db::Edge e (*p1, *p2);

    bool first_or_last = fallback || (e.double_length () > circle_segment_thr * ep.double_length () || ep.double_length () > circle_segment_thr * e.double_length ());

    if (first_or_last && db::sprod (ep, e) > cos_thr * e.double_length () * ep.double_length ()) {
      double ls = (db::vprod_sign (ep, e) > 0 ? ls_inner : ls_outer);
      if (! fallback && e.double_length () < circle_segment_thr * ls && ep.double_length () > circle_segment_thr * ls) {
        found = true;
        break;
      } else if (fallback && (ep.dx () == 0 || ep.dy () == 0)) {
        found = true;
        break;
      }
    }

    p0 = p1;
    p1 = p2;

  } while (p0 != from);

  if (! found) {
    return false;
  }

  //  create a list of new points without the rounded corners
  new_pts.clear ();

  db::Polygon::polygon_contour_iterator pfirst = p0;
  p2 = p1;
  bool in_corner = false;
  double ls_corner = 0.0;
  db::Edge elast;
  db::Polygon::polygon_contour_iterator plast;
  double asum = 0.0;
  unsigned int nseg = 0;

  double rxi_sum = 0.0;
  double rxo_sum = 0.0;
  double da_sum = 0.0;
  int n_corners = 0;
  int ni_corners = 0;
  int no_corners = 0;

  do {

    ++p2;
    if (p2 == to) {
      p2 = from;
    }

    db::Edge ep (*p0, *p1);
    db::Edge e (*p1, *p2);

    bool first_or_last = fallback || (e.double_length () > circle_segment_thr * ep.double_length () || ep.double_length () > circle_segment_thr * e.double_length ());

    if (db::sprod (ep, e) > cos_thr * e.double_length () * ep.double_length ()) { 

      double ls = db::vprod_sign (ep, e) > 0 ? ls_inner : ls_outer;

      if ((! fallback && first_or_last && e.double_length () < circle_segment_thr * ls && ep.double_length () > circle_segment_thr * ls) ||
          (fallback && (ep.dx () == 0 || ep.dy () == 0))) {

        if (! in_corner) {
          elast = ep;
          plast = p1;
          asum = db::vprod (*p1, *p2);
          nseg = 1;
          ls_corner = ls;
        } 
        in_corner = true;

      } else if ((!fallback && first_or_last && e.double_length () > circle_segment_thr * ls_corner && ep.double_length () < circle_segment_thr * ls_corner) ||
                 (fallback && (e.dx () == 0 || e.dy () == 0))) {

        if (in_corner) {

          std::pair<bool, db::Point> cp = elast.cut_point (e);
          if (! cp.first) {

            //  We have a full 180 degree bend without a stop (actually two corners).
            //  Use the segment in between that is perpendicular to the start and end segment as stop edge.
            db::Polygon::polygon_contour_iterator pp1 = plast;
            db::Polygon::polygon_contour_iterator pp2 = pp1;
            double asum_part = 0.0;
            unsigned int nseg_part = 0;

            while (pp1 != p1) {

              ++pp2;
              if (pp2 == to) {
                pp2 = from;
              }

              e = db::Edge (*pp1, *pp2);
              if (db::sprod_sign (elast, e) == 0) {
                break;
              }

              asum_part += db::vprod (*pp1, *pp2);
              ++nseg_part;

              pp1 = pp2;

            }

            cp = elast.cut_point (e);
            if (! cp.first) {
              return false;
            }

            new_pts.push_back (cp.second);

            ++nseg_part;

            asum -= asum_part;
            asum -= db::vprod (e.p1 (), e.p2 ());
            nseg -= nseg_part;

            asum_part += db::vprod (cp.second, elast.p2 ());
            asum_part += db::vprod (*pp1, cp.second);

            double sin_atot = db::vprod (elast, e);
            double cos_atot = db::sprod (elast, e);
            double atot = fabs (atan2 (sin_atot, cos_atot));

            double rx = sqrt (fabs (asum_part) * 0.5 / (tan (atot * 0.5) - tan (atot * 0.5 / nseg_part) * nseg_part));
            double da = atot / nseg_part;

            if (sin_atot > 0.0) {
              rxi_sum += rx;
              ++ni_corners;
            } else {
              rxo_sum += rx;
              ++no_corners;
            }

            da_sum += da;
            ++n_corners;

            elast = e;

            e = db::Edge (*p1, *p2);
            cp = elast.cut_point (e);
            if (! cp.first) {
              return false;
            }

          }

          new_pts.push_back (cp.second);

          asum += db::vprod (cp.second, elast.p2 ());
          asum += db::vprod (*p1, cp.second);

          ++nseg;

          double sin_atot = db::vprod (elast, e);
          double cos_atot = db::sprod (elast, e);
          double atot = fabs (atan2 (sin_atot, cos_atot));

          double rx = sqrt (fabs (asum) * 0.5 / (tan (atot * 0.5) - tan (atot * 0.5 / nseg) * nseg));
          double da = atot / nseg;

          if (sin_atot > 0.0) {
            rxi_sum += rx;
            ++ni_corners;
          } else {
            rxo_sum += rx;
            ++no_corners;
          }

          da_sum += da;
          ++n_corners;

        }
        in_corner = false;

      } else if (in_corner) {

        asum += db::vprod (*p1, *p2);
        ++nseg;

      } else {

        new_pts.push_back (*p1);

      }

    } else {
      new_pts.push_back (*p1);
    }

    p0 = p1;
    p1 = p2;

  } while (p0 != pfirst);

  if (n_corners < 2) {

    return false;

  } else {
    n = floor (2.0 * M_PI / (da_sum / n_corners) + 0.5);
    if (ni_corners > 0) {
      rinner = floor ((rxi_sum / ni_corners * 0.5) + 0.5) * 2;
    } 
    if (no_corners > 0) {
      router = floor ((rxo_sum / no_corners * 0.5) + 0.5) * 2;
    }
    return true;
  }
}

static void extract_rad (std::vector <db::Polygon> &poly, double &rinner, double &router, unsigned int &n)
{
  std::vector <db::Point> new_pts;

  for (std::vector<db::Polygon>::iterator p = poly.begin (); p != poly.end (); ++p) {

    db::Polygon new_poly;

    new_pts.clear ();
    if (! extract_rad_from_contour (p->begin_hull (), p->end_hull (), rinner, router, n, new_pts) && 
        ! extract_rad_from_contour (p->begin_hull (), p->end_hull (), rinner, router, n, new_pts, true)) {
      //  ultimate fallback: assign original contour
      new_poly.assign_hull (p->begin_hull (), p->end_hull (), false /*don't compress*/);
    } else {
      new_poly.assign_hull (new_pts.begin (), new_pts.end (), true /*compress*/);
    }

    for (unsigned int h = 0; h < p->holes (); ++h) {

      new_pts.clear ();
      if (! extract_rad_from_contour (p->begin_hole (h), p->end_hole (h), rinner, router, n, new_pts) && 
          ! extract_rad_from_contour (p->begin_hole (h), p->end_hole (h), rinner, router, n, new_pts, true)) {
        //  ultimate fallback: assign original contour
        new_poly.insert_hole (p->begin_hole (h), p->end_hole (h), false /*don't compress*/);
      } else {
        new_poly.insert_hole (new_pts.begin (), new_pts.end (), true /*compress*/);
      }

    }

    p->swap (new_poly);


  } 
}

static void compute_rounded (db::Polygon::polygon_contour_iterator from, db::Polygon::polygon_contour_iterator to, std::vector <db::Point> &new_pts, double rinner, double router, unsigned int n)
{
  db::Polygon::polygon_contour_iterator p0 = from;
  db::Polygon::polygon_contour_iterator p1 = p0;
  ++p1;
  if (p1 == to) {
    p1 = from;
  }
  db::Polygon::polygon_contour_iterator p2 = p1;

  do {

    ++p2;
    if (p2 == to) {
      p2 = from;
    }

    if (! db::Edge (*p0, *p1).parallel (db::Edge (*p1, *p2))) {

      db::DPoint e1 = (db::DPoint (*p1) - db::DPoint (*p0)) * (1.0 / (*p0).double_distance (*p1));
      db::DPoint e2 = (db::DPoint (*p2) - db::DPoint (*p1)) * (1.0 / (*p1).double_distance (*p2));

      double sin_a = db::vprod (e1, e2);
      double cos_a = db::sprod (e1, e2);
      double a = fabs (atan2 (sin_a, cos_a));

      double r = sin_a > 0.0 ? rinner : router;

      if ((*p0).double_distance (*p1) < 2.0 * r || (*p2).double_distance (*p1) < 2.0 * r) {
        throw tl::Exception (tl::translate ("Can't apply corner rounding on contours with segments shorter than twice the radius"));
      }

      if (r > 0.0) {

        if (cos_a < -db::epsilon * e1.double_distance () * e2.double_distance ()) {
          throw tl::Exception (tl::translate ("Can't apply corner rounding on contours with sharp angles (more than 90 degree)"));
        }

        db::DPoint q0 = db::DPoint (*p1) - e1 * (tan (a * 0.5) * r);
        db::DPoint n1;
        if (sin_a > 0) {
          n1 = db::DPoint (e1.y (), -e1.x ());
        } else {
          n1 = db::DPoint (-e1.y (), e1.x ());
        }
        db::DPoint pr = q0 - n1 * r;

        double ares = (2.0 * M_PI) / double (n);
        unsigned int nseg = floor (a / ares + 0.5);
        if (nseg == 0) {
          new_pts.push_back (*p1);
        } else {

          double da = a / floor (a / ares + 0.5);
          db::DPoint pl = pr;
          for (double aa = 0.0; aa < a - 1e-6; aa += da) {

            db::DPoint q1 = pr + n1 * (r * cos (aa + da)) + e1 * (r * sin (aa + da));

            //  do a interpolation by computing the crossing point of the tangents of the
            //  circle at a and a+da. This scheme guarantees a low distortion of the original
            //  polygon and enables reverting back to the original polygon to some degree.
            db::DPoint qm = (q0 + q1) * 0.5;
            db::DPoint q = qm + (qm - pr) * (q0.sq_distance (qm) / pr.sq_distance (qm));

            new_pts.push_back (db::Point::from_double (q));

            q0 = q1;

          }

        }

      } else {
        new_pts.push_back (*p1);
      }

    } else {
      new_pts.push_back (*p1);
    }

    p0 = p1;
    p1 = p2;

  } while (p0 != from);
}

static void apply_rounding (std::vector <db::Polygon> &poly, double rinner, double router, unsigned int n)
{
  std::vector <db::Point> new_pts;

  for (std::vector <db::Polygon>::iterator p = poly.begin (); p != poly.end (); ++p) {

    db::Polygon new_poly;

    new_pts.clear ();
    compute_rounded (p->begin_hull (), p->end_hull (), new_pts, rinner, router, n);
    new_poly.assign_hull (new_pts.begin (), new_pts.end (), false /*don't compress*/);

    for (unsigned int h = 0; h < p->holes (); ++h) {
      new_pts.clear ();
      compute_rounded (p->begin_hole (h), p->end_hole (h), new_pts, rinner, router, n);
      new_poly.insert_hole (new_pts.begin (), new_pts.end (), false /*don't compress*/);
    }

    p->swap (new_poly);

  }
}

void
MainService::cm_round_corners ()
{
  tl_assert (view ()->is_editable ());

  int cv_index = -1;
  int layer_index = -1;

  std::vector<edt::Service *> edt_services = view ()->get_plugins <edt::Service> ();

  std::vector<db::Polygon> primary;

  //  get (common) cellview index of the primary selection
  for (std::vector<edt::Service *>::const_iterator es = edt_services.begin (); es != edt_services.end (); ++es) {
    for (edt::Service::obj_iterator s = (*es)->selection ().begin (); s != (*es)->selection ().end (); ++s) {

      if (! s->is_cell_inst () && (s->shape ().is_polygon () || s->shape ().is_path () || s->shape ().is_box ())) {

        if (cv_index >= 0 && cv_index != int (s->cv_index ())) {
          throw tl::Exception (tl::translate ("Selection originates from different layouts - cannot compute result in this case."));
        }
        cv_index = int (s->cv_index ());

        if (layer_index >= 0 && layer_index != int (s->layer ())) {
          throw tl::Exception (tl::translate ("Selection originates from different layers - cannot compute result in this case."));
        }
        layer_index = int (s->layer ());

        primary.push_back (db::Polygon ());
        s->shape ().polygon (primary.back ());
        primary.back ().transform (s->trans ());

      }

    }
  }

  if (cv_index < 0 || layer_index < 0) {
    throw tl::Exception (tl::translate ("Selection does not contain polygons"));
  }

  db::EdgeProcessor ep;
  std::vector <db::Polygon> out;
  ep.merge (primary, out, 0 /*min_wc*/, false /*resolve holes*/, true /*min coherence*/);

  unsigned int n = 100;
  double rinner = 0.0, router = 0.0;
  extract_rad (out, rinner, router, n);

  const lay::CellView &cv = view ()->cellview (cv_index);
  double dbu = cv->layout ().dbu ();

  rinner *= dbu;
  router *= dbu;

  RoundCornerOptionsDialog dialog (view ());
  if (! dialog.exec (cv->layout (), router, rinner, n)) {
    return;
  }

  apply_rounding (out, rinner / dbu, router / dbu, n);

  //  remove holes (result in primary)
  primary.clear ();
  ep.merge (out, primary, 0 /*min_wc*/, true /*resolve holes*/, true /*min coherence*/);

  view ()->cancel_edits ();
  manager ()->transaction (tl::translate ("Corner rounding operation on selection"));

  //  Delete the current selection
  for (std::vector<edt::Service *>::const_iterator es = edt_services.begin (); es != edt_services.end (); ++es) {
    for (edt::Service::obj_iterator s = (*es)->selection ().begin (); s != (*es)->selection ().end (); ++s) {
      if (! s->is_cell_inst () && (s->shape ().is_polygon () || s->shape ().is_path () || s->shape ().is_box ())) {
        db::Cell &cell = view ()->cellview (s->cv_index ())->layout ().cell (s->cell_index ());
        if (cell.shapes (s->layer ()).is_valid (s->shape ())) {
          cell.shapes (s->layer ()).erase_shape (s->shape ());
        }
      }
    }
  }

  //  The selection is no longer valid
  view ()->clear_selection ();

  //  Insert the new shapes on top level
  std::vector<ObjectInstPath> new_selection;
  new_selection.reserve (primary.size ());

  //  create the new shapes 
  db::Shapes &target_shapes = cv->layout ().cell (cv.cell_index ()).shapes (layer_index);
  for (std::vector <db::Polygon>::const_iterator p = primary.begin (); p != primary.end (); ++p) {
    db::Shape new_shape = target_shapes.insert (*p);
    new_selection.push_back (ObjectInstPath ());
    new_selection.back ().set_topcell (cv.cell_index ());
    new_selection.back ().set_cv_index (cv_index);
    new_selection.back ().set_layer (layer_index);
    new_selection.back ().set_shape (new_shape);
  }
  
  //  set the new selection on the polygon service (because now we have polygons)
  for (std::vector<edt::Service *>::const_iterator es = edt_services.begin (); es != edt_services.end (); ++es) {
    if (dynamic_cast <edt::PolygonService *> (*es) != 0) {
      (*es)->set_selection (new_selection.begin (), new_selection.end ());
      break;
    }
  }

  manager ()->commit ();
}

void
MainService::cm_size ()
{
  tl_assert (view ()->is_editable ());

  int cv_index = -1;
  int layer_index = -1;

  std::vector<edt::Service *> edt_services = view ()->get_plugins <edt::Service> ();

  std::vector<db::Polygon> primary;

  //  get (common) cellview index of the primary selection
  for (std::vector<edt::Service *>::const_iterator es = edt_services.begin (); es != edt_services.end (); ++es) {
    for (edt::Service::obj_iterator s = (*es)->selection ().begin (); s != (*es)->selection ().end (); ++s) {

      if (! s->is_cell_inst () && (s->shape ().is_polygon () || s->shape ().is_path () || s->shape ().is_box ())) {

        if (cv_index >= 0 && cv_index != int (s->cv_index ())) {
          throw tl::Exception (tl::translate ("Selection originates from different layouts - cannot compute result in this case."));
        }
        cv_index = int (s->cv_index ());

        if (layer_index >= 0 && layer_index != int (s->layer ())) {
          throw tl::Exception (tl::translate ("Selection originates from different layers - cannot compute result in this case."));
        }
        layer_index = int (s->layer ());

        primary.push_back (db::Polygon ());
        s->shape ().polygon (primary.back ());
        primary.back ().transform (s->trans ());

      }

    }
  }

  if (cv_index < 0 || layer_index < 0) {
    throw tl::Exception (tl::translate ("Selection does not contain polygons"));
  }

  bool ok = false;
  QString s = QInputDialog::getText (view (), 
                                     tl::translate ("Sizing").c_str (), 
                                     tl::translate ("Sizing (in micron, positive or negative). Two values (dx, dy) for anisotropic sizing.").c_str (), 
                                     QLineEdit::Normal, "0.0", 
                                     &ok);

  if (!ok) {
    return;
  }

  double dx = 0.0, dy = 0.0;
  tl::Extractor ex (s.ascii ());
  ex.read (dx);
  if (ex.test (",")) {
    ex.read (dy);
  } else {
    dy = dx;
  }

  const lay::CellView &cv = view ()->cellview (cv_index);
  double dbu = cv->layout ().dbu ();
  db::Coord idx = db::coord_traits<db::Coord>::rounded (dx / dbu);
  db::Coord idy = db::coord_traits<db::Coord>::rounded (dy / dbu);

  std::vector <db::Polygon> out;
  db::EdgeProcessor ep;
  ep.size (primary, idx, idy, out, 2 /*mode, TODO: make variable*/, true /*resolve holes*/, true /*min coherence*/);

  view ()->cancel_edits ();
  manager ()->transaction (tl::translate ("Sizing operation on selection"));

  //  Delete the current selection
  for (std::vector<edt::Service *>::const_iterator es = edt_services.begin (); es != edt_services.end (); ++es) {
    for (edt::Service::obj_iterator s = (*es)->selection ().begin (); s != (*es)->selection ().end (); ++s) {
      if (! s->is_cell_inst () && (s->shape ().is_polygon () || s->shape ().is_path () || s->shape ().is_box ())) {
        db::Cell &cell = view ()->cellview (s->cv_index ())->layout ().cell (s->cell_index ());
        if (cell.shapes (s->layer ()).is_valid (s->shape ())) {
          cell.shapes (s->layer ()).erase_shape (s->shape ());
        }
      }
    }
  }

  //  The selection is no longer valid
  view ()->clear_selection ();

  //  Insert the new shapes on top level
  std::vector<ObjectInstPath> new_selection;
  new_selection.reserve (out.size ());

  //  create the new shapes 
  db::Shapes &target_shapes = cv->layout ().cell (cv.cell_index ()).shapes (layer_index);
  for (std::vector <db::Polygon>::const_iterator p = out.begin (); p != out.end (); ++p) {
    db::Shape new_shape = target_shapes.insert (*p);
    new_selection.push_back (ObjectInstPath ());
    new_selection.back ().set_topcell (cv.cell_index ());
    new_selection.back ().set_cv_index (cv_index);
    new_selection.back ().set_layer (layer_index);
    new_selection.back ().set_shape (new_shape);
  }
  
  //  set the new selection on the polygon service (because now we have polygons)
  for (std::vector<edt::Service *>::const_iterator es = edt_services.begin (); es != edt_services.end (); ++es) {
    if (dynamic_cast <edt::PolygonService *> (*es) != 0) {
      (*es)->set_selection (new_selection.begin (), new_selection.end ());
      break;
    }
  }

  manager ()->commit ();
}

void
MainService::boolean_op (int mode)
{
  tl_assert (view ()->is_editable ());

  int cv_index = -1;
  int layer_index = -1;

  std::vector<edt::Service *> edt_services = view ()->get_plugins <edt::Service> ();

  std::vector<db::Polygon> primary;

  //  get (common) cellview index of the primary selection
  for (std::vector<edt::Service *>::const_iterator es = edt_services.begin (); es != edt_services.end (); ++es) {
    for (edt::Service::obj_iterator s = (*es)->selection ().begin (); s != (*es)->selection ().end (); ++s) {

      if (s->seq () == 0 && ! s->is_cell_inst () && (s->shape ().is_polygon () || s->shape ().is_path () || s->shape ().is_box ())) {

        if (cv_index >= 0 && cv_index != int (s->cv_index ())) {
          throw tl::Exception (tl::translate ("Primary selection originates from different layouts - cannot compute result in this case."));
        }
        cv_index = int (s->cv_index ());

        if (layer_index >= 0 && layer_index != int (s->layer ())) {
          throw tl::Exception (tl::translate ("Primary selection originates from different layers - cannot compute result in this case."));
        }
        layer_index = int (s->layer ());

        primary.push_back (db::Polygon ());
        s->shape ().polygon (primary.back ());
        primary.back ().transform (s->trans ());

      }

    }
  }

  if (cv_index < 0 || layer_index < 0) {
    throw tl::Exception (tl::translate ("Primary selection does not contain polygons"));
  }

  std::vector<db::Polygon> secondary;

  const lay::CellView &cv = view ()->cellview (cv_index);
  double dbu_prim = cv->layout ().dbu ();

  //  get the secondary selection
  for (std::vector<edt::Service *>::const_iterator es = edt_services.begin (); es != edt_services.end (); ++es) {
    for (edt::Service::obj_iterator s = (*es)->selection ().begin (); s != (*es)->selection ().end (); ++s) {

      if (s->seq () > 0 && ! s->is_cell_inst () && (s->shape ().is_polygon () || s->shape ().is_path () || s->shape ().is_box ())) {

        double dbu_sec = view ()->cellview (s->cv_index ())->layout ().dbu ();

        secondary.push_back (db::Polygon ());
        s->shape ().polygon (secondary.back ());

        //  HINT: this transforms the shape without any grid snapping precautions ..
        secondary.back ().transform (db::CplxTrans (dbu_sec / dbu_prim) * s->trans ());

      }

    }
  }

  std::vector <db::Polygon> out;
  db::EdgeProcessor ep;
  ep.boolean (primary, secondary, out, mode);

  view ()->cancel_edits ();
  manager ()->transaction (tl::translate ("Boolean operation on selection"));

  //  Delete the current selection
  for (std::vector<edt::Service *>::const_iterator es = edt_services.begin (); es != edt_services.end (); ++es) {
    for (edt::Service::obj_iterator s = (*es)->selection ().begin (); s != (*es)->selection ().end (); ++s) {
      if (! s->is_cell_inst () && (s->shape ().is_polygon () || s->shape ().is_path () || s->shape ().is_box ())) {
        db::Cell &cell = view ()->cellview (s->cv_index ())->layout ().cell (s->cell_index ());
        if (cell.shapes (s->layer ()).is_valid (s->shape ())) {
          cell.shapes (s->layer ()).erase_shape (s->shape ());
        }
      }
    }
  }

  //  The selection is no longer valid
  view ()->clear_selection ();

  //  Insert the new shapes on top level
  std::vector<ObjectInstPath> new_selection;
  new_selection.reserve (out.size ());

  //  create the new shapes 
  db::Shapes &target_shapes = cv->layout ().cell (cv.cell_index ()).shapes (layer_index);
  for (std::vector <db::Polygon>::const_iterator p = out.begin (); p != out.end (); ++p) {
    db::Shape new_shape = target_shapes.insert (*p);
    new_selection.push_back (ObjectInstPath ());
    new_selection.back ().set_topcell (cv.cell_index ());
    new_selection.back ().set_cv_index (cv_index);
    new_selection.back ().set_layer (layer_index);
    new_selection.back ().set_shape (new_shape);
  }
  
  //  set the new selection on the polygon service (because now we have polygons)
  for (std::vector<edt::Service *>::const_iterator es = edt_services.begin (); es != edt_services.end (); ++es) {
    if (dynamic_cast <edt::PolygonService *> (*es) != 0) {
      (*es)->set_selection (new_selection.begin (), new_selection.end ());
      break;
    }
  }

  manager ()->commit ();
}

void 
MainService::cm_union ()
{
  boolean_op (db::BooleanOp::Or);
}

void 
MainService::cm_intersection ()
{
  boolean_op (db::BooleanOp::And);
}

void 
MainService::cm_difference ()
{
  boolean_op (db::BooleanOp::ANotB);
}

static 
db::DVector compute_alignment_vector (const db::DBox &prim_box, const db::DBox &box, int hmode, int vmode)
{
  double dx = 0.0;
  if (hmode == 1) {
    dx = prim_box.left () - box.left ();
  } else if (hmode == 2) {
    dx = prim_box.center ().x () - box.center ().x ();
  } else if (hmode == 3) {
    dx = prim_box.right () - box.right ();
  }

  double dy = 0.0;
  if (vmode == 1) {
    dy = prim_box.top () - box.top ();
  } else if (vmode == 2) {
    dy = prim_box.center ().y () - box.center ().y ();
  } else if (vmode == 3) {
    dy = prim_box.bottom () - box.bottom ();
  }

  return db::DVector (dx, dy);
}

static db::DBox 
inst_bbox (const db::DCplxTrans &tr, lay::LayoutView *view, int cv_index, const db::InstElement &inst_element, bool visible_only)
{
  db::DBox box;

  const db::Layout &layout = view->cellview (cv_index)->layout ();

  if (visible_only) {
    for (lay::LayerPropertiesConstIterator l = view->begin_layers (); ! l.at_end (); ++l) {
      if (l->cellview_index () == cv_index && l->visible (true)) {
        db::box_convert<db::CellInst> bc (layout, l->layer_index ());
        box += tr * inst_element.bbox (bc);
      }
    }
  } else {
    db::box_convert<db::CellInst> bc (layout);
    box += tr * inst_element.bbox (bc);
  }

  return box;
}

void 
MainService::cm_align ()
{
  tl_assert (view ()->is_editable ());

  std::vector<edt::Service *> edt_services = view ()->get_plugins <edt::Service> ();

  AlignOptionsDialog dialog (view ());
  if (! dialog.exec (view (), m_align_hmode, m_align_vmode, m_align_visible_layers)) {
    return;
  }

  db::DBox prim_box;

  //  get (common) bbox index of the primary selection
  for (std::vector<edt::Service *>::const_iterator es = edt_services.begin (); es != edt_services.end (); ++es) {
    for (edt::Service::obj_iterator s = (*es)->selection ().begin (); s != (*es)->selection ().end (); ++s) {

      if (s->seq () == 0) {

        const db::Layout &layout = view ()->cellview (s->cv_index ())->layout ();
        db::DCplxTrans tr = db::DCplxTrans (layout.dbu ()) * db::DCplxTrans (s->trans ());

        if (! s->is_cell_inst ()) {
          prim_box += tr * s->shape ().bbox ();
        } else {
          prim_box += inst_bbox (tr, view (), s->cv_index (), s->back (), m_align_visible_layers);
        }

      }

    }
  }

  if (! prim_box.empty ()) {

    view ()->cancel_edits ();
    manager ()->transaction (tl::translate ("Alignment"));

    //  do the alignment
    for (std::vector<edt::Service *>::const_iterator es = edt_services.begin (); es != edt_services.end (); ++es) {

      //  create a transformation vector that describes each shape's transformation
      std::vector <db::DCplxTrans> tv;
      tv.reserve ((*es)->selection ().size ());

      for (edt::Service::obj_iterator s = (*es)->selection ().begin (); s != (*es)->selection ().end (); ++s) {

        db::DVector v;

        if (s->seq () > 0) {

          db::Layout &layout = view ()->cellview (s->cv_index ())->layout ();
          db::DCplxTrans tr = db::DCplxTrans (layout.dbu ()) * db::DCplxTrans (s->trans ());

          if (! s->is_cell_inst ()) {

            db::DBox box = tr * s->shape ().bbox ();
            v = compute_alignment_vector (prim_box, box, m_align_hmode, m_align_vmode);

          } else {

            db::DBox box = inst_bbox (tr, view (), s->cv_index (), s->back (), m_align_visible_layers);
            v = compute_alignment_vector (prim_box, box, m_align_hmode, m_align_vmode);

          }

        }

        tv.push_back (db::DCplxTrans (db::DTrans (v)));

      }

      //  use the "transform" method to transform the shapes and instances (with individual transformations)
      (*es)->transform (db::DCplxTrans () /*dummy*/, &tv);

    }

    manager ()->commit ();

  }
}

void 
MainService::cm_change_layer ()
{
  tl_assert (view ()->is_editable ());

  int cv_index = -1;

  std::vector<edt::Service *> edt_services = view ()->get_plugins <edt::Service> ();

  //  get (common) cellview index of the selected shapes
  for (std::vector<edt::Service *>::const_iterator es = edt_services.begin (); es != edt_services.end (); ++es) {
    for (edt::Service::obj_iterator s = (*es)->selection ().begin (); s != (*es)->selection ().end (); ++s) {
      if (! s->is_cell_inst ()) {
        if (cv_index >= 0 && cv_index != int (s->cv_index ())) {
          throw tl::Exception (tl::translate ("Selections originate from different layouts - cannot switch layer in this case."));
        }
        cv_index = int (s->cv_index ());
      }
    }
  }

  if (cv_index >= 0) {

    //  HINT: the following code is copied from ShapeEditService::get_edit_layer - it should
    //  be at some common place

    lay::LayerPropertiesConstIterator cl = view ()->current_layer ();
    if (cl.is_null ()) {
      throw tl::Exception (tl::translate ("Please select a layer first").c_str ());
    }
    
    if (cv_index != cl->cellview_index ()) {
      throw tl::Exception (tl::translate ("Shapes cannot be moved to a different layout").c_str ());
    }

    const lay::CellView &cv = view ()->cellview (cv_index);
    int layer = cl->layer_index ();

    if (! cv.is_valid ()) {
      throw tl::Exception (tl::translate ("Please select a cell first").c_str ());
    }

    if (layer < 0 || ! cv->layout ().is_valid_layer ((unsigned int) layer)) {

      if (cl->has_children ()) {
        throw tl::Exception (tl::translate ("Please select a valid drawing layer first").c_str ());
      } else {

        //  create this layer now
        const lay::ParsedLayerSource &source = cl->source (true /*real*/);

        db::LayerProperties db_lp;
        if (source.has_name ()) {
          db_lp.name = source.name ();
        }
        db_lp.layer = source.layer ();
        db_lp.datatype = source.datatype ();

        cv->layout ().insert_layer (db_lp);

        //  update the layer index inside the layer view
        cl->realize_source ();
          
        //  Hint: we could have taken the new index from insert_layer, but this 
        //  is a nice test:
        layer = cl->layer_index ();
        tl_assert (layer >= 0);

      }

    }

    view ()->cancel_edits ();

    manager ()->transaction (tl::translate ("Change layer"));

    db::Layout &layout = view ()->cellview (cv_index)->layout ();

    //  Insert and delete the shape. This exploits the fact, that a shape can be erased multiple times -
    //  this is important since the selection potentially contains the same shape multiple times.
    for (std::vector<edt::Service *>::const_iterator es = edt_services.begin (); es != edt_services.end (); ++es) {
      for (edt::Service::obj_iterator s = (*es)->selection ().begin (); s != (*es)->selection ().end (); ++s) {
        if (!s->is_cell_inst () && int (s->layer ()) != layer) {
          db::Cell &cell = layout.cell (s->cell_index ());
          if (cell.shapes (s->layer ()).is_valid (s->shape ())) {
            cell.shapes (layer).insert (s->shape ());
            cell.shapes (s->layer ()).erase_shape (s->shape ());
          }
        }
      }
    }
    
    //  The selection is no longer valid
    view ()->clear_selection ();

    manager ()->commit ();

  } else {
    throw tl::Exception (tl::translate ("Nothing selected to switch layers for"));
  }

}

void  
MainService::cm_edit_options ()
{
  show_editor_options_dialog ();
}

/**
 *  @brief Implementation of the insert notification object 
 *
 *  The basic purpose of this object is to provide a new selection for the
 *  new shapes inserted.
 */
class NewObjectsSelection
 : public db::ClipboardDataInsertReceiver 
{
public:
  NewObjectsSelection (int cv_index, db::cell_index_type topcell, lay::LayoutView *view)
    : m_cv_index (cv_index), m_topcell (topcell), mp_view (view)
  {
    mp_polygon_service = view->get_plugin <edt::PolygonService> ();
    mp_box_service = view->get_plugin <edt::BoxService> ();
    mp_text_service = view->get_plugin <edt::TextService> ();
    mp_path_service = view->get_plugin <edt::PathService> ();
    mp_inst_service = view->get_plugin <edt::InstService> ();
    tl_assert (mp_polygon_service);
    tl_assert (mp_box_service);
    tl_assert (mp_text_service);
    tl_assert (mp_path_service);
    tl_assert (mp_inst_service);
  }

  void finish ()
  {
    mp_polygon_service->selection_to_view ();
    mp_box_service->selection_to_view ();
    mp_text_service->selection_to_view ();
    mp_path_service->selection_to_view ();
    mp_inst_service->selection_to_view ();
  }

  void shape_inserted (db::cell_index_type cell, int layer, const db::Shape &shape) 
  {
    edt::ObjectInstPath sel;
    sel.set_cv_index (m_cv_index);
    sel.set_topcell (m_topcell);
    sel.set_layer (layer);
    sel.set_shape (shape);

    if (m_topcell != cell) {
      return; // HINT insertion into subcell not supported currently
    }

    if (shape.is_polygon ()) {
      mp_polygon_service->add_selection (sel);
    } else if (shape.is_box ()) {
      mp_box_service->add_selection (sel);
    } else if (shape.is_text ()) {
      mp_text_service->add_selection (sel);
    } else if (shape.is_path ()) {
      mp_path_service->add_selection (sel);
    } 
  }

  void instance_inserted (db::cell_index_type cell, const db::Instance &instance) 
  {
    edt::ObjectInstPath sel;
    sel.set_cv_index (m_cv_index);
    sel.set_topcell (m_topcell);
    sel.add_path (db::InstElement (instance, db::CellInstArray::iterator ()));

    if (m_topcell != cell) {
      return; // HINT insertion into subcell not supported currently
    }

    mp_inst_service->add_selection (sel);
  }

private:
  edt::PolygonService *mp_polygon_service;
  edt::BoxService *mp_box_service;
  edt::TextService *mp_text_service;
  edt::PathService *mp_path_service;
  edt::InstService *mp_inst_service;
  int m_cv_index;
  db::cell_index_type m_topcell;
  lay::LayoutView *mp_view;
};

void 
MainService::paste ()
{
  if (view ()->is_editable ()) {

    int cv_index = view ()->active_cellview_index ();
    const lay::CellView &cv = view ()->cellview (cv_index);

    NewObjectsSelection insert_notification (cv_index, cv.cell_index (), view ());

    std::vector<unsigned int> new_layers;

    //  paste the content into the active cellview.
    for (db::Clipboard::iterator c = db::Clipboard::instance ().begin (); c != db::Clipboard::instance ().end (); ++c) {
      const db::ClipboardValue<edt::ClipboardData> *value = dynamic_cast<const db::ClipboardValue<edt::ClipboardData> *> (*c);
      if (value) {

        if (! cv.is_valid ()) {
          throw tl::Exception (tl::translate ("No cell selected to paste something into"));
        }

        std::vector<unsigned int> nl = value->get ().insert (cv->layout (), cv.context_trans ().inverted (), &cv->layout ().cell (cv.cell_index ()), 0 /*new_tops*/, &insert_notification);
        new_layers.insert (new_layers.end (), nl.begin (), nl.end ());

      }
    }

    //  Add new layers to the view if required.
    view ()->add_new_layers (new_layers, cv_index);
    view ()->update_content ();

    //  make the selection visible 
    insert_notification.finish ();

  }
}

} // namespace edt


