
/*

  KLayout Layout Viewer
  Copyright (C) 2006-2007 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#include "edtPropertiesPages.h"
#include "edtUtils.h"
#include "edtDialogs.h"
#include "layLayoutView.h"
#include "tlString.h"

#include <string>

namespace edt
{

// -------------------------------------------------------------------------
//  ShapePropertiesPage implementation

ShapePropertiesPage::ShapePropertiesPage (edt::Service *service, QWidget *parent)
  : lay::PropertiesPage (parent, service), 
    mp_service (service), m_enable_cb_callback (true)
{
  m_selection_ptrs.reserve (service->selection ().size ());
  for (edt::Service::obj_iterator s = service->selection ().begin (); s != service->selection ().end (); ++s) {
    m_selection_ptrs.push_back (s);
  }
  m_index = 0;
  m_prop_id = 0;
  mp_service->clear_highlights ();
}

ShapePropertiesPage::~ShapePropertiesPage ()
{
  mp_service->restore_highlights ();
}

void 
ShapePropertiesPage::setup ()
{
  connect (dbu_checkbox (), SIGNAL (toggled (bool)), this, SLOT (display_mode_changed (bool)));
  connect (abs_checkbox (), SIGNAL (toggled (bool)), this, SLOT (display_mode_changed (bool)));

  m_enable_cb_callback = false;
  dbu_checkbox ()->setChecked (mp_service->view ()->dbu_coordinates ());
  abs_checkbox ()->setChecked (mp_service->view ()->absolute_coordinates ());
  m_enable_cb_callback = true;
}

void 
ShapePropertiesPage::back ()
{
  m_index = m_selection_ptrs.size ();
}

void 
ShapePropertiesPage::front ()
{
  m_index = 0;
}

bool 
ShapePropertiesPage::at_begin () const
{
  return (m_index == 0);
}

bool 
ShapePropertiesPage::at_end () const
{
  return (m_index == m_selection_ptrs.size ());
}

void 
ShapePropertiesPage::operator-- ()
{
  --m_index;
}

void 
ShapePropertiesPage::operator++ ()
{
  ++m_index;
}

void 
ShapePropertiesPage::leave ()
{
  mp_service->clear_highlights ();
}

bool 
ShapePropertiesPage::dbu_units () const
{
  return dbu_checkbox ()->isChecked ();
}

bool 
ShapePropertiesPage::abs_trans () const
{
  return abs_checkbox ()->isChecked ();
}

db::CplxTrans
ShapePropertiesPage::trans () const
{
  if (abs_trans ()) {
    return m_selection_ptrs[m_index]->trans ();
  } else {
    return db::CplxTrans ();
  }
}

void 
ShapePropertiesPage::display_mode_changed (bool)
{
  if (! m_enable_cb_callback) {
    return;
  }

  update_shape ();
}

void 
ShapePropertiesPage::update ()
{
  mp_service->highlight (m_index);

  update_shape ();
}

void 
ShapePropertiesPage::apply ()
{
  edt::Service::obj_iterator pos = m_selection_ptrs [m_index];

  const lay::CellView &cv = mp_service->view ()->cellview (pos->cv_index ());

  tl_assert (! pos->is_cell_inst ());

  db::Shapes &shapes = cv->layout ().cell (pos->cell_index ()).shapes (pos->layer ());
  double dbu = cv->layout ().dbu ();

  if (pos->shape ().is_array_member ()) {
    throw tl::Exception (tl::translate ("Shape array members cannot be changed"));
  }

  db::Shape new_shape = do_apply (shapes, pos->shape (), dbu);

  if (new_shape != pos->shape ()) {

    //  change selection to new shape
    ObjectInstPath new_sel (*pos);
    new_sel.set_shape (new_shape);

    mp_service->select (*pos, lay::Editable::Reset);
    mp_service->select (new_sel, lay::Editable::Add);

    // locate the new selection
    for (pos = mp_service->selection ().begin (); pos != mp_service->selection ().end () && *pos != new_sel; ++pos) 
      ;
    tl_assert (pos != mp_service->selection ().end ());
    m_selection_ptrs [m_index] = pos;

    mp_service->selection_to_view ();

  } 

  update ();
}

void 
ShapePropertiesPage::update_shape ()
{
  edt::Service::obj_iterator pos = m_selection_ptrs [m_index];

  const lay::CellView &cv = mp_service->view ()->cellview (pos->cv_index ());
  double dbu = cv->layout ().dbu ();

  tl_assert (! pos->is_cell_inst ());

  std::string layer (tl::translate ("Layer "));
  layer += cv->layout ().get_properties (pos->layer ()).to_string ();
  layer += ", ";
  layer += tl::translate ("Cell ");
  layer += cv->layout ().cell_name (pos->cell_index ());

  m_prop_id = pos->shape ().prop_id ();

  do_update (pos->shape (), dbu, layer);
}

void
ShapePropertiesPage::show_inst ()
{
  InstantiationForm inst_form (this);
  inst_form.show (mp_service->view (), *m_selection_ptrs [m_index]);
}

void
ShapePropertiesPage::show_props ()
{
  edt::Service::obj_iterator pos = m_selection_ptrs [m_index];

  UserPropertiesForm props_form (this);
  if (props_form.show (mp_service->view (), pos->cv_index (), m_prop_id)) {
    if (! readonly () && m_prop_id != pos->shape ().prop_id ()) {

      const lay::CellView &cv = mp_service->view ()->cellview (pos->cv_index ());
      db::Shapes &shapes = cv->layout ().cell (pos->cell_index ()).shapes (pos->layer ());

      db::Transaction transaction (cv->layout ().manager (), tl::translate ("Change shape properties"));

      db::Shape new_shape = shapes.replace_prop_id (pos->shape (), m_prop_id);

      if (new_shape != pos->shape ()) {

        //  change selection to new shape
        ObjectInstPath new_sel (*pos);
        new_sel.set_shape (new_shape);

        mp_service->select (*pos, lay::Editable::Reset);
        mp_service->select (new_sel, lay::Editable::Add);

        // locate the new selection
        for (pos = mp_service->selection ().begin (); pos != mp_service->selection ().end () && *pos != new_sel; ++pos) 
          ;
        tl_assert (pos != mp_service->selection ().end ());
        m_selection_ptrs [m_index] = pos;

        mp_service->selection_to_view ();

      } 

      update ();

    }
  }
}

bool 
ShapePropertiesPage::readonly ()
{
  return ! mp_service->view ()->is_editable ();
}

// -------------------------------------------------------------------------
//  helper functions to convert coordinates

static std::string
coord_to_string (double dc, double dbu, bool du)
{
  if (du) {
    return tl::db_to_string (dc);
  } else {
    return tl::micron_to_string (dc * dbu);
  }
}

static db::Coord
coord_from_dcoord (double d, double dbu, bool du, const db::CplxTrans &t)
{
  if (du) {
    dbu = 1.0;
  }
  
  db::DCoord dc = t.ctrans (d / dbu);
  db::Coord c = db::Coord (floor (dc + 0.5));
  if (fabs (dc - c) > 1e-6) { 
    if (! du) {
      throw tl::Exception (tl::translate ("%g is not a multiple of the database unit (%g)").c_str (), d, dbu);
    } else {
      throw tl::Exception (tl::translate ("%g is not a multiple of the database unit").c_str (), d);
    }
  }

  return c;
}

static db::Point
coord_from_dpoint (const db::DPoint &dp, double dbu, bool du, const db::CplxTrans &t)
{
  if (du) {
    dbu = 1.0;
  }
  
  db::DPoint dpt = t * (dp * (1.0 / dbu));

  db::Coord cx = coord_from_dcoord (dpt.x (), 1.0, false, db::CplxTrans ());
  db::Coord cy = coord_from_dcoord (dpt.y (), 1.0, false, db::CplxTrans ());

  return db::Point (cx, cy);
}

static db::Coord
coord_from_string (const char *txt, double dbu, bool du, const db::CplxTrans &t)
{
  double d = 0.0;
  tl::from_string (txt, d);
  return coord_from_dcoord (d, dbu, du, t);
}

static std::string
coords_to_string (const db::DPoint &dp, double dbu, bool du, const char *sep = "\t")
{
  return coord_to_string (dp.x (), dbu, du) + sep + coord_to_string (dp.y (), dbu, du);
}

// -------------------------------------------------------------------------
//  PolygonPropertiesPage implementation

PolygonPropertiesPage::PolygonPropertiesPage (edt::Service *service, QWidget *parent)
  : ShapePropertiesPage (service, parent)
{
  setupUi (this);
  setup ();

  connect (inst_pb, SIGNAL (clicked ()), this, SLOT (show_inst ()));
  connect (prop_pb, SIGNAL (clicked ()), this, SLOT (show_props ()));
}

void 
PolygonPropertiesPage::do_update (const db::Shape &shape, double dbu, const std::string &lname)
{
  layer_lbl->setText (lname.c_str ());

  db::Polygon poly;
  shape.polygon (poly);

  std::string ptlist;
  ptlist.reserve (4096);

  db::CplxTrans t = trans ();
  bool du = dbu_units ();

  bool first = true;
  for (db::Polygon::polygon_contour_iterator pt = poly.begin_hull (); pt != poly.end_hull (); ++pt) {
    if (! first) {
      ptlist += "\n";
    } else {
      first = false;
    }
    ptlist += coords_to_string (t * *pt, dbu, du);
  }

  pointListEdit->setText (ptlist.c_str ());
  pointCountLabel->setText (tl::sprintf (tl::translate ("(%lu points)").c_str (), poly.vertices ()).c_str());
}

db::Shape
PolygonPropertiesPage::do_apply (db::Shapes &shapes, const db::Shape &shape, double dbu)
{
  QString text (pointListEdit->text ());
  tl::Extractor ex (text.ascii ());

  std::vector <db::Point> points;

  db::CplxTrans t = trans ().inverted ();
  bool du = dbu_units ();

  while (! ex.at_end ()) {

    double dx = 0.0, dy = 0.0;
    ex.read (dx);
    ex.read (dy);

    points.push_back (coord_from_dpoint (db::DPoint (dx, dy), dbu, du, t));

  }

  if (points.size () < 3) {
    throw tl::Exception (tl::translate ("Polygon must have at least three points"));
  }

  db::Polygon poly;
  poly.assign_hull (points.begin (), points.end (), false /*not compressed*/);

  return shapes.replace (shape, poly);
}

// -------------------------------------------------------------------------
//  BoxPropertiesPage implementation

BoxPropertiesPage::BoxPropertiesPage (edt::Service *service, QWidget *parent)
  : ShapePropertiesPage (service, parent)
{
  setupUi (this);
  setup ();

  connect (inst_pb, SIGNAL (clicked ()), this, SLOT (show_inst ()));
  connect (prop_pb, SIGNAL (clicked ()), this, SLOT (show_props ()));
}

void 
BoxPropertiesPage::do_update (const db::Shape &shape, double dbu, const std::string &lname)
{
  layer_lbl->setText (lname.c_str ());

  db::Box box;
  shape.box (box);

  db::CplxTrans t = trans ();
  bool du = dbu_units ();

  x1_le->setText (coord_to_string ((t (box.lower_left ())).x (), dbu, du).c_str ());
  y1_le->setText (coord_to_string ((t (box.lower_left ())).y (), dbu, du).c_str ());
  x2_le->setText (coord_to_string ((t (box.upper_right ())).x (), dbu, du).c_str ());
  y2_le->setText (coord_to_string ((t (box.upper_right ())).y (), dbu, du).c_str ());
  w_le->setText (coord_to_string (t.ctrans (box.width ()), dbu, du).c_str ());
  h_le->setText (coord_to_string (t.ctrans (box.height ()), dbu, du).c_str ());
}

db::Shape
BoxPropertiesPage::do_apply (db::Shapes &shapes, const db::Shape &shape, double dbu)
{
  double x1 = 0.0, y1 = 0.0, x2 = 0.0, y2 = 0.0;
  tl::from_string (x1_le->text ().ascii (), x1);
  tl::from_string (y1_le->text ().ascii (), y1);
  tl::from_string (x2_le->text ().ascii (), x2);
  tl::from_string (y2_le->text ().ascii (), y2);

  db::CplxTrans t = trans ().inverted ();
  bool du = dbu_units ();

  db::Box box (coord_from_dpoint (db::DPoint (x1, y1), dbu, du, t), coord_from_dpoint (db::DPoint (x2, y2), dbu, du, t));
  return shapes.replace (shape, box);
}

// -------------------------------------------------------------------------
//  TextPropertiesPage implementation

TextPropertiesPage::TextPropertiesPage (edt::Service *service, QWidget *parent)
  : ShapePropertiesPage (service, parent)
{
  setupUi (this);
  setup ();

  connect (inst_pb, SIGNAL (clicked ()), this, SLOT (show_inst ()));
  connect (prop_pb, SIGNAL (clicked ()), this, SLOT (show_props ()));
}

void 
TextPropertiesPage::do_update (const db::Shape &shape, double dbu, const std::string &lname)
{
  layer_lbl->setText (lname.c_str ());

  db::Text text;
  shape.text (text);

  db::CplxTrans t = trans ();
  bool du = dbu_units ();
  db::DPoint dp = t * text.trans ().disp ();

  text_le->setText (text.string ());
  x_le->setText (coord_to_string (dp.x (), dbu, du).c_str ());
  y_le->setText (coord_to_string (dp.y (), dbu, du).c_str ());
  if (text.size () != 0) {
    size_le->setText (coord_to_string (t.ctrans (text.size ()), dbu, du).c_str ());
  } else {
    size_le->setText ("");
  }
  orient_cbx->setCurrentIndex (text.trans ().rot ());
}

db::Shape
TextPropertiesPage::do_apply (db::Shapes &shapes, const db::Shape &shape, double dbu)
{
  db::CplxTrans t = trans ().inverted ();
  bool du = dbu_units ();

  double x = 0.0, y = 0.0;
  tl::from_string (x_le->text ().ascii (), x);
  tl::from_string (y_le->text ().ascii (), y);

  db::Point tp = coord_from_dpoint (db::DPoint (x, y), dbu, du, t);
  db::Trans tt (orient_cbx->currentIndex (), tp);

  db::Text text;
  if (! size_le->text ().isEmpty ()) {
    db::Coord size = coord_from_string (size_le->text ().ascii (), dbu, du, t);
    text = db::Text (text_le->text ().ascii (), tt, size);
  } else {
    text = db::Text (text_le->text ().ascii (), tt);
  }

  db::Shape new_shape = shapes.replace (shape, text);
  return new_shape;
}

// -------------------------------------------------------------------------
//  PathPropertiesPage implementation

PathPropertiesPage::PathPropertiesPage (edt::Service *service, QWidget *parent)
  : ShapePropertiesPage (service, parent)
{
  setupUi (this);
  setup ();

  connect (inst_pb, SIGNAL (clicked ()), this, SLOT (show_inst ()));
  connect (prop_pb, SIGNAL (clicked ()), this, SLOT (show_props ()));
}

void 
PathPropertiesPage::do_update (const db::Shape &shape, double dbu, const std::string &lname)
{
  layer_lbl->setText (lname.c_str ());

  db::Path path;
  shape.path (path);

  std::string ptlist;
  ptlist.reserve (4096);

  db::CplxTrans t = trans ();
  bool du = dbu_units ();

  bool first = true;
  for (db::Path::iterator pt = path.begin (); pt != path.end (); ++pt) {
    if (! first) {
      ptlist += "\n";
    } else {
      first = false;
    }
    ptlist += coords_to_string (t * *pt, dbu, du);
  }
  ptlist_le->setText (ptlist.c_str ());

  width_le->setText (coord_to_string (t.ctrans (path.width ()), dbu, du).c_str ());
  start_ext_le->setText (coord_to_string (t.ctrans (path.extensions ().first), dbu, du).c_str ());
  end_ext_le->setText (coord_to_string (t.ctrans (path.extensions ().second), dbu, du).c_str ());
  round_cb->setChecked (path.round ());
}

db::Shape
PathPropertiesPage::do_apply (db::Shapes &shapes, const db::Shape &shape, double dbu)
{
  db::CplxTrans t = trans ().inverted ();
  bool du = dbu_units ();

  QString text (ptlist_le->text ());
  tl::Extractor ex (text.ascii ());

  std::vector <db::Point> points;

  while (! ex.at_end ()) {

    double dx = 0.0, dy = 0.0;
    ex.read (dx);
    ex.read (dy);

    points.push_back (coord_from_dpoint (db::DPoint (dx, dy), dbu, du, t));

  }

  if (points.size () < 2) {
    throw tl::Exception (tl::translate ("Path must have at least two points"));
  }

  db::Coord w = coord_from_string (width_le->text ().ascii (), dbu, du, t);
  db::Coord se = coord_from_string (start_ext_le->text ().ascii (), dbu, du, t);
  db::Coord ee = coord_from_string (end_ext_le->text ().ascii (), dbu, du, t);

  db::Path path (points.begin (), points.end (), w, se, ee, round_cb->isChecked ());
  return shapes.replace (shape, path);
}

// -------------------------------------------------------------------------
//  EditablePathPropertiesPage implementation

EditablePathPropertiesPage::EditablePathPropertiesPage (edt::Service *service, QWidget *parent)
  : ShapePropertiesPage (service, parent)
{
  setupUi (this);
  setup ();

  connect (inst_pb, SIGNAL (clicked ()), this, SLOT (show_inst ()));
  connect (prop_pb, SIGNAL (clicked ()), this, SLOT (show_props ()));
  connect (type_cb, SIGNAL (currentIndexChanged (int)), this, SLOT (type_selected (int)));
}

void 
EditablePathPropertiesPage::do_update (const db::Shape &shape, double dbu, const std::string &lname)
{
  layer_lbl->setText (lname.c_str ());

  db::Path path;
  shape.path (path);

  std::string ptlist;
  ptlist.reserve (4096);

  db::CplxTrans t = trans ();
  bool du = dbu_units ();

  bool first = true;
  for (db::Path::iterator pt = path.begin (); pt != path.end (); ++pt) {
    if (! first) {
      ptlist += "\n";
    } else {
      first = false;
    }
    ptlist += coords_to_string (t * *pt, dbu, du);
  }
  ptlist_le->setText (ptlist.c_str ());

  db::Coord w = path.width ();
  db::Coord se = path.extensions ().first;
  db::Coord ee = path.extensions ().second;

  width_le->setText (coord_to_string (t.ctrans (w), dbu, du).c_str ());

  start_ext_le->setText (coord_to_string (t.ctrans (se), dbu, du).c_str ());
  end_ext_le->setText (coord_to_string (t.ctrans (ee), dbu, du).c_str ());

  start_ext_le->setEnabled (false);
  end_ext_le->setEnabled (false);

  if (se == 0 && ee == 0) {
    type_cb->setCurrentIndex (0); // flush
  } else if (se == w / 2 && ee == w / 2 && path.round ()) {
    type_cb->setCurrentIndex (3); // round
  } else if (se == w / 2 && ee == w / 2) {
    type_cb->setCurrentIndex (1); // square
  } else {
    type_cb->setCurrentIndex (2); // variable
    start_ext_le->setEnabled (true);
    end_ext_le->setEnabled (true);
  }
}

db::Shape
EditablePathPropertiesPage::do_apply (db::Shapes &shapes, const db::Shape &shape, double dbu)
{
  db::CplxTrans t = trans ().inverted ();
  bool du = dbu_units ();

  QString text (ptlist_le->text ());
  tl::Extractor ex (text.ascii ());

  std::vector <db::Point> points;

  while (! ex.at_end ()) {

    double dx = 0.0, dy = 0.0;
    ex.read (dx);
    ex.read (dy);

    points.push_back (coord_from_dpoint (db::DPoint (dx, dy), dbu, du, t));

  }

  if (points.size () < 2) {
    throw tl::Exception (tl::translate ("Path must have at least two points"));
  }

  db::Coord w = coord_from_string (width_le->text ().ascii (), dbu, du, t);

  db::Coord se = 0, ee = 0;
  bool round = false;

  switch (type_cb->currentItem ()) {
  case 0: // flush
    break;
  case 1: // square
    se = ee = w / 2;
    break;
  case 2: // variable
    se = coord_from_string (start_ext_le->text ().ascii (), dbu, du, t);
    ee = coord_from_string (end_ext_le->text ().ascii (), dbu, du, t);
    break;
  case 3: // round
    se = ee = w / 2;
    round = true;
    break;
  } 

  db::Path path (points.begin (), points.end (), w, se, ee, round);
  return shapes.replace (shape, path);
}

void
EditablePathPropertiesPage::type_selected (int t)
{
  start_ext_le->setEnabled (t == 2);
  end_ext_le->setEnabled (t == 2);
}

}

