
/*

  KLayout Layout Viewer
  Copyright (C) 2006-2007 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#include "gsi.h"
#include "dbReader.h"
#include "dbLoadLayoutOptions.h"

namespace gsi
{
  static db::LDPair 
  ldpair_from_lp (const db::LayerProperties &lp)
  {
    return db::LDPair (lp.layer, lp.datatype);
  }

  static db::LayerProperties 
  lp_from_ldpair (const db::LDPair &ld)
  {
    db::LayerProperties lp;
    lp.layer = ld.layer;
    lp.datatype = ld.datatype;
    return lp;
  }

  static bool
  lm_is_mapped (const db::LayerMap *layer_map, const db::LayerProperties &lp)
  {
    return layer_map->logical (ldpair_from_lp (lp)).first;
  }

  static int 
  lm_logical (const db::LayerMap *layer_map, const db::LayerProperties &lp)
  {
    std::pair<bool, unsigned int> lm = layer_map->logical (ldpair_from_lp (lp));
    return lm.first ? int (lm.second) : -1;
  }

  static db::LayerProperties 
  lm_mapping (const db::LayerMap *layer_map, unsigned int l)
  {
    return lp_from_ldpair (layer_map->mapping (l));
  }

  static void
  lm_map (db::LayerMap *layer_map, const db::LayerProperties &lp, unsigned int l)
  {
    layer_map->map (ldpair_from_lp (lp), l);
  }

  static void
  lm_map_interval (db::LayerMap *layer_map, const db::LayerProperties &lp1, const db::LayerProperties &lp2, unsigned int l)
  {
    layer_map->map (ldpair_from_lp (lp1), ldpair_from_lp (lp2), l);
  }

  static void
  lm_map_string (db::LayerMap *layer_map, std::string &s, unsigned int l)
  {
    layer_map->map (s, l);
  }

  template <>
  Class<db::LayerMap> Class<db::LayerMap>::decl ("LayerMap", 
    gsi::method_ext ("is_mapped?", &lm_is_mapped, 
      "@brief Check, if a given physical layer is mapped.\n"
      "@args layer\n"
      "@args layer The physical layer specified with an \\LayerInfo object.\n"
      "@return True, if the layer is mapped."
    ) +
    gsi::method_ext ("logical", &lm_logical, 
      "@brief Returns the logical layer (the layer index in the layout object) for a given physical layer.\n"
      "@args layer\n"
      "@param layer The physical layer specified with an \\LayerInfo object.\n"
      "@return The logical layer index or -1 if the layer is not mapped."
    ) +
    gsi::method ("mapping_str", &db::LayerMap::mapping_str, 
      "@brief Returns the mapping string for a given logical layer.\n"
      "@args log_layer\n"
      "@param log_layer The logical layer for which the mapping is requested.\n"
      "@return A string describing the mapping."
      "\n"
      "The mapping string is compatible with the string that the \"map\" method accepts.\n"
    ) +
    gsi::method_ext ("mapping", &lm_mapping, 
      "@brief Returns the mapped physical layer for a given logical layer.\n"
      "@args log_layer\n"
      "@param log_layer The logical layer for which the mapping is requested.\n"
      "@return A \\LayerInfo object which is the physical layer mapped to the logical layer."
      "\n"
      "In general, there may be more than one physical layer mapped\n"
      "to one logical layer. This method will return a single one of\n"
      "them. It will return the one with the lowest layer and datatype.\n"
    ) +
    gsi::method_ext ("map", &lm_map, 
      "@brief Maps a physical layer to a logical one.\n"
      "@args phys_layer,log_layer\n"
      "@param phys_layer The physical layer (a \\LayerInfo object).\n"
      "@param log_layer The logical layer to which the physical layer is mapped.\n"
      "\n"
      "In general, there may be more than one physical layer mapped\n"
      "to one logical layer. This method will add the given physical layer to the mapping for the logical layer.\n"
    ) +
    gsi::method_ext ("map", &lm_map_interval, 
      "@brief Maps a physical layer interval to a logical one.\n"
      "@args pl_start,pl_stop,log_layer\n"
      "@param pl_start The first physical layer (a \\LayerInfo object).\n"
      "@param pl_stop The last physical layer (a \\LayerInfo object).\n"
      "@param log_layer The logical layer to which the physical layers are mapped.\n"
      "\n"
      "This method maps an interval of layers l1..l2 and datatypes d1..d2 to the mapping for the "
      "given logical layer. l1 and d1 are given by the pl_start argument, while l2 and d2 are given by "
      "the pl_stop argument."
    ) +
    gsi::method_ext ("map", &lm_map_string, 
      "@brief Maps a physical layer given by a string to a logical one.\n"
      "@args map_expr,log_layer\n"
      "@param map_expr The string describing the physical layer to map.\n"
      "@param log_layer The logical layer to which the physical layers are mapped.\n"
      "\n"
      "The string expression is constructed using the syntax: \n"
      "\"list[/list][;..]\" for layer/datatype pairs. \"list\" is a \n"
      "sequence of numbers, separated by comma values or a range \n"
      "separated by a hyphen. Examples are: \"1/2\", \"1-5/0\", \"1,2,5/0\",\n"
      "\"1/5;5/6\".\n"
    ) +
    gsi::method ("clear", &db::LayerMap::clear, 
      "@brief Clears the map.\n"
    ),
    "@brief An object representing an arbitrary mapping of physical layers to logical layers\n"
    "\n"
    "\"Physical\" layers are stream layers or other separated layers in a CAD file. \"Logical\" layers "
    "are the layers present in a \\Layout object. Logical layers are represented by an integer index while "
    "physical layers are given by a layer and datatype number or name. A logical layer is created automatically "
    "in the layout on reading if it does not exist yet.\n"
    "\n"
    "The mapping describes an association of a set of physical layers to a set of logical ones, where multiple\n"
    "physical layers can be mapped to a single logical one, which effectively merges the layers.\n"
    "\n"
    "This class has been introduced in version 0.18."
  );

  template <>
  Class<db::LoadLayoutOptions> Class<db::LoadLayoutOptions>::decl ("LoadLayoutOptions", 
    gsi::method ("set_layer_map", &db::LoadLayoutOptions::set_layer_map, 
      "@brief Set a layer map\n"
      "@args map, create_other_layers\n"
      "This sets a layer mapping for the reader. The \"create_other_layers\" specifies whether to create layers that are not "
      "in the mapping and automatically assign layers to them.\n"
      "@param map The layer map to set."
      "@param create_other_layers The flag telling whether other layer should be created also. Set to false if just the layers in the mapping table should be read."
    ) +
    gsi::method ("select_all_layers", &db::LoadLayoutOptions::select_all_layers, 
      "@brief Select all layers\n"
      "\n"
      "This disables any layer map and enables reading of all layers while\n"
      "new layers will be created when required.\n"
    ) +
    gsi::method ("layer_map", (db::LayerMap &(db::LoadLayoutOptions::*)()) &db::LoadLayoutOptions::layer_map,
      "@brief Access to the layer map member.\n"
      "@return The layer map."
    ) +
    gsi::method ("is_creating_other_layers?", (bool (db::LoadLayoutOptions::*)() const) &db::LoadLayoutOptions::create_other_layers,
      "@brief Tell whether other layers should be created.\n"
      "@return True, if other layers should be created."
    ) +
    gsi::method ("create_other_layers=", (void (db::LoadLayoutOptions::*)(bool)) &db::LoadLayoutOptions::create_other_layers,
      "@brief Specifies whether other layers should be created.\n"
      "@args create\n"
      "@param create True, if other layers should be created."
    ) +
    gsi::method ("is_text_enabled?", (bool (db::LoadLayoutOptions::*)() const) &db::LoadLayoutOptions::enable_text_objects,
      "@brief Tell whether text objects should be read.\n"
      "@return True, if text objects should be read."
    ) +
    gsi::method ("text_enabled=", (void (db::LoadLayoutOptions::*)(bool)) &db::LoadLayoutOptions::enable_text_objects,
      "@brief Specifies whether text objects should be read.\n"
      "@args enabled\n"
      "@param enabled True, if text objects should be read."
    ) +
    gsi::method ("is_properties_enabled?", (bool (db::LoadLayoutOptions::*)() const) &db::LoadLayoutOptions::enable_properties,
      "@brief Tell whether properties should be read.\n"
      "@return True, if properties should be read."
    ) +
    gsi::method ("properties_enabled=", (void (db::LoadLayoutOptions::*)(bool)) &db::LoadLayoutOptions::enable_properties,
      "@brief Specifies whether properties should be read.\n"
      "@args enabled\n"
      "@param enabled True, if properties should be read."
    ),
    "@brief Layout reader options\n"
    "\n"
    "This object describes various layer reader options used for loading layouts.\n"
    "\n"
    "This class has been introduced in version 0.18.\n"
  );

  static db::LayerMap
  load_without_options (db::Layout *layout, const std::string &filename)
  {
    tl::InputZLibFile stream (filename);
    db::Reader reader (stream);
    return reader.read (*layout);
  }

  static db::LayerMap
  load_with_options (db::Layout *layout, const std::string &filename, const db::LoadLayoutOptions &options)
  {
    tl::InputZLibFile stream (filename);
    db::Reader reader (stream);
    return reader.read (*layout, options);
  }

  //  extend the layout class by two reader methods
  static
  gsi::Class<db::Layout> layout_reader_decl ("Layout",
    gsi::method_ext ("read", &load_without_options,
      "@brief Load the layout from the given file\n"
      "@args filename\n"
      "The format of the file is determined automatically and automatic unzipping is provided. "
      "No particular options can be specified.\n"
      "@param filename The name of the file to load.\n"
      "@return A layer map that contains the mapping used by the reader including the layers that have been created."
      "\n"
      "This method has been added in version 0.18."
    ) +
    gsi::method_ext ("read", &load_with_options,
      "@brief Load the layout from the given file with options\n"
      "@args filename,options\n"
      "The format of the file is determined automatically and automatic unzipping is provided. "
      "In this version, some reader options can be specified. "
      "@param filename The name of the file to load.\n"
      "@param options The options object specifying further options for the reader.\n"
      "@return A layer map that contains the mapping used by the reader including the layers that have been created."
      "\n"
      "This method has been added in version 0.18."
    ),
    ""
  );

}

