
/*

  KLayout Layout Viewer
  Copyright (C) 2006-2007 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#ifndef _HDR_gsiMethods
#define _HDR_gsiMethods

#include "tlString.h"

//  For a comprehensive documentation see gsi.h

namespace gsi
{

struct NoArgumentsAllowedException
  : public tl::Exception
{
  NoArgumentsAllowedException ()
    : tl::Exception (tl::translate ("Method does not allow arguments"))
  { }
};

struct NeedsArgumentsException
  : public tl::Exception
{
  NeedsArgumentsException (unsigned int got, unsigned int want)
    : tl::Exception (tl::sprintf (tl::translate ("Method requires %d arguments, got %d").c_str (), want, got))
  { }
};

struct IncompatibleReturnTypeException
  : public tl::Exception
{
  IncompatibleReturnTypeException (const ArgType &got, const ArgType &want)
    : tl::Exception (tl::translate ("Incompatible return types: got '") + got.to_string () + tl::translate ("', want '") + want.to_string () + "'")
  { }
};

/**
 *  @brief Basic declaration of a method
 *
 *  A class declaration collects objects of this kind to represent methods and their specific
 *  implementation.
 *  This class is implemented in various specific ways that bind the abstract call to a specifc
 *  C++ method.
 */
class MethodBase 
{
public:
  typedef std::vector<ArgType>::const_iterator argument_iterator;

  MethodBase (const std::string &name, const std::string &doc, bool c, bool s)
    : m_name (name), m_doc (doc), m_const (c), m_static (s), m_argsize (0), m_retsize (0)
  { 
    // .. nothing yet ..
  }

  virtual ~MethodBase () { }

  const std::string &name () const
  {
    return m_name;
  }

  const std::string &doc () const
  {
    return m_doc;
  }

  virtual MethodBase *clone () const = 0;

  virtual void call (void *cls, SerialArgs &args, SerialArgs &ret) const = 0;

  virtual bool is_callback () const 
  {
    return false;
  }

  virtual void set_callback (void *v, const Callback &cb) const 
  {
    // .. nothing yet ..
  }

  argument_iterator begin_arguments () const
  {
    return m_arg_types.begin ();
  }

  argument_iterator end_arguments () const
  {
    return m_arg_types.end ();
  }

  const ArgType &ret_type () const
  {
    return m_ret_type;
  }

  bool is_const () const
  {
    return m_const;
  }

  bool is_static () const
  {
    return m_static;
  }

  void check_no_args () const
  {
    if (end_arguments () != begin_arguments ()) {
      throw NoArgumentsAllowedException ();
    }
  }

  void check_num_args (unsigned int num) const
  {
    if (std::distance (begin_arguments (), end_arguments ()) != int (num)) {
      throw NeedsArgumentsException (num, std::distance (begin_arguments (), end_arguments ()));
    }
  }

  void check_return_type (const ArgType &a) const
  {
    if (m_ret_type != a) {
      throw IncompatibleReturnTypeException (a, m_ret_type);
    }
  }

  unsigned int argsize () const
  {
    return m_argsize;
  }

  unsigned int retsize () const
  {
    return m_retsize;
  }

protected:
  template <class X> 
  void add_arg () 
  {
    ArgType a;
    a.template init<X> ();
    m_arg_types.push_back (a);
    m_argsize += item_size<X> ();
  }

  template <class X> 
  void set_return () 
  {
    m_ret_type.template init<X> ();
    m_retsize += item_size<X> ();
  }

  template <class X> 
  void set_return_new () 
  {
    m_ret_type = ArgType (T_object_new, false, true, false, &Class<X>::decl);
    m_retsize += item_size<X *> ();
  }

  void set_const (bool c)
  {
    m_const = c;
  }

private:
  std::string m_name, m_doc;
  std::vector<ArgType> m_arg_types;
  ArgType m_ret_type;
  bool m_const;
  bool m_static;
  unsigned int m_argsize;
  unsigned int m_retsize;
};

/**
 *  @brief A collection of methods
 *
 *  The basic purpose of this object is to provide the + operator that allows to concatenate 
 *  method declarations in the class declaration.
 */
class Methods
{
public:
  typedef std::vector<MethodBase *>::const_iterator iterator;

  Methods ()
    : m_methods ()
  {
    // .. nothing yet ..
  }

  Methods (MethodBase *m)
    : m_methods ()
  {
    m_methods.push_back (m);
  }

  Methods (const Methods &d)
  {
    operator= (d);
  }

  Methods &operator= (const Methods &d)
  {
    if (this != &d) {
      clear ();
      m_methods.reserve (d.m_methods.size ());
      for (std::vector<MethodBase *>::const_iterator m = d.m_methods.begin (); m != d.m_methods.end (); ++m) {
        m_methods.push_back ((*m)->clone ());
      }
    }
    return *this;
  }

  ~Methods ()
  {
    clear ();
  }

  void clear ()
  {
    for (std::vector<MethodBase *>::iterator m = m_methods.begin (); m != m_methods.end (); ++m) {
      delete *m;
    }
    m_methods.clear ();
  }

  //  HINT: this is not the usual + semantics but this is more effective
  Methods &operator+ (const Methods &m)
  {
    return operator+= (m);
  }

  Methods &operator+= (const Methods &m)
  {
    for (std::vector<MethodBase *>::const_iterator mm = m.m_methods.begin (); mm != m.m_methods.end (); ++mm) 
    {
      add_method ((*mm)->clone ());
    }
    return *this;
  }

  iterator begin () const
  {
    return m_methods.begin ();
  }

  iterator end () const
  {
    return m_methods.end ();
  }

  void add_method (MethodBase *method)
  {
    m_methods.push_back (method);
  }

  size_t size () const
  {
    return m_methods.size ();
  }

public:
  std::vector<MethodBase *> m_methods;
};

inline Methods operator+ (const Methods &a, const Methods &b)
{
  return Methods (a) + b;
}

template <class X>
class MethodSpecificBase 
  : public MethodBase
{
public:
  MethodSpecificBase (const std::string &name, const std::string &doc, bool c, bool s, Callback X::*cb)
    : MethodBase (name, doc, c, s),
      m_cb (cb)
  { 
    // .. nothing yet ..
  }

  virtual void set_callback (void *v, const Callback &cb) const
  {
    typedef typename non_const_x<X>::nc_x nc_x;
    nc_x *x = (nc_x *)v;
    x->*m_cb = cb;
  }

  virtual bool is_callback () const
  {
    return m_cb != 0;
  }

  Callback X::*callback () const
  {
    return m_cb;
  }

private:
  Callback X::*m_cb;
};

class StaticMethodBase 
  : public MethodBase
{
public:
  StaticMethodBase (const std::string &name, const std::string &doc)
    : MethodBase (name, doc, false, true)
  { 
    // .. nothing yet ..
  }
};

// 0 argument

#define _COUNT        0
#define _NAME(x)      x##0
#define _TMPLARG
#define _TMPLARGPART 
#define _FUNCARGLIST  
#define _FUNCARGLISTPART  
#define _ADDARGS      
#define _GETARGVARS   
#define _ARGVARLIST   
#define _ARGVARLISTPART

#include "gsiMethodsVar.h"

#undef _ARGVARLISTPART
#undef _ARGVARLIST   
#undef _GETARGVARS   
#undef _ADDARGS
#undef _FUNCARGLISTPART  
#undef _FUNCARGLIST
#undef _TMPLARG
#undef _TMPLARGPART
#undef _NAME
#undef _COUNT

// 1 argument

#define _COUNT        1
#define _NAME(x)      x##1
#define _TMPLARG      class A1
#define _TMPLARGPART  , class A1
#define _FUNCARGLIST  A1
#define _FUNCARGLISTPART  , A1
#define _ADDARGS      this->template add_arg<A1> ();
#define _GETARGVARS   A1 a1 = args.template get_value<A1> (typename type_traits<A1>::tag ());
#define _ARGVARLIST   a1
#define _ARGVARLISTPART   , a1

#include "gsiMethodsVar.h"

#undef _ARGVARLISTPART
#undef _ARGVARLIST   
#undef _GETARGVARS   
#undef _ADDARGS
#undef _FUNCARGLISTPART  
#undef _FUNCARGLIST
#undef _TMPLARGPART
#undef _TMPLARG
#undef _NAME
#undef _COUNT

// 2 arguments

#define _COUNT        2
#define _NAME(x)      x##2
#define _TMPLARG      class A1, class A2
#define _TMPLARGPART  , class A1, class A2
#define _FUNCARGLIST  A1, A2
#define _FUNCARGLISTPART  , A1, A2
#define _ADDARGS      this->template add_arg<A1> (); \
                      this->template add_arg<A2> (); 
#define _GETARGVARS   A1 a1 = args.template get_value<A1> (typename type_traits<A1>::tag ()); \
                      A2 a2 = args.template get_value<A2> (typename type_traits<A2>::tag ());
#define _ARGVARLIST   a1, a2
#define _ARGVARLISTPART   , a1, a2

#include "gsiMethodsVar.h"

#undef _ARGVARLISTPART
#undef _ARGVARLIST   
#undef _GETARGVARS   
#undef _ADDARGS
#undef _FUNCARGLISTPART  
#undef _FUNCARGLIST
#undef _TMPLARGPART
#undef _TMPLARG
#undef _NAME
#undef _COUNT

// 3 arguments

#define _COUNT        3
#define _NAME(x)      x##3
#define _TMPLARG      class A1, class A2, class A3
#define _TMPLARGPART  , class A1, class A2, class A3
#define _FUNCARGLIST  A1, A2, A3
#define _FUNCARGLISTPART  , A1, A2, A3
#define _ADDARGS      this->template add_arg<A1> (); \
                      this->template add_arg<A2> (); \
                      this->template add_arg<A3> (); 
#define _GETARGVARS   A1 a1 = args.template get_value<A1> (typename type_traits<A1>::tag ()); \
                      A2 a2 = args.template get_value<A2> (typename type_traits<A2>::tag ()); \
                      A3 a3 = args.template get_value<A3> (typename type_traits<A3>::tag ());
#define _ARGVARLIST   a1, a2, a3
#define _ARGVARLISTPART   , a1, a2, a3

#include "gsiMethodsVar.h"

#undef _ARGVARLISTPART
#undef _ARGVARLIST   
#undef _GETARGVARS   
#undef _ADDARGS
#undef _FUNCARGLISTPART  
#undef _FUNCARGLIST
#undef _TMPLARGPART
#undef _TMPLARG
#undef _NAME
#undef _COUNT

// 4 arguments

#define _COUNT        4
#define _NAME(x)      x##4
#define _TMPLARG      class A1, class A2, class A3, class A4
#define _TMPLARGPART  , class A1, class A2, class A3, class A4
#define _FUNCARGLIST  A1, A2, A3, A4
#define _FUNCARGLISTPART  , A1, A2, A3, A4
#define _ADDARGS      this->template add_arg<A1> (); \
                      this->template add_arg<A2> (); \
                      this->template add_arg<A3> (); \
                      this->template add_arg<A4> (); 
#define _GETARGVARS   A1 a1 = args.template get_value<A1> (typename type_traits<A1>::tag ()); \
                      A2 a2 = args.template get_value<A2> (typename type_traits<A2>::tag ()); \
                      A3 a3 = args.template get_value<A3> (typename type_traits<A3>::tag ()); \
                      A4 a4 = args.template get_value<A4> (typename type_traits<A4>::tag ());
#define _ARGVARLIST   a1, a2, a3, a4
#define _ARGVARLISTPART   , a1, a2, a3, a4

#include "gsiMethodsVar.h"

#undef _ARGVARLISTPART
#undef _ARGVARLIST   
#undef _GETARGVARS   
#undef _ADDARGS
#undef _FUNCARGLISTPART  
#undef _FUNCARGLIST
#undef _TMPLARGPART
#undef _TMPLARG
#undef _NAME
#undef _COUNT

// 5 arguments

#define _COUNT        5
#define _NAME(x)      x##5
#define _TMPLARG      class A1, class A2, class A3, class A4, class A5
#define _TMPLARGPART  , class A1, class A2, class A3, class A4, class A5
#define _FUNCARGLIST  A1, A2, A3, A4, A5
#define _FUNCARGLISTPART  , A1, A2, A3, A4, A5
#define _ADDARGS      this->template add_arg<A1> (); \
                      this->template add_arg<A2> (); \
                      this->template add_arg<A3> (); \
                      this->template add_arg<A4> (); \
                      this->template add_arg<A5> (); 
#define _GETARGVARS   A1 a1 = args.template get_value<A1> (typename type_traits<A1>::tag ()); \
                      A2 a2 = args.template get_value<A2> (typename type_traits<A2>::tag ()); \
                      A3 a3 = args.template get_value<A3> (typename type_traits<A3>::tag ()); \
                      A4 a4 = args.template get_value<A4> (typename type_traits<A4>::tag ()); \
                      A5 a5 = args.template get_value<A5> (typename type_traits<A5>::tag ());
#define _ARGVARLIST   a1, a2, a3, a4, a5
#define _ARGVARLISTPART   , a1, a2, a3, a4, a5

#include "gsiMethodsVar.h"

#undef _ARGVARLISTPART
#undef _ARGVARLIST   
#undef _GETARGVARS   
#undef _ADDARGS
#undef _FUNCARGLISTPART  
#undef _FUNCARGLIST
#undef _TMPLARGPART
#undef _TMPLARG
#undef _NAME
#undef _COUNT

// 6 arguments

#define _COUNT        6
#define _NAME(x)      x##6
#define _TMPLARG      class A1, class A2, class A3, class A4, class A5, class A6
#define _TMPLARGPART  , class A1, class A2, class A3, class A4, class A5, class A6
#define _FUNCARGLIST  A1, A2, A3, A4, A5, A6
#define _FUNCARGLISTPART  , A1, A2, A3, A4, A5, A6
#define _ADDARGS      this->template add_arg<A1> (); \
                      this->template add_arg<A2> (); \
                      this->template add_arg<A3> (); \
                      this->template add_arg<A4> (); \
                      this->template add_arg<A5> (); \
                      this->template add_arg<A6> (); 
#define _GETARGVARS   A1 a1 = args.template get_value<A1> (typename type_traits<A1>::tag ()); \
                      A2 a2 = args.template get_value<A2> (typename type_traits<A2>::tag ()); \
                      A3 a3 = args.template get_value<A3> (typename type_traits<A3>::tag ()); \
                      A4 a4 = args.template get_value<A4> (typename type_traits<A4>::tag ()); \
                      A5 a5 = args.template get_value<A5> (typename type_traits<A5>::tag ()); \
                      A6 a6 = args.template get_value<A6> (typename type_traits<A6>::tag ());
#define _ARGVARLIST   a1, a2, a3, a4, a5, a6
#define _ARGVARLISTPART   , a1, a2, a3, a4, a5, a6

#include "gsiMethodsVar.h"

#undef _ARGVARLISTPART
#undef _ARGVARLIST   
#undef _GETARGVARS   
#undef _ADDARGS
#undef _FUNCARGLISTPART  
#undef _FUNCARGLIST
#undef _TMPLARGPART
#undef _TMPLARG
#undef _NAME
#undef _COUNT

// 7 arguments

#define _COUNT        7
#define _NAME(x)      x##7
#define _TMPLARG      class A1, class A2, class A3, class A4, class A5, class A6, class A7
#define _TMPLARGPART  , class A1, class A2, class A3, class A4, class A5, class A6, class A7
#define _FUNCARGLIST  A1, A2, A3, A4, A5, A6, A7
#define _FUNCARGLISTPART  , A1, A2, A3, A4, A5, A6, A7
#define _ADDARGS      this->template add_arg<A1> (); \
                      this->template add_arg<A2> (); \
                      this->template add_arg<A3> (); \
                      this->template add_arg<A4> (); \
                      this->template add_arg<A5> (); \
                      this->template add_arg<A6> (); \
                      this->template add_arg<A7> (); 
#define _GETARGVARS   A1 a1 = args.template get_value<A1> (typename type_traits<A1>::tag ()); \
                      A2 a2 = args.template get_value<A2> (typename type_traits<A2>::tag ()); \
                      A3 a3 = args.template get_value<A3> (typename type_traits<A3>::tag ()); \
                      A4 a4 = args.template get_value<A4> (typename type_traits<A4>::tag ()); \
                      A5 a5 = args.template get_value<A5> (typename type_traits<A5>::tag ()); \
                      A6 a6 = args.template get_value<A6> (typename type_traits<A6>::tag ()); \
                      A7 a7 = args.template get_value<A7> (typename type_traits<A7>::tag ());
#define _ARGVARLIST   a1, a2, a3, a4, a5, a6, a7
#define _ARGVARLISTPART   , a1, a2, a3, a4, a5, a6, a7

#include "gsiMethodsVar.h"

#undef _ARGVARLISTPART
#undef _ARGVARLIST   
#undef _GETARGVARS   
#undef _ADDARGS
#undef _FUNCARGLISTPART  
#undef _FUNCARGLIST
#undef _TMPLARGPART
#undef _TMPLARG
#undef _NAME
#undef _COUNT

// 8 arguments

#define _COUNT        8
#define _NAME(x)      x##8
#define _TMPLARG      class A1, class A2, class A3, class A4, class A5, class A6, class A7, class A8
#define _TMPLARGPART  , class A1, class A2, class A3, class A4, class A5, class A6, class A7, class A8
#define _FUNCARGLIST  A1, A2, A3, A4, A5, A6, A7, A8
#define _FUNCARGLISTPART  , A1, A2, A3, A4, A5, A6, A7, A8
#define _ADDARGS      this->template add_arg<A1> (); \
                      this->template add_arg<A2> (); \
                      this->template add_arg<A3> (); \
                      this->template add_arg<A4> (); \
                      this->template add_arg<A5> (); \
                      this->template add_arg<A6> (); \
                      this->template add_arg<A7> (); \
                      this->template add_arg<A8> (); 
#define _GETARGVARS   A1 a1 = args.template get_value<A1> (typename type_traits<A1>::tag ()); \
                      A2 a2 = args.template get_value<A2> (typename type_traits<A2>::tag ()); \
                      A3 a3 = args.template get_value<A3> (typename type_traits<A3>::tag ()); \
                      A4 a4 = args.template get_value<A4> (typename type_traits<A4>::tag ()); \
                      A5 a5 = args.template get_value<A5> (typename type_traits<A5>::tag ()); \
                      A6 a6 = args.template get_value<A6> (typename type_traits<A6>::tag ()); \
                      A7 a7 = args.template get_value<A7> (typename type_traits<A7>::tag ()); \
                      A8 a8 = args.template get_value<A8> (typename type_traits<A8>::tag ());
#define _ARGVARLIST   a1, a2, a3, a4, a5, a6, a7, a8
#define _ARGVARLISTPART   , a1, a2, a3, a4, a5, a6, a7, a8

#include "gsiMethodsVar.h"

#undef _ARGVARLISTPART
#undef _ARGVARLIST   
#undef _GETARGVARS   
#undef _ADDARGS
#undef _FUNCARGLISTPART  
#undef _FUNCARGLIST
#undef _TMPLARGPART
#undef _TMPLARG
#undef _NAME
#undef _COUNT

// 9 arguments

#define _COUNT        9
#define _NAME(x)      x##9
#define _TMPLARG      class A1, class A2, class A3, class A4, class A5, class A6, class A7, class A8, class A9
#define _TMPLARGPART  , class A1, class A2, class A3, class A4, class A5, class A6, class A7, class A8, class A9
#define _FUNCARGLIST  A1, A2, A3, A4, A5, A6, A7, A8, A9
#define _FUNCARGLISTPART  , A1, A2, A3, A4, A5, A6, A7, A8, A9
#define _ADDARGS      this->template add_arg<A1> (); \
                      this->template add_arg<A2> (); \
                      this->template add_arg<A3> (); \
                      this->template add_arg<A4> (); \
                      this->template add_arg<A5> (); \
                      this->template add_arg<A6> (); \
                      this->template add_arg<A7> (); \
                      this->template add_arg<A8> (); \
                      this->template add_arg<A9> (); 
#define _GETARGVARS   A1 a1 = args.template get_value<A1> (typename type_traits<A1>::tag ()); \
                      A2 a2 = args.template get_value<A2> (typename type_traits<A2>::tag ()); \
                      A3 a3 = args.template get_value<A3> (typename type_traits<A3>::tag ()); \
                      A4 a4 = args.template get_value<A4> (typename type_traits<A4>::tag ()); \
                      A5 a5 = args.template get_value<A5> (typename type_traits<A5>::tag ()); \
                      A6 a6 = args.template get_value<A6> (typename type_traits<A6>::tag ()); \
                      A7 a7 = args.template get_value<A7> (typename type_traits<A7>::tag ()); \
                      A8 a8 = args.template get_value<A8> (typename type_traits<A8>::tag ()); \
                      A9 a9 = args.template get_value<A9> (typename type_traits<A9>::tag ());
#define _ARGVARLIST   a1, a2, a3, a4, a5, a6, a7, a8, a9
#define _ARGVARLISTPART   , a1, a2, a3, a4, a5, a6, a7, a8, a9

#include "gsiMethodsVar.h"

#undef _ARGVARLISTPART
#undef _ARGVARLIST   
#undef _GETARGVARS   
#undef _ADDARGS
#undef _FUNCARGLISTPART  
#undef _FUNCARGLIST
#undef _TMPLARGPART
#undef _TMPLARG
#undef _NAME
#undef _COUNT

// 10 arguments

#define _COUNT        10
#define _NAME(x)      x##10
#define _TMPLARG      class A1, class A2, class A3, class A4, class A5, class A6, class A7, class A8, class A9, class A10
#define _TMPLARGPART  , class A1, class A2, class A3, class A4, class A5, class A6, class A7, class A8, class A9, class A10
#define _FUNCARGLIST  A1, A2, A3, A4, A5, A6, A7, A8, A9, A10
#define _FUNCARGLISTPART  , A1, A2, A3, A4, A5, A6, A7, A8, A9, A10
#define _ADDARGS      this->template add_arg<A1> (); \
                      this->template add_arg<A2> (); \
                      this->template add_arg<A3> (); \
                      this->template add_arg<A4> (); \
                      this->template add_arg<A5> (); \
                      this->template add_arg<A6> (); \
                      this->template add_arg<A7> (); \
                      this->template add_arg<A8> (); \
                      this->template add_arg<A9> (); \
                      this->template add_arg<A10> (); 
#define _GETARGVARS   A1 a1 = args.template get_value<A1> (typename type_traits<A1>::tag ()); \
                      A2 a2 = args.template get_value<A2> (typename type_traits<A2>::tag ()); \
                      A3 a3 = args.template get_value<A3> (typename type_traits<A3>::tag ()); \
                      A4 a4 = args.template get_value<A4> (typename type_traits<A4>::tag ()); \
                      A5 a5 = args.template get_value<A5> (typename type_traits<A5>::tag ()); \
                      A6 a6 = args.template get_value<A6> (typename type_traits<A6>::tag ()); \
                      A7 a7 = args.template get_value<A7> (typename type_traits<A7>::tag ()); \
                      A8 a8 = args.template get_value<A8> (typename type_traits<A8>::tag ()); \
                      A9 a9 = args.template get_value<A9> (typename type_traits<A9>::tag ()); \
                      A10 a10 = args.template get_value<A10> (typename type_traits<A10>::tag ());
#define _ARGVARLIST   a1, a2, a3, a4, a5, a6, a7, a8, a9, a10
#define _ARGVARLISTPART   , a1, a2, a3, a4, a5, a6, a7, a8, a9, a10

#include "gsiMethodsVar.h"

#undef _ARGVARLISTPART
#undef _ARGVARLIST   
#undef _GETARGVARS   
#undef _ADDARGS
#undef _FUNCARGLISTPART  
#undef _FUNCARGLIST
#undef _TMPLARGPART
#undef _TMPLARG
#undef _NAME
#undef _COUNT

// 11 arguments

#define _COUNT        11
#define _NAME(x)      x##11
#define _TMPLARG      class A1, class A2, class A3, class A4, class A5, class A6, class A7, class A8, class A9, class A10, class A11
#define _TMPLARGPART  , class A1, class A2, class A3, class A4, class A5, class A6, class A7, class A8, class A9, class A10, class A11
#define _FUNCARGLIST  A1, A2, A3, A4, A5, A6, A7, A8, A9, A10, A11
#define _FUNCARGLISTPART  , A1, A2, A3, A4, A5, A6, A7, A8, A9, A10, A11
#define _ADDARGS      this->template add_arg<A1> (); \
                      this->template add_arg<A2> (); \
                      this->template add_arg<A3> (); \
                      this->template add_arg<A4> (); \
                      this->template add_arg<A5> (); \
                      this->template add_arg<A6> (); \
                      this->template add_arg<A7> (); \
                      this->template add_arg<A8> (); \
                      this->template add_arg<A9> (); \
                      this->template add_arg<A10> (); \
                      this->template add_arg<A11> (); 
#define _GETARGVARS   A1 a1 = args.template get_value<A1> (typename type_traits<A1>::tag ()); \
                      A2 a2 = args.template get_value<A2> (typename type_traits<A2>::tag ()); \
                      A3 a3 = args.template get_value<A3> (typename type_traits<A3>::tag ()); \
                      A4 a4 = args.template get_value<A4> (typename type_traits<A4>::tag ()); \
                      A5 a5 = args.template get_value<A5> (typename type_traits<A5>::tag ()); \
                      A6 a6 = args.template get_value<A6> (typename type_traits<A6>::tag ()); \
                      A7 a7 = args.template get_value<A7> (typename type_traits<A7>::tag ()); \
                      A8 a8 = args.template get_value<A8> (typename type_traits<A8>::tag ()); \
                      A9 a9 = args.template get_value<A9> (typename type_traits<A9>::tag ()); \
                      A10 a10 = args.template get_value<A10> (typename type_traits<A10>::tag ()); \
                      A11 a11 = args.template get_value<A11> (typename type_traits<A11>::tag ());
#define _ARGVARLIST   a1, a2, a3, a4, a5, a6, a7, a8, a9, a10, a11
#define _ARGVARLISTPART   , a1, a2, a3, a4, a5, a6, a7, a8, a9, a10, a11

#include "gsiMethodsVar.h"

#undef _ARGVARLISTPART
#undef _ARGVARLIST   
#undef _GETARGVARS   
#undef _ADDARGS
#undef _FUNCARGLISTPART  
#undef _FUNCARGLIST
#undef _TMPLARGPART
#undef _TMPLARG
#undef _NAME
#undef _COUNT

}

#endif


