
/*

  KLayout Layout Viewer
  Copyright (C) 2006-2007 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#ifndef _HDR_gsiObject
#define _HDR_gsiObject

//  For a comprehensive documentation see gsi.h

namespace gsi
{

template <class X> class Class;

/**
 *  @brief If being used for "copy_semantics" in ClassTag, tells the GSI framework that an object can be copied
 */
struct CanCopySemanticsTag { 
  CanCopySemanticsTag () { } 
};

/**
 *  @brief If being used for "copy_semantics" in ClassTag, tells the GSI framework that an object cannot be copied
 */
struct CantCopySemanticsTag { 
  CantCopySemanticsTag () { } 
};

/**
 *  @brief If being used for "copy_semantics" in ClassTag, tells the GSI framework that an object can be default-created
 */
struct CanDefaultCreateSemanticsTag { 
  CanDefaultCreateSemanticsTag () { } 
};

/**
 *  @brief If being used for "copy_semantics" in ClassTag, tells the GSI framework that an object cannot be default-created
 */
struct CantDefaultCreateSemanticsTag { 
  CantDefaultCreateSemanticsTag () { }
};

/**
 *  @brief The default properties of a class
 */
template<class X> 
struct ClassTag 
{
  typedef CanCopySemanticsTag copy_semantics;
  typedef CanDefaultCreateSemanticsTag create_semantics;
};

/**
 *  @brief A lower base class for Object<X>
 */
class ObjectBase
{
public:
  ObjectBase () 
    : mp_data (0)
  {
    //  .. nothing yet ..
  }

  ~ObjectBase ()
  {
    mp_data = 0;
  }

  void *data () const
  {
    return mp_data;
  }

protected:
  void *mp_data;
};

/**
 *  @brief A generic base class
 *
 *  Objects that are derived from this template (A being derived from gsi::Object<A>)
 *  offer additional functionality over objects not derived from this base class such as
 *  notification when the C++ object is destroyed and persistency of the client-side object
 *  state.
 */
#ifdef HAVE_RUBY
template <class X>
class Object
  : public ObjectBase
{
public:
  //  TODO: currently the client data is not copied. As a consequence, copying the C++
  //  objects does not copy the Ruby part of the object but rather leaves it unitialized.
  //  Since however, the client-side data feature is supposed to be used to non-copying 
  //  objects mainly, this problem is not considered serious for now.

  Object () 
  { 
    Class<X>::decl.attach (this);
  }

  Object (const Object &) 
  { 
    Class<X>::decl.attach (this);
  }

  Object &operator= (const Object &)
  {
    return *this;
  }

  ~Object ()
  {
    Class<X>::decl.detach (this);
  }

private:
  friend class Class<X>;
};
#else
template <class X>
class Object
  : public ObjectBase
{
  friend class Class<X>;
};
#endif

}

#endif

