
/*

  KLayout Layout Viewer
  Copyright (C) 2006-2007 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/



#ifndef HDR_imgObject
#define HDR_imgObject

#include "dbUserObject.h"
#include "dbBox.h"
#include "dbTrans.h"
#include "tlDataMapping.h"
#include "layViewOp.h"

#include <string>
#include <vector>

#include <QtGui/QColor>

namespace img {
  
class DataHeader;

/**
 *  @brief A structure describing the data mapping of an image object
 *
 *  Data mapping is the process of transforming the data into RGB pixel values.
 */
struct DataMapping
{
public:
  /**
   *  @brief The constructor
   */
  DataMapping ();

  /**
   *  @brief Equality
   */
  bool operator== (const DataMapping &d) const;

  /**
   *  @brief Less operator
   */
  bool operator< (const DataMapping &d) const;

  /**
   *  @brief The false color mapping nodes
   *
   *  Each node is a pair or x-value (normalized to a range of 0..1) and a corresponding color.
   *  The list should have an element with x value of 0.0 and one with an x value of 1.0.
   */
  std::vector< std::pair<double, QColor> > false_color_nodes;

  /**
   *  @brief The brightness value
   *
   *  The brightness is a double value between -1.0 and 1.0. 
   *  Neutral brightness is 0.0.
   */
  double brightness;

  /**
   *  @brief The contrast value
   *
   *  The contrast is a double value between -1.0 and 1.0. 
   *  Neutral contrast is 0.0.
   */
  double contrast;

  /**
   *  @brief The gamma value
   */
  double gamma;

  /**
   *  @brief The red channel gain
   *
   *  This value is the multiplier by which the red channel is scaled after applying 
   *  false color transformation and contrast/brightness/gamma.
   *
   *  1.0 is a neutral value. The gain should be >=0.0.
   */
  double red_gain;

  /**
   *  @brief The green channel gain
   *
   *  This value is the multiplier by which the green channel is scaled after applying 
   *  false color transformation and contrast/brightness/gamma.
   *
   *  1.0 is a neutral value. The gain should be >=0.0.
   */
  double green_gain;

  /**
   *  @brief The blue channel gain
   *
   *  This value is the multiplier by which the blue channel is scaled after applying 
   *  false color transformation and contrast/brightness/gamma.
   *
   *  1.0 is a neutral value. The gain should be >=0.0.
   */
  double blue_gain;

  /**
   *  @brief Create a tl::DataMapping object that represents this data mapping
   *
   *  @param xmin The x value corresponding to the lower limit
   *  @param xmax The x value corresponding to the upper limit
   *  @param channel 0=red, 1=green, 2=blue
   *  @return A new'd DataMapping object
   */
  tl::DataMappingBase *create_data_mapping (bool monochrome, double xmin, double xmax, unsigned int channel) const;
};

/**
 *  @brief A image object
 * 
 *  This class implements the actual image.
 *  Since this class derives from db::UserObjectBase, these objects
 *  can be stored within the database.
 */  
class Object 
  : public db::DUserObjectBase
{
public:
  typedef db::coord_traits<coord_type> coord_traits;

  /**
   *  @brief Default constructor
   *
   *  This will create an empty image with no defined bounding box and a unit transformation.
   */
  Object ();

  /**
   *  @brief Destructor
   */
  ~Object ();

  /**
   *  @brief Constructor for monochrome or color images with zero pixel values
   *
   *  This constructor creates an image object from a data set describing one monochrome channel
   *  or three color channels.
   *  Each channel consists of an array of x*y values where the first "x" values describe the first (lowest!) row
   *  and so on. Note, that the rows are oriented in the mathematical sense (first one is the lowest) contrary to 
   *  the common convention for image data.
   *  The data fields can be accessed with the "data", "set_data", "pixel" or "set_pixel" methods.
   *
   *  @param w The width of the image
   *  @param h The height of the image
   *  @param trans The transformation from pixel space to micron space
   *  @param color True to create a color image.
   */
  Object (size_t w, size_t h, const db::DCplxTrans &trans, bool color);

  /**
   *  @brief Constructor for a monochrome image with the given pixel values
   *
   *  This constructor creates an image from the given pixel values. The values have to be organized
   *  line by line. Each line must consist of "w" values where the first value is the leftmost pixel.
   *  Note, that the rows are oriented in the mathematical sense (first one is the lowest) contrary to 
   *  the common convention for image data.
   *
   *  @param w The width of the image
   *  @param h The height of the image
   *  @param trans The transformation from pixel space to micron space
   *  @param d The data set which will become owned by the image
   */
  Object (size_t w, size_t h, const db::DCplxTrans &trans, float *d);

  /**
   *  @brief Constructor for a monochrome image with the given pixel values
   *
   *  This constructor creates an image from the given pixel values. The values have to be organized
   *  line by line. Each line must consist of "w" values where the first value is the leftmost pixel.
   *  Note, that the rows are oriented in the mathematical sense (first one is the lowest) contrary to 
   *  the common convention for image data.
   *
   *  @param w The width of the image
   *  @param h The height of the image
   *  @param trans The transformation from pixel space to micron space
   *  @param d The data set
   */
  Object (size_t w, size_t h, const db::DCplxTrans &trans, const std::vector <double> &d);

  /**
   *  @brief Constructor for a color image with the given pixel values
   *
   *  This constructor creates an image from the given pixel values. The values have to be organized
   *  line by line and separated by color channel. Each line must consist of "w" values where the first value is the leftmost pixel.
   *  Note, that the rows are oriented in the mathematical sense (first one is the lowest) contrary to 
   *  the common convention for image data.
   *
   *  @param w The width of the image
   *  @param h The height of the image
   *  @param trans The transformation from pixel space to micron space
   *  @param red The red channel data set which will become owned by the image
   *  @param green The green channel data set which will become owned by the image
   *  @param blue The blue channel data set which will become owned by the image
   */
  Object (size_t w, size_t h, const db::DCplxTrans &trans, float *red, float *green, float *blue);

  /**
   *  @brief Constructor for a monochrome image with the given pixel values
   *
   *  This constructor creates an image from the given pixel values. The values have to be organized
   *  line by line and separated by color channel. Each line must consist of "w" values where the first value is the leftmost pixel.
   *  Note, that the rows are oriented in the mathematical sense (first one is the lowest) contrary to 
   *  the common convention for image data.
   *
   *  @param w The width of the image
   *  @param h The height of the image
   *  @param trans The transformation from pixel space to micron space
   *  @param red The red channel data set 
   *  @param green The green channel data set
   *  @param blue The blue channel data set
   */
  Object (size_t w, size_t h, const db::DCplxTrans &trans, const std::vector <double> &red, const std::vector <double> &green, const std::vector <double> &blue);

  /**
   *  @brief Constructor from a image file 
   *
   *  This constructor creates an image object from a file (which can have any format supported by Qt) and 
   *  a transformation. The image will originally be put to position 0, 0 (lower left corner) and each pixel
   *  will have a size of 1. The transformation describes how to transform this image into micron space.
   */
  Object (const std::string &filename, const db::DCplxTrans &trans);

  /**
   *  @brief Copy constructor
   */
  Object (const img::Object &d);

  /**
   *  @brief Assignment
   */
  Object &operator= (const img::Object &d);

  /**
   *  @brief The generic "equal" operator
   *
   *  This method reimplements the db::DUserObjectBase interface.
   */
  virtual bool equals (const db::DUserObjectBase *d) const;

  /**
   *  @brief The generic "less" operator
   *
   *  This method reimplements the db::DUserObjectBase interface.
   */
  virtual bool less (const db::DUserObjectBase *d) const;

  /**
   *  @brief Get the class Id which internally distinguishes the img::Object's class from other classes
   *
   *  This method reimplements the db::DUserObjectBase interface.
   */
  virtual unsigned int class_id () const;

  /**
   *  @brief Clone the object
   *
   *  This method reimplements the db::DUserObjectBase interface.
   */
  virtual db::DUserObjectBase *clone () const;

  /**
   *  @brief Get the bounding box
   *
   *  This method reimplements the db::DUserObjectBase interface.
   */
  virtual db::DBox box () const;

  /**
   *  @brief Transform with a given complex transformation
   *
   *  This method reimplements the db::DUserObjectBase interface.
   */
  virtual void transform (const db::DCplxTrans &t);

  /**
   *  @brief Transform with a given standard transformation
   *
   *  This method reimplements the db::DUserObjectBase interface.
   */
  virtual void transform (const db::DTrans &t);

  /**
   *  @brief Transform with a given fixpoint transformation
   *
   *  This method reimplements the db::DUserObjectBase interface.
   */
  virtual void transform (const db::fixpoint_trans &t);

  /**
   *  @brief Return the transformed object
   */
  template <class Trans>
  img::Object transformed (const Trans &t) const
  {
    img::Object obj (*this);
    obj.transform (t);
    return obj;
  }

  /**
   *  @brief In-place move
   */
  Object &move (const db::DPoint &p)
  {
    transform (db::DTrans (p));
    return *this;
  }

  /**
   *  @brief Return the moved object 
   */
  Object moved (const db::DPoint &p) const
  {
    img::Object d (*this);
    d.move (p);
    return d;
  }

  /**
   *  @brief Accessor to the width property
   */
  size_t width () const;

  /**
   *  @brief Accessor to the height property
   */
  size_t height () const;

  /**
   *  @brief Get the number of entries in the data set
   */
  size_t data_length () const;

  /**
   *  @brief Get the name of the file currently loaded
   */
  const std::string &filename () const
  {
    return m_filename;
  }

  /**
   *  @brief Accessor to the is_empty property
   *
   *  @return True, if the image is empty
   */
  bool is_empty () const;

  /**
   *  @brief Accessor to the is_color property
   *
   *  @return True, if the image is a color image
   */
  bool is_color () const;

  /**
   *  @brief Set one pixel (monochrome)
   *
   *  @param x The x coordinate of the pixel (0..width()-1)
   *  @param y The y coordinate of the pixel (mathematical order: 0 is the lowest, 0..height()-1)
   *  @param v The value
   *
   *  If the component index, x or y value exceeds the image bounds of the image is a color image,
   *  this method does nothing.
   */
  void set_pixel (size_t x, size_t y, double v);

  /**
   *  @brief Set one pixel (color)
   *
   *  @param x The x coordinate of the pixel (0..width()-1)
   *  @param y The y coordinate of the pixel (mathematical order: 0 is the lowest, 0..height()-1)
   *  @param red The red component
   *  @param green The green component
   *  @param blue The blue component
   *
   *  If the component index, x or y value exceeds the image bounds of the image is not a color image,
   *  this method does nothing.
   */
  void set_pixel (size_t x, size_t y, double red, double green, double blue);

  /**
   *  @brief Accessor to one pixel (monchrome)
   *
   *  @param x The x coordinate of the pixel (0..width()-1)
   *  @param y The y coordinate of the pixel (mathematical order: 0 is the lowest, 0..height()-1)
   *
   *  If the x or y value exceeds the image bounds or the image is not a monochrome image, this method 
   *  returns 0.0.
   */
  double pixel (size_t x, size_t y) const;

  /**
   *  @brief Accessor to one pixel (monochrome and color)
   *
   *  @param x The x coordinate of the pixel (0..width()-1)
   *  @param y The y coordinate of the pixel (mathematical order: 0 is the lowest, 0..height()-1)
   *  @param component 0 for red, 1 for green, 2 for blue.
   *
   *  If the component index, x or y value exceeds the image bounds, this method 
   *  returns 0.0. For monochrome images, the component index is ignored.
   */
  double pixel (size_t x, size_t y, unsigned int component) const;

  /**
   *  @brief Accessor to the image data field
   *
   *  @return The pointer to an float field with w*h entries or 0 if the image is a color image.
   */
  const float *data () const;

  /**
   *  @brief Accessor to the image data field (color component)
   *
   *  @return The pointer to an float field with w*h entries or 0 if the image is not a color image.
   *
   *  @param component 0 for red, 1 for green and 2 for blue component.
   *  @return The pointer to an float field with w*h entries or 0 if the image is not a color image.
   */
  const float *data (unsigned int component) const;

  /**
   *  @brief Write accessor to the image data field (monochrome)
   *
   *  This method must be passed a new'd pointer to a float array with a length of width * height.
   *  The pointer to the data set becomes owned by the image object.
   */
  void set_data (size_t width, size_t height, float *d);

  /**
   *  @brief Write accessor to the image data field (monochrome)
   *
   *  This variant takes the data from a field of double values.
   */
  void set_data (size_t width, size_t height, const std::vector<double> &d);

  /**
   *  @brief Write accessor to the image data field (color)
   *
   *  This method must be passed new'd pointers to float arrays with a length of "data_length()".
   *  The pointers to the data set become owned by the image object.
   */
  void set_data (size_t width, size_t height, float *red, float *green, float *blue);

  /**
   *  @brief Write accessor to the image data field (monochrome)
   *
   *  This variant takes the data from three fields of double values.
   */
  void set_data (size_t width, size_t height, const std::vector<double> &red, const std::vector<double> &green, const std::vector<double> &blue);

  /** 
   *  @brief Set the pixel width
   *
   *  The pixel width determines the width of on pixel in the original space which is transformed to
   *  micron space with the transformation.
   */
  void set_pixel_width (double w)
  {
    m_pixel_width = w;
  }

  /**
   *  @brief Get the pixel width
   *
   *  See \set_pixel_width for a description of that property.
   */
  double pixel_width () const
  {
    return m_pixel_width;
  }

  /** 
   *  @brief Set the pixel height
   *
   *  The pixel height determines the height of on pixel in the original space which is transformed to
   *  micron space with the transformation.
   */
  void set_pixel_height (double h)
  {
    m_pixel_height = h;
  }

  /**
   *  @brief Get the pixel height
   *
   *  See \set_pixel_height for a description of that property.
   */
  double pixel_height () const
  {
    return m_pixel_height;
  }

  /**
   *  @brief Set the transformation
   *
   *  This transformation converts pixel coordinates (0,0 being the lower left corner and each pixel having the dimension of pixel_width and pixel_height)
   *  to micron coordinates. The coordinate of the pixel is the lower left corner of the pixel.
   */
  void set_trans (const db::DCplxTrans &trans)
  {
    m_trans = trans;
  }

  /**
   *  @brief Return the pixel-to-micron transformation
   *
   *  This transformation converts pixel coordinates (0,0 being the lower left corner and each pixel having the dimension of pixel_width and pixel_height)
   *  to micron coordinates. The coordinate of the pixel is the lower left corner of the pixel.
   */
  const db::DCplxTrans &trans () const
  {
    return m_trans;
  }

  /**
   *  @brief Get the upper limit of the values in the data set
   *
   *  This value determines the upper end of the data mapping (i.e. white value etc.).
   *  It does not necessarily correspond to the minimum value of the data set but it must be
   *  larger than that.
   */
  double min_value () const
  {
    return m_min_value;
  }

  /** 
   *  @brief Set the minimum value
   *
   *  See the \min_value method for the description of the minimum value property.
   */
  void set_min_value (double h)
  {
    invalidate_pixel_data ();
    m_min_value = h;
  }

  /**
   *  @brief Get the upper limit of the values in the data set
   *
   *  This value determines the upper end of the data mapping (i.e. white value etc.).
   *  It does not necessarily correspond to the maximum value of the data set but it must be
   *  larger than that.
   */
  double max_value () const
  {
    return m_max_value;
  }

  /** 
   *  @brief Set the maximum value
   *
   *  See the \max_value method for the description of the maximum value property.
   */
  void set_max_value (double h)
  {
    invalidate_pixel_data ();
    m_max_value = h;
  }

  /**  
   *  @brief Get the Id 
   *  
   *  The Id is an arbitrary integer that can be used to track the evolution of an
   *  image object. The Id is not changed when the object is edited.
   */
  int id () const
  {
    return m_id;
  }

  /**  
   *  @brief Set the Id 
   *  
   *  The Id is an arbitrary integer that can be used to track the evolution of an
   *  image object. The Id is not changed when the object is edited.
   */
  void id (int _id) 
  {
    m_id = _id;
  }

  /**
   *  @brief Get the data mapping
   */
  const DataMapping &data_mapping () const
  {
    return m_data_mapping;
  }

  /**
   *  @brief Set the data mapping
   */
  void set_data_mapping (const DataMapping &dm) 
  {
    invalidate_pixel_data ();
    m_data_mapping = dm;
  }

  /**
   *  @brief Get the RGB pixel data sets obtained by applying the LUT's
   */
  const lay::color_t *pixel_data () const
  {
    validate_pixel_data ();
    return mp_pixel_data;
  }

  /**
   *  @brief Load the data from the given file
   *
   *  @param adjust_min_max True, if min and max values shall be adjusted
   */
  void load_data (const std::string &filename, bool adjust_min_max = true);

  /**
   *  @brief Equality
   */
  bool operator== (const img::Object &d) const;

  /**
   *  @brief Inequality
   */
  bool operator!= (const img::Object &d) const
  {
    return !operator== (d);
  }

  /**
   *  @brief The class name for the generic user object factory 
   */
  virtual const char *class_name () const;

  /**
   *  @brief Fill from a string
   *
   *  This method needs to be implemented mainly if the object is to be created from the
   *  generic factory.
   */
  virtual void from_string (const char *);

  /**
   *  @brief Convert to a string
   *
   *  This method needs to be implemented mainly if the object is to be created from the
   *  generic factory.
   */
  virtual std::string to_string () const;

  /**
   *  @brief Return the memory used in bytes
   */
  virtual size_t mem_used () const;

  /**
   *  @brief Return the memory required in bytes
   */
  virtual size_t mem_reqd () const 
  {
    return mem_used ();
  }

private:
  std::string m_filename;
  db::DCplxTrans m_trans;
  double m_pixel_width;
  double m_pixel_height;
  DataHeader *mp_data;
  int m_id;
  double m_min_value, m_max_value;
  bool m_min_value_set, m_max_value_set;
  DataMapping m_data_mapping;
  mutable const lay::color_t *mp_pixel_data;

  void release ();
  void invalidate_pixel_data ();
  void validate_pixel_data () const;
  void allocate (bool color);
  void read_file ();
};

}

#endif

