
/*

  KLayout Layout Viewer
  Copyright (C) 2006-2007 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#ifndef HDR_layLayerControlPanel
#define HDR_layLayerControlPanel


#include <vector>
#include <string>
#include <set>
#include <algorithm>

#include <QtGui/QFrame>
#include <QtGui/QLabel>
#include <QtGui/QPainter>
#include <QtGui/QPaintEvent>
#include <QtGui/QTreeView>

#include "layCanvasPlane.h"
#include "layViewOp.h"
#include "layLayoutView.h"
#include "layColorPalette.h"
#include "layStipplePalette.h"
#include "layDitherPattern.h"
#include "layLayerProperties.h"
#include "dbObject.h"
#include "tlObserver.h"
#include "tlDeferredExecution.h"

class QTreeView;
class QModelIndex;
class QMenu;

namespace lay 
{

class LCPActiveLabel;
class LCPTreeWidget;
class LayerTreeModel;

/**
 *  @brief A helper class that provides a frame with background
 *
 *  In Qt3, the QFrame was painting it's own background.
 *  In Qt4 it does not. Therefore this class is provided.
 */
class FrameWithBackground : public QFrame
{
public:
  FrameWithBackground (QWidget *parent, const char *name = 0) 
    : QFrame (parent, name)
  {
    //  nothing yet.
  }
  
  void paintEvent (QPaintEvent *pe)
  {
    {
      QPainter painter (this);
      painter.fillRect (pe->rect (), palette ().brush (backgroundRole ()));
    }
    QFrame::paintEvent (pe);
  }
};

/**
 *  @brief A helper class that provides a label with background
 *
 *  In Qt3, the QLabel was painting it's own background.
 *  In Qt4 it does not. Therefore this class is provided.
 */
class LabelWithBackground : public QLabel
{
public:
  LabelWithBackground (QWidget *parent, const char *name = 0) 
    : QLabel (parent, name)
  {
    //  nothing yet.
  }
  
  void paintEvent (QPaintEvent *pe)
  {
    {
      QPainter painter (this);
      painter.fillRect (pe->rect (), palette ().brush (backgroundRole ()));
    }
    QLabel::paintEvent (pe);
  }
};

/**
 *  @brief Qt signal remitter class
 *
 *  The purpose of this class is to re-emit a signal at the
 *  input with a integer argument that is given to the class
 *  instance.
 */
class LCPRemitter
  : public QObject
{
Q_OBJECT

public:
  LCPRemitter (int index, QObject *parent, const char *name = 0);

public slots:
  void the_slot ();

signals:
  void the_signal (int index);

private:
  int m_index;
};

/** 
 *  @brief A color palette
 *
 *  This class implements a color chooser palette.
 *  The only signal emitted by this class is the color_selected
 *  signal which is emitted if a color (either from the
 *  palette or the "more colors" dialog) is selected.
 */
class LCPColorPalette
  : public QFrame
{
Q_OBJECT

public:
  LCPColorPalette (QWidget *parent, const char *name = 0); 

  void set_palette (const lay::ColorPalette &p);

public slots:
  void button_clicked (int index);

signals:
  void color_selected (QColor);
  void color_brightness_selected (int);

private:
  lay::ColorPalette m_palette;
  std::vector <LCPActiveLabel *> m_color_buttons;
};

/** 
 *  @brief A animation palette
 *
 *  This class implements a animation chooser palette.
 *  The signal emitted by this class is the animation_selected
 *  signal which is emitted if a animation mode is selected.
 */
class LCPAnimationPalette
  : public QFrame
{
Q_OBJECT

public:
  LCPAnimationPalette (QWidget *parent, const char *name = 0); 

public slots:
  void button_clicked (int index);

signals:
  void animation_selected (int mode);
};

/** 
 *  @brief A style palette
 *
 *  This class implements a style chooser palette.
 *  The signals emitted by this class is the width_selected
 *  signal which is emitted if a width is selected and
 *  the marked_selected which is emitted if one of the marked button
 *  was pressed.
 */
class LCPStylePalette
  : public QFrame
{
Q_OBJECT

public:
  LCPStylePalette (QWidget *parent, const char *name = 0); 

public slots:
  void button_clicked (int index);

signals:
  void width_selected (int width);
  void marked_selected (bool marked);
};

/** 
 *  @brief A dither pattern palette
 *
 *  This class implements a dither pattern palette.
 *  The only signal emitted by this class is the dither_selected
 *  signal which is emitted if a pattern is selected.
 */
class LCPDitherPalette
  : public QFrame
{
Q_OBJECT

public:
  LCPDitherPalette (QWidget *parent, lay::LayoutView *view, const char *name = 0); 

  void set_palette (const lay::StipplePalette &p);

public slots:
  void button_clicked (int index);

signals:
  void dither_selected (int dither_index);
  void pattern_changed (const lay::DitherPattern &);

private:
  lay::StipplePalette m_palette;
  lay::LayoutView *mp_view;
  std::vector <LCPActiveLabel *> m_stipple_buttons;

  void create_pixmap_for (LCPActiveLabel *b, int n);
};

/**
 *  @brief A palette with the buttons for visibility settings
 *
 *  This palette exhibits two button groups: for visibility
 *  and transparency. Two signals are emitted: 
 *  visibility_change if "show" or "hide" are clicked,
 *  transparency_changed if "transparent" or "opaque" are clicked.
 */
class LCPVisibilityPalette
  : public QFrame
{
Q_OBJECT

public:
  LCPVisibilityPalette (QWidget *parent, const char *name = 0); 

public slots:
  void button_clicked (int index);

signals:
  void visibility_change (bool visible);
  void transparency_change (bool transparent);
};

/**
 *  @brief A lightweight button class
 *
 *  The lightweight buttons are implemented from labels that
 *  have a "push-down" behaviour. The only signal emitted is
 *  "clicked" with an integer value that can be assigned to
 *  the button.
 */
class LCPActiveLabel 
  : public LabelWithBackground
{
Q_OBJECT 

public:
  LCPActiveLabel (int index, QWidget *parent, const char *name = "button");

  virtual void mousePressEvent (QMouseEvent *e);
  virtual void mouseReleaseEvent (QMouseEvent *e);

signals: 
  void clicked (int);

private:
  int m_index;
  bool m_grabbed;
};

/**
 *  @brief A layer tree widget helper class
 *
 *  A specialization of the TreeView that allows us to control
 *  sizeHint which otherwise is pretty large (around 100 pixel per column)
 *  and attaches the layer tree model to the view.
 */

class LCPTreeWidget : public QTreeView
{
Q_OBJECT 

public:
  LCPTreeWidget (QWidget *parent, lay::LayerTreeModel *model, const char *name);
  ~LCPTreeWidget ();

  virtual QSize sizeHint () const;

  //  overload the double click event, because the default behaviour is
  //  expanding/collapsing the items
  virtual void mouseDoubleClickEvent (QMouseEvent *event);

  void set_selection (const std::vector<lay::LayerPropertiesConstIterator> &sel);
  void set_current (const lay::LayerPropertiesConstIterator &sel);
  void collapse_all ();
  void expand_all ();

signals:
  void double_clicked (const QModelIndex &, Qt::KeyboardModifiers);

private:
  lay::LayerTreeModel *mp_model;
};

/**
 *  @brief The layer control panel 
 *
 *  The layer control panel has a layer list and four panels 
 *  for colors, dither pattern and visibility.
 *
 *  The class communicates with a Layout interface for
 *  retrieval and changing of layer properties.
 */
class LayerControlPanel
  : public QFrame, 
    public db::Object,
    public tl::Observer
{
Q_OBJECT 

public:
  enum SortOrder { ByName, ByIndexLayerDatatype, ByIndexDatatypeLayer, ByLayerDatatypeIndex, ByDatatypeLayerIndex };
  enum RegroupMode { RegroupByIndex, RegroupByDatatype, RegroupByLayer, RegroupFlatten };

  /**
   *  @brief Constructor
   *
   *  @param layout The Layout interface that the layer
   *         control panel is attached to
   *  @param parent The Qt parent widget
   *  @param name The layer control panel's widget name
   */
  LayerControlPanel (lay::LayoutView *view, db::Manager *manager, QWidget *parent = 0, const char *name = "control_panel");

  /**
   *  @brief Destructor
   */
  ~LayerControlPanel ();

  /** 
   *  @brief Return true, if the tree view has the focus
   */
  bool has_focus () const;

  /**
   *  @brief Perform the layer control panel's initialisations on the main menu
   */
  static void init_menu (lay::AbstractMenu &menu);

  /**
   *  @brief Tell, if there is something to copy
   */
  bool has_selection () const;

  /**
   *  @brief Cut to clipboards
   */
  void cut ();

  /**
   *  @brief Copy to clipboards
   */
  void copy ();

  /**
   *  @brief Paste from clipboard
   */
  void paste ();

  /**
   *  @brief The sizeHint implementation for Qt layout management
   */
  virtual QSize sizeHint () const;

  /** 
   *  @brief The Qt resize event
   */
  virtual void resizeEvent (QResizeEvent *re);

  /** 
   *  @brief The Qt resize function that does the layout management
   */
  virtual void resize (int w, int h);

  /** 
   *  @brief The Qt geometry setting function that also does the layout management
   */
  virtual void setGeometry (int x, int y, int w, int h);

  /**
   *  @brief Changing of the color palette
   */
  void set_palette (const lay::ColorPalette &p);

  /**
   *  @brief Changing of the stipple palette
   */
  void set_palette (const lay::StipplePalette &p);

  /**
   *  @brief Changing of the background color
   */
  void set_background_color (QColor c);

  /**
   *  @brief Changing of the text color
   */
  void set_text_color (QColor c);

  /**
   *  @brief Set the "hide empty layers" flag
   *
   *  If this flag is set, empty layers (either completely empty or no shapes in view - depending on the test_shapes_in_view flag)
   *  are not shown. Otherwise they are shown "faded".
   */
  void set_hide_empty_layers (bool f);

  /**
   *  @brief Get the "hide empty layers" flag
   */
  bool hide_empty_layers () 
  {
    return m_hide_empty_layers;
  }

  /**
   *  @brief Set the "test_shapes_in_view" flag
   *
   *  If this flag is set, layers without a shape in the viewport are shown "empty".
   */
  void set_test_shapes_in_view (bool f);

  /**
   *  @brief Get the "test shapes in view" flag
   */
  bool test_shapes_in_view () 
  {
    return m_test_shapes_in_view;
  }

  /**
   *  @brief Set the animation phase
   */
  void set_phase (int phase);

  /**
   *  @brief Tell, if the model has been updated already (true) or if it is still under construction (false)
   */
  bool model_updated ();

  /**
   *  @brief Inform of coming changes
   *  
   *  This method may be called before changes are made to the cell list.
   *  "end_updates" resets this state.
   */
  void begin_updates ();

  /**
   *  @brief Cancel the "begin_update" state
   */
  void cancel_updates ();

  /**
   *  @brief Tells that updates started with begin_updates() have been finished
   *
   *  In result, this method will perform all actions to update the display
   *  and reset the "in_update" state.
   */
  void end_updates ();

  /**
   *  @brief Set the non-empty layers (the "uint" for the layer iterators) for the "test shapes is view" mode
   */
  void set_non_empty_layers (const std::set <size_t> &non_empty_layers);

  /**
   *  @brief Return the selected layers
   */
  std::vector<lay::LayerPropertiesConstIterator> selected_layers () const;

  /**
   *  @brief Return the current layer index
   *
   *  Will return a "null" iterator if no layer is selected currently.
   */
  lay::LayerPropertiesConstIterator current_layer () const;

  /**
   *  @brief Sort the layer list in the given order 
   */
  void sort_layers (SortOrder order);

  /**
   *  @brief Regroup the layer list in the given way 
   */
  void regroup_layers (RegroupMode mode);

  /**
   *  @brief Implementation of the undo operations
   */
  virtual void undo (db::Op *op) throw ();

  /**
   *  @brief Implementation of the redo operations
   */
  virtual void redo (db::Op *op) throw ();

  /**
   *  @brief Signals that the layer list has changed
   */
  virtual void signal_observed_int (int f);

  /**
   *  @brief Signals that the cellview has changed
   */
  virtual void signal_cv_changed ();

  /**
   *  @brief Signals that the viewport has changed
   */
  virtual void signal_vp_changed ();

  using QFrame::setGeometry;

signals:
  void color_changed ();
  void transparency_changed ();
  void stipple_changed ();
  void visibility_changed ();
  void marked_changed ();
  void width_changed ();
  void animation_changed ();
  void order_changed ();

public slots:
  void cm_select_all ();
  void cm_hide ();
  void cm_show ();
  void cm_rename ();
  void cm_delete ();
  void cm_insert ();
  void cm_group ();
  void cm_ungroup ();
  void cm_source ();
  void cm_sort_by_name ();
  void cm_sort_by_ild ();
  void cm_sort_by_idl ();
  void cm_sort_by_ldi ();
  void cm_sort_by_dli ();
  void cm_regroup_by_index ();
  void cm_regroup_by_datatype ();
  void cm_regroup_by_layer ();
  void cm_regroup_flatten ();
  void cm_expand_all ();
  void cm_add_missing ();
  void cm_remove_unused ();
  void panel_button_clicked (int index);
  void fill_color_changed (QColor c);
  void frame_color_changed (QColor c);
  void fill_color_brightness (int delta);
  void frame_color_brightness (int delta);
  void dither_changed (int di);
  void dither_pattern_changed (const lay::DitherPattern &pattern);
  void visibility_changed (bool visible);
  void transparency_changed (bool transparent);
  void width_changed (int width);
  void marked_changed (bool marked);
  void animation_changed (int mode);
  void double_clicked (const QModelIndex &index, Qt::KeyboardModifiers modifiers);
  void context_menu (const QPoint &pt);  
  void up_clicked ();
  void upup_clicked ();
  void down_clicked ();
  void downdown_clicked ();

private:
  QFrame *mp_layer_frame;
  LCPTreeWidget *mp_layer_list;
  LayerTreeModel *mp_model;
  std::vector <std::pair <QWidget *, QWidget *> > m_tool_panels;
  std::vector <lay::CanvasPlane> m_bitmaps;
  lay::LayoutView *mp_view;
  bool m_needs_update;
  bool m_force_update_hidden_flags;
  LCPDitherPalette *mp_dither_palette;
  LCPColorPalette *mp_palette;
  LCPColorPalette *mp_frame_palette;
  std::set<unsigned int> m_expanded;
  bool m_in_update;
  int m_phase;
  tl::DeferredMethod<LayerControlPanel> m_do_update_content_dm;
  tl::ObserverDelegate<LayerControlPanel> m_cv_observer;
  tl::ObserverDelegate<LayerControlPanel> m_vp_observer;
  bool m_hide_empty_layers;
  bool m_test_shapes_in_view;

  void rearrange (int w, int h);
  void add_panel (QWidget *panel_widget, const char *text);

  template <class Op>
  void foreach_selected (const Op &op);

  void set_selection (const std::vector<lay::LayerPropertiesConstIterator> &new_sel);
  void clear_selection ();

  void do_move (int mode);

  void move_algo (std::vector<lay::LayerPropertiesConstIterator>::const_iterator from, 
                  std::vector<lay::LayerPropertiesConstIterator>::const_iterator to, 
                  lay::LayerPropertiesConstIterator parent,
                  lay::LayerPropertiesIterator new_parent,
                  std::vector<lay::LayerPropertiesConstIterator> &new_sel,
                  int mode);

  void save_expanded ();
  void restore_expanded ();

  void do_update_content ();
  void do_delete ();
  void do_copy ();
  void recover ();
};

} // namespace lay

#endif

