
/*

  KLayout Layout Viewer
  Copyright (C) 2006-2007 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#include <QtCore/QEvent>
#include <QtGui/QPixmap>
#include <QtGui/QBitmap>
#include <QtGui/QPainter>
#include <QtGui/QApplication>
#include <QtCore/QBuffer>

#include "tlTimer.h"
#include "tlLog.h"
#include "tlAssert.h"
#include "layLayoutCanvas.h"
#include "layRedrawThread.h"
#include "layLayoutView.h"
#include "layMarker.h"
#include "gtf.h"

#include "layBitmapsToImage.h"

#include <strstream>

namespace lay
{


#if defined (_WIN32) || defined (Q_WS_MAC)
// Windows build runs well with single buffering
// and Qt built in second buffering
// #  define SINGLE_BUFFERED_MASK
// The same is true vor MAC OS X
#  define SINGLE_BUFFERED
#else
// X11 runs better with the SINGLE_BUFFERED_MASK scheme
// #  define SINGLE_BUFFERED
#  define SINGLE_BUFFERED_MASK
#endif

LayoutCanvas::LayoutCanvas (QWidget *parent, lay::LayoutView *view, const char *name)
  : lay::ViewObjectWidget (parent, name), 
    mp_view (view),
    mp_image (0), mp_image_bg (0), mp_pixmap (0), 
    m_background (0), m_foreground (0), m_active (0),
    m_need_redraw (false),
    m_redraw_clearing (false),
    m_update_image (true),
    m_do_update_image_dm (this, &LayoutCanvas::do_update_image),
    m_do_update_image_always_dm (this, &LayoutCanvas::do_update_image_always)
{
  //  some reasonable initializations for the size
  m_viewport.set_size (100, 100); 

  mp_redraw_thread = new lay::RedrawThread (this, view);

  setBackgroundMode (Qt::NoBackground);
  set_colors (paletteBackgroundColor (),
              palette ().color (QPalette::Normal, QColorGroup::Text),
              palette ().color (QPalette::Normal, QColorGroup::Mid));
#ifndef SINGLE_BUFFERED
  setAttribute (Qt::WA_PaintOnScreen);
#endif
  setAttribute (Qt::WA_NoSystemBackground);
}

LayoutCanvas::~LayoutCanvas ()
{
  m_viewport_changed_observers.remove_observers ();

  if (mp_image) {
    delete mp_image;
    mp_image = 0;
  }
  if (mp_image_bg) {
    delete mp_image_bg;
    mp_image_bg = 0;
  }
  if (mp_pixmap) {
    delete mp_pixmap;
    mp_pixmap = 0;
  }
  if (mp_redraw_thread) {
    delete mp_redraw_thread;
    mp_redraw_thread = 0;
  }

  clear_fg_bitmaps ();
}

void 
LayoutCanvas::set_colors (QColor background, QColor foreground, QColor active)
{
  m_background = background.rgb ();
  m_foreground = foreground.rgb ();
  m_active = active.rgb ();
      
  //  force regeneration of background image ..
  if (mp_image_bg) {
    delete mp_image_bg;
  }
  mp_image_bg = 0;

  update_image ();
}

void
LayoutCanvas::set_view_ops (std::vector <lay::ViewOp> &view_ops)
{
  if (view_ops != m_view_ops) {
    m_view_ops.swap (view_ops);
    update_image ();
  }
}

void
LayoutCanvas::set_dither_pattern (const lay::DitherPattern &p)
{
  if (p != m_dither_pattern) {
    m_dither_pattern = p;
    update_image ();
  }
}

void
LayoutCanvas::prepare_drawing ()
{
  if (m_need_redraw) {

    BitmapViewObjectCanvas::set_size (m_viewport.width (), m_viewport.height ());

    if (! mp_image ||
        (unsigned int) mp_image->width () != m_viewport.width () || 
        (unsigned int) mp_image->height () != m_viewport.height ()) {
      if (mp_image) {
        delete mp_image;
      }
      mp_image = new QImage (m_viewport.width (), m_viewport.height (), QImage::Format_RGB32);
      if (mp_pixmap) {
        delete mp_pixmap;
        mp_pixmap = 0;
      }
    }

    mp_image->fill (m_background);

    if (m_redraw_clearing) {
      mp_redraw_thread->start (m_layers, m_viewport, mp_view->synchroneous ());
    } else {
      mp_redraw_thread->restart (m_need_redraw_layer);
    }

    m_need_redraw = false;
    m_update_image = true;

  }
}

void
LayoutCanvas::update_image ()
{
  // this will make the image being redone (except for background objects which will
  // only be redrawn on touch_bg)
  m_update_image = true;

  update (); // produces a paintEvent()
}

void
LayoutCanvas::free_resources ()
{
  if (mp_pixmap) {
    delete mp_pixmap;
    mp_pixmap = 0;
  }
}

void
LayoutCanvas::paintEvent (QPaintEvent *p)
{
  //  this is the update image request
  tl::SelfTimer timer_info (tl::verbosity () >= 31, tl::translate ("PaintEvent"));

  //  if required, start the redraw thread ..
  prepare_drawing ();

  if (mp_image) {
    
    //  check, if the background needs to be updated 
    if (m_update_image || needs_update_bg ()) {

      tl_assert (mp_image->width () == int (m_viewport.width ()));
      tl_assert (mp_image->height () == int (m_viewport.height ()));

      if (needs_update_bg () || ! mp_image_bg) {

        //  clear the image and paint the background objects
        mp_image->fill (m_background);
        do_render_bg (m_viewport, *this);

        //  save the current background image
        if (mp_image_bg) {
          delete mp_image_bg;
        }
        mp_image_bg = new QImage (*mp_image);

      } else {
        //  else reuse the saved image
        *mp_image = *mp_image_bg;
      }

      to_image (m_view_ops, dither_pattern (), this, *mp_image, m_viewport.width (), m_viewport.height ());

      if (mp_pixmap) {
        delete mp_pixmap;
        mp_pixmap = 0;
      }

      //  give the view a chance to react on the change of the image
      if (mp_view) {
        mp_view->image_changed ();
      }

      m_update_image = false;

    }

    //  prepare a buffer
    size_t nbytes = (m_viewport.width () + 7) / 8 * m_viewport.height ();
    unsigned char *p_data = new unsigned char [nbytes];
    memset (p_data, 0, nbytes);

    //  create a base pixmap consisting of the layout with background
    //  and static foreground objects

    if (! mp_pixmap || needs_update_static () || 
        mp_image->size ().width () != mp_pixmap->size ().width () ||
        mp_image->size ().height () != mp_pixmap->size ().height ()) {

      if (mp_pixmap) {
        delete mp_pixmap;
      } 

      //  render the foreground parts ..
      clear_fg_bitmaps ();
      do_render (m_viewport, *this, true);

      mp_pixmap = new QPixmap ();
      if (fg_bitmaps () > 0) {
        QImage full_image (*mp_image);
        bitmaps_to_image (fg_view_op_vector (), fg_bitmap_vector (), dither_pattern (), &full_image, m_viewport.width (), m_viewport.height (), false, &m_mutex);
        *mp_pixmap = full_image; // Qt 4.6.0 workaround
      } else {
        *mp_pixmap = *mp_image;
      }

    }

    //  erase any previous data
    clear_fg_bitmaps ();

    //  render foreground content
    do_render (m_viewport, *this, false);

#if defined(SINGLE_BUFFERED)

    //  Single-buffering scheme - somewhat more flickering 
    //  but less resource requirements.

    //  produce the pixmap first and then overdraw with dynamic content
    bitBlt (this, 0, 0, mp_pixmap, 0, 0, mp_pixmap->size ().width (), mp_pixmap->size ().height ());

    //  dynamic bitmaps are always drawn in transparent mode ..
    QPainter painter (this);
    painter.setBackgroundMode (Qt::TransparentMode);

    //  create QBitmap objects and paint them for the bitmaps
    //  present
    for (unsigned int n = 0; n < fg_bitmaps (); ++n) {
      if (fg_bitmap (n) != 0) {
        memset (p_data, 0, nbytes);
        bitmap_to_bitmap (fg_style (n), *fg_bitmap (n), p_data, m_viewport.width (), m_viewport.height (), dither_pattern ());
        QBitmap bitmap (m_viewport.width (), m_viewport.height (), p_data, true);
        painter.setPen (QRgb (fg_style (n).ormask ()));
        painter.drawPixmap (0, 0, bitmap);
      }
    }

#elif defined(SINGLE_BUFFERED_MASK)

    //  Single-buffering scheme with mask - somewhat more flickering 
    //  but less resource requirements.

    //  combine the bitmaps into a mask
    for (unsigned int n = 0; n < fg_bitmaps (); ++n) {
      if (fg_bitmap (n) != 0) {
        bitmap_to_bitmap (fg_style (n), *fg_bitmap (n), p_data, m_viewport.width (), m_viewport.height (), dither_pattern ());
      }
    }

    //  invert the mask
    unsigned char *d = p_data;
    for (unsigned int i = 0; i < nbytes; ++i) {
      *d++ ^= 0xff;
    }

    //  create the mask
    QBitmap mask (m_viewport.width (), m_viewport.height (), p_data, true);
    mp_pixmap->setMask (mask);

    //  produce the pixmap with the mask
    bitBlt (this, 0, 0, mp_pixmap, 0, 0, mp_pixmap->size ().width (), mp_pixmap->size ().height ());

    //  dynamic bitmaps are always drawn in transparent mode ..
    QPainter painter (this);
    painter.setBackgroundMode (Qt::TransparentMode);

    //  create QBitmap objects and paint them for the bitmaps
    //  present
    for (unsigned int n = 0; n < fg_bitmaps (); ++n) {
      if (fg_bitmap (n) != 0) {
        memset (p_data, 0, nbytes);
        bitmap_to_bitmap (fg_style (n), *fg_bitmap (n), p_data, m_viewport.width (), m_viewport.height (), dither_pattern ());
        QBitmap bitmap (m_viewport.width (), m_viewport.height (), p_data, true);
        painter.setPen (QRgb (fg_style (n).ormask ()));
        painter.drawPixmap (0, 0, bitmap);
      }
    }

#else

    //  Double-buffering scheme - somewhat less flickering but
    //  only efficient if the QImage has the precise size and
    //  at the cost of twice the X resource requirements.

    //  create an intermediate pixmap
    QPixmap px (*mp_pixmap);

    //  dynamic bitmaps are always drawn in transparent mode ..
    QPainter painter (&px);
    painter.setBackgroundMode (Qt::TransparentMode);

    //  create QBitmap objects and paint them for the bitmaps
    //  present
    for (unsigned int n = 0; n < fg_bitmaps (); ++n) {
      if (fg_bitmap (n) != 0) {
        memset (p_data, 0, nbytes);
        bitmap_to_bitmap (fg_style (n), *fg_bitmap (n), p_data, m_viewport.width (), m_viewport.height (), dither_pattern ());
        QBitmap bitmap (m_viewport.width (), m_viewport.height (), p_data, true);
        painter.setPen (QRgb (fg_style (n).ormask ()));
        painter.drawPixmap (0, 0, bitmap);
      }
    }

    //  produce the combined pixmap on the window
    bitBlt (this, 0, 0, &px, 0, 0, mp_pixmap->size ().width (), mp_pixmap->size ().height ());

#endif

    //  delete buffer
    delete [] p_data;

    //  erase dynamic bitmaps 
    clear_fg_bitmaps ();

    QApplication::syncX ();

  }

}

class DetachedViewObjectCanvas
  : public BitmapViewObjectCanvas
{
public:
  DetachedViewObjectCanvas (QColor bg, QColor fg, QColor ac, unsigned int width, unsigned int height, QImage *img)
    : BitmapViewObjectCanvas (width, height),
      m_bg (bg), m_fg (fg), m_ac (ac), mp_image (img)
  {
    //  .. nothing yet ..
  }

  ~DetachedViewObjectCanvas ()
  {
    clear_fg_bitmaps ();
  }

  QColor background_color () const
  {
    return m_bg;
  }

  QColor foreground_color () const
  {
    return m_fg;
  }

  QColor active_color () const
  {
    return m_ac;
  }

  virtual QImage &bg_image () 
  {
    return *mp_image;
  }

  void transfer_to_image (const lay::DitherPattern &dp, unsigned int width, unsigned int height) 
  {
    bitmaps_to_image (fg_view_op_vector (), fg_bitmap_vector (), dp, mp_image, width, height, false, 0);
    clear_fg_bitmaps ();
  }

private:
  QColor m_bg, m_fg, m_ac;
  QImage *mp_image;
};

QImage 
LayoutCanvas::image (unsigned int width, unsigned int height) 
{
  QImage img (width, height, QImage::Format_RGB32);
  img.fill (m_background);
  
  //  provide canvas objects for the layout bitmaps and the foreground/background objects
  BitmapRedrawThreadCanvas rd_canvas;
  DetachedViewObjectCanvas vo_canvas (background_color (), foreground_color (), active_color (), width, height, &img);

  //  compute the new viewport 
  Viewport vp (width, height, m_viewport.target_box ());

  lay::RedrawThread redraw_thread (&rd_canvas, mp_view);

  //  render the layout
  redraw_thread.start (m_layers, vp, true /*synchroneous*/);
  redraw_thread.stop (); // safety

  //  paint the background objects. It uses "img" to paint on.
  do_render_bg (vp, vo_canvas);

  rd_canvas.to_image (m_view_ops, dither_pattern (), this, img, width, height);

  //  render the foreground parts ..
  do_render (vp, vo_canvas, true);
  vo_canvas.transfer_to_image (dither_pattern (), width, height);

  do_render (vp, vo_canvas, false);
  vo_canvas.transfer_to_image (dither_pattern (), width, height);

  return img;
}

QImage 
LayoutCanvas::screenshot () 
{
  //  if required, start the redraw thread ..
  prepare_drawing ();

  QImage img (m_viewport.width (), m_viewport.height (), QImage::Format_RGB32);
  img.fill (m_background);

  DetachedViewObjectCanvas vo_canvas (background_color (), foreground_color (), active_color (), m_viewport.width (), m_viewport.height (), &img);

  //  and paint the background objects. It uses mp_image to paint on.
  do_render_bg (m_viewport, vo_canvas);

  to_image (m_view_ops, dither_pattern (), this, img, m_viewport.width (), m_viewport.height ());

  //  render the foreground parts ..
  do_render (m_viewport, vo_canvas, true);
  vo_canvas.transfer_to_image (dither_pattern (), m_viewport.width (), m_viewport.height ());

  do_render (m_viewport, vo_canvas, false);
  vo_canvas.transfer_to_image (dither_pattern (), m_viewport.width (), m_viewport.height ());

  return img;
}

void 
LayoutCanvas::resizeEvent (QResizeEvent *)
{
  m_viewport.set_size (width (), height ());
  mouse_event_trans (m_viewport.trans ());
  redraw_all ();
  m_viewport_changed_observers.signal_observers ();
}

void 
LayoutCanvas::update_viewport ()
{
  mouse_event_trans (m_viewport.trans ());
  for (service_iterator svc = begin_services (); svc != end_services (); ++svc) {
    svc->update ();
  }
  redraw_all ();
  m_viewport_changed_observers.signal_observers ();
}

void
LayoutCanvas::zoom_box (const db::DBox &box)
{
  m_viewport.set_box (box);
  update_viewport ();
}

void
LayoutCanvas::zoom_trans (const db::DCplxTrans &trans)
{
  m_viewport.set_trans (trans);
  update_viewport ();
}

void 
LayoutCanvas::do_update_image_always ()
{
  update_image ();
}

void
LayoutCanvas::do_update_image ()
{
  update_image ();
}

bool
LayoutCanvas::event (QEvent *e) 
{
  if (e->type () == QEvent::MaxUser) {

    //  GTF probe event
    //  record the contents (the screenshot) as ASCII text
    if (gtf::Recorder::instance () && gtf::Recorder::instance ()->recording ()) {
      gtf::Recorder::instance ()->probe (this, tl::Variant (1, screenshot (), tl::Variant::has_equal_tag ()));
    }

    e->accept ();
    return true;

  } else {
    return QWidget::event (e);
  }
}

void
LayoutCanvas::redraw_all ()
{
  mp_redraw_thread->stop ();

  if (! m_need_redraw) {
    m_redraw_clearing = false;
    m_need_redraw_layer.clear ();
  }

  m_need_redraw = true;
  m_redraw_clearing = true;

  //  redraw the background elements
  touch_bg ();

  update (); // produces a paintEvent()
}

void
LayoutCanvas::redraw_new (std::vector<lay::RedrawLayerInfo> &layers)
{
  m_layers.swap (layers);
  redraw_all ();
}

void
LayoutCanvas::redraw_selected (const std::vector<int> &layers)
{
  mp_redraw_thread->stop ();

  if (! m_need_redraw) {
    m_redraw_clearing = false;
    m_need_redraw_layer.clear ();
  }

  m_need_redraw = true;
  m_need_redraw_layer = layers;

  update (); // produces a paintEvent()
}

void
LayoutCanvas::change_visibility (const std::vector <bool> &visible)
{
  mp_redraw_thread->stop ();
  mp_redraw_thread->change_visibility (visible);

  if (! m_need_redraw) {
    m_redraw_clearing = false;
    m_need_redraw_layer.clear ();
  }

  m_need_redraw = true;
  m_need_redraw_layer.clear ();

  update (); // produces a paintEvent()
}

void
LayoutCanvas::stop_redraw ()
{
  mp_redraw_thread->stop ();
}

void
LayoutCanvas::update_drawings ()
{
  update_image ();
}

void 
LayoutCanvas::signal_transfer_done () 
{
  m_do_update_image_dm ();
}

void 
LayoutCanvas::signal_end_of_drawing () 
{
  m_do_update_image_always_dm ();
}

}  // namespace lay

