
/*

  KLayout Layout Viewer
  Copyright (C) 2006-2007 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/



#ifndef HDR_layMainWindow
#define HDR_layMainWindow

#include <QtGui/QStackedWidget>
#include <QtGui/QMainWindow>
#include <QtGui/QDialog>
#include <QtCore/QTimer>

#include <vector>

#include "dbManager.h"
#include "dbLoadLayoutOptions.h"
#include "layAbstractMenu.h"
#include "layLayoutView.h"
#include "layPlugin.h"
#include "layProgress.h"
#include "tlException.h"
#include "tlDeferredExecution.h"
#include "gsiObject.h"

class QTabBar;
class QToolBar;
class QLabel;
class QAction;
class QMenu;
class QStackedWidget;

class SettingsFormImpl;

namespace rba {
  class ConsoleImpl;
}

namespace Ui {
  class HelpAboutDialog;
}

namespace lay {

class ViewWidgetStack;
class ProgressReporter;
class ProgressWidget;
class FileDialog;
class AbstractMenu;
class SaveLayoutOptionsDialog;
class LoadLayoutOptionsDialog;
class CellView;

class HelpAboutDialog
  : public QDialog
{
public:
  HelpAboutDialog (QWidget *parent); 
  ~HelpAboutDialog ();

private:
  Ui::HelpAboutDialog *mp_ui;
};

class MainWindow
  : public QMainWindow,
    public lay::CoordTracker,
    public lay::ProgressBar,
    public lay::PluginRoot,
    public gsi::Object<MainWindow>
{
Q_OBJECT
public:

  /**
   *  @brief The (only) instance of the main window
   */
  static MainWindow &instance ();

  /**
   *  @brief Constructor
   */
  MainWindow (QApplication *app = 0, const char *name = "main_window");

  /** 
   *  @brief Destructor
   */
  ~MainWindow ();

  /**
   *  @brief Get the main menu abstraction object
   */
  AbstractMenu &menu ()
  {
    return *mp_menu;
  }

  /**
   *  @brief Close all views
   */
  void close_all ();

  /**
   *  @brief Create a new view 
   *
   *  @return The index of the newly created view
   */
  int create_view ();

  /**
   *  @brief Display a status message
   */
  void message (const std::string &s, int ms);

  /**
   *  @brief Save the session to a file
   */
  void save_session (const std::string &fn);

  /**
   *  @brief Restore the session from a file
   */
  void restore_session (const std::string &fn);

  /** 
   *  @brief Reload the given cellview into the current view
   *
   *  The cellview is given by index in the current view's cellview list.
   */
  void reload_layout (unsigned int cv_index);

  /** 
   *  @brief Load a (new) file into the layout
   *
   *  The mode param controls how to open the file:
   *    0 = replace current view, 1 = create new view, 2 = add to current view
   */
  const lay::CellView &load_layout (const std::string &filename, int mode = 0);

  /** 
   *  @brief Create new, empty layout
   *
   *  The mode param controls how to create the layout
   *    0 = replace current view, 1 = create new view, 2 = add to current view
   */
  const lay::CellView &create_layout (int mode = 0);

  /**
   *  @brief Clone the current view
   */
  void clone_current_view ();

  /**
   *  @brief Load a layer definition file
   *
   *  If "all_views" is true, the layer file is loaded on all views. If false,
   *  it is loaded only on the current view.
   */
  void load_layer_properties (const std::string &fn, bool all_views);

  /** 
   *  @brief Grid accessor: get the grid value
   *
   *  This is the version delivering a double value in micron.
   */
  double grid_micron () const;

  /**
   *  @brief Hierarchy level selection setter
   */
  void set_hier_levels (std::pair<int, int> l);

  /**
   *  @brief Hierarchy level selection setter
   */
  void set_min_hier_levels (int l)
  {
    set_hier_levels (std::pair<int, int> (l, get_max_hier_levels ()));
  }

  /**
   *  @brief Hierarchy level selection setter
   */
  void set_max_hier_levels (int l)
  {
    set_hier_levels (std::pair<int, int> (get_min_hier_levels (), l));
  }

  /**
   *  @brief Hierarchy level selection getter
   */
  std::pair<int, int> get_hier_levels () const;

  /**
   *  @brief Hierarchy level selection getter
   */
  int get_min_hier_levels () const
  {
    return get_hier_levels ().first;
  }

  /**
   *  @brief Hierarchy level selection getter
   */
  int get_max_hier_levels () const
  {
    return get_hier_levels ().second;
  }

  /**
   *  @brief Return true, if there is a "last" display state
   */
  bool has_last_display_state ();

  /**
   *  @brief Return true, if there is a "next" display state
   */
  bool has_next_display_state ();

  /**
   *  @brief Display the current position
   */
  void current_pos (double x, double y, bool dbu_units);

  /**
   *  @brief Clear the current position
   */
  void clear_current_pos ();

  /**
   *  @brief Implementation of the lay::ProgressBar interface: set the text to display
   */
  void set_progress_text (const std::string &text); 

  /**
   *  @brief Implementation of the lay::ProgressBar interface: set the value to display
   */
  void set_progress_value (double v, const std::string &value);

  /**
   *  @brief Implementation of the lay::ProgressBar interface: Make the progress widget visible or invisible
   */
  void show_progress_bar (bool show);

  /**
   *  @brief Enable or disable the actions for edit functions 
   * 
   *  This method is used by non-modal dialogs that want to suppress any editing
   *  activities (like browsers) while they are open.
   */
  void enable_edits (bool enable);

  /** 
   *  @brief Implementation of the PluginRoot interface
   */
  bool configure (const std::string &name, const std::string &value);

  /** 
   *  @brief Implementation of the PluginRoot interface
   */
  void config_finalize ();

  /**
   *  @brief Tell, if exit was called
   *
   *  This method returns true if exit () has been called. This state is never reset.
   *  The main purpose is to detect "exit ()" calls before the event loop is executed.
   */
  bool exited () 
  {
    return m_exited;
  }

  /**
   *  @brief Enable synchroneous mode or disable it
   *
   *  In synchroneous mode, the program waits until the drawing has finished.
   *  If this mode is disabled, drawing is done in the background thread.
   *  Synchroneous mode is mainly intended for automation purposes.
   */
  void set_synchroneous (bool sync_mode);

  /**
   *  @brief Tell the state of the synchroneous mode flag
   */
  bool synchroneous () const
  {
    return m_synchroneous;
  }

  /**
   *  @brief Retrieve index of the current view
   *
   *  @return the current view's index or -1 if no index is available
   */
  int current_view_index () const;

  /**
   *  @brief Retrieve the current view
   */
  lay::LayoutView *current_view ();

  /**
   *  @brief Retrieve the current view (const mode)
   */
  const lay::LayoutView *current_view () const;

  /**
   *  @brief Retrieve current the current view
   */
  lay::LayoutView *view (int index);

  /**
   *  @brief Retrieve current the current view (const mode)
   */
  const lay::LayoutView *view (int index) const;

  /**
   *  @brief Return the number of number of views
   */
  unsigned int views () const
  {
    return mp_views.size ();
  }

  /**
   *  @brief Select the view with the given index
   */
  void select_view (int index);

  /**
   *  @brief Select the given mode
   *
   *  @param The index of the mode to select
   */
  void select_mode (unsigned int m);

  /**
   *  @brief Issue a plugin menu item with the given symbol
   */
  void generic_menu (const std::string &symbol);

  /**
   *  @brief Get the action for a slot
   */
  lay::Action &action_for_slot (const char *slot);

  /**
   *  @brief Create a configuration action with the given title, parameter name and value
   *
   *  The action will be owned by the main window but can be deleted to remove the action from the main window.
   */
  lay::Action *create_config_action (const std::string &title, const std::string &cname, const std::string &cvalue);
  
  /**
   *  @brief Create a configuration action with the given parameter name and value
   *
   *  The action will be owned by the main window but can be deleted to remove the action from the main window.
   *  This version is provided for applications, where the title is set later.
   */
  lay::Action *create_config_action (const std::string &cname, const std::string &cvalue);
  
  /**
   *  @brief Register a configuration action with the given name
   */
  void register_config_action (const std::string &name, lay::ConfigureAction *action);
  
  /**
   *  @brief Unregister a configuration action with the given name
   */
  void unregister_config_action (const std::string &name, lay::ConfigureAction *action);
  
  /**
   *  @brief Open a new layout in mode 'mode'
   *
   *  If mode is 2, the layout is opened in the current view in addition to the existing ones.
   *  If mode is 1, the layout is opened in a new view. If mode is 0, the layout is opened in 
   *  the current view deleting the existing ones.
   */
  void open (int mode);

  /**
   *  @brief Add an observer for the current view
   *  
   *  If the current view changes, this observer is triggered. The integer slot of the
   *  observer will receive the number of the view active before. The current view's reference
   *  is already updated when this event is issued.
   */
  void add_current_view_observer (tl::Observer &observer)
  {
    m_current_view_observers.add_observer (observer);
  }

  /**
   *  @brief Remove an observer for the current view
   */
  void remove_current_view_observer (tl::Observer &observer)
  {
    m_current_view_observers.remove_observer (observer);
  }

  /**
   *  @brief Add an observer for a new view
   *  
   *  If a new view is created, this observer will receive a signal.
   *  The integer slot of this observer will receive the index of the newly created view.
   */
  void add_new_view_observer (tl::Observer &observer)
  {
    m_new_view_observers.add_observer (observer);
  }

  /**
   *  @brief Remove an observer for a new view
   */
  void remove_new_view_observer (tl::Observer &observer)
  {
    m_new_view_observers.remove_observer (observer);
  }

  /**
   *  @brief Add an entry to the MRU list
   */
  void add_mru (const std::string &fn);

  /**
   *  @brief Get the current reader options
   */
  db::LoadLayoutOptions &reader_options () 
  {
    return m_reader_options;
  }

  /**
   *  @brief Get the current reader options (const version)
   */
  const db::LoadLayoutOptions &reader_options () const
  {
    return m_reader_options;
  }

  /**
   *  @brief Reimplementation of the dragEnterEvent event handler
   */
  void dragEnterEvent(QDragEnterEvent *event);

  /**
   *  @brief Reimplementation of the dropEvent event handler
   */
  void dropEvent(QDropEvent *event);

public slots:
  void popup_menu_show ();
  void cancel ();
  void redraw ();
  void exit ();
  void close_current_view ();
  void mode ();
  void enable ();
  void goto_bookmark ();
  void open_recent ();
  void generic_menu ();
  void view_selected (int index);
  void view_title_changed ();
  void bookmark_menu_show ();
  void file_menu_show ();

  //  menu callbacks for main window
  void cm_undo ();
  void cm_redo ();
  void cm_delete ();
  void cm_show_properties ();
  void cm_copy ();
  void cm_paste ();
  void cm_cut ();
  void cm_zoom_fit_sel ();
  void cm_zoom_fit ();
  void cm_zoom_in ();
  void cm_zoom_out ();
  void cm_pan_up ();
  void cm_pan_down ();
  void cm_pan_left ();
  void cm_pan_right ();
  void cm_save_session ();
  void cm_restore_session ();
  void cm_setup ();
  void cm_save_as ();
  void cm_save ();
  void cm_reload ();
  void cm_close ();
  void cm_clone ();
  void cm_layout_props ();
  void cm_inc_max_hier ();
  void cm_dec_max_hier ();
  void cm_max_hier ();
  void cm_max_hier_0 ();
  void cm_max_hier_1 ();
  void cm_last_display_state ();
  void cm_next_display_state ();
  void cm_cancel ();
  void cm_redraw ();
  void cm_screenshot ();
  void cm_save_layer_props ();
  void cm_load_layer_props ();
  void cm_save_bookmarks ();
  void cm_load_bookmarks ();
  void cm_select_cell ();
  void cm_select_current_cell ();
  void cm_exit ();
  void cm_bookmark_view ();
  void cm_manage_bookmarks ();
  void cm_goto_position ();
  void cm_help_about ();
  void cm_console ();
  void cm_open_too ();
  void cm_open_new_view ();
  void cm_open ();
  void cm_reader_options ();
  void cm_new_layout ();
  void cm_adjust_origin ();
  void cm_new_cell ();
  void cm_new_layer ();
  void cm_clear_layer ();
  void cm_delete_layer ();
  void cm_edit_layer ();
  void cm_edit_boolean ();
  void cm_edit_size ();
  void cm_edit_merge ();
  void cm_sel_flip_x ();
  void cm_sel_flip_y ();
  void cm_sel_rot_cw ();
  void cm_sel_rot_ccw ();
  void cm_sel_free_rot ();
  void cm_sel_scale ();
  void cm_sel_move ();

  //  forwarded to the current view: layer list context menu
  void cm_lv_hide ();
  void cm_lv_show ();
  void cm_lv_rename ();
  void cm_lv_select_all ();
  void cm_lv_delete ();
  void cm_lv_insert ();
  void cm_lv_group ();
  void cm_lv_ungroup ();
  void cm_lv_source ();
  void cm_lv_sort_by_name ();
  void cm_lv_sort_by_ild ();
  void cm_lv_sort_by_idl ();
  void cm_lv_sort_by_ldi ();
  void cm_lv_sort_by_dli ();
  void cm_lv_regroup_by_index ();
  void cm_lv_regroup_by_datatype ();
  void cm_lv_regroup_by_layer ();
  void cm_lv_regroup_flatten ();
  void cm_lv_expand_all ();
  void cm_lv_add_missing ();
  void cm_lv_remove_unused ();

  // forwarded to the current view: cell list context menu
  void cm_cell_delete ();
  void cm_cell_rename ();
  void cm_cell_flatten ();
  void cm_cell_copy ();
  void cm_cell_cut ();
  void cm_cell_paste ();
  void cm_cell_select ();
  void cm_open_current_cell ();
  void cm_save_current_cell_as ();
  void cm_cell_hide ();
  void cm_cell_show ();
  void cm_cell_show_all ();

protected slots:
  void menu_changed ();
  void message_timer ();

protected:
  void update_content ();
  void do_update_menu ();

private:
  //  Main menu
  AbstractMenu *mp_menu;
  QMenu *mp_goto_bookmark_menu;
  QTabBar *mp_tab_bar;
  QToolBar *mp_tool_bar;
  ViewWidgetStack *mp_view_stack;
  lay::FileDialog *mp_bookmarks_fdia;
  lay::FileDialog *mp_session_fdia;
  lay::FileDialog *mp_lprops_fdia;
  lay::FileDialog *mp_screenshot_fdia;
  lay::FileDialog *mp_layout_fdia;
  lay::SaveLayoutOptionsDialog *mp_layout_save_options;
  lay::LoadLayoutOptionsDialog *mp_layout_load_options;
  std::vector<lay::Action> m_mode_menu_actions;
  unsigned int m_mode;
  std::vector< std::pair<unsigned int, lay::Action> > m_plugin_enable_actions;
  std::vector<bool> m_plugin_enabled;
  SettingsFormImpl *mp_setup_form;
  std::vector <lay::LayoutView *> mp_views;
  size_t m_current_view;
  int m_open_mode;
  std::vector<std::string> m_mru;
  QStatusBar *mp_status_bar;
  QStackedWidget *mp_right_stack;
  ProgressWidget *mp_progress_widget;
  QFrame *mp_cp_frame;
  QLabel *mp_cpx_label, *mp_cpy_label;
  QLabel *mp_msg_label;
  bool m_edits_enabled;
  bool m_disable_tab_selected;
  bool m_exited;
  tl::DeferredMethod<MainWindow> m_do_update_menu_dm;
  QTimer m_message_timer;

  db::LoadLayoutOptions m_reader_options;
  double m_grid_micron;
  std::vector<double> m_default_grids;
  bool m_default_grids_updated;
  std::vector<lay::Action *> m_default_grid_actions;
  double m_new_cell_window_size;
  std::string m_new_cell_cell_name;
  double m_default_dbu;
  bool m_synchronized_views;
  bool m_synchroneous;
  int m_disabled_edits;
  bool m_work_in_progress;
  QApplication *mp_app;
  rba::ConsoleImpl *mp_console;
  std::string m_current_session;

  std::map<std::string, lay::Action> m_actions_for_slot;
  std::map<std::string, std::vector<lay::ConfigureAction *> > m_configuration_actions;
  tl::Collection <lay::ConfigureAction> m_ca_collection;

  //  the object manager (undo/redo mechanism and others)
  db::Manager m_manager;

  lay::ProgressReporter *mp_pr;

  tl::Observed m_current_view_observers;
  tl::Observed m_new_view_observers;

  void init_menu ();

  /** 
   *  @brief Event filter
   *  
   *  In order to be able to filter out mouse events in "work_in_progress" mode,
   *  we implement an event filter here.
   */
  bool eventFilter (QObject *watched, QEvent *event);

  /**
   *  @brief Handle QCloseEvent events
   */
  void closeEvent (QCloseEvent *event);

  /**
   *  @brief Collect the dirty files in a list
   *
   *  @return The number of dirty files
   *  @param dirty_files A string containing the files
   */
  int dirty_files (std::string &dirty_files);

  /**
   *  @brief Check, if we can close this window
   */
  bool can_close ();

  /**
   *  @brief Create a new view 
   *
   *  @return The index of the newly created view
   */
  int do_create_view ();

  const lay::CellView &create_or_load_layout (const std::string *filename, int mode);
  void do_save (bool as);
};

}

namespace gsi {
  template <> struct ClassTag<lay::MainWindow> {
    typedef CantCopySemanticsTag copy_semantics;
    typedef CantDefaultCreateSemanticsTag create_semantics;
  };
}

#endif
