
/*

  KLayout Layout Viewer
  Copyright (C) 2006-2007 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#include <QtCore/QEvent>
#include <QtGui/QApplication>

#include "layRedrawThread.h"
#include "tlLog.h"
#include "tlAssert.h"
#include "dbHershey.h"
#include "dbShape.h"

namespace lay 
{

//  time delay until the first snapshot is taken
const int first_snapshot_delay = 20;

// -------------------------------------------------------------

RedrawThread::RedrawThread (lay::RedrawThreadCanvas *canvas, lay::LayoutView *view)
  : QThread (),
    tl::Observer (),
    m_initial_update (false),
    mp_layout (0),
    mp_cell_var_cache (0),
    m_cache_hits (0),
    m_cache_misses (0),
    m_cv_index (-1),
    mp_canvas (canvas),
    mp_view (view),
    m_start_recursion_sentinel (false),
    m_request_stop (false),
    m_request_terminate (false),
    m_redraw (false),
    m_test_count (0),
    m_from_level (0), m_to_level (0),
    m_from_level_default (0), m_to_level_default (0),
    m_drawing_context (false),
    m_nlayers (0),
    m_box_text_transform (false),
    m_box_font (0),
    m_min_size_for_label (1),
    m_text_font (0),
    m_text_visible (false),
    m_text_lazy_rendering (false),
    m_show_properties (false),
    m_apply_text_trans (false),
    m_default_text_size (0.0),
    m_drop_small_cells (false),
    m_drop_small_cells_value (0),
    m_drop_small_cells_cond (lay::LayoutView::DSC_Min),
    m_draw_array_border_instances (false),
    m_layer (0),
    m_level (0),
    m_custom_draw_boxes (false),
    m_inv_prop_sel (false),
    m_ci (0),
    m_prop_id (0),
    m_clock (tl::Clock::current ()),
    m_width (0), m_height (0),
    m_boxes_already_drawn (false), 
    m_custom_already_drawn (false),
    m_synchroneous (false)
{
  for (unsigned int i = 0; i < sizeof (m_planes) / sizeof (m_planes[0]); ++i) {
    m_planes[i] = 0;
  }
}

RedrawThread::~RedrawThread ()
{
  //  terminate thread if already running
  if (running ()) {
    //  wake up the threads, so the they can process the m_request_terminate 
    //  signal
    m_request_stop = true;
    m_request_terminate = true;
    m_restart_cond.wakeAll ();
    //  no terminate ();: do a smooth shutdown.
    wait ();
  }

  for (unsigned int i = 0; i < sizeof (m_planes) / sizeof (m_planes[0]); ++i) {
    if (m_planes[i]) {
      delete m_planes[i];
      m_planes[i] = 0;
    }
  }
}

void RedrawThread::stop ()
{
  m_request_stop = true;

  //  wait until the drawing thread has finished and unlocks the
  //  main lock.
  m_main_lock.lock ();
  m_main_lock.unlock ();

  //  release all cellview references
  //  do this in the main thread since this way there is no problem 
  //  we thread-safetiness of reference counting in the cellview references.
  m_cellviews.clear ();
}

void RedrawThread::signal_observed ()
{
  if (running () && tl::verbosity () >= 30) {
    tl::info << tl::translate ("Layout changed: redraw thread stopped");
  }

  //  if something changed on the layouts we observe, stop the redraw thread
  stop ();
}

void
RedrawThread::start (const std::vector <lay::RedrawLayerInfo> &layers, const lay::Viewport &vp, bool synchroneous)
{
  m_vp_trans = vp.trans ();
  m_width = vp.width ();
  m_height = vp.height ();

  std::vector<int> restart;
  do_start (true, &layers, restart, synchroneous);
}

void  
RedrawThread::restart (const std::vector<int> &restart)
{
  do_start (false, 0, restart, m_synchroneous);
}

void  
RedrawThread::change_visibility (const std::vector<bool> &visibility)
{
  for (unsigned int i = 0; i < visibility.size () && i < m_layers.size (); ++i) {
    m_layers [i].visible = visibility [i];
  }
}

void 
RedrawThread::do_start (bool clear, const std::vector <lay::RedrawLayerInfo> *layers, const std::vector<int> &restart, bool synchroneous)
{
  m_synchroneous = synchroneous;

  //  We need this recursion sentinel because start may be called recursively
  //  via Layout::update which may open a progress widget and do processEvents.
  //  This may result in a repaint for example issuing a "start" call.
  //  Recursion however is forbidden because of the deadlocking locks used originally
  //  to lock out the redraw thread.
  if (m_start_recursion_sentinel) {
    return;
  }
  m_start_recursion_sentinel = true;

  //  stop anything that runs already
  m_request_stop = true;

  //  wait until the drawing thread has finished and unlocks the
  //  main lock.
  m_main_lock.lock ();
 
  //  reset stop request flag
  m_request_stop = false;

  //  the redraw flag tells the redraw thread to immediately 
  //  redraw.
  m_redraw = true;

  if (! running ()) {
    //  start the thread if not done already.
    QThread::start ();
  }

  {
    if (tl::verbosity () >= 30) {
      tl::info << tl::translate ("Preparing to draw");
    }
    tl::SelfTimer timer (tl::verbosity () >= 31, tl::translate ("Preparing to draw"));

    //  detach from all layout objects 
    tl::Observer::detach_all ();

    //  Update all relevant layout objects.
    for (unsigned int i = 0; i < mp_view->cellviews (); ++i) {
      const lay::CellView &cv = mp_view->cellview (i);
      if (cv.is_valid () && ! cv->layout ().under_construction () && ! (cv->layout ().manager () && cv->layout ().manager ()->transacting ())) {
        cv->layout ().update ();
        //  attach to the layout object to receive change notifications to stop the redraw thread
        cv->layout ().add_hier_observer (*this);
        cv->layout ().add_bboxes_observer (*this);
      }
    }
    mp_view->annotation_shapes ().update ();
    //  attach to the layout object to receive change notifications to stop the redraw thread
    mp_view->annotation_shapes ().add_hier_observer (*this);  //  not really required, since the shapes have no hierarchy, but for completeness ..
    mp_view->annotation_shapes ().add_bboxes_observer (*this);
    mp_view->add_cellviews_about_to_change_observer (*this);
    mp_view->add_cellview_about_to_change_observer (*this);

    //  copy everything that we need so there is no need to access 
    //  members of lay::LayoutView in the drawing thread.
    //  Note: copying the cellviews will create new references to the
    //  layout objects. These are not automatically freed when the 
    //  drawing ends but rather on "stop". The advantage of this is
    //  that, since "stop" is called from the main thread like "start",
    //  we don't challenge lay::CellView's MT compliance.
    m_from_level_default = mp_view->get_hier_levels ().first;
    m_to_level_default = mp_view->get_hier_levels ().second;
    m_min_size_for_label = mp_view->min_inst_label_size ();
    m_box_text_transform = mp_view->cell_box_text_transform ();
    m_box_font = mp_view->cell_box_text_font ();
    m_text_font = mp_view->text_font ();
    m_text_visible = mp_view->text_visible ();
    m_text_lazy_rendering = mp_view->text_lazy_rendering ();
    m_show_properties = mp_view->show_properties_as_text ();
    m_apply_text_trans = mp_view->apply_text_trans ();
    m_default_text_size = mp_view->default_text_size ();
    m_drop_small_cells = mp_view->drop_small_cells ();
    m_drop_small_cells_value = mp_view->drop_small_cells_value ();
    m_drop_small_cells_cond = mp_view->drop_small_cells_cond ();
    m_draw_array_border_instances = mp_view->draw_array_border_instances ();
    m_test_count = 0;
    m_initial_update = true;

    m_trans = db::CplxTrans ();
    mp_prop_sel = 0;
    m_inv_prop_sel = false;
    m_custom_draw_boxes = false;

    m_hidden_cells = mp_view->hidden_cells ();

    m_cellviews.clear ();
    m_cellviews.reserve (mp_view->cellviews ());
    for (unsigned int i = 0; i < mp_view->cellviews (); ++i) {
      m_cellviews.push_back (mp_view->cellview (i));
    }

    if (clear) {
      m_layers = *layers;
    }

    m_nlayers = m_layers.size ();

    m_box_variants = mp_view->cv_transform_variants ();

    m_ci = 0;
    m_prop_id = 0;

    m_draw_queue.clear ();

    if (mp_view->cellviews () > 0) {

      if (clear) {
        mp_canvas->prepare (m_nlayers * planes_per_layer + special_planes_before + special_planes_after, m_width, m_height, mp_view->drawings ());
        m_boxes_already_drawn = false;
        m_custom_already_drawn = false;
      } else {
        for (std::vector<int>::const_iterator l = restart.begin (); l != restart.end (); ++l) {
          if (*l >= 0 && *l < int (m_layers.size ())) {
            m_layers [*l].enabled = true;
          } else if (*l == draw_boxes_queue_entry) {
            m_boxes_already_drawn = false;
          } else if (*l == draw_custom_queue_entry) {
            m_custom_already_drawn = false;
          }
        }
      }

      //  set up the drawing queue: controls the sequence of things to 
      //  draw. First there are the visible layers, then there are the
      //  cell boundaries. Then come the invisible layers.

      m_draw_queue.reserve (m_nlayers + special_queue_entries);

      //  decoration drawing
      if (! m_custom_already_drawn) {
        m_draw_queue.push_back (draw_custom_queue_entry);
      }

      for (unsigned int i = 0; i < m_nlayers; ++i) {
        if (m_layers [i].visible && m_layers [i].enabled) {
          m_draw_queue.push_back (int (i));
        }
      }

      //  cell box drawing
      if (! m_boxes_already_drawn) {
        m_draw_queue.push_back (draw_boxes_queue_entry);
      }

    } else {
      mp_canvas->prepare (1, m_width, m_height, mp_view->drawings ());
    }

  }

  //  setup the wait condition for the initial transfer
  m_initial_wait_lock.lock ();

  //  provide a signal for the waiting thread to restart
  m_restart_cond.wakeAll ();
  //  release the locks and let the thread do it's work
  m_main_lock.unlock ();

  //  wait for the first transfer 
  m_initial_wait_cond.wait (&m_initial_wait_lock);
  m_initial_update = false;
  m_initial_wait_lock.unlock ();

  m_start_recursion_sentinel = false;
}

bool 
RedrawThread::cell_var_cached () 
{
  if (mp_cell_var_cache) {
    //  Use the native transformation (just including cell instantiation components) to enable
    //  fuzzy comparison of floating-point coordinates: this requires a well-defined unit system to
    //  allow the definition of an uncertainty value.
    db::CplxTrans db_trans ((m_vp_trans * mp_layout->dbu ()).inverted () * db::DCplxTrans (m_trans));
    if (mp_cell_var_cache->find (std::make_pair (db_trans, m_ci)) != mp_cell_var_cache->end ()) {
      ++m_cache_hits;
      return true;
    }
    ++m_cache_misses;
    mp_cell_var_cache->insert (std::make_pair (db_trans, m_ci));
  }
  return false;
}

void 
RedrawThread::iterate_variants (void (RedrawThread::*what) ())
{
  m_level = 0;

  //  save current state
  db::CplxTrans trans = m_trans;
  int from_level = m_from_level;
  int to_level = m_to_level;
  
  //  if a context path is given, we adjust the levels such that the target (not the context 
  //  cell) is drawn and the context cell is visible through addressing negative levels. The
  //  iterate_variants_rec methods takes care of converting the negative hierarchy levels into
  //  traversals along the context path bottom up.
  size_t ctx_path_length = m_cellviews [m_cv_index].specific_path ().size ();
  if (ctx_path_length > 0) {
    m_from_level -= ctx_path_length;
    m_to_level -= ctx_path_length;
    m_trans = m_trans * m_cellviews [m_cv_index].context_trans ();
  }

  if (m_from_level_default < 0 || ctx_path_length > 0) {

    //  if we start from above the hierarchy, we need to establish a 
    //  cell variant cache to at least avoid painting the current cell
    //  multiple times.
    std::set <std::pair <db::CplxTrans, db::cell_index_type>, lay::CellVariantCacheCompare> cell_var_cache;
    mp_cell_var_cache = &cell_var_cache;

    //  Use the cell variant cache to exclude the basic instance from the 
    //  drawing in the first pass.
    cell_var_cache.insert (std::make_pair (m_cellviews [m_cv_index].context_trans (), m_ci));

    m_cache_hits = m_cache_misses = 0;

    //  draw the context for the current instance
    m_drawing_context = true;
    iterate_variants_rec (what, true);

    cell_var_cache.clear ();

    //  draw the current instance without context (using a minimum of from_level=0 for this)
    mp_cell_var_cache = 0;
    m_drawing_context = false;
    int fl = m_from_level;
    if (m_from_level < 0) {
      m_from_level = 0;
    }
    iterate_variants_rec (what, false);
    m_from_level = fl;

    if (tl::verbosity () >= 40) {
      tl::info << tl::translate ("Cell variant cache hits/misses: ") << m_cache_hits << "/" << m_cache_misses;
    }

  } else {
    mp_cell_var_cache = 0;
    m_drawing_context = false;
    iterate_variants_rec (what, false);
  }

  //  restore state
  m_trans = trans;
  m_from_level = from_level;
  m_to_level = to_level;
}

void 
RedrawThread::iterate_variants_rec (void (RedrawThread::*what) (), bool spread)
{
  db::Cell::parent_inst_iterator p = mp_layout->cell (m_ci).begin_parent_insts ();
  int context_path_length = int (m_cellviews [m_cv_index].specific_path ().size ());

  if ((spread || m_level > m_from_level) && m_level + context_path_length > 0) {

    //  pull an specific instance from the instance stack and move this one up
    --m_level;

    const db::InstElement &ie = m_cellviews [m_cv_index].specific_path ().end () [m_level];

    db::cell_index_type ci = m_ci;
    db::CplxTrans trans = m_trans;
    db::Box vp = m_vp;

    if (m_level + context_path_length > 0) {
      m_ci = m_cellviews [m_cv_index].specific_path ().end () [m_level - 1].inst_ptr.cell_index ();
    } else {
      m_ci = m_cellviews [m_cv_index].ctx_cell_index ();
    }

    db::CplxTrans t (ie.complex_trans ());
    m_trans = trans * t.inverted ();
    m_vp = db::Box::from_double (t * vp);
    iterate_variants_rec (what, spread);

    m_trans = trans;
    m_vp = vp;
    m_ci = ci;

    ++m_level;

  } else if (m_level > (spread ? (m_from_level_default - context_path_length) : m_from_level) && ! p.at_end ()) {

    //  one level up ..
    --m_level;

    db::CplxTrans trans = m_trans;
    db::Box vp = m_vp;
    db::cell_index_type ci = m_ci;

    while (! p.at_end ()) {

      db::Cell::cell_inst_array_type pi = (*p).inst ();

      m_ci = pi.object ().cell_index ();

      for (db::Cell::cell_inst_array_type::iterator pp = pi.begin (); ! pp.at_end (); ++pp) {

        db::CplxTrans t (pi.complex_trans (*pp));
        m_trans = trans * t;
        m_vp = db::Box::from_double (t.inverted () * vp);
        iterate_variants_rec (what, spread);

      }

      ++p;

    }

    m_trans = trans;
    m_vp = vp;
    m_ci = ci;

    ++m_level;

  } else {

    db::Coord lim = std::numeric_limits<db::Coord>::max ();
    db::DBox world (m_trans * db::Box (db::Point (-lim, -lim), db::Point (lim, lim)));
    m_vp = db::Box::from_double (m_trans.inverted () * (world & db::DBox (db::DPoint (0.0, 0.0), db::DPoint (m_width, m_height))));
    m_vp &= mp_layout->cell (m_ci).bbox (); // this avoids problems with accessing designs through very large viewports

    (this->*what) ();

  }

}

void 
RedrawThread::run ()
{
  while (! m_request_terminate) {

    //  lock against setup code
    m_main_lock.lock ();

    m_redraw = false;
    m_custom_draw_boxes = false;

    try {

      tl::SelfTimer timer (tl::verbosity () >= 21, "Redrawing");

      //  allow a very short time to pass before we issue the
      //  first update event.
      m_clock = tl::Clock::current () - tl::Clock ((update_interval - first_snapshot_delay) * 0.001);

      //  initialize the drawing planes
      for (unsigned int i = 0; i < (unsigned int) planes_per_layer; ++i) {
        if (m_planes[i]) {
          delete m_planes[i];
        }
        m_planes[i] = mp_canvas->create_drawing_plane ();
      }

      //  give the renderer some initial capacity.
      mp_canvas->renderer ().reserve_edges (100);

      //  work the drawing queue ..
      for (std::vector <int>::const_iterator q = m_draw_queue.begin (); q != m_draw_queue.end (); ++q) {

        m_level = 0;
        m_from_level = m_from_level_default;
        m_to_level = m_to_level_default;

        if (*q >= 0) {

          //  draw a layer
          for (unsigned int i = 0; i < (unsigned int) planes_per_layer; ++i) {
            m_planes[i]->clear ();
          }

          //  HINT: the order in which the planes are delivered (the index stored in the first member of the pair below)
          //  must correspond with the order by which the ViewOp's are created inside LayoutView::set_view_ops
          m_buffers.clear ();
          for (unsigned int i = 0; i < (unsigned int) planes_per_layer / 2; ++i) {

            unsigned int i1 = *q * (planes_per_layer / 2) + special_planes_before + i;
            mp_canvas->initialize_plane (m_planes[i], i1); 
            m_buffers.push_back (std::make_pair (i1, m_planes [i]));

            unsigned int i2 = (*q + m_nlayers) * (planes_per_layer / 2) + special_planes_before + i;
            mp_canvas->initialize_plane (m_planes [i + planes_per_layer / 2], i2); 
            m_buffers.push_back (std::make_pair (i2, m_planes [i + planes_per_layer / 2]));

          }

          const RedrawLayerInfo &li = m_layers [*q];

          if (li.cellview_index >= 0) {

            //  determine layout and cell associated with this layer ..
            const lay::CellView &cv = m_cellviews [li.cellview_index];
            if (cv.is_valid () && ! cv->layout ().under_construction () && ! (cv->layout ().manager () && cv->layout ().manager ()->transacting ())) {

              mp_layout = &cv->layout ();
              m_cv_index = li.cellview_index;
              m_ci = cv.cell_index ();

              int ctx_path_length = int (m_cellviews [m_cv_index].specific_path ().size ());

              if (li.hier_levels.has_from_level ()) {
                m_from_level = li.hier_levels.from_level (ctx_path_length);
              }
              if (li.hier_levels.has_to_level ()) {
                m_to_level = li.hier_levels.to_level (ctx_path_length);
              }

              mp_prop_sel = &li.prop_sel;
              m_inv_prop_sel = li.inverse_prop_sel;
              if (mp_prop_sel->empty () && m_inv_prop_sel) {
                //  no property selection
                mp_prop_sel = 0;
              }

              if (li.layer_index >= 0) {

                m_layer = li.layer_index;
             
                if (tl::verbosity () >= 30) {
                  tl::info << tl::translate ("Drawing layer: ") << mp_layout->get_properties (m_layer).name;
                }
                tl::SelfTimer timer (tl::verbosity () >= 31, tl::translate ("Drawing layer"));

                for (std::vector<db::CplxTrans>::const_iterator t = li.trans.begin (); t != li.trans.end (); ++t) {
                  m_trans = (m_vp_trans * mp_layout->dbu ()) * db::DCplxTrans (*t);
                  m_mi_cache.clear ();
                  iterate_variants (&RedrawThread::draw_layer);
                }

              } else if (li.cell_frame) {

                //  if no specific layer is assigned, draw cell boxes with the style given
                if (tl::verbosity () >= 30) {
                  tl::info << tl::translate ("Drawing custom frames");
                }
                tl::SelfTimer timer (tl::verbosity () >= 31, tl::translate ("Drawing frames"));

                for (std::set< std::pair<db::CplxTrans, int> >::const_iterator b = m_box_variants.begin (); b != m_box_variants.end (); ++b) {
                  if (b->second == li.cellview_index) {
                    m_trans = (m_vp_trans * mp_layout->dbu ()) * db::DCplxTrans (b->first);
                    m_custom_draw_boxes = true;
                    iterate_variants (&RedrawThread::draw_boxes);
                  }
                }

              }

              mp_prop_sel = 0;
              m_inv_prop_sel = false;

            }

          }

        } else if (*q == draw_boxes_queue_entry) {

          //  draw the bounding boxes 
          if (tl::verbosity () >= 30) {
            tl::info << tl::translate ("Drawing frames");
          }
          tl::SelfTimer timer (tl::verbosity () >= 31, tl::translate ("Drawing frames"));

          m_planes [0]->clear ();
          m_planes [1]->clear ();

          m_buffers.clear ();

          mp_canvas->initialize_plane (m_planes[0], 0); 
          m_buffers.push_back (std::make_pair (0, m_planes[0]));

          mp_canvas->initialize_plane (m_planes[1], 1); 
          m_buffers.push_back (std::make_pair (1, m_planes[1]));

          for (std::set< std::pair<db::CplxTrans, int> >::const_iterator b = m_box_variants.begin (); b != m_box_variants.end (); ++b) {

            const lay::CellView &cv = m_cellviews [b->second];
            if (cv.is_valid () && ! cv->layout ().under_construction () && ! (cv->layout ().manager () && cv->layout ().manager ()->transacting ())) {

              mp_layout = &cv->layout ();
              m_ci = cv.cell_index ();
              m_cv_index = b->second;

              m_trans = (m_vp_trans * mp_layout->dbu ()) * db::DCplxTrans (b->first);
              m_custom_draw_boxes = false;

              iterate_variants (&RedrawThread::draw_boxes);

            }

          }

        } else if (*q == draw_custom_queue_entry) {

          //  draw the decorations
          if (tl::verbosity () >= 30) {
            tl::info << tl::translate ("Drawing decorations");
          }
          tl::SelfTimer timer (tl::verbosity () >= 31, tl::translate ("Drawing decorations"));

          m_planes [0]->clear ();

          m_buffers.clear ();
          mp_canvas->initialize_plane (m_planes[0], m_layers.size () * planes_per_layer + special_planes_before); 
          m_buffers.push_back (std::make_pair (m_layers.size () * planes_per_layer + special_planes_before, m_planes [0]));

          unsigned int nd = 0;
          for (lay::Drawings::iterator d = mp_view->drawings ()->begin (); d != mp_view->drawings ()->end (); ++d, ++nd) {

            //  temporarily create bitmap objects, paint on them,
            //  transfer them to the canvas and clear them again.
            //  This operation is not interrupted by any "test_snapshot".

            std::vector <lay::CanvasPlane *> tmp_planes;
            tmp_planes.reserve (d->num_planes ());
            for (unsigned int i = 0; i < d->num_planes (); ++i) {
              tmp_planes.push_back (mp_canvas->create_drawing_plane ());
              mp_canvas->initialize_plane (tmp_planes.back (), nd, i); 
            }

            //  currently, all cellviews are painted over each other ..
            for (unsigned int i = 0; i < m_cellviews.size (); ++i) {
              test_snapshot (); 
              const lay::CellView &cv = m_cellviews [i];
              if (cv.is_valid () && ! cv->layout ().under_construction () && ! (cv->layout ().manager () && cv->layout ().manager ()->transacting ())) {
                db::CplxTrans trans (m_vp_trans * cv->layout ().dbu ());
                d->paint_cv_on_planes (cv, trans, tmp_planes);
              }
            }

            //  do the non-cv-related painting
            test_snapshot (); 
            d->paint_on_planes (m_vp_trans, tmp_planes, mp_canvas->renderer ());

            for (unsigned int i = 0; i < d->num_planes (); ++i) {
              mp_canvas->set_drawing_plane (nd, i, tmp_planes [i]); 
            }

            while (! tmp_planes.empty ()) {
              delete tmp_planes.back ();
              tmp_planes.pop_back ();
            }

          }

        }

        transfer ();
        m_buffers.clear ();

        //  mark this entry as already drawn
        if (*q == draw_custom_queue_entry) {
          m_custom_already_drawn = true;
        } else if (*q == draw_boxes_queue_entry) {
          m_boxes_already_drawn = true;
        } else if (*q >= 0 && *q < int (m_layers.size ())) {
          m_layers [*q].enabled = false;
        }

      }

      //  post an event to actually draw
      if (m_initial_update) {
        m_initial_wait_lock.lock ();
        m_initial_wait_cond.wakeAll ();
        m_initial_wait_lock.unlock ();
        m_initial_update = false;
      } else {
        mp_canvas->signal_end_of_drawing ();
      }

    } catch (...) {

      //  simply catch all exceptions - they are used to abort.

      //  wake up any waiting start() method
      if (m_initial_update) {
        m_initial_wait_lock.lock ();
        m_initial_wait_cond.wakeAll ();
        m_initial_wait_lock.unlock ();
        m_initial_update = false;
      } 

    }

    //  release all cellview references here.
    m_cellviews.clear ();

    m_main_lock.unlock ();

    //  if the redraw flag is set now, a start () has happened
    //  between the last unlock and we do not need to do the 
    //  next lock/unlock pair to wait.
    if (! m_redraw && ! m_request_terminate) {
      //  this will block until the redraw is restarted ..
      m_restart_lock.lock ();
      m_restart_cond.wait (& m_restart_lock);
      m_restart_lock.unlock ();
    }

    for (unsigned int i = 0; i < sizeof (m_planes) / sizeof (m_planes[0]); ++i) {
      if (m_planes[i]) {
        delete m_planes[i];
        m_planes[i] = 0;
      }
    }
  } 
}

static bool
cells_in (const db::Layout *layout, const db::Cell &cell, 
          const std::set <db::cell_index_type> &selected,
          int levels,
          std::set <std::pair <int, db::cell_index_type> > &cache)
{
  if (selected.find (cell.cell_index ()) != selected.end ()) {
    return true;
  }
  if (levels > 0) {
    for (db::Cell::child_cell_iterator c = cell.begin_child_cells (); ! c.at_end (); ++c) {
      if (cache.find (std::make_pair (levels, *c)) == cache.end ()) {
        if (cells_in (layout, layout->cell (*c), selected, levels - 1, cache)) {
          return true;
        }
        cache.insert (std::make_pair (levels, *c));
      }
    }
  } 
  return false;
}

static bool 
need_draw_box (const db::Layout *layout, const db::Cell &cell, 
               int level, int to_level, 
               const std::vector <std::set <db::cell_index_type> > &hidden_cells, unsigned int cv_index) 
{
  if (level > to_level) {
    return false;
  }
  if (hidden_cells.size () > cv_index && ! hidden_cells [cv_index].empty ()) {
    std::set <std::pair <int, db::cell_index_type> > cache;
    if (cells_in (layout, cell, hidden_cells [cv_index], to_level - level, cache)) {
      return true;
    }
  }
  return int (cell.hierarchy_levels ()) + level >= to_level;
}

bool
RedrawThread::drop_cell (const db::Cell &cell)
{
  db::DBox bbox = m_trans * cell.bbox ();

  double value = 0;
  if (m_drop_small_cells_cond == lay::LayoutView::DSC_Min) {
    value = std::min (bbox.width (), bbox.height ());
  } else if (m_drop_small_cells_cond == lay::LayoutView::DSC_Max) {
    value = std::max (bbox.width (), bbox.height ());
  } else {
    value = bbox.width () + bbox.height ();
  }

  return (value < double (m_drop_small_cells_value));
}

void 
RedrawThread::draw_boxes ()
{
  lay::Renderer &r = mp_canvas->renderer ();

  //  do not draw, if there is nothing to draw
  if (mp_layout->cells () <= m_ci || m_vp.empty ()) {
    return;
  }

  const db::Cell &cell = mp_layout->cell (m_ci);

  //  we will never come to a valid level ..
  if (! need_draw_box (mp_layout, cell, m_level, m_to_level, m_hidden_cells, m_cv_index)) {
    return;
  }
  if (cell_var_cached ()) {
    return;
  }

  //  For small bboxes, the cell outline can be reduced ..
  db::Box bbox = cell.bbox ();

  if (bbox.empty ()) {

    //  no shapes there and below ..

  } else if (m_drop_small_cells && drop_cell (cell)) {

    //  small cell dropped

  } else if (m_level == m_to_level || (m_cv_index < int (m_hidden_cells.size ()) && m_hidden_cells [m_cv_index].find (m_ci) != m_hidden_cells [m_cv_index].end ())) {

    //  paint the box on this level
    draw_cell (bbox, mp_layout->cell_name (m_ci), m_prop_id);

  } else if (m_level < m_to_level) {

    db::DBox dbbox = m_trans * bbox;
    if (dbbox.width () < 1.5 && dbbox.height () < 1.5) {

      //  the cell is a very small box and we know there must be 
      //  some level at which to draw the boundary: just draw it 
      //  here and stop looking further down ..
      draw_cell (bbox, "", 0);

    } else {

      db::box_convert <db::CellInst> bc (*mp_layout);
     
      db::CplxTrans trans = m_trans;
      db::Box vp = m_vp;
      db::cell_index_type ci = m_ci;
      db::PropertiesRepository::properties_id_type prop_id = m_prop_id;
      ++m_level;

      //  dive down into the hierarchy ..
      db::Cell::touching_iterator inst = cell.begin_touching (m_vp); 
      while (! inst.at_end ()) {

        const db::CellInstArray &cell_inst = inst->cell_inst ();
        db::PropertiesRepository::properties_id_type cell_inst_prop = inst->prop_id ();

        db::Point a, b;
        unsigned long amax, bmax; 
        bool simplify = false;
        if (cell_inst.is_regular_array (a, b, amax, bmax)) {

          db::DBox inst_box;
          if (cell_inst.is_complex ()) {
            inst_box = trans * (cell_inst.complex_trans () * mp_layout->cell (cell_inst.object ().cell_index ()).bbox ());
          } else {
            inst_box = trans * mp_layout->cell (cell_inst.object ().cell_index ()).bbox ();
          }
          if (((a.x () == 0 && b.y () == 0) || (a.y () == 0 && b.x () == 0)) && 
              inst_box.width () < 1.5 && inst_box.height () < 1.5 && 
              (amax <= 1 || trans.ctrans (a.distance ()) < 1.5) && 
              (bmax <= 1 || trans.ctrans (b.distance ()) < 1.5)) {
            simplify = true;
          }

        }

        if (simplify) {

          //  The array can be simplified if there are levels below to draw
          if (need_draw_box (mp_layout, mp_layout->cell (cell_inst.object ().cell_index ()), m_level, m_to_level, m_hidden_cells, m_cv_index)) {
            
            db::box_convert <db::CellInst> bc (*mp_layout);
            r.draw (cell_inst.bbox (bc), trans, m_drawing_context ? m_planes[0] : m_planes[1], 0, 0, 0);

          }

        } else {

          //  The array (or single instance) must be iterated instance
          //  by instance
          for (db::CellInstArray::iterator p = cell_inst.begin_touching (vp, bc); ! p.at_end (); ++p) {
        
            test_snapshot (); 
            db::CplxTrans t (cell_inst.complex_trans (*p));
            m_trans = trans * t;
            m_vp = db::Box::from_double (t.inverted () * vp);
            m_ci = cell_inst.object ().cell_index ();
            m_prop_id = cell_inst_prop;
            draw_boxes ();

          }

        }

        ++inst;
      
      }

      --m_level;
      m_trans = trans;
      m_vp = vp;
      m_ci = ci;
      m_prop_id = prop_id;

    }

  }

}

/**
 *  @brief A helper function to determine if there are any shapes on the cell below to a certain hierarchy level
 *
 *  @return 0, if there are no shapes
 *  @return 1, if there are only contour-only shapes
 *  @return 2, if there are texts or properties to draw anywhere
 */
static int 
any_shapes (RedrawThread::MicroInstanceCache &cache, 
            const db::Layout &layout, db::cell_index_type cell_index, unsigned int layer, 
            unsigned long levels, const std::set<db::cell_index_type> *hidden_cells,
            const std::set<db::PropertiesRepository::properties_id_type> *prop_sel, bool inv_prop_sel)
{
  //  if the cell is "hidden", it does not need to be drawn
  if (hidden_cells && hidden_cells->find (cell_index) != hidden_cells->end ()) {
    return 0;
  }

  //  the cache contains all cells that do not have any shapes
  RedrawThread::MicroInstanceCache::const_iterator c = cache.find (std::make_pair (cell_index, levels));
  if (c != cache.end ()) {
    return c->second;
  }

  int ret = 0;

  const db::Cell &cell = layout.cell (cell_index);

  if (! cell.shapes (layer).begin (db::ShapeIterator::Texts).at_end ()) {

    ret = 2;

  } else if (! cell.shapes (layer).begin (db::ShapeIterator::AllWithProperties, prop_sel, inv_prop_sel).at_end ()) {

    //  shapes with properties count like text objects ..
    ret = 2;

  } else if (! cell.shapes (layer).begin (db::ShapeIterator::Polygons | db::ShapeIterator::Edges | db::ShapeIterator::Paths | db::ShapeIterator::Boxes, prop_sel, inv_prop_sel).at_end ()) {

    ret = 1;

  } else if (levels > 1) {
    
    for (db::Cell::child_cell_iterator cc = cell.begin_child_cells (); ! cc.at_end (); ++cc) {
      int r = any_shapes (cache, layout, *cc, layer, levels - 1, hidden_cells, prop_sel, inv_prop_sel);
      if (r > 0) {
        ret = r;
        break;
      }
    }

  }

  cache.insert (std::make_pair (std::make_pair (cell_index, levels), ret));

  return ret;

}

void 
RedrawThread::draw_layer ()
{
  lay::Renderer &r = mp_canvas->renderer ();

  lay::CanvasPlane *fill   = m_drawing_context ? m_planes[0] : m_planes[4];
  lay::CanvasPlane *frame  = m_drawing_context ? m_planes[1] : m_planes[5];
  lay::CanvasPlane *text   = m_drawing_context ? m_planes[2] : m_planes[6];
  lay::CanvasPlane *vertex = m_drawing_context ? m_planes[3] : m_planes[7];

  test_snapshot (); 

  //  do not draw, if there is nothing to draw
  if (mp_layout->cells () <= m_ci || m_vp.empty ()) {
    return;
  }
  if (cell_var_cached ()) {
    return;
  }

  const db::Cell &cell = mp_layout->cell (m_ci);

  //  For small bboxes, the cell outline can be reduced ..
  db::Box bbox = cell.bbox (m_layer);

  if (m_drop_small_cells && drop_cell (cell)) {

    //  small cell dropped

  } else if (! bbox.empty ()) {

    bool hidden = (m_cv_index < int (m_hidden_cells.size ()) && m_hidden_cells [m_cv_index].find (m_ci) != m_hidden_cells [m_cv_index].end ());
    bool need_to_dive = (m_level + 1 < m_to_level) && ! hidden;

    //  draw this level
    if (m_level >= m_from_level && m_level < m_to_level && ! hidden) {

      //  configure renderer to text drawing ..
      r.draw_texts (m_text_visible);
      r.draw_properties (m_show_properties);
      r.default_text_size (db::Coord (m_default_text_size / mp_layout->dbu ()));
      r.set_font (db::Font (m_text_font));
      r.apply_text_trans (m_apply_text_trans);

      db::DBox dbbox_tot = m_trans * cell.bbox ();
      bool text_simplified = m_text_lazy_rendering && (dbbox_tot.width () <= 8.0 || dbbox_tot.height () <= 8.0);
          
      const db::Shapes &shapes = cell.shapes (m_layer);

      db::DBox dbbox = m_trans * bbox;
      if ((dbbox.width () < 2.5 && dbbox.height () < 1.5) || 
          (dbbox.width () < 1.5 && dbbox.height () < 2.5)) {

        //  Here we first analyze the cell to find out 
        //  how far we need to go while maintaining the precision:

        int a;
        if (int (m_hidden_cells.size ()) <= m_cv_index || m_hidden_cells [m_cv_index].empty ()) {
          if (! mp_prop_sel && (text_simplified || ! m_text_visible)) {
            //  if there are no cells hiddden and there is no need to determine whether there are texts
            //  to draw (since these are switched off etc.), we do not need to employ the complicated
            //  "any_shapes" function. If a property selection is applied however, we need "any_shapes".
            a = 1;
          } else {
            a = any_shapes (m_mi_cache, *mp_layout, m_ci, m_layer, m_to_level - m_level, 0, mp_prop_sel, m_inv_prop_sel);
          }
        } else {
          a = any_shapes (m_mi_cache, *mp_layout, m_ci, m_layer, m_to_level - m_level, & m_hidden_cells [m_cv_index], mp_prop_sel, m_inv_prop_sel);
        }

        if (a > 0) {
          //  any shapes here: paint bbox for simplification
          r.clear ();
          r.insert (dbbox);
          r.render_contour (*frame);
          r.render_vertices (*vertex);
        }

        //  texts not visible or no texts below - do not dive further into hierarchy
        if (text_simplified || ! m_text_visible || a < 2) {
          need_to_dive = false;
        }

      } else {

        db::ShapeIterator shape (shapes.begin_touching (m_vp, db::ShapeIterator::Boxes | db::ShapeIterator::Polygons | db::ShapeIterator::Edges | db::ShapeIterator::Paths, mp_prop_sel, m_inv_prop_sel));
        while (! shape.at_end ()) {
          test_snapshot (); 
          r.draw (*shape, &mp_layout->properties_repository (), m_trans, fill, frame, vertex, text);
          ++shape;
        }

      }

      //  In lazy text rendering mode, all texts are only rendered if the cell is "reasonable large". 
      //  Empirically, a reasonable minimum dimension of 8x8 pixels was determined.
      //  otherwise just a few texts are rendered.
      if (m_text_visible) {

        //  this is the number of texts to draw in lazy text rendering mode
        size_t ntexts = 2;
        if (! text_simplified) {
          ntexts = std::numeric_limits <size_t>::max ();
        }

        db::ShapeIterator shape = shapes.begin_touching (m_vp, db::ShapeIterator::Texts, mp_prop_sel, m_inv_prop_sel);
        while (! shape.at_end () && ntexts > 0) {

          test_snapshot (); 

          r.draw (*shape, &mp_layout->properties_repository (), m_trans, fill, frame, vertex, text);
          ++shape;

          --ntexts;

        }

      }

    }

    db::box_convert <db::CellInst> bc (*mp_layout, m_layer);

    //  dive down into the hierarchy ..
    if (need_to_dive) {

      db::CplxTrans trans = m_trans;
      db::Box vp = m_vp;
      db::cell_index_type ci = m_ci;
      ++m_level;

      db::Cell::touching_iterator inst = cell.begin_touching (vp); 
      while (! inst.at_end ()) {

        const db::CellInstArray &cell_inst = inst->cell_inst ();
        ++inst;

        m_ci = cell_inst.object ().cell_index ();
        bool hidden = (m_cv_index < int (m_hidden_cells.size ()) && m_hidden_cells [m_cv_index].find (m_ci) != m_hidden_cells [m_cv_index].end ());

        db::Box cell_box = mp_layout->cell (m_ci).bbox (m_layer);
        if (! cell_box.empty () && ! hidden) {

          db::Point a, b;
          unsigned long amax, bmax; 
          bool simplify = false;
          bool anything = true;

          if (cell_inst.is_regular_array (a, b, amax, bmax)) {

            db::DBox inst_box;
            if (cell_inst.is_complex ()) {
              inst_box = trans * (cell_inst.complex_trans () * cell_box);
            } else {
              inst_box = trans * cell_box;
            }

            if (((a.x () == 0 && b.y () == 0) || (a.y () == 0 && b.x () == 0)) && 
                inst_box.width () < 1.5 && inst_box.height () < 1.5 && 
                (amax <= 1 || trans.ctrans (a.distance ()) < 1.5) && 
                (bmax <= 1 || trans.ctrans (b.distance ()) < 1.5)) {

              int as;
              if (int (m_hidden_cells.size ()) <= m_cv_index) {
                as = any_shapes (m_mi_cache, *mp_layout, m_ci, m_layer, m_to_level - m_level, 0, mp_prop_sel, m_inv_prop_sel);
              } else {
                as = any_shapes (m_mi_cache, *mp_layout, m_ci, m_layer, m_to_level - m_level, & m_hidden_cells [m_cv_index], mp_prop_sel, m_inv_prop_sel);
              }
              if (as == 0) {
                anything = false;
              } else if ((m_text_visible && as == 1) || (! m_text_visible && as >= 1)) {
                //  only for simple shapes this array instance can be simplified ...
                simplify = true;
              }

            }

          }

          if (simplify) {

            //  The array can be simplified ..
            r.draw (cell_inst.bbox (bc), trans, 0, 0, 0, 0);

            r.render_contour (*frame);
            r.render_fill (*frame);
            r.render_contour (*vertex);
            r.render_fill (*vertex);

          } else if (anything) {

            for (db::CellInstArray::iterator p = cell_inst.begin_touching (vp, bc); ! p.at_end (); ++p) {

              if (! m_draw_array_border_instances || 
                  p.index_a () == 0 || (unsigned long)p.index_a () == amax - 1 || p.index_b () == 0 || (unsigned long)p.index_b () == amax - 1) {

                db::CplxTrans t (cell_inst.complex_trans (*p));
                m_trans = trans * t;
                m_vp = db::Box::from_double (t.inverted () * vp);
                draw_layer ();

              } 

            }

          }

        }

      }

      --m_level;
      m_trans = trans;
      m_vp = vp;
      m_ci = ci;

    }

  }

}

void 
RedrawThread::draw_cell (const db::Box &box, const std::string &txt, db::PropertiesRepository::properties_id_type prop_id)
{
  lay::Renderer &r = mp_canvas->renderer ();
  r.clear ();

  db::DBox dbox = m_trans * box;
  if (m_trans.is_ortho ()) {
    r.insert (dbox);
  } else {
    db::Polygon poly (box);
    for (db::Polygon::polygon_edge_iterator e = poly.begin_edge (); ! e.at_end (); ++e) {
      r.insert (m_trans * *e);
    }
  }

  if (dbox.width () > m_min_size_for_label && dbox.height () > m_min_size_for_label) {
    r.insert (dbox, txt, 
                db::Font (m_box_font), 
                db::HAlignCenter, 
                db::VAlignCenter, 
                //  TODO: apply "real" transformation?
                db::fixpoint_trans (m_box_text_transform ? m_trans.fp_trans ().rot () : db::fixpoint_trans::r0));
  }

  //  TODO: there should not be two methods of drawing instances.
  //  In addition, instance drawing should be a method of the renderer
  if (m_custom_draw_boxes) {

    r.render_fill     (*m_planes[m_drawing_context ? 0 : 4]);
    r.render_contour  (*m_planes[m_drawing_context ? 1 : 5]);
    r.render_texts    (*m_planes[m_drawing_context ? 2 : 6]);
    r.render_vertices (*m_planes[m_drawing_context ? 3 : 7]);

    if (m_prop_id != 0 && m_show_properties) {
      r.draw_propstring (m_prop_id, &mp_layout->properties_repository (), dbox.p1 (), m_planes [m_drawing_context ? 2 : 6], m_trans);
    }

  } else {

    r.render_texts   (*m_planes [m_drawing_context ? 0 : 1]);
    r.render_contour (*m_planes [m_drawing_context ? 0 : 1]);

    if (m_prop_id != 0 && m_show_properties) {
      r.draw_propstring (m_prop_id, &mp_layout->properties_repository (), dbox.p1 (), m_planes [m_drawing_context ? 0 : 1], m_trans);
    }

  }

}

void 
RedrawThread::test_snapshot ()
{
  if (m_request_stop) {
    throw std::exception ();
  }

  if (! m_synchroneous) {
    if (m_test_count == 0) {

      m_test_count = 100; // TODO: make variable?

      tl::Clock c = tl::Clock::current ();
      if ((c - m_clock).seconds () > update_interval * 0.001) {
        transfer ();
        if (m_initial_update) {
          m_initial_wait_lock.lock ();
          m_initial_wait_cond.wakeAll ();
          m_initial_wait_lock.unlock ();
          m_initial_update = false;
        } else {
          mp_canvas->signal_transfer_done ();
        }
        m_clock = c;
      }

    } else {
      --m_test_count;
    }
  }
}

void RedrawThread::transfer ()
{
  for (std::vector<std::pair<unsigned int, lay::CanvasPlane *> >::iterator b = m_buffers.begin (); b != m_buffers.end (); ++b) {
    mp_canvas->set_plane (b->first, b->second);
  }
}

} // namespace lay
