
/*

  KLayout Layout Viewer
  Copyright (C) 2006-2007 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#include <QtGui/QGridLayout>
#include <QtGui/QMenu>
#include <QtGui/QColorDialog>
#include <QtGui/QPainter>
#include <QtGui/QPixmap>

#include "layWidgets.h"

#include "layLayoutView.h"
#include "layDialogs.h"
#include "layExceptions.h"
#include "dbLayout.h"

#include "tlInternational.h"

#include <vector>

namespace lay
{

// -------------------------------------------------------------
//  CellViewSelectionComboBox implementation

struct CellViewSelectionComboBoxPrivateData
{
  const lay::LayoutView *layout_view;
};

CellViewSelectionComboBox::CellViewSelectionComboBox (QWidget *parent)
{
  mp_private = new CellViewSelectionComboBoxPrivateData ();
  mp_private->layout_view = 0;
}

CellViewSelectionComboBox::~CellViewSelectionComboBox ()
{
  delete mp_private;
  mp_private = 0;
}

void 
CellViewSelectionComboBox::set_layout_view (const lay::LayoutView *layout_view)
{
  if (layout_view == mp_private->layout_view) {
    return;
  }

  mp_private->layout_view = layout_view;

  clear ();
  for (unsigned int cv = 0; cv < layout_view->cellviews (); ++cv) {
    addItem ((layout_view->cellview (cv)->name () + ", Cell '" 
              + layout_view->cellview (cv)->layout ().cell_name (layout_view->cellview (cv).cell_index ()) + "'").c_str ());
  }

  set_current_cv_index (layout_view->cellviews () > 0 ? 0 : -1);
}

void 
CellViewSelectionComboBox::set_current_cv_index (int cv)
{
  setCurrentIndex (cv);
}

int 
CellViewSelectionComboBox::current_cv_index () const
{
  return currentIndex ();
}

// -------------------------------------------------------------
//  LayerSelectionComboBox implementation

struct LayerSelectionComboBoxPrivateData
{
  std::vector <std::pair <db::LayerProperties, unsigned int> > layers;
  const db::Layout *layout;
  lay::LayoutView *view;
  int cv_index;
};

LayerSelectionComboBox::LayerSelectionComboBox (QWidget *parent)
{
  mp_private = new LayerSelectionComboBoxPrivateData ();
  mp_private->layout = 0;
  mp_private->view = 0;
  mp_private->cv_index = -1;

  connect (this, SIGNAL (activated (int)), this, SLOT (item_selected (int)));
}

LayerSelectionComboBox::~LayerSelectionComboBox ()
{
  delete mp_private;
  mp_private = 0;
}

void 
LayerSelectionComboBox::item_selected (int index)
{
BEGIN_PROTECTED

  if (mp_private->view != 0 && index == count () - 1) {

    setCurrentItem (-1);

    const lay::CellView &cv = mp_private->view->cellview (mp_private->cv_index);
    db::LayerProperties lp;

    if (! mp_private->view->current_layer ().is_null ()) {
      int li = mp_private->view->current_layer ()->layer_index ();
      if (li >= 0) {
        lp = mp_private->view->cellview (mp_private->view->current_layer ()->cellview_index ())->layout ().get_properties (li);
      }
    }

    lay::NewLayerPropertiesDialog prop_dia (this);
    if (prop_dia.exec (cv, lp)) {

      for (unsigned int l = 0; l < cv->layout ().layers (); ++l) {
        if (cv->layout ().is_valid_layer (l) && cv->layout ().get_properties (l).log_equal (lp)) {
          throw tl::Exception (tl::translate ("A layer with that signature already exists: ") + lp.to_string ());
        }
      }

      mp_private->view->manager ()->transaction (tl::translate ("New layer")); 

      unsigned int l = cv->layout ().insert_layer (lp);
      std::vector <unsigned int> nl;
      nl.push_back (l);
      mp_private->view->add_new_layers (nl, mp_private->cv_index);
      mp_private->view->update_content ();

      mp_private->view->manager ()->commit ();

      insertItem (index, QString (lp.to_string ().c_str ()));
      setCurrentItem (index);

      mp_private->layers.push_back (std::make_pair (lp, l));

    }

  }

END_PROTECTED;
}

void 
LayerSelectionComboBox::set_view (lay::LayoutView *view, int cv_index)
{
  if (view == 0 || cv_index < 0) {
    set_layout (0);
    return;
  }

  mp_private->layout = &view->cellview (cv_index)->layout ();
  mp_private->view = view;
  mp_private->cv_index = cv_index;

  int i = currentIndex ();
  db::LayerProperties props;
  if (i >= 0 && i < int (mp_private->layers.size ())) {
    props = mp_private->layers [i].first;
  }

  mp_private->layers.clear ();

  int new_layer = -1;
  for (unsigned int l = 0; l < mp_private->layout->layers (); ++l) {
    if (mp_private->layout->is_valid_layer (l)) {
      mp_private->layers.push_back (std::make_pair (mp_private->layout->get_properties (l), l));
      if (props == mp_private->layers.back ().first) {
        new_layer = int (l);
      }
    }
  }

  std::sort (mp_private->layers.begin (), mp_private->layers.end ());

  clear ();
  for (std::vector <std::pair <db::LayerProperties, unsigned int> >::iterator ll = mp_private->layers.begin (); ll != mp_private->layers.end (); ++ll) {
    addItem (QString (ll->first.to_string ().c_str ()));
  }
  addItem (QString (tl::translate ("New Layer ..").c_str ()));

  set_current_layer (new_layer);
}

void 
LayerSelectionComboBox::set_layout (const db::Layout *layout)
{
  mp_private->layout = layout;
  mp_private->view = 0;
  mp_private->cv_index = -1;

  int i = currentIndex ();
  db::LayerProperties props;
  if (i >= 0 && i < int (mp_private->layers.size ())) {
    props = mp_private->layers [i].first;
  }

  mp_private->layers.clear ();
  clear ();

  if (layout != 0) {

    int new_layer = -1;
    for (unsigned int l = 0; l < layout->layers (); ++l) {
      if (layout->is_valid_layer (l)) {
        mp_private->layers.push_back (std::make_pair (layout->get_properties (l), l));
        if (props == mp_private->layers.back ().first) {
          new_layer = int (l);
        }
      }
    }

    std::sort (mp_private->layers.begin (), mp_private->layers.end ());

    for (std::vector <std::pair <db::LayerProperties, unsigned int> >::iterator ll = mp_private->layers.begin (); ll != mp_private->layers.end (); ++ll) {
      addItem (QString (ll->first.to_string ().c_str ()));
    }

    set_current_layer (new_layer);

  } else {
    set_current_layer (-1);
  }
}

void 
LayerSelectionComboBox::set_current_layer (int l)
{
  if (l < 0) {
    setCurrentIndex (-1);
  } else {
    for (std::vector <std::pair <db::LayerProperties, unsigned int> >::iterator ll = mp_private->layers.begin (); ll != mp_private->layers.end (); ++ll) {
      if (ll->second == (unsigned int) l) {
        setCurrentIndex (std::distance (mp_private->layers.begin (), ll));
      }
    }
  }
}

int 
LayerSelectionComboBox::current_layer () const
{
  int i = currentIndex ();
  if (i < 0 || i > int (mp_private->layers.size ())) {
    return -1;
  } else {
    return int (mp_private->layers [i].second);
  }
}

// -------------------------------------------------------------
//  SimpleColorButton implementation

SimpleColorButton::SimpleColorButton (QWidget *parent, const char *name)
  : QPushButton (parent, name)
{
  connect (this, SIGNAL (clicked ()), this, SLOT (selected ()));
}

SimpleColorButton::SimpleColorButton (QPushButton *&to_replace, const char *name)
  : QPushButton (to_replace->parentWidget (), name)
{
  //  If the push button was part of a layout, replace it.
  //  This is somewhat tricky because there is no common method of
  //  the layout managers to replace a widget.

  QLayout *ly = to_replace->parentWidget ()->layout ();
  if (ly) {

    QBoxLayout *bx_ly = dynamic_cast <QBoxLayout *> (ly);
    if (bx_ly) {
      int i = ly->indexOf (to_replace);
      bx_ly->insertWidget (i, this);
    }

    QGridLayout *grid_ly = dynamic_cast <QGridLayout *> (ly);
    if (grid_ly) {
      int i = ly->indexOf (to_replace);
      int row = 0, column = 0;
      int row_span = 0, column_span = 0;
      grid_ly->getItemPosition (i, &row, &column, &row_span, &column_span);
      grid_ly->addWidget (this, row, column, row_span, column_span);
    }

  }

  delete to_replace;
  to_replace = 0;

  connect (this, SIGNAL (clicked ()), this, SLOT (selected ()));
}

void
SimpleColorButton::set_color (QColor c)
{
  set_color_internal (c);
}

void 
SimpleColorButton::set_color_internal (QColor c)
{
  m_color = c;

  QFontMetrics fm (font (), this);
  QRect rt (fm.boundingRect (tl::translate ("Auto").c_str ())); // dummy text to be compliant with the other color button
  QPixmap pxmp (rt.width () + 24, rt.height () + 8);

  QPainter pxpainter (&pxmp);
  QColor text_color = palette ().color (QPalette::Active, QPalette::Text);
  pxpainter.setPen (QPen (text_color));
  pxpainter.setBrush (QBrush (c.isValid () ? c : QColor (128, 128, 128)));
  QRect r (0, 0, pxmp.width () - 1, pxmp.height () - 1);
  pxpainter.drawRect (r);

  setIconSize (pxmp.size ());
  setPixmap (pxmp);
}

QColor 
SimpleColorButton::get_color () const
{
  return m_color;
}

void
SimpleColorButton::selected ()
{
  QColor c = QColorDialog::getColor (get_color (), this);
  if (c.isValid ()) {
    set_color (c);
    emit color_changed (m_color);
  }
}

// -------------------------------------------------------------
//  ColorButton implementation

ColorButton::ColorButton (QWidget *parent, const char *name)
  : QPushButton (parent, name)
{
  QMenu *menu = new QMenu (this);
  menu->addAction (tl::translate ("Automatic").c_str (), this, SLOT (auto_selected ()));
  menu->addAction (tl::translate ("Choose ...").c_str (), this, SLOT (other_selected ()));
  setMenu (menu);
}

ColorButton::ColorButton (QPushButton *&to_replace, const char *name)
  : QPushButton (to_replace->parentWidget (), name)
{
  //  If the push button was part of a layout, replace it.
  //  This is somewhat tricky because there is no common method of
  //  the layout managers to replace a widget.

  QLayout *ly = to_replace->parentWidget ()->layout ();
  if (ly) {

    QBoxLayout *bx_ly = dynamic_cast <QBoxLayout *> (ly);
    if (bx_ly) {
      int i = ly->indexOf (to_replace);
      bx_ly->insertWidget (i, this);
    }

    QGridLayout *grid_ly = dynamic_cast <QGridLayout *> (ly);
    if (grid_ly) {
      int i = ly->indexOf (to_replace);
      int row = 0, column = 0;
      int row_span = 0, column_span = 0;
      grid_ly->getItemPosition (i, &row, &column, &row_span, &column_span);
      grid_ly->addWidget (this, row, column, row_span, column_span);
    }

  }

  delete to_replace;
  to_replace = 0;

  QMenu *menu = new QMenu (this);
  menu->addAction (tl::translate ("Automatic").c_str (), this, SLOT (auto_selected ()));
  menu->addAction (tl::translate ("Choose ...").c_str (), this, SLOT (other_selected ()));
  setMenu (menu);
}

void
ColorButton::set_color (QColor c)
{
  set_color_internal (c);
}

void 
ColorButton::set_color_internal (QColor c)
{
  m_color = c;

  QFontMetrics fm (font (), this);
  QRect rt (fm.boundingRect (tl::translate ("Auto").c_str ()));
  QPixmap pxmp (rt.width () + 24, rt.height () + 8);

  QPainter pxpainter (&pxmp);
  QColor text_color = palette ().color (QPalette::Active, QPalette::Text);
  pxpainter.setPen (QPen (text_color));
  pxpainter.setBrush (QBrush (c.isValid () ? c : QColor (128, 128, 128)));
  QRect r (0, 0, pxmp.width () - 1, pxmp.height () - 1);
  pxpainter.drawRect (r);

  if (! c.isValid ()) {
    pxpainter.setFont (font ());
    pxpainter.setPen (QPen (text_color));
    pxpainter.drawText (r, Qt::AlignHCenter | Qt::AlignVCenter | Qt::TextSingleLine, tl::translate ("Auto").c_str ());
  }

  setIconSize (pxmp.size ());
  setPixmap (pxmp);
}

QColor 
ColorButton::get_color () const
{
  return m_color;
}

void
ColorButton::auto_selected ()
{
  set_color (QColor ());
  emit color_changed (m_color);
}

void
ColorButton::other_selected ()
{
  QColor c = QColorDialog::getColor (get_color (), this);
  if (c.isValid ()) {
    set_color (c);
    emit color_changed (m_color);
  }
}


}

