
/*

  KLayout Layout Viewer
  Copyright (C) 2006-2007 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#include "rdbMarkerBrowserDialog.h"
#include "rdb.h"
#include "rdbReader.h"
#include "layLayoutView.h"
#include "layExceptions.h"
#include "layFileDialog.h"
#include "ConfigurationDialogImpl.h"

#include <QtGui/QMessageBox>

#include <memory>

namespace rdb
{

extern std::string cfg_rdb_context_mode;
extern std::string cfg_rdb_window_mode;
extern std::string cfg_rdb_window_dim;
extern std::string cfg_rdb_max_marker_count;

MarkerBrowserDialog::MarkerBrowserDialog (lay::MainWindow *main, lay::LayoutView *vw)
  : lay::Browser (main, vw),
    Ui::MarkerBrowserDialog (),
    m_context (rdb::AnyCell),
    m_window (rdb::FitMarker),
    m_window_dim (0.0),
    m_max_marker_count (0),
    m_cv_index (-1),
    m_rdb_index (-1),
    m_cellviews_changed_delegate (this, &MarkerBrowserDialog::cellviews_changed),
    m_cellview_changed_delegate (this, &MarkerBrowserDialog::cellview_changed),
    m_rdbs_changed_delegate (this, &MarkerBrowserDialog::rdbs_changed)
{
  Ui::MarkerBrowserDialog::setupUi (this);

  if (view ()) {
    view ()->add_cellview_list_observer (m_cellviews_changed_delegate);
    view ()->add_cellview_observer (m_cellview_changed_delegate);
    view ()->add_rdb_list_changed_observer (m_rdbs_changed_delegate);
  }

  connect (layout_cb, SIGNAL (activated (int)), this, SLOT (cv_index_changed (int)));
  connect (rdb_cb, SIGNAL (activated (int)), this, SLOT (rdb_index_changed (int)));
  connect (open_pb, SIGNAL (clicked ()), this, SLOT (open_clicked ()));
  connect (saveas_pb, SIGNAL (clicked ()), this, SLOT (saveas_clicked ()));
  connect (reload_pb, SIGNAL (clicked ()), this, SLOT (reload_clicked ()));
  connect (unload_pb, SIGNAL (clicked ()), this, SLOT (unload_clicked ()));
  connect (configure_pb, SIGNAL (clicked ()), this, SLOT (configure_clicked ()));

  cellviews_changed ();
}

MarkerBrowserDialog::~MarkerBrowserDialog ()
{
  if (view ()) {
    view ()->remove_cellview_list_observer (m_cellviews_changed_delegate);
  }
}

void
MarkerBrowserDialog::configure_clicked ()
{
  ConfigurationDialogImpl config_dialog (this, main_window (), "MarkerBrowserPlugin");
  config_dialog.exec ();
}

void
MarkerBrowserDialog::unload_clicked ()
{
BEGIN_PROTECTED

  if (m_rdb_index < int (view ()->num_rdbs ()) && m_rdb_index >= 0) {

    rdb::Database *rdb = view ()->get_rdb (m_rdb_index);
    if (rdb && rdb->is_modified ()) {

      QMessageBox msgbox (QMessageBox::Question, tl::translate ("Unload Without Closing").c_str (),
                                                 tl::translate ("The database contains unsaved changes.\nPress 'Continue' to continue anyway or 'Cancel' for not unloading the database.").c_str ());
      QPushButton *ok = msgbox.addButton (tl::translate ("Continue").c_str (), QMessageBox::AcceptRole);
      msgbox.setDefaultButton (msgbox.addButton (QMessageBox::Cancel));

      msgbox.exec ();

      if (msgbox.clickedButton () != ok) {
        return;
      }

    }

    int new_rdb_index = m_rdb_index;

    view ()->remove_rdb (m_rdb_index);

    // try to use another rbd ...
    if (new_rdb_index >= int (view ()->num_rdbs ())) {
      --new_rdb_index;
    }
    if (new_rdb_index < int (view ()->num_rdbs ()) && new_rdb_index >= 0) {
      rdb_index_changed (new_rdb_index);      
    }

  }

END_PROTECTED
}

void
MarkerBrowserDialog::saveas_clicked ()
{
BEGIN_PROTECTED

  if (m_rdb_index < int (view ()->num_rdbs ()) && m_rdb_index >= 0) {

    rdb::Database *rdb = view ()->get_rdb (m_rdb_index);
    if (rdb) {

      //  prepare and open the file dialog
      lay::FileDialog save_dialog (this, tl::translate ("Marker Database File"), "KLayout RDB files (*.lyrdb)");
      std::string fn (rdb->filename ());
      if (save_dialog.get_save (fn)) {

        rdb->save (fn);
        rdb->reset_modified ();

      }

    }

  }

END_PROTECTED
}

void
MarkerBrowserDialog::reload_clicked ()
{
BEGIN_PROTECTED
  if (m_rdb_index < int (view ()->num_rdbs ()) && m_rdb_index >= 0) {

    rdb::Database *rdb = view ()->get_rdb (m_rdb_index);
    if (rdb && ! rdb->filename ().empty ()) {

      browser_frame->set_rdb (0);
      rdb->load (rdb->filename ());
      browser_frame->set_rdb (rdb);

    }

  }
END_PROTECTED
}

void
MarkerBrowserDialog::open_clicked ()
{
BEGIN_PROTECTED

  //  collect the formats available ...
  std::string fmts = tl::translate ("All files (*)");
  for (tl::Registrar<rdb::FormatDeclaration>::iterator rdr = tl::Registrar<rdb::FormatDeclaration>::begin (); rdr != tl::Registrar<rdb::FormatDeclaration>::end (); ++rdr) {
    fmts += ";;" + rdr->file_format ();
  }

  //  prepare and open the file dialog
  lay::FileDialog open_dialog (this, tl::translate ("Marker Database File"), fmts);
  if (open_dialog.get_open (m_open_filename)) {

    std::auto_ptr <rdb::Database> db (new rdb::Database ());
    db->load (m_open_filename);

    int rdb_index = view ()->add_rdb (db.release ());
    rdb_cb->setCurrentIndex (rdb_index);
    //  it looks like the setCurrentIndex does not issue this signal:
    rdb_index_changed (rdb_index);

  }

END_PROTECTED
}

bool 
MarkerBrowserDialog::configure (const std::string &name, const std::string &value)
{
  bool need_update = false;
  bool taken = true;

  if (name == cfg_rdb_context_mode) {

    context_mode_type context = m_context;
    MarkerBrowserContextModeConverter ().from_string (value, context);
    need_update = lay::test_and_set (m_context, context);

  } else if (name == cfg_rdb_window_mode) {

    window_type window = m_window;
    MarkerBrowserWindowModeConverter ().from_string (value, window);
    need_update = lay::test_and_set (m_window, window);

  } else if (name == cfg_rdb_window_dim) {

    double wdim = m_window_dim;
    tl::from_string (value, wdim);
    if (fabs (wdim - m_window_dim) > 1e-6) {
      m_window_dim = wdim;
      need_update = true;
    }

  } else if (name == cfg_rdb_max_marker_count) {

    unsigned int mc = 0;
    tl::from_string (value, mc);
    need_update = lay::test_and_set (m_max_marker_count, mc);

  } else {
    taken = false;
  }

  if (active () && need_update) {
    browser_frame->set_max_marker_count (m_max_marker_count);
    browser_frame->set_window (m_window, m_window_dim, m_context);
  }

  return taken;
}

void 
MarkerBrowserDialog::rdbs_changed ()
{
  int rdb_index = -1;

  rdb_cb->clear ();

  for (unsigned int i = 0; i < view ()->num_rdbs (); ++i) {
    const rdb::Database *rdb = view ()->get_rdb (i);
    rdb_cb->addItem (QString (rdb->name ().c_str ()));
    if (rdb->name () == m_rdb_name) {
      rdb_index = i;
    }
  }

  rdb_cb->setCurrentIndex (m_rdb_index);
  rdb_index_changed (rdb_index);
}

void 
MarkerBrowserDialog::cellview_changed ()
{
  browser_frame->update_markers ();
}

void 
MarkerBrowserDialog::cellviews_changed ()
{
  int cv_index = -1;

  layout_cb->clear ();

  for (unsigned int i = 0; i < view ()->cellviews (); ++i) {
    const lay::CellView &cv = view ()->cellview (i);
    layout_cb->addItem (QString (cv->name ().c_str ()));
    if (cv.is_valid () && cv->name () == m_layout_name) {
      cv_index = i;
    }
  }

  layout_cb->setCurrentIndex (cv_index);
  cv_index_changed (cv_index);
}

void 
MarkerBrowserDialog::rdb_index_changed (int index)
{
  if (m_rdb_index != index) {
    m_rdb_index = index;
    if (active ()) {
      update_content ();
    }
  }
}

void 
MarkerBrowserDialog::cv_index_changed (int index)
{
  if (m_cv_index != index) {
    m_cv_index = index;
    if (active ()) {
      update_content ();
    }
  }
}

void 
MarkerBrowserDialog::activated ()
{
  //  Switch to the active cellview index when no valid one is set.
  lay::CellView cv = view ()->cellview (m_cv_index);
  if (! cv.is_valid ()) {
    m_cv_index = view ()->active_cellview_index ();
  }

  if (m_rdb_index < 0 && view ()->get_rdb (0) != 0) {

    m_rdb_name = view ()->get_rdb (0)->name ();
    rdbs_changed ();

  } else {
    update_content ();
  }
}

void
MarkerBrowserDialog::update_content ()
{
  rdb::Database *rdb = view ()->get_rdb (m_rdb_index);
  central_stack->setCurrentIndex (rdb != 0 ? 0 : 1);
  
  saveas_pb->setEnabled (rdb != 0);
  unload_pb->setEnabled (rdb != 0);
  reload_pb->setEnabled (rdb != 0);

  //  KLUDGE: this will rebuild the internal lists several times ..
  browser_frame->set_rdb (rdb);
  browser_frame->set_max_marker_count (m_max_marker_count);
  browser_frame->set_window (m_window, m_window_dim, m_context);
  browser_frame->set_view (view (), m_cv_index);

  lay::CellView cv = view ()->cellview (m_cv_index);
  m_layout_name = std::string ();
  if (cv.is_valid ()) {
    m_layout_name = cv->name ();
  }

  if (layout_cb->currentIndex () != m_cv_index) {
    layout_cb->setCurrentIndex (m_cv_index);
  }

  if (rdb_cb->currentIndex () != m_rdb_index) {
    rdb_cb->setCurrentIndex (m_rdb_index);
  }
}

void 
MarkerBrowserDialog::deactivated ()
{
  browser_frame->set_rdb (0);
  browser_frame->set_view (0, 0);
}

void 
MarkerBrowserDialog::menu_activated (const std::string &symbol)
{
  if (symbol == "marker_browser::show") {
    view ()->deactivate_all_browsers ();
    activate ();
  } else {
    lay::Browser::menu_activated (symbol);
  }
}

}

