
/*

  KLayout Layout Viewer
  Copyright (C) 2006-2007 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#include <stdio.h>

#include "tlLog.h"
#include "gtf.h"

#include <QtCore/QMutex>

namespace tl
{

static int m_verbosity_level = 0;

void 
verbosity (int level)
{
  m_verbosity_level = level;
}

int 
verbosity ()
{
  return m_verbosity_level;
}

// ------------------------------------------------
//  Channel implementation

Channel::Channel () 
  : mp_lock (new QMutex ()), m_lock_count (0), m_no_endl (false)
{
  //  .. nothing else ..
}

Channel::~Channel ()
{
  delete mp_lock;
  mp_lock = 0;
}

ChannelProxy 
Channel::issue_proxy ()
{
  m_no_endl = false;
  if (m_lock_count == 0) {
    mp_lock->lock ();
  }
  begin ();
  return ChannelProxy (this);
}

void 
Channel::release_proxy ()
{
  if (! m_no_endl) {
    endl (); //  this helps remembering that there is something implicit going on here ..
  }
  m_no_endl = false;
  end ();
  if (--m_lock_count == 0) {
    mp_lock->unlock ();
  }
}

ChannelEndl endl;
ChannelNoendl noendl;

// ------------------------------------------------
//  ChannelProxy implementation

ChannelProxy::ChannelProxy (Channel *channel)
  : mp_channel (channel)
{
  ++mp_channel->m_lock_count;
}

ChannelProxy::~ChannelProxy ()
{
  mp_channel->release_proxy ();
}

ChannelProxy &
ChannelProxy::operator= (const ChannelProxy &d)
{
  if (mp_channel != d.mp_channel) {
    --mp_channel->m_lock_count;
    mp_channel = d.mp_channel;
    ++mp_channel->m_lock_count;
  }
  return *this;
}

ChannelProxy::ChannelProxy (const ChannelProxy &d)
  : mp_channel (d.mp_channel)
{
  ++mp_channel->m_lock_count;
}

// ------------------------------------------------
//  InfoChannel implementation

InfoChannel::InfoChannel ()
{
  //  .. nothing yet ..
}

InfoChannel::~InfoChannel ()
{
  //  .. nothing yet ..
}

void 
InfoChannel::puts (const char *s)
{
  fprintf (stdout, "%s", s);
}

void 
InfoChannel::endl ()
{
  fprintf (stdout, "\n");
}

void 
InfoChannel::end ()
{
  fflush (stdout);
}

void 
InfoChannel::begin ()
{
  //  .. nothing yet ..
}

InfoChannel info;

// ------------------------------------------------
//  WarningChannel implementation

WarningChannel::WarningChannel ()
{
  //  .. nothing yet ..
}

WarningChannel::~WarningChannel ()
{
  //  .. nothing yet ..
}

void 
WarningChannel::puts (const char *s)
{
  fprintf (stdout, "%s", s);
}

void 
WarningChannel::endl ()
{
  fprintf (stdout, "\n");
}

void 
WarningChannel::end ()
{
  fflush (stdout);
}

void 
WarningChannel::begin ()
{
  fprintf (stdout, "Warning: ");
}


/// The static instance of the warning channel
WarningChannel warn;

// ------------------------------------------------
//  ErrorChannel implementation

ErrorChannel::ErrorChannel ()
{
  //  .. nothing yet ..
}

ErrorChannel::~ErrorChannel ()
{
  //  .. nothing yet ..
}

void 
ErrorChannel::puts (const char *s)
{
  if (gtf::Recorder::instance ()) {
    gtf::Recorder::instance ()->errlog_puts (s);
  }
  fprintf (stderr, "%s", s);
}

void 
ErrorChannel::endl ()
{
  if (gtf::Recorder::instance ()) {
    gtf::Recorder::instance ()->errlog_endl ();
  }
  fprintf (stderr, "\n");
}

void 
ErrorChannel::end ()
{
  if (gtf::Recorder::instance ()) {
    gtf::Recorder::instance ()->errlog_end ();
  }
  fflush (stderr);
}

void 
ErrorChannel::begin ()
{
  if (gtf::Recorder::instance ()) {
    gtf::Recorder::instance ()->errlog_begin ();
  }
  fprintf (stderr, "*** ERROR: ");
}

/// The static instance of the error channel
ErrorChannel error;

} // namespace tl

