
/*

  KLayout Layout Viewer
  Copyright (C) 2006-2007 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#ifndef HDR_tlProgress
#define HDR_tlProgress

#include <string>
#include "tlException.h"

namespace tl
{

class Progress;

class ProgressAdaptor
{
public:  
  virtual ~ProgressAdaptor () { }

  virtual void register_object (Progress *progress) = 0;
  virtual void unregister_object (Progress *progress) = 0;
  virtual void trigger (Progress *progress) = 0;
  virtual void yield (Progress *progress) = 0;
};

/**
 *  @brief The "break" exception class
 *
 *  Exceptions of this kind are thrown if a break was encountered
 */

class BreakException
  : public tl::Exception
{
public:
  BreakException () : tl::Exception ("Operation cancelled") { }
};

/**
 *  @brief A "progress" reporter class 
 *
 *  A progress can be reported in two ways: as
 *  a relative value with a target "max" values and as
 *  an absolute value. 
 *  This implementation delegates the actual implementation
 *  to an "adaptor" class that handles the actual display
 *  of the progress.
 *  This class also allows for some kind of background processing:
 *  once the set() or operator++() methods are called, the 
 *  adaptor's yield() method is called once that "yield_interval" 
 *  calls have passed. The yield_interval value should be choosen
 *  sufficiently large so that no performance penalty is to be paid
 *  for this. In addition, each set() or operator++() tests whether
 *  the operation may have been cancelled and may throw an 
 *  BreakException in this case.
 */

class Progress
{
public:
  /**
   *  @brief Constructor
   *
   *  This initializes the progress reporter object.
   *  The string passed is the initial description and title string.
   * 
   *  @param desc The description and title string
   *  @param max_count The limit "max" value. 0 for absolute display of values.
   *  @param yield_interval See above.
   */
  Progress (const std::string &desc, size_t max_count = 0, size_t yield_interval = 1000);

  /**
   *  @brief The destructor
   */
  ~Progress ();

  /**
   *  @brief Set the resolution for the output (granularity)
   *
   *  This setting is independent of the format and controls the
   *  update frequency. It is the step by which the ratio 
   *  count/unit must increase before a new value is reported.
   */
  void set_res (double res)
  {
    m_res = res;
  }

  /** 
   *  @brief Set the format of the output.
   *
   *  This is a sprintf format string with the value being
   *  passed as a single double argument.
   */
  void set_format (const std::string &fmt)
  {
    m_format = fmt;
  }
  
  /**
   *  @brief Set the unit
   *
   *  This is the unit of the output. The unit is the 
   *  value by which the current count is divided to render
   *  the value passed to the format string.
   */
  void set_unit (size_t unit)
  {
    m_unit = unit;
  }

  /**
   *  @brief Set the description text
   */
  void set_desc (const std::string &desc);

  /**
   *  @brief Render a string representing the actual value
   *
   *  This method uses the unit and format.
   */
  std::string formatted_value () const;

  /**
   *  @brief Render the value as a double value.
   */
  double value () const;

  /**
   *  @brief Render the description string
   */
  const std::string &desc () const
  {
    return m_desc;
  }

  /**
   *  @brief Render the title string
   */
  const std::string &title () const
  {
    return m_title;
  }

  /**
   *  @brief Used to register an adaptor
   */
  static void register_adaptor (tl::ProgressAdaptor *pa);

  /**
   *  @brief Used by the adaptor to signal a break condition
   */
  void signal_break ();

  /**
   *  @brief Increment the count
   */
  Progress &operator++ () throw (tl::BreakException)
  {
    return set (m_count + 1);
  }

  /**
   *  @brief Set the count absolutely
   */
  Progress &set (size_t count) throw (tl::BreakException);

private:
  std::string m_desc;
  std::string m_title;
  std::string m_format;
  double m_res;
  size_t m_interval_count;
  size_t m_count;
  double m_unit;
  size_t m_yield_interval;
  double m_last_value;
  bool m_cancelled;
};

}

#endif

