
/*

  KLayout Layout Viewer
  Copyright (C) 2006-2007 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#include "tlVariant.h"
#include "tlString.h"

#include <string.h>

namespace tl
{

// --------------------------------------------------------------------
//  Implementation of tl::Variant

Variant::Variant () 
  : m_type (t_nil), m_string (0)
{
  // .. nothing yet ..
}

Variant::Variant (const std::string &s) 
  : m_type (t_string)
{
  m_string = new char [s.size () + 1];
  strcpy (m_string, s.c_str ());
}

Variant::Variant (const char *s) 
  : m_type (t_string)
{
  m_string = new char [strlen (s) + 1];
  strcpy (m_string, s);
}

Variant::Variant (double d)
  : m_type (t_double), m_string (0)
{
  m_var.m_double = d;
}

Variant::Variant (long l)
  : m_type (t_long), m_string (0)
{
  m_var.m_long = l;
}

Variant::Variant (long l, bool /*dummy*/)
  : m_type (t_id), m_string (0)
{
  m_var.m_long = l;
}

Variant::Variant (const Variant &v)
  : m_type (t_nil), m_string (0)
{
  operator= (v);
}

Variant::~Variant ()
{
  reset ();
}

void
Variant::reset ()
{
  if (m_string) {
    delete [] m_string;
  }
  m_string = 0;
  if (m_type == t_list) {
    delete m_var.m_list;
  } else if (is_user ()) {
    delete m_var.mp_user;
  }
  m_type = t_nil;
}

Variant &
Variant::operator= (const char *s)
{
  reset ();
  m_type = t_string;
  m_string = new char [strlen (s) + 1];
  strcpy (m_string, s);
  return *this;
}

Variant &
Variant::operator= (const std::string &s)
{
  reset ();
  m_type = t_string;
  m_string = new char [s.size () + 1];
  strcpy (m_string, s.c_str ());
  return *this;
}

Variant &
Variant::operator= (double d)
{
  reset ();
  m_type = t_double;
  m_var.m_double = d;
  return *this;
}

Variant &
Variant::operator= (long l)
{
  reset ();
  m_type = t_long;
  m_var.m_long = l;
  return *this;
}

Variant &
Variant::operator= (const Variant &v)
{
  if (this != &v) {

    reset ();

    m_type = v.m_type;
    if (m_type == t_double) {
      m_var.m_double = v.m_var.m_double;
    } else if (m_type == t_long || m_type == t_id) {
      m_var.m_long = v.m_var.m_long;
    } else if (m_type == t_string) {
      m_string = new char [strlen (v.m_string) + 1];
      strcpy (m_string, v.m_string);
    } else if (m_type == t_list) {
      m_var.m_list = new std::vector<tl::Variant> (*v.m_var.m_list);
    } else if (is_user ()) {
      m_var.mp_user = v.m_var.mp_user->clone ();
    }

  } 
  return *this;
}

bool 
Variant::operator== (const tl::Variant &d) const
{
  if (m_type != d.m_type) {
    return false;
  }

  if (m_type == t_nil) {
    return true;
  } else if (m_type == t_long || m_type == t_id) {
    return m_var.m_long == d.m_var.m_long;
  } else if (m_type == t_double) {
    return m_var.m_double == d.m_var.m_double;
  } else if (m_type == t_string) {
    return strcmp (m_string, d.m_string) == 0;
  } else if (m_type == t_list) {
    return *m_var.m_list == *d.m_var.m_list;
  } else if (is_user ()) {
    return m_var.mp_user->equal (d.m_var.mp_user);
  } else {
    return false;
  }
}

bool 
Variant::operator< (const tl::Variant &d) const
{
  if (m_type != d.m_type) {
    return int (m_type) < int (d.m_type);
  }

  if (m_type == t_nil) {
    return false;
  } else if (m_type == t_long || m_type == t_id) {
    return m_var.m_long < d.m_var.m_long;
  } else if (m_type == t_double) {
    return m_var.m_double < d.m_var.m_double;
  } else if (m_type == t_string) {
    return strcmp (m_string, d.m_string) < 0;
  } else if (m_type == t_list) {
    return *m_var.m_list < *d.m_var.m_list;
  } else if (is_user ()) {
    return m_var.mp_user->less (d.m_var.mp_user);
  } else {
    return false;
  }
}

const char *
Variant::to_string () const
{
  if (! m_string) {

    std::string r;

    if (m_type == t_nil) {
      r = "";
    } else if (m_type == t_double) {
      r = tl::to_string (m_var.m_double);
    } else if (m_type == t_long) {
      r = tl::to_string (m_var.m_long);
    } else if (m_type == t_list) {
      for (std::vector<tl::Variant>::const_iterator v = m_var.m_list->begin (); v != m_var.m_list->end (); ++v) {
        if (v != m_var.m_list->begin ()) {
          r += ",";
        }
        r += v->to_string ();
      }
    } else if (m_type == t_id)  {
      r = "{id" + tl::to_string (m_var.m_long) + "}";
    } else {
      r = "{user type}";
    }

    m_string = new char [r.size () + 1];
    strcpy (m_string, r.c_str ());

  }

  return m_string;
}

long 
Variant::to_long () const
{
  if (m_type == t_nil) {
    return 0;
  } else if (m_type == t_double) {
    return long (m_var.m_double);
  } else if (m_type == t_long) {
    return m_var.m_long;
  } else if (m_type == t_string) {
    long l = 0;
    tl::from_string (m_string, l);
    return l;
  } else {
    return 0;
  }
}

double 
Variant::to_double () const
{
  if (m_type == t_nil) {
    return 0;
  } else if (m_type == t_double) {
    return m_var.m_double;
  } else if (m_type == t_long) {
    return double (m_var.m_long);
  } else if (m_type == t_string) {
    double d = 0;
    tl::from_string (m_string, d);
    return d;
  } else {
    return 0;
  }
}

tl::Variant
Variant::empty_list ()
{
  static std::vector<tl::Variant> empty_list;
  return tl::Variant (empty_list.begin (), empty_list.end ());
}

std::string 
Variant::to_parsable_string () const
{
  if (is_long ()) {
    return "#" + tl::to_string (to_long ());
  } else if (is_double ()) {
    return "##" + tl::to_string (to_double ());
  } else if (is_string ()) {
    return tl::to_word_or_quoted_string (to_string ());
  } else if (is_list ()) {
    std::string r = "(";
    for (tl::Variant::iterator l = begin (); l != end (); ++l) {
      if (l != begin ()) {
        r += ",";
      }
      r += l->to_parsable_string ();
    }
    r += ")";
    return r;
  } else if (is_id ()) {
    return "{id" + tl::to_string (to_long ()) + "}";
  } else if (is_user ()) {
    return "{user type}";
  } else {
    return "";
  }
}

// ----------------------------------------------------------------------------------
//  Extractor implementation

template <>
void extractor_impl (tl::Extractor &ex, tl::Variant &v)
{
  if (! test_extractor_impl (ex, v)) {
    ex.error (tl::translate ("Expected a value specification"));
  }
}

template <>
bool test_extractor_impl (tl::Extractor &ex, tl::Variant &v)
{
  std::string s;

  if (ex.test ("##")) {

    double x = 0;
    ex.read (x);
    v = x;
    return true;

  } else if (ex.test ("#")) {

    long x = 0;
    ex.read (x);
    v = x;
    return true;

  } else if (ex.test ("(")) {

    std::vector<tl::Variant> values;
    if (! ex.test (")")) {
      while (true) {
        values.push_back (tl::Variant ());
        ex.read (values.back ());
        if (ex.test (",")) {
          //  .. continue
        } else {
          ex.expect (")");
          break;
        }
      }
    }
    v = tl::Variant (values.begin (), values.end ());
    return true;

  } else if (ex.try_read_word_or_quoted (s)) {

    v = tl::Variant (s);
    return true;

  } else {

    return false;

  }
}

} // namespace tl

