
/*

  KLayout Layout Viewer
  Copyright (C) 2006-2007 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#include "tlXMLParser.h"
#include "tlString.h"
#include "tlLog.h"
#include "tlAssert.h"

#ifndef USE_PARSIFAL
#  include <QtCore/QFile>
#endif

namespace tl
{

#ifdef USE_PARSIFAL

// --------------------------------------------------------------------
//  XMLStringSource implementation

XMLStringSource::XMLStringSource (const std::string &string)
  : m_string (string), mp_ptr (0)
{
  mp_ptr = m_string.c_str ();
}

XMLStringSource::~XMLStringSource ()
{
  //  .. nothing yet ..
}

int 
XMLStringSource::read (char *buffer, int requested) throw (XMLException)
{
  //  copy the number of requested bytes into the buffer ..
  int c = 0;
  while (*mp_ptr && c < requested) {
    *buffer++ = *mp_ptr++;
    ++c;
  }
  return c;
}

// --------------------------------------------------------------------
//  XMLStringSource implementation

XMLFileSource::XMLFileSource (const std::string &path)
  : mp_file (NULL)
{
  mp_file = fopen (path.c_str (), "r");
  if (mp_file == NULL) {
    throw tl::Exception (tl::translate ("Unable to open file: ") + path);
  }
}

XMLFileSource::~XMLFileSource ()
{
  if (mp_file != NULL) {
    fclose (mp_file);
    mp_file = NULL;
  }
}

int 
XMLFileSource::read (char *buffer, int requested) throw (XMLException)
{
  tl_assert (mp_file != NULL);
  int read = fread (buffer, 1, requested, mp_file);
  if (ferror (mp_file)) {
    throw tl::XMLException (tl::translate ("File read error"));
  }
  return read;
}

// --------------------------------------------------------------------
//  XMLAttributeAccessor implementation

std::string 
XMLAttributeAccessor::lname () const
{
  return to_string (mp_it->mp_ptr->localName);
}

std::string 
XMLAttributeAccessor::qname () const
{
  return to_string (mp_it->mp_ptr->qname);
}

std::string 
XMLAttributeAccessor::uri () const
{
  return to_string (mp_it->mp_ptr->uri);
}

std::string 
XMLAttributeAccessor::prefix () const
{
  return to_string (mp_it->mp_ptr->prefix);
}

std::string 
XMLAttributeAccessor::value () const
{
  return to_string (mp_it->mp_ptr->value);
}


// --------------------------------------------------------------------
//  XMLAttributes implementation

XMLAttributes::XMLAttributes (const LPXMLPARSER iparser, const LPXMLVECTOR vector)
  : mp_iparser (iparser), mp_vector (vector)
{
  //  nothing else
}

XMLAttributeIterator 
XMLAttributes::begin () const
{
  return XMLAttributeIterator (mp_vector, (unsigned int) 0);
}

XMLAttributeIterator 
XMLAttributes::end () const
{
  return XMLAttributeIterator (mp_vector, mp_vector->length);
}

XMLAttributeIterator 
XMLAttributes::by_name (const std::string &name) const
{
  const XMLCH *n;
  from_string (name, n);
  const LPXMLRUNTIMEATT attr = XMLParser_GetNamedItem (mp_iparser, n);
  if (attr) {
    return XMLAttributeIterator (mp_vector, attr);
  } else {
    return end ();
  }
}


// --------------------------------------------------------------------
//  XMLParser implementation

static int 
s_input_source (BYTE *buf, int requested, int *actual, void *inputData)
{
  XMLParser *parser = (XMLParser *) inputData;
  XMLSource *src = parser->source ();

  try {

    *actual = src->read ((char *) buf, requested);
    return (*actual < requested) ? BIS_EOF : 0;

  } catch (XMLLocatedException &ex) {
    parser->set_error (new XMLLocatedException (ex));
  } catch (XMLException &ex) {
    parser->set_error (new XMLLocatedException (ex.raw_msg (), parser->current_line (), parser->current_column ()));
  } catch (Exception &ex) {
    parser->set_error (new XMLLocatedException (ex.msg (), parser->current_line (), parser->current_column ()));
  }

  return BIS_ERR_INPUT;
}

static void 
s_error_handler (LPXMLPARSER iparser)
{
  if (iparser->ErrorCode != ERR_XMLP_ABORT) {
    XMLParser *parser = (XMLParser *) iparser->UserData;
    parser->set_error (new XMLLocatedException (to_string (iparser->ErrorString), iparser->ErrorLine, iparser->ErrorColumn));
  }
}

int
XMLParser::s_start_element (void *userData, const XMLCH *uri, const XMLCH *lname, 
			     const XMLCH *qname, LPXMLVECTOR a)
{
  XMLParser *parser = (XMLParser *) userData;
  XMLHandler *handler = parser->handler ();

  try {

    XMLAttributes attrs (parser->mp_iparser, a);
    handler->start_element (to_string (uri), to_string (lname), to_string (qname), attrs);
    return XML_OK;

  } catch (XMLLocatedException &ex) {
    parser->set_error (new XMLLocatedException (ex));
  } catch (XMLException &ex) {
    parser->set_error (new XMLLocatedException (ex.raw_msg (), parser->current_line (), parser->current_column ()));
  }

  return XML_ABORT;
}

static int
s_end_element (void *userData, const XMLCH *uri, const XMLCH *lname, 
			   const XMLCH *qname)
{
  XMLParser *parser = (XMLParser *) userData;
  XMLHandler *handler = parser->handler ();

  try {

    handler->end_element (to_string (uri), to_string (lname), to_string (qname));
    return XML_OK;

  } catch (XMLLocatedException &ex) {
    parser->set_error (new XMLLocatedException (ex));
  } catch (XMLException &ex) {
    parser->set_error (new XMLLocatedException (ex.raw_msg (), parser->current_line (), parser->current_column ()));
  }

  return XML_ABORT;
}

static int
s_characters (void *userData, const XMLCH *buf, int size)
{
  XMLParser *parser = (XMLParser *) userData;
  XMLHandler *handler = parser->handler ();

  try {

    handler->characters (to_string (buf, size));
    return XML_OK;

  } catch (XMLLocatedException &ex) {
    parser->set_error (new XMLLocatedException (ex));
  } catch (XMLException &ex) {
    parser->set_error (new XMLLocatedException (ex.raw_msg (), parser->current_line (), parser->current_column ()));
  }

  return XML_ABORT;
}

static int
s_ignorable_whitespace (void *userData, const XMLCH *buf, int size)
{
  XMLParser *parser = (XMLParser *) userData;
  XMLHandler *handler = parser->handler ();

  try {

    handler->ignorable_whitespace (to_string (buf, size));
    return XML_OK;

  } catch (XMLLocatedException &ex) {
    parser->set_error (new XMLLocatedException (ex));
  } catch (XMLException &ex) {
    parser->set_error (new XMLLocatedException (ex.raw_msg (), parser->current_line (), parser->current_column ()));
  }

  return XML_ABORT;
}

static int
s_start_document (void *userData)
{
  XMLParser *parser = (XMLParser *) userData;
  XMLHandler *handler = parser->handler ();

  try {

    handler->start_document ();
    return XML_OK;

  } catch (XMLLocatedException &ex) {
    parser->set_error (new XMLLocatedException (ex));
  } catch (XMLException &ex) {
    parser->set_error (new XMLLocatedException (ex.raw_msg (), parser->current_line (), parser->current_column ()));
  }

  return XML_ABORT;
}

static int
s_end_document (void *userData)
{
  XMLParser *parser = (XMLParser *) userData;
  XMLHandler *handler = parser->handler ();

  try {

    handler->end_document ();
    return XML_OK;

  } catch (XMLLocatedException &ex) {
    parser->set_error (new XMLLocatedException (ex));
  } catch (XMLException &ex) {
    parser->set_error (new XMLLocatedException (ex.raw_msg (), parser->current_line (), parser->current_column ()));
  }

  return XML_ABORT;
}


XMLParser::XMLParser ()
  : mp_iparser (0), mp_source (0), mp_handler (0), mp_error (0)
{
  tl_assert (XMLParser_Create (&mp_iparser) != 0);

  mp_iparser->UserData = this;
  mp_iparser->errorHandler = &s_error_handler;
  mp_iparser->startElementHandler = &s_start_element;
  mp_iparser->endElementHandler = &s_end_element;
  mp_iparser->charactersHandler = &s_characters;
  mp_iparser->ignorableWhitespaceHandler = &s_ignorable_whitespace;
  mp_iparser->startDocumentHandler = &s_start_document;
  mp_iparser->endDocumentHandler = &s_end_document;
}

XMLParser::~XMLParser ()
{
  if (mp_iparser) {
    XMLParser_Free (mp_iparser);
    mp_iparser = 0;
  }
  set_error (0);
}

void
XMLParser::set_error (XMLLocatedException *ex)
{
  if (mp_error) {
    delete mp_error;
    mp_error = 0;
  }
  mp_error = ex;
}

void 
XMLParser::parse (XMLSource &source, XMLHandler &handler) throw (XMLException)
{
  mp_source = &source;
  mp_handler = &handler;
  set_error (0);

  if (! XMLParser_Parse (mp_iparser, &s_input_source, this, NULL /*encoding*/)) {

    //  produce an error if we got one.
    mp_source = 0;
    mp_handler = 0;
    throw XMLLocatedException (*error ());

  }

  mp_source = 0;
  mp_handler = 0;
  tl_assert (error () == 0);
}

int 
XMLParser::current_line ()
{
  //  :HINT: this information may not be precise - see libparsifal documentation
  return XMLParser_GetCurrentLine (mp_iparser);
}

int 
XMLParser::current_column ()
{
  //  :HINT: this information may not be precise - see libparsifal documentation
  return XMLParser_GetCurrentColumn (mp_iparser);
}

#else

// --------------------------------------------------------------------
//  XMLStringSource implementation

XMLStringSource::XMLStringSource (const std::string &string)
{
  mp_source = new QXmlInputSource ();
  mp_source->setData (QByteArray (string.c_str ()));
}

XMLStringSource::~XMLStringSource ()
{
  delete mp_source;
  mp_source = 0;
}

// --------------------------------------------------------------------
//  XMLStringSource implementation

XMLFileSource::XMLFileSource (const std::string &path)
{
  mp_file = new QFile (path.c_str ());
  mp_source = new QXmlInputSource (mp_file);
  if (! mp_file->exists ()) {
    throw tl::XMLException (tl::translate ("File does not exist: ") + path);
  }
}

XMLFileSource::~XMLFileSource ()
{
  delete mp_source;
  mp_source = 0;
  delete mp_file;
  mp_file = 0;
}

// --------------------------------------------------------------------
//  XMLParser implementation

XMLParser::XMLParser ()
{
  mp_reader = new QXmlSimpleReader ();
}

XMLParser::~XMLParser ()
{
  delete mp_reader;
  mp_reader = 0;
}

void 
XMLParser::parse (XMLSource &source, QXmlDefaultHandler &handler) throw (XMLException)
{
  mp_reader->setContentHandler (&handler);
  mp_reader->setErrorHandler (&handler);

  mp_reader->parse (source.source (), false /*=not incremental*/);
}


#endif

// -----------------------------------------------------------------
//  The C++ structure definition interface (for use cases see tlXMLParser.ut)

//  XMLElementProxy implementation

XMLElementProxy::XMLElementProxy (const XMLElementProxy &d)
  : mp_ptr (d.mp_ptr->clone ())
{
  //  .. nothing yet ..
}

XMLElementProxy::XMLElementProxy (const XMLElementBase &d)
  : mp_ptr (d.clone ())
{
  //  .. nothing yet ..
}

XMLElementProxy::XMLElementProxy (XMLElementBase *d)
  : mp_ptr (d)
{
  //  .. nothing yet ..
}

XMLElementProxy::~XMLElementProxy ()
{
  delete mp_ptr;
  mp_ptr = 0;
}

//  XMLElementBase implementation

void 
XMLElementBase::write_indent (std::ostream &os, int indent)
{
  for (int i = 0; i < indent; ++i) {
    os << " ";
  }
}

void 
XMLElementBase::write_string (std::ostream &os, const std::string &s)
{
  for (const char *cp = s.c_str (); *cp; ++cp) {
    if (*cp == '&') {
      os << "&amp;";
    } else if (*cp == '<') {
      os << "&lt;";
    } else if (*cp == '<') {
      os << "&gt;";
    } else if (*cp < ' ') {
      os << "&#" << int ((unsigned char) *cp) << ";";
    } else {
      os << *cp;
    }
  }
}

//  XMLStructureHandler implementation

#ifdef USE_PARSIFAL

XMLStructureHandler::XMLStructureHandler (const XMLElementBase *root, XMLReaderState *reader_state) 
  : tl::XMLHandler (), mp_root (root), mp_state (reader_state)
{ 
  // .. nothing yet ..
}

void 
XMLStructureHandler::start_element (const std::string &uri, const std::string &lname, const std::string &qname, const tl::XMLAttributes &atts) throw (tl::XMLException)
{
  const XMLElementBase *new_element = 0;
  const XMLElementBase *parent = 0;

  if (m_stack.size () == 0) {
    if (! mp_root->check_name (uri, lname, qname)) {
      throw tl::XMLException (tl::translate ("Root element must be ") + mp_root->name ());
    }
    new_element = mp_root;
  } else {
    parent = m_stack.back ();
    for (XMLElementBase::iterator c = parent->begin (); c != parent->end (); ++c) {
      if ((*c)->check_name (uri, lname, qname)) {
        new_element = (*c).get ();
        break;
      }
    }
    if (! new_element) {
      throw tl::XMLException (tl::translate ("Unknown element/as child of: ") + qname + "/" + m_stack.back ()->name ());
    }
  }

  try {
    new_element->create (parent, *mp_state);
  } catch (tl::XMLException &ex) {
    throw;
  } catch (tl::Exception &ex) {
    throw tl::XMLException (ex.msg ());
  }
  m_stack.push_back (new_element);
}

void 
XMLStructureHandler::end_element (const std::string &uri, const std::string &lname, const std::string &qname) throw (tl::XMLException)
{
  const XMLElementBase *element = m_stack.back ();
  m_stack.pop_back ();

  try {
    if (m_stack.size () == 0) {
      element->finish (0, *mp_state);
    } else {
      element->finish (m_stack.back (), *mp_state);
    }
  } catch (tl::XMLException &ex) {
    throw;
  } catch (tl::Exception &ex) {
    throw tl::XMLException (ex.msg ());
  }
}

void 
XMLStructureHandler::characters (const std::string &t) throw (tl::XMLException)
{
  try {
    m_stack.back ()->cdata (t, *mp_state);
  } catch (tl::XMLException &ex) {
    throw;
  } catch (tl::Exception &ex) {
    throw tl::XMLException (ex.msg ());
  }
}

#else

XMLStructureHandler::XMLStructureHandler (const XMLElementBase *root, XMLReaderState *reader_state) 
  : QXmlDefaultHandler (), mp_root (root), mp_locator (0), mp_state (reader_state)
{ 
  // .. nothing yet ..
}

void
XMLStructureHandler::setDocumentLocator (QXmlLocator *locator)
{
  mp_locator = locator;
}

bool 
XMLStructureHandler::startElement (const QString &qs_uri, const QString &qs_lname, const QString &qs_qname, const QXmlAttributes &atts)
{
  const XMLElementBase *new_element = 0;
  const XMLElementBase *parent = 0;

  std::string uri (qs_uri.ascii ());
  std::string lname (qs_lname.ascii ());
  std::string qname (qs_qname.ascii ());

  try {

    if (m_stack.size () == 0) {
      if (! mp_root->check_name (uri, lname, qname)) {
        throw tl::XMLException (tl::translate ("Root element must be ") + mp_root->name ());
      }
      new_element = mp_root;
    } else {
      parent = m_stack.back ();
      for (XMLElementBase::iterator c = parent->begin (); c != parent->end (); ++c) {
        if ((*c)->check_name (uri, lname, qname)) {
          new_element = (*c).get ();
          break;
        }
      }
      if (! new_element) {
        throw tl::XMLException (tl::translate ("Unknown element/as child of: ") + qname + "/" + m_stack.back ()->name ());
      }
    }

    new_element->create (parent, *mp_state);

  } catch (tl::XMLException &ex) {
    throw tl::XMLLocatedException (ex.raw_msg (), mp_locator->lineNumber (), mp_locator->columnNumber ());
  } catch (tl::Exception &ex) {
    throw tl::XMLLocatedException (ex.msg (), mp_locator->lineNumber (), mp_locator->columnNumber ());
  }
  m_stack.push_back (new_element);

  //  successful
  return true;
}

bool 
XMLStructureHandler::endElement (const QString &, const QString &, const QString &)
{
  const XMLElementBase *element = m_stack.back ();
  m_stack.pop_back ();

  try {
    if (m_stack.size () == 0) {
      element->finish (0, *mp_state);
    } else {
      element->finish (m_stack.back (), *mp_state);
    }
  } catch (tl::XMLException &ex) {
    throw tl::XMLLocatedException (ex.raw_msg (), mp_locator->lineNumber (), mp_locator->columnNumber ());
  } catch (tl::Exception &ex) {
    throw tl::XMLLocatedException (ex.msg (), mp_locator->lineNumber (), mp_locator->columnNumber ());
  }

  //  successful
  return true;
}

bool 
XMLStructureHandler::characters (const QString &t)
{
  try {
    m_stack.back ()->cdata (t.ascii (), *mp_state);
  } catch (tl::XMLException &ex) {
    throw tl::XMLLocatedException (ex.raw_msg (), mp_locator->lineNumber (), mp_locator->columnNumber ());
  } catch (tl::Exception &ex) {
    throw tl::XMLLocatedException (ex.msg (), mp_locator->lineNumber (), mp_locator->columnNumber ());
  }

  //  successful
  return true;
}

bool  
XMLStructureHandler::error (const QXmlParseException &ex)
{
  throw tl::XMLLocatedException (ex.message ().ascii (), ex.lineNumber (), ex.columnNumber ());
}

bool  
XMLStructureHandler::fatalError (const QXmlParseException &ex)
{
  throw tl::XMLLocatedException (ex.message ().ascii (), ex.lineNumber (), ex.columnNumber ());
}

bool  
XMLStructureHandler::warning (const QXmlParseException &ex)
{
  tl::XMLLocatedException lex (ex.message ().ascii (), ex.lineNumber (), ex.columnNumber ());
  tl::warn << lex.msg ();
  //  continue
  return true;
}

#endif

// --------------------------------------------------------------------
//  XMLReaderState implementation

XMLReaderState::XMLReaderState ()
{
  //  .. nothing yet ..
}

XMLReaderState::~XMLReaderState ()
{
  for (std::vector <XMLReaderProxyBase *>::const_iterator o = m_objects.begin (); o != m_objects.end (); ++o) {
    (*o)->release ();
    delete *o;
  }
  m_objects.clear ();
}

// --------------------------------------------------------------------
//  XMLWriterState implementation

XMLWriterState::XMLWriterState ()
{
  //  .. nothing yet ..
}

}

